/* hp3000_gic_lpa.c: HP 3000 2611/13/17/18/19 "Amigo" Line Printer simulator

   Copyright (c) 2024, J. David Bryan

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
   AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
   ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
   WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of the author shall not be used
   in advertising or otherwise to promote the sale, use or other dealings in
   this Software without prior written authorization from the author.

   LPA          HP 2611/13/17/18/19 "Amigo" Line Printer simulator

   16-Sep-24    JDB     First release
   18-Jun-24    JDB     Passes the On-Line HP Line Printers Verification (D466A)
   07-May-24    JDB     Created

   References:
     - 26069A HPIB-to-Differential Translator Kit Installation and Service Manual
         (26069-90901 March 1980)
     - Line Printer Operating and Programming Manual
         (30209-90008, June 1976)
     - IEEE Standard Digital Interface for Programmable Instrumentation
         (IEEE-488A-1980, September 1979)


   The 26069A HPIB-to-Differential Translator allows the HP 2611, 2613, 2617,
   2618, and 2619 line printers with differential interfaces to be connected to
   the HP 3000 General I/O Channel (GIC).  This device translates the HP-IB
   signals supplied by the GIC to the parallel differential signals required by
   the printers.  It uses MPE driver "HIOLPRT2" (module 21).  Each translator
   supports a single line printer.  Up to four translators are supported on the
   Series 44 and above.

   The translator consists of a single Printed Circuit Assembly (PCA) containing
   a Z-80 microprocessor, four 2K x 8 MOS ROMs, one 1K x 8 bipolar ROM, four 1K
   x 4 MOS static RAMs, and a HP-custom Computer to HP-IB Interface (CHI) chip.

   Hardware configuration settings include the HP-IB bus address and 132 or 136
   column print line length.  A momentary switch that initiates an on-board
   self-test is provided, as are eight LEDs for self-test results.  Self-test
   can also be initiated programmatically.

   The translator uses the "Amigo" command protocol.  Implemented originally for
   connection to the HP 300 ("Amigo") computer, the protocol includes several
   commands related to HP-IB operation.  They include the Amigo Identify and
   Amigo Clear sequences, Read and Write Loopback channel tests, device self
   tests, and Device Specified Jump commands.

   The commands implemented by the translator are:

     Pri  Sec  Operation
     ---  ---  -------------------------------
     UNT  MSA  Identify
      14   -   Universal Device Clear
      04   -   Selected Device Clear

     MLA   00  Print data
     MLA   02  Enable/Disable Escape Sequences
     MLA   10  Amigo Clear
     MLA   1E  Write Loopback
     MLA   1F  Initiate Self-Test

     MTA   0E  Read Status
     MTA   10  Device-Specified Jump
     MTA   1E  Read Loopback
     MTA   1F  Read Self-Test Result

   The HP-IB bus transaction sequences for each of the commands are as follows:

     Identify
     --------

       ATN  UNT     Untalk
       ATN  MSA     My secondary address
            DAB     ID data byte #1
       EOI  DAB     ID data byte #2
       ATN  OTA     Talk 30


     Universal Device Clear
     ----------------------

       ATN  PCG     Primary command 14H
            ppe     Parallel poll enabled conditionally


     Selected Device Clear
     ---------------------

       ATN  MLA     My listen address
       ATN  PCG     Primary command 04H
            ppe     Parallel poll enabled conditionally


     Amigo Clear
     -----------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 10H
       EOI  DAB     Command parity checking enabled/disabled
       ATN  SDC     Selected device clear
            ppe     Parallel poll enabled conditionally
       ATN  UNL     Unlisten


     Device-Specified Jump
     ---------------------

       ATN  MTA     My talk address
       ATN  SCG     Secondary command 10H
       EOI  DAB     Reporting byte
            ppe     Parallel poll enabled conditionally
       ATN  UNT     Untalk


     Print Data
     ----------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 00H
            ppd     Parallel poll disabled
            DAB     Write data byte #1
            ...     (more data bytes)
       EOI  DAB     Write data byte #n
            ppe     Parallel poll enabled conditionally
       ATN  UNL     Unlisten


     Enable/Disable Escape Sequences
     -------------------------------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 02H
       EOI  DAB     Escape recognition action
            ppe     Parallel poll enabled conditionally
       ATN  UNL     Unlisten


     Read Status
     -----------

       ATN  MTA     My talk address
       ATN  SCG     Secondary command 0EH
       EOI  DAB     Status byte
            ppe     Parallel poll enabled conditionally
       ATN  UNT     Untalk


     Write Loopback and Read Loopback
     --------------------------------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 1EH
            DAB     Loopback data byte #1
            ...
       EOI  DAB     Loopback data byte #n
            ppe     Parallel poll enabled conditionally
       ATN  UNL     Unlisten

       ATN  MTA     My talk address
       ATN  SCG     Secondary command 1EH
            DAB     Loopback data byte #1
            ...
       EOI  DAB     Loopback data byte #n
            ppe     Parallel poll enabled conditionally
       ATN  UNT     Untalk


     Initiate Self-Test and Read Self-Test Result
     --------------------------------------------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 1FH
            DAB     Execute test #n
            ppe     Parallel poll enabled conditionally
       ATN  UNL     Unlisten

       ATN  MTA     My talk address
       ATN  SCG     Secondary command 1FH
       EOI  DAB     Self-test result byte
            ppe     Parallel poll enabled conditionally
       ATN  UNT     Untalk

   Unlike other protocols, a parallel poll response is not asserted to indicate
   command completion.  Instead, a poll response is asserted only when a
   128-byte bus buffer is available and either the printer is online or the DSJ
   value indicates that status must be read due to an error or initial power-on.
   Note that the first alternative will assert a poll response for an offline-
   to-online transition if there is a free buffer and no printer fault.  The
   second alternative asserts a poll response when DSJ > 0 and maintains it
   until DSJ is cleared by a Read Status command.  This gives the host a chance
   to see the error status before the poll response is denied to inhibit further
   print commands.  In tabular form:

     Buffer               Poll
     Avail   Online  DSJ  Response
     ------  ------  ---  --------
       No      --     -   Denied
      Yes      No     0   Denied
      Yes      No    1-3  Asserted
      Yes     Yes     -   Asserted

   The Universal Device Clear, Selected Device Clear, and Amigo Clear reset the
   print buffers, so they assert a poll response if the printer is ready.

   Printer data is buffered by eight 128-byte blocks of the translator RAM.
   These are filled and emptied in round-robin order.  A parallel poll response
   indicates that up to 128 bytes may be sent without danger of overrunning the
   translator.  As a consequence, the MPE driver sends Print Data commands in
   Burst Mode using 128-byte bursts. If a poll response is received after a
   burst, then another burst will be sent.

   The printer hardware also contains a print buffer that is large enough to
   hold an entire print line.  So the print process transfers data from system
   memory via the GIC to a translator buffer, then via the translator to the
   printer buffer, and finally to the print hammers that form the characters on
   the printed page.

   If enabled, the translator interprets two escape sequences on behalf of the
   connected printer to initiate paper motion.  They are:

     ESC & l nnn U -- slew by line count nnn.
     ESC & l nnn V -- slew by VFU channel nnn.

   These commands are converted into paper instructions and sent to the printer.
   All other escapes are absorbed and discarded silently.

   The escapes may be sent with spaces embedded in the numeric parameter; the
   translator ignores these.  If escapes are disabled, then the characters are
   sent to the printer as data.

   While the printers accept paper instructions for slewing up to 15 lines, the
   translator accepts line counts up to 63.  It handles this by issuing multiple
   slew commands until the full line count is achieved.

   A paper instruction to slew to a VFU channel that is not punched or to a VFU
   channel other than those defined for the printer will cause a tape fault.

   The Read Status command returns one status byte in the following format:

       7   6 | 5   4   3 | 2   1   0
     +---+---+---+---+---+---+---+---+
     | L | B | - | E | P | - | O | F |
     +---+---+---+---+---+---+---+---+

   Where:

     L = Printer online
     B = I/O buffer ready
     E = Protocol error
     P = Bus parity error
     O = Paper out
     F = Power fail

   The following hardware signals are exchanged between the translator and the
   printer:

     Src  Signal  Description
     ---  ------  ------------------------------------------------
      P   BUFRDY  Printer is online and not printing or slewing
      P   ONLINE  Printer is READY and online
      P   READY   Power is on, no alarm condition, VFU initialized
      P   CH9     VFU channel 9 is punched
      P   CH12    VFU channel 12 is punched
      P   DEMAND  Print online and ready for a character

      I   MASCLR  Master clear
      I   STROBE  Data is valid on D7-D1 lines
      I   PINST   Paper instruction
      I   D1      Data bit 1
      I   D2      Data bit 2
      I   D3      Data bit 3
      I   D4      Data bit 4
      I   D5      Data bit 5
      I   D6      Data bit 6
      I   D7      Data bit 7

   The READY and ONLINE signals indicate the current state of the printer.
   READY asserts when printer power is on, no alarm condition (paper out, tape
   format error) exits, and the VFU has been initialized.  ONLINE asserts when
   READY is asserted and the Online button is pressed.  Therefore:

     -- Printer --  -- Read Status --
     ONLINE  READY  Online  Paper Out  Meaning
     ------  -----  ------  ---------  ------------------------
     False   False    0         1      Paper out or VFU error
     False   True     0         0      Paper loaded and offline
     True    False    1         1      (prohibited)
     True    True     1         0      Paper loaded and online

   The complement of the READY signal is reflected in the Paper Out status bit,
   while the ONLINE signal is reflected in the Online status bit.  In
   simulation, paper faults (print unit not attached) and tape faults (commanded
   VFU channel is not punched) deny the READY condition, while the presence of
   the OFFLINE bit in the unit flags denies the ONLINE condition.

   The translator does not communicate the VFU Channel 9 and 12 signals through
   the status byte, although they are presented to the translator by the
   printer.  The BUFRDY signal from the printer is similarly ignored.

   When the ON/OFFLINE button on the printer is pressed, the printer will not go
   offline (i.e., deny the ONLINE signal) if there are characters in the print
   buffer.  Instead, the offline condition is held off until an internal "allow
   offline" signal asserts.  This occurs when the print buffer is empty and the
   print cycle is inactive, which has implications for the SET OFFLINE and
   DETACH commands if they are issued while the print buffer contains data or
   the printer unit is busy executing a print action.

   The SET LPA OFFLINE and DETACH LPA commands check for data in the print
   buffer or a print operation in progress.  If either condition is true, they
   set a deferred-action flag and display "Command not completed."  A SHOW LPA
   will show that the device is still online and attached.  Once simulation is
   resumed and the print operation completes, the printer is set offline or
   detached as requested.  No console message reports this, as it is assumed
   that the executing program will detect the condition and report accordingly.
   A subsequent SHOW LPA will indicate the new status.

   A SET LPA ONLINE command when the deferred-action flag is set simply clears
   the flag, which cancels the pending offline or detach condition.

   A RESET LPA command also clears the deferred-action flag and so clears any
   pending offline or detach.  However, it also clears the translator and
   printer buffers and terminates any print action in progress, so a subsequent
   SET LPA OFFLINE or DETACH LPA will succeed immediately.

   The simulator provides one SIMH unit that represents the printer and a second
   SIMH unit that represents the translator's bus controller.  All commands are
   executed by the bus controller; the printer unit is responsible only for
   moving characters from the bus buffer to the print buffer, and then to the
   simulated paper on receipt of a paper instruction.

   The bus controller has eight possible states, as follows:

     Command_Ready     -- Waiting for a new command to be received.  This is the
                          idle state for the bus controller.

     Parameter_Wait    -- Waiting for a command parameter to be received.  For
                          example, Set Escape is followed by a parameter byte
                          that determines if escape recognition is enabled or
                          disabled, and Amigo Clear is followed by an option
                          byte that specifies whether bus parity is or is not to
                          be checked.

     Command_Send      -- Waiting to send a command byte, for example the status
                          byte for the Read Status command.  The bus ATN and
                          NRFD signals must be denied before a byte is sourced.

     Command_Receive   -- Waiting to receive a command byte, for example the
                          bytes of a Write Loopback command.  When a data byte
                          is accepted, the simulator asserts NRFD to hold off
                          the talker until the bus service is entered in this
                          state.

     Execution_Receive -- Waiting to receive a data byte for a Print Data
                          command.  As in the Command_Receive state, NRFD is
                          asserted to hold off the talker until the bus event is
                          serviced.

     Completion_Wait   -- Waiting for a command to complete before optionally
                          asserting a parallel poll response.  The bus
                          controller enters this state at the conclusion of all
                          commands.

     Error_Source      -- Waiting to send an error byte in response to an
                          invalid or undefined talk secondary.  If an illegal
                          secondary is received, entry in the bus controller
                          event service sources a single byte of value 1 with
                          EOI to satisfy the inbound transfer.

     Error_Sink        -- Waiting to receive an error byte in response to an
                          invalid or undefined listen secondary.  If such a
                          secondary is received, the bus controller event
                          service will receive and discard as many bytes as the
                          host sends until one tagged with EOI is received.

   The bus controller initiates all commands.  For bus commands such as Amigo
   Identify or Device Specified Jump, only the bus controller unit is scheduled.
   For Print Data commands, both the bus controller and printer units are
   scheduled, operating in tandem to transfer data between the bus, bus buffers,
   and the printer buffer.

   A Print Data command begins with the bus unit accepting data and adding it to
   the next available bus buffer.  When EOI is asserted with a data byte, or
   when the buffer is full, the printer unit is scheduled to transfer the data
   from the bus buffer to the print buffer and, if a paper instruction command
   is included, from the print buffer to the attached printer file.  Additional
   Print Data commands are accepted concurrent with print unit operations until
   all eight bus buffers are occupied.  At that point, the bus unit denies the
   parallel poll response to indicate that additional data cannot be accepted
   until the printer empties one of the bus buffers.  The bus and print units
   continue to operate in parallel, with the bus buffers filling and emptying in
   a circular queue, until all print data has been transferred to the attached
   printer file.

   Two output modes are provided by the simulator.  In compact mode, a printed
   line is terminated by a CR LF pair, but subsequent line spacing is performed
   by LFs alone.  A top-of-form request emits a FF character.  This mode is
   appropriate when the printer output file will be sent to a physical printer
   connected to the host.  In expanded mode, paper advance is handled solely by
   emitting CR LF pairs. This mode is used where the printer output file will be
   saved or manipulated as a text file.

   The simulation provides both realistic and optimized ("fast") timing modes.
   The setting determines the scheduling delays for the SIMH unit representing
   the print mechanism as well as the printer and bus controllers.


   Implementation notes:

    1. In hardware, a paper-out condition is noted, but the line printer does
       not go offline until the top of the next form is reached.  This ensures
       that the current page is completed first.  By contrast, a torn-paper
       condition causes the printer to go offline at the completion of the
       current line.  In simulation, a DETACH is handled as a torn-paper
       condition.

    2. Slewing in expanded mode is performed by appending CR LF pairs to the
       printer buffer and then writing the combined buffer to the printer output
       file.  The size of the printer buffer must accommodate the largest print
       line (136 characters) plus the largest possible slew (144 lines * 2
       characters per line).

    3. The 26069A PCA hardware contains an HP-IB address switch, a Self-Test
       switch, and a 132/136-column jumper.  Only the address switch is modeled
       in simulation.  The self-test switch is irrelevant to simulation, and the
       column jumper is not needed, as selecting the printer model automatically
       determines the number of print columns supported.
*/



#include <ctype.h>

#include "hp3000_defs.h"
#include "hp3000_io.h"
#include "hp3000_imb.h"                         /* must be included after hp3000_io.h */
#include "hp3000_gic.h"



/* Program constants */

#define MAX_LPA_COUNT       4                   /* count of potential LPA instances */

#define LPA_AUXUNITS        1                   /* the number of bus controller units */

#define UNIT_COUNT          (1 + LPA_AUXUNITS)  /* the total number of units */

#define LP_UPTR             (&dptr->units [0])  /* the line printer unit pointer */
#define BUS_UPTR            (&dptr->units [1])  /* the bus controller unit pointer */

#define F_SWITCH            SWMASK ('F')        /* force an immediate detach */
#define P_SWITCH            SWMASK ('P')        /* perform a power-on reset */

#define ENABLE_ESC          0200                /* enable escape sequence recognition code */
#define DISABLE_ESC         0202                /* disable escape sequence recognition code */


/* Printer program constants */

#define CR                  '\r'                /* carriage return */
#define LF                  '\n'                /* line feed */
#define FF                  '\f'                /* form feed */
#define ESC                 '\033'              /* escape */
#define DEL                 '\177'              /* delete */

#define PAPER_CMD           0200u               /* printer paper command indicator */
#define DATA_MASK           0177u               /* printer uses only 7 bits for data */
#define FORMAT_VFU          0100u               /* printer VFU selector */
#define FORMAT_MASK         0117u               /* printer format command mask for 12-channel VFU */
#define FORMAT_VFU_8_MASK   0107u               /* printer format command mask for 8-channel VFU */

#define FORMAT_SUPPRESS     0000u               /* format code to slew 0 lines */
#define FORMAT_VFU_CHAN_1   0100u               /* format code to slew to VFU channel 1 */
#define FORMAT_VFU_BIAS     0077u               /* bias converting from format code to channel number */

#define VFU_MAX             144                 /* maximum number of VFU form lines */
#define VFU_SIZE            (VFU_MAX + 1)       /* size of the VFU array */
#define LINE_SIZE           256                 /* size of the character array used to read the VFU file */

#define VFU_WIDTH           12                  /* maximum number of VFU channels */

#define VFU_CHANNEL_1       BIT (11)            /* top of form */
#define VFU_CHANNEL_2       BIT (10)            /* bottom of form */
#define VFU_CHANNEL_3       BIT (9)             /* single space */
#define VFU_CHANNEL_4       BIT (8)             /* double space */
#define VFU_CHANNEL_5       BIT (7)             /* triple space */
#define VFU_CHANNEL_6       BIT (6)             /* half page */
#define VFU_CHANNEL_7       BIT (5)             /* quarter page */
#define VFU_CHANNEL_8       BIT (4)             /* sixth page */
#define VFU_CHANNEL_9       BIT (3)             /* bottom of form */
#define VFU_CHANNEL_10      BIT (2)             /* (unassigned) */
#define VFU_CHANNEL_11      BIT (1)             /* (unassigned) */
#define VFU_CHANNEL_12      BIT (0)             /* (unassigned) */

#define CHARS_MAX           136                 /* maximum number of characters buffered by the printers */
#define SLEW_MAX            15                  /* maximum number of lines to slew in a single command */

#define PB_SIZE             (CHARS_MAX + VFU_MAX * 2)   /* print buffer = max chars + max VFU * 2 (CR LF) */


/* Printer types.


   Implementation notes:

    1. The HP 2607 printer is not supported by the translator or by MPE driver
       HIOLPRT2.  However, we define the corresponding model symbol here to
       preserve commonality with the LP device code.
*/

typedef enum {                                  /* printer model identifiers */
    HP_2607,                                    /*   HP 2607A */
    HP_2611,                                    /*   HP 2611A */
    HP_2613,                                    /*   HP 2613A */
    HP_2617,                                    /*   HP 2617A */
    HP_2618,                                    /*   HP 2618A */
    HP_2619                                     /*   HP 2619A */
    } PRINTER_TYPE;


/* Device flags and accessors.

   DEV_V_UF + 15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
             +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
             | - | - | - | - | - | - | - | - |  address  |    channel    | R |
             +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     R = real-time mode is enabled
*/

#define DEV_ADDRESS         DF_BITS (7, 5)      /* HP-IB controller bus address */
#define DEV_CHANNEL         DF_BITS (4, 1)      /* GIC channel address */
#define DEV_REALTIME        DF_BIT  (0)         /* (R) real-time mode is selected */

#define ADDRESS(w)          DF_BITS_TO (7, 5, w)
#define CHANNEL(w)          DF_BITS_TO (4, 1, w)

#define TO_ADDRESS(w)       TO_DF_BITS (7, 5, w)
#define TO_CHANNEL(w)       TO_DF_BITS (4, 1, w)


/* Unit flags and accessors.

   UNIT_V_UF + 15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
              +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
              | - | - | - | - | - | - | - | - | - | - | - | O | E |   model   |
              +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     O     = the printer is offline
     E     = print using expanded output
     model = the DRIVE_TYPE enumeration constant for the simulated drive
*/

#define UNIT_ONLINE         0
#define UNIT_OFFLINE        UF_BIT  (4)         /* (O) the printer is offline */
#define UNIT_EXPAND         UF_BIT  (3)         /* (E) print using expanded format */
#define UNIT_MODEL          UF_BITS (2, 0)      /* the printer model ID */

#define MODEL(w)            UF_BITS_TO (2, 0, w)

#define TO_MODEL(w)         TO_UF_BITS (2, 0, w)

#define UNIT_2607           TO_MODEL (HP_2607)
#define UNIT_2611           TO_MODEL (HP_2611)
#define UNIT_2613           TO_MODEL (HP_2613)
#define UNIT_2617           TO_MODEL (HP_2617)
#define UNIT_2618           TO_MODEL (HP_2618)
#define UNIT_2619           TO_MODEL (HP_2619)

#define Instance            buf                 /* the unit field for the instance ID of the device */


/* Printer locality states */

typedef enum {                                  /* printer locality */
    Offline = 0,                                /*   printer is going offline */
    Online  = 1                                 /*   printer is going online */
    } LOCALITY;

static const char *locality_names [] = {        /* printer locality names, indexed by LOCALITY */
    "offline",                                  /*   printer is offline */
    "online"                                    /*   printer is online */
    };

static const char *condition_names [] = {       /* object condition names, indexed by t_bool */
    "disabled",                                 /*   condition is disabled */
    "enabled"                                   /*   condition is enabled */
    };


/* Printer properties table.

   This table contains the characteristics that vary between printer models.
   The "char_set" field values reflect printer Option 001, 96/128-character set.
   The "not_ready" field indicates whether a paper fault sets a separate
   not-ready status or simply takes the printer offline.  The "autoprints" field
   indicates whether the printer automatically prints the buffer contents when
   another character is added to a full buffer, or whether the additional
   character is discarded.  The "fault_at_eol" field indicates whether a paper
   fault is reported at the end of any line or only at the top of the next form.
*/

typedef struct {
    uint32  line_length;                        /* the maximum number of print positions */
    uint32  char_set;                           /* the size of the character set */
    uint32  vfu_channels;                       /* the number of VFU channels */
    t_bool  not_ready;                          /* TRUE if the printer reports a separate not ready status */
    t_bool  overprints;                         /* TRUE if the printer supports overprinting */
    t_bool  autoprints;                         /* TRUE if the printer automatically prints on buffer overflow */
    t_bool  fault_at_eol;                       /* TRUE if a paper fault is reported at the end of any line */
    } PRINTER_PROPS;

static const PRINTER_PROPS print_props [] = {   /* printer properties, indexed by PRINTER_TYPE */
/*     line    char    VFU      not     over    auto   fault  */
/*    length   set   channels  ready   prints  prints  at EOL */
/*    ------  -----  --------  ------  ------  ------  ------ */
    {  132,   128,      8,     FALSE,  FALSE,  TRUE,   FALSE  },    /* HP_2607 */
    {  132,    96,     12,     TRUE,   TRUE,   FALSE,  TRUE   },    /* HP_2611 */
    {  136,    96,     12,     TRUE,   TRUE,   FALSE,  TRUE   },    /* HP_2613 */
    {  136,    96,     12,     TRUE,   TRUE,   FALSE,  TRUE   },    /* HP_2617 */
    {  132,    96,     12,     TRUE,   TRUE,   FALSE,  TRUE   },    /* HP_2618 */
    {  132,    96,     12,     TRUE,   TRUE,   FALSE,  TRUE   }     /* HP_2619 */
    };


/* Delay properties table.

   To support the realistic timing mode, the delay properties table contains
   timing specifications for the supported printers.  The times represent the
   delays for mechanical and electronic operations.  Delay values are in event
   tick counts; macros are used to convert from times to ticks.


   Implementation notes:

    1. Although all of the printers operate more slowly with a 96/128-character
       set installed than with a 64-character set, the times reflect the smaller
       set size.  Also, some models provide different print rates, depending on
       how many and/or which characters are printed.  These variations are not
       simulated.
*/

typedef struct {
    int32  buffer_load;                         /* per-character transfer time */
    int32  print;                               /* print time */
    int32  advance;                             /* paper advance time per line */
    } DELAY_PROPS;

static const DELAY_PROPS real_times [] = {      /* real-time delays, indexed by PRINTER_TYPE */
   /*  buffer                paper   */
   /*   load      print     advance  */
   /* ---------  --------  --------- */
    { uS (12.6), mS (260), mS (40.1) },         /* HP_2607  200 lines per minute */
    { uS (1.75), mS ( 92), mS (4.17) },         /* HP_2611  600 lines per minute */
    { uS (1.75), mS (183), mS (8.33) },         /* HP_2613  300 lines per minute */
    { uS (1.75), mS ( 86), mS (6.67) },         /* HP_2617  600 lines per minute */
    { uS (1.75), mS ( 38), mS (4.76) },         /* HP_2618 1250 lines per minute */
    { uS (1.75), mS ( 55), mS (4.17) }          /* HP_2619 1000 lines per minute */
    };

static const DELAY_PROPS initial_times = {      /* initial optimized times */
   /*  buffer                paper   */
   /*   load      print     advance  */
   /* ---------  --------  --------- */
      uS (1.00), mS (  1), uS (  50)            /* FASTTIME delays */
    };


/* Bus controller commands */

typedef enum {                                  /* bus controller commands */
    Invalid = 0,                                /*   invalid command (default for reset) */
    Device_Clear,                               /*   PCG x4 */
    Print_Data,                                 /*   MLA 00 */
    Set_Escape,                                 /*   MLA 02 */
    Amigo_Clear,                                /*   MLA 10 */
    Write_Loopback,                             /*   MLA 1E */
    Initiate_Self_Test,                         /*   MLA 1F */

    Read_Status,                                /*   MTA 0E */
    Device_Specified_Jump,                      /*   MTA 10 */
    Read_Loopback,                              /*   MTA 1E */
    Read_Self_Test,                             /*   MLA 1F */
    Amigo_Identify                              /*   UNT MSA */
    } CN_COMMAND;

/* Bus controller command names */

static const char * const command_names [] = {  /* command names, indexed by CN_COMMAND */
    "invalid command",
    "Device Clear",
    "Print Data",
    "Set Escape",
    "Amigo Clear",
    "Write Loopback",
    "Initiate Self-Test",
    "Read Status",
    "Device Specified Jump",
    "Read Loopback",
    "Read Self-Test",
    "Amigo Identify"
    };


/* Bus controller state declarations.

   The bus controller states determine the actions performed by the bus event
   service routine.  All commands begin and end in the Command_Ready state.  The
   intermediate state transitions for each command are as follows:

     Command                 State Transitions
     ----------------------  ------------------------------------
     Universal Device Clear  Completion_Wait
     Selected Device Clear   Completion_Wait
     Print Data              Execution_Receive -> Completion_Wait
     Set Escape              Parameter_Wait    -> Completion_Wait
     Amigo Clear             Parameter_Wait    -> Completion_Wait
     Write Loopback          Command_Receive   -> Completion_Wait
     Initiate Self Test      Parameter_Wait    -> Completion_Wait
     Read Status             Command_Send      -> Completion_Wait
     Device Specified Jump   Command_Send      -> Completion_Wait
     Read Loopback           Command_Send      -> Completion_Wait
     Read Self Test          Command_Send      -> Completion_Wait
     Amigo Identify          Command_Send      -> Completion_Wait
*/

typedef enum {                                  /* translator state */
    Command_Ready = 0,                          /*   waiting for a command secondary (default for reset) */
    Parameter_Wait,                             /*   waiting for a parameter reception */
    Command_Send,                               /*   sending command status */
    Command_Receive,                            /*   receiving command data */
    Execution_Receive,                          /*   receiving write data */
    Completion_Wait,                            /*   waiting for command completion */
    Error_Source,                               /*   sending bytes for error recovery */
    Error_Sink                                  /*   receiving bytes for error recovery */
    } CN_STATE;

/* Bus controller state names */

static const char * const state_names [] = {    /* state names, indexed by CN_STATE */
    "Command Ready",
    "Parameter Wait",
    "Command Send",
    "Command Receive",
    "Execution Receive",
    "Completion Wait",
    "Error Source",
    "Error Sink"
    };


/* Bus controller command properties */

typedef struct {                                /* command properties */
    CN_STATE  next_state;                       /*   the next command state */
    uint32    parameter_count;                  /*   the number of command parameters expected */
    t_bool    use_controller;                   /*   TRUE if the command schedules the bus controller unit */
    t_bool    reported;                         /*   TRUE if the command provides a custom trace report */
    } CMD_PROPERTIES;

static const CMD_PROPERTIES cmd_props [] = {            /* command properties, indexed by CN_COMMAND */
/*                        Param    Use    Custom */
/*    Next State          Count   Cntlr   Report */
/*    ------------------  ------  ------  ------ */
    { Command_Ready,         0,   FALSE,  TRUE   },     /*   invalid command */
    { Completion_Wait,       0,   FALSE,  TRUE   },     /*   Device_Clear */
    { Execution_Receive,     0,   FALSE,  FALSE  },     /*   Print_Data */
    { Parameter_Wait,        1,   TRUE,   TRUE   },     /*   Set_Escape */
    { Parameter_Wait,        1,   FALSE,  TRUE   },     /*   Amigo_Clear */
    { Command_Receive,     256,   TRUE,   FALSE  },     /*   Write_Loopback */
    { Parameter_Wait,        1,   TRUE,   TRUE   },     /*   Initiate_Self_Test */
    { Command_Send,          1,   TRUE,   FALSE  },     /*   Read_Status */
    { Command_Send,          1,   TRUE,   TRUE   },     /*   Device_Specified_Jump */
    { Command_Send,          0,   TRUE,   FALSE  },     /*   Read_Loopback */
    { Command_Send,          1,   TRUE,   FALSE  },     /*   Read_Self_Test */
    { Command_Send,          2,   TRUE,   TRUE   }      /*   Amigo_Identify */
    };


/* Bus controller delay properties */

typedef enum {                                  /* bus controller delay reasons */
    Activation_Delay = 0,                       /*   the delay before command activation begins */
    Status_Delay,                               /*   the delay between returned status bytes */
    Data_Delay,                                 /*   the delay between returned data bytes */
    Load_Delay,                                 /*   the calculated delay to load a character into the printer */
    Print_Delay                                 /*   the calculated delay to print and advance the paper */
    } DELAY;

typedef struct {                                /* bus delay structure */
    const char  *name;                          /*   delay name */
    int32       ticks;                          /*   delay in event ticks */
    } BUS_DELAY;

static const BUS_DELAY bus_delays [] = {        /* bus delays, indexed by DELAY */
    { "activation", uS (4.0) },
    { "status",     uS (2.0) },
    { "data",       uS (1.2) },
    { "loading",    uS (0.0) },                 /* the delay in the "wait" field" is calculated */
    { "printing",   uS (0.0) }                  /* the delay in the "wait" field" is calculated */
    };


/* Escape parser states.

   The escape sequence parser decodes received character strings of the form:

     ESC & l nnn V

   ...where "nnn" is a sequence of decimal digits, optionally with leading or
   trailing blanks.  The sequence ends with an uppercase letter, either "U" or
   "V".  The parser checks for the tokens in sequence; if an invalid token is
   detected, error recovery consists of absorbing characters until an uppercase
   alphabetic character is encountered.

   When escape sequence recognition is enabled, the characters forming the
   sequence are absorbed internally.  When recognition is disabled, all
   characters are passed to the print buffer.
*/

typedef enum {                                  /* escape sequence parser states */
    Match_Disabled = 0,                         /*   parsing is disabled (default for reset) */
    Match_ESC,                                  /*   seeking an ESC character */
    Match_Amp,                                  /*   seeking an ampersand character */
    Match_L,                                    /*   seeking a lowercase "L" character */
    Match_Number,                               /*   seeking numeric digits and a trailing "U" or "V" */
    Match_Error                                 /*   discarding until a trailing uppercase character */
    } PARSER_STATE;


/* Device-Specified Jump states.

   The 26069A service manual does not describe how the translator assigns DSJ
   values.  The HIOLPRT2 driver expects the DSJ return status to range between 0
   and 3, with 0 representing "OK" status.  However, all three error status
   returns are lumped together and result in a Read Status command, followed by
   a channel program halt.  Looking at other printers does not provide a clue as
   to how the DSJ value is assigned.

   DSJ values are intended to provide a "quick status" without having to issue a
   Read Status and then decode the status bits.  As it doesn't actually matter
   to the driver how the errors are assigned, we elect to assign them as
   follows:

     0 = Printer is ready to accept data
     1 = Printer error (not ONLINE, not READY)
     2 = Power fail
     3 = Bus error (protocol, parity)

   These Read Status bits produce the indicated DSJ values:

     Status Bit        DSJ
     ----------------  ---
     (not) Online       1
     I/O buffer ready   -
     Protocol error     3
     Bus parity error   3
     Paper out          1
     Power fail         2
*/

typedef enum {                                  /* DSJ state values */
    DSJ_No_Error    = 0,                        /*   the printer is ready */
    DSJ_Print_Error = 1,                        /*   a printer error has occurred */
    DSJ_Power_Error = 2,                        /*   a power failure has occurred */
    DSJ_Bus_Error   = 3                         /*   a bus error has occurred */
    } DSJ_STATE;


/* Status byte accessors.

       7   6 | 5   4   3 | 2   1   0
     +---+---+---+---+---+---+---+---+
     | L | B | - | E | P | - | O | F |
     +---+---+---+---+---+---+---+---+

   Where:

     L = Online
     B = I/O buffer ready
     E = Protocol error
     P = Bus parity error
     O = Paper out
     F = Power fail

   The Online, I/O Buffer Ready, and Paper Out status bits are set or cleared by
   the occurrence of the relevant conditions.  The Protocol Error, Bus Parity
   Error, and Power Fail bits are set when the conditions are noted and are
   cleared by a Read Status or Device Clear command.


   Implementation notes:

    1. The translator does not pass the CH9 and CH12 signals from the printer
       VFU to the status byte.  However, we define the corresponding status
       symbols here to preserve commonality with the LP device code.
*/

#define ST_ONLINE           BIT (7)             /* (L) Online */
#define ST_BUFFER_READY     BIT (6)             /* (B) I/O buffer ready */
#define ST_PROTOCOL_ERROR   BIT (4)             /* (E) Protocol error */
#define ST_PARITY_ERROR     BIT (3)             /* (P) Bus parity error */
#define ST_PAPER_OUT        BIT (1)             /* (O) Paper out */
#define ST_POWER_FAIL       BIT (0)             /* (F) Power fail */

#define ST_DYNAMIC          (ST_ONLINE | ST_BUFFER_READY | ST_PAPER_OUT)
#define ST_STATIC           (ST_PROTOCOL_ERROR | ST_PARITY_ERROR | ST_POWER_FAIL)
#define ST_BUS_ERROR        (ST_PROTOCOL_ERROR | ST_PARITY_ERROR)

#define ST_VFU_9            0                   /* VFU channel 9 is punched */
#define ST_VFU_12           0                   /* VFU channel 12 is punched */

/* Status byte formatter declarations */

static const BITSET_NAME status_names [] = {    /* status byte names */
    "\1online\0offline",                        /*   bit  7 */
    "buffer ready",                             /*   bit  6 */
    NULL,                                       /*   bit  5 */
    "protocol error",                           /*   bit  4 */
    "bus parity error",                         /*   bit  3 */
    NULL,                                       /*   bit  2 */
    "paper out",                                /*   bit  1 */
    "power fail"                                /*   bit  0 */
    };

static const BITSET_FORMAT status_format =      /* names, offset, direction, alternates, bar */
    { FMT_INIT (status_names, 0, msb_first, has_alt, no_bar) };


/* HP-IB translator state declarations.

   The translator state structure contains the state of the HP-IB, the bus
   controller state (including the bus buffers, associated counts and indices,
   and the escape sequence parser state), printer state (including the print
   buffer), printer VFU, and the operation delay times.


   Implementation notes:

    1. The bits of the bus address bitmaps are in C order, i.e., bit 0 is the
       LSB.  The bits of the parallel poll response set are in HP-IB order,
       i.e., the bit corresponding to bus address 0 is the MSB.

    2. The talker field contains the bit corresponding to the talker address.
       This is analogous to the listeners field, which contains the bits
       corresponding to the addresses of the listeners.  Untalk and Unlisten are
       represented by bit 31 in the fields.

    3. The bus data buffers consist of eight 128-byte arrays.  The corresponding
       buffer counts reflect the current occupancy of each buffer.  An empty
       (free) buffer is indicated by a count of zero.

    4. The parameter field is used to hold the 256 data bytes received for a
       Write Loopback command plus the DSJ value, as DSJ is executed between the
       Write Loopback and Read Loopback commands and so must preserve the
       loopback data.  The size is rounded up to end on a 32-bit boundary.

       The alternative of using two of the eight bus buffers to hold the
       loopback data was not chosen, as there is no indication in the service
       manual of what consequences this might entail (e.g., what would happen if
       the buffers were occupied with unprinted data).

    5. The "overprint_char" field is defined as a "uint32" value to avoid field
       alignment issues within the structure.  It holds an 8-bit character.

    6. The "punched_char" and "unpunched_char" fields are defined as "int32"
       values to be compatible with the parameter supplied to the "fputc"
       routine.

    7. The "dlyptr" field points at either the "fast_times" field or at the
       structure element within the "real_times" array corresponding to the
       configured printer model, depending on whether FASTTIME or REALTIME mode,
       respectively, is selected.
*/

typedef struct {                                /* channel state */
    BUS_CABLE          bus;                     /*   the current HP-IB state */
    uint32             present;                 /*   the bus address bitmap of devices present on the bus */
    uint32             parity;                  /*   the bus address bitmap of devices checking parity */
    uint32             listeners;               /*   the bus address bitmap of the listeners */
    uint32             talker;                  /*   the bus address bit of the talker */
    t_bool             poll_active;             /*   TRUE if a parallel poll is in progress */
    BUS_DATA           response_set;            /*   the HP-IB address bitmap of parallel poll responses */

    CN_STATE           state;                   /*   the bus controller state */
    CN_COMMAND         command;                 /*   the bus controller command */
    t_bool             active;                  /*   TRUE if a bus transaction is active */
    uint8              parameter [260];         /*   the command parameters */
    uint32             param_index;             /*   the command parameters index */
    uint32             parameter_count;         /*   the count of command parameters */
    DSJ_STATE          dsj;                     /*   the Device Specified Jump state */
    uint8              buffer [8] [128];        /*   the bus data buffers */
    uint32             count [8];               /*   the bus data buffer counts */
    uint32             fill;                    /*   the current filling buffer number */
    uint32             findex;                  /*   the current filling buffer index */
    uint32             empty;                   /*   the current emptying buffer number */
    uint32             eindex;                  /*   the current emptying buffer index */
    PARSER_STATE       parser;                  /*   the current escape parser state */
    uint32             value;                   /*   the current escape parameter value */

    uint8              status;                  /*   the line printer status */
    t_bool             paper_fault;             /*   TRUE if the printer is out of paper */
    t_bool             tape_fault;              /*   TRUE if there is no punch in a commanded VFU channel */
    t_bool             offline_pending;         /*   TRUE if an offline request is waiting for the printer to finish */
    uint32             current_line;            /*   the current form line */
    uint32             form_length;             /*   the number of lines on the form */
    uint8              printer [PB_SIZE];       /*   the printer character and paper advance buffer */
    uint32             pindex;                  /*   the current print buffer index */
    uint32             overprint_char;          /*   the character to use if overprinted */
    uint32             ovp_index;               /*   the current overprint buffer index */

    uint32             VFU [VFU_SIZE];          /*   the vertical format unit tape content */
    char               vfu_title [LINE_SIZE];   /*   the descriptive title of the tape currently in the VFU */
    int32              punched_char;            /*   the character to display if VFU channel is punched */
    int32              unpunched_char;          /*   the character to display if VFU channel is not punched */

    DELAY_PROPS        fast_times;              /*   the optimized printer times, indexed by DELAY */
    const DELAY_PROPS  *dlyptr;                 /*   the pointer to the current printer times */
    } LPA_STATE;


/* Multiple instance structure declaration */

typedef struct {                                /* LPA instance structure */
    DEVICE     *dptr;                           /*   a pointer to the device instance */
    LPA_STATE  *sptr;                           /*   a pointer to the translator state instance */
    } INSTANCE;


/* HP-IB translator local state declarations */

static LPA_STATE lpa_state;


/* HP-IB translator local SCP support routine declarations */

static t_stat lp_service     (UNIT *uptr);
static t_stat bus_service    (UNIT *uptr);

static t_stat lpa_reset      (DEVICE *dptr);
static t_bool lpa_initialize (DEVICE *dptr, DIB *dibptr);
static uint32 lpa_adjust     (DEVICE **dvptr, int32 delta);
static t_stat lpa_attach     (UNIT   *uptr, char *cptr);
static t_stat lpa_detach     (UNIT   *uptr);

static t_stat set_model      (UNIT *uptr, int32 value, char  *cptr, void *desc);
static t_stat set_on_offline (UNIT *uptr, int32 value, char  *cptr, void *desc);
static t_stat set_vfu        (UNIT *uptr, int32 value, char  *cptr, void *desc);
static t_stat show_vfu       (FILE *st,   UNIT  *uptr, int32 value, void *desc);


/* HP-IB translator local bus routine declarations */

static REPORTER lpa_report;
static ACCEPTOR lpa_hpib_accept;


/* HP-IB translator local utility routine declarations */

static void   clear_controller  (UNIT   *uptr, LPA_STATE *lpa);
static void   process_secondary (uint32 lpa_id, uint32 address, BUS_CABLE bus);
static void   process_data      (uint32 lpa_id, uint32 address, BUS_CABLE bus);
static void   execute_command   (uint32 lpa_id, uint32 address, BUS_CABLE bus);
static void   end_transaction   (uint32 lpa_id, BUS_DATA bus_group);
static void   set_response      (uint32 lpa_id, uint32 address, FLIP_FLOP response);
static t_stat activate_unit     (uint32 lpa_id, UNIT *uptr, DELAY delay);
static t_bool set_alarm         (uint32 lpa_id, UNIT *uptr, PRINTER_TYPE model);
static t_bool set_locality      (uint32 lpa_id, UNIT *uptr, LOCALITY printer_state);
static t_stat load_vfu          (uint32 lpa_id, UNIT *uptr, FILE *vf);
static int32  read_vfu_line     (FILE   *vf, char *line, uint32 size);
static void   report_error      (FILE   *stream);
static void   trace_state       (DEVICE *dptr, uint32 address, CN_STATE entry_state, CN_STATE exit_state);


/* HP-IB registration declaration */

static const BUS_REGISTER lpa_registration = {  /* HP-IB registration */
    LPA_Index,                                  /*   the dispatcher table index */
    { lpa_report,                               /*   a pointer to the protocol reporting routine */
      lpa_hpib_accept }                         /*   a pointer to the bus acceptor routine */
    };


/* HP-IB translator SCP data structure declarations */


/* Device information block */

static DIB lpa_dib = {                          /* LPA Device Information Block */
    NULL,                                       /*   the device's IOP interface function pointer */
    lpa_initialize,                             /*   the execution initialization function pointer */
    DEVNO_UNUSED,                               /*   the device number */
    SRNO_UNUSED,                                /*   the service request number */
    INTPRI_UNUSED,                              /*   the interrupt priority */
    INTMASK_UNUSED,                             /*   the interrupt mask */
    0,                                          /*   the card index */
    CLEAR,                                      /*   the interrupt request flip-flop state */
    CLEAR,                                      /*   the interrupt active flip-flop state */
    FALSE,                                      /*   TRUE if channel service has been requested */
    NULL,                                       /*   the device's IMB interface function pointer */
    CHAN_UNUSED,                                /*   the IMB channel number */
    SLOT_UNUSED,                                /*   the IMB slot number */
    CPU_IS_IMB                                  /*   the supported system configurations */
    };


/* Unit list.

   Two SIMH units are used by the translator.  The first unit corresponds to the
   connected line printer and is responsible for loading characters into the
   printer buffer, printing them, and advancing the paper.  Attaching a file to
   this unit corresponds to loading paper into the printer.  The second unit
   corresponds to the bus controller in the translator; it is responsible for
   interacting with the HP-IB and transferring characters from the bus to the
   eight bus buffers.  Neither unit may be disabled, although the LPA device
   itself can be.
*/

#define UNIT_FLAGS          (UNIT_2617 | UNIT_OFFLINE | UNIT_EXPAND | UNIT_ATTABLE)

static UNIT lpa_unit [UNIT_COUNT] = {
    { UDATA (lp_service,  UNIT_FLAGS, 0) },     /* the line printer unit */
    { UDATA (bus_service, UNIT_DIS,   0) }      /* the bus controller unit */
    };

static char *unit_names [] = {                  /* the unit names, indexed by unit number */
    "printer",
    "bus"
    };


/* Register list.

   We provide register access to the optimized timing table to allow the user to
   alter these if desired.  The HP-IB controller hardware does not have
   externally accessible registers, but we provide a few "pseudo-registers" to
   provide a view into the controller state.
*/

static REG lpa_reg [] = {
/*    Macro   Name     Location                          Radix  Width  Offset    Depth          Flags       */
/*    ------  -------  --------------------------------  -----  -----  ------  ----------  ---------------- */
    { YRDATA (DEVS,    lpa_state.present,                         8),                      PV_RZRO | REG_RO },
    { FLDATA (XFER,    lpa_state.active,                                 0),               PV_RZRO | REG_RO },
    { FLDATA (PP,      lpa_state.poll_active,                            0),               PV_LEFT | REG_RO },
    { YRDATA (PPR,     lpa_state.response_set,                    8),                      PV_RZRO          },

    { DRDATA (BSTATE,  lpa_state.state,                           4),                      PV_LEFT          },
    { DRDATA (BCMD,    lpa_state.command,                         4),                      PV_LEFT          },
    { DRDATA (DSJ,     lpa_state.dsj,                             2),                      PV_LEFT          },
    { BRDATA (COUNT,   lpa_state.count,                   10,     8,               8),     PV_LEFT          },
    { DRDATA (PARSER,  lpa_state.parser,                          3),                      PV_LEFT          },

    { YRDATA (STATUS,  lpa_state.status,                          8),                      PV_RZRO          },

    { FLDATA (PFAULT,  lpa_state.paper_fault,                            0)                                 },
    { FLDATA (TFAULT,  lpa_state.tape_fault,                             0)                                 },
    { FLDATA (OLPEND,  lpa_state.offline_pending,                        0)                                 },

    { DRDATA (PRLINE,  lpa_state.current_line,                    8),                      PV_LEFT          },
    { BRDATA (PRTBUF,  lpa_state.printer,                  8,     8,           PB_SIZE),   PV_RZRO | REG_A  },
    { DRDATA (BUFIDX,  lpa_state.pindex,                          8),                      PV_LEFT          },
    { ORDATA (OVPCHR,  lpa_state.overprint_char,                  8),                      PV_RZRO | REG_A  },

    { DRDATA (FORMLN,  lpa_state.form_length,                     8),                      PV_LEFT | REG_RO },
    { ORDATA (PUNCHR,  lpa_state.punched_char,                    8),                      PV_RZRO | REG_A  },
    { ORDATA (UNPCHR,  lpa_state.unpunched_char,                  8),                      PV_RZRO | REG_A  },

    { DRDATA (BTIME,   lpa_state.fast_times.buffer_load,         24),                      PV_LEFT | REG_NZ },
    { DRDATA (PTIME,   lpa_state.fast_times.print,               24),                      PV_LEFT | REG_NZ },
    { DRDATA (STIME,   lpa_state.fast_times.advance,             24),                      PV_LEFT | REG_NZ },

    { SVDATA (STATE,   lpa_state)                                                                           },

    { NULL }
    };


/* Modifier list */

#define ETAB_DEVICE         (MTAB_EDV)
#define ETAB_CHANNEL        (MTAB_EDV | MTAB_ERX (10) | MTAB_EMN (2))
#define ETAB_ADDRESS        (MTAB_EDV | MTAB_ERX (10))

static MTAB lpa_mod [] = {

/*    Mask Value     Match/Max      Print String         Match String  Valid  Display  Descr  4.x    Expansion Flags */
/*    -------------  -------------  -------------------  ------------  -----  -------  -----  -----  --------------- */
    { DEV_REALTIME,  DEV_REALTIME,  "realistic timing",  "REALTIME",   NULL,  NULL,    NULL,  NULL,  ETAB_DEVICE     },
    { DEV_REALTIME,       0,        "fast timing",       "FASTTIME",   NULL,  NULL,    NULL,  NULL,  ETAB_DEVICE     },
    { DEV_CHANNEL,       15,        "channel ",          "CHANNEL",    NULL,  NULL,    NULL,  NULL,  ETAB_CHANNEL    },
    { DEV_ADDRESS,        7,        "bus=",              "BUS",        NULL,  NULL,    NULL,  NULL,  ETAB_ADDRESS    },

/*    Mask Value     Match Value    Print String        Match String  Validation        Display  Descriptor */
/*    -------------  -------------  ------------------  ------------  ----------------  -------  ---------- */
/*  { UNIT_MODEL,    UNIT_2607,     "2607",             "2607",       &set_model,       NULL,    NULL       }, */
    { UNIT_MODEL,    UNIT_2611,     "2611",             "2611",       &set_model,       NULL,    NULL       },
    { UNIT_MODEL,    UNIT_2613,     "2613",             "2613",       &set_model,       NULL,    NULL       },
    { UNIT_MODEL,    UNIT_2617,     "2617",             "2617",       &set_model,       NULL,    NULL       },
    { UNIT_MODEL,    UNIT_2618,     "2618",             "2618",       &set_model,       NULL,    NULL       },
    { UNIT_MODEL,    UNIT_2619,     "2619",             "2619",       &set_model,       NULL,    NULL       },

    { UNIT_OFFLINE,  UNIT_OFFLINE,  "offline",          "OFFLINE",    &set_on_offline,  NULL,    NULL       },
    { UNIT_OFFLINE,  0,             "online",           "ONLINE",     &set_on_offline,  NULL,    NULL,      },

    { UNIT_EXPAND,   UNIT_EXPAND,   "expanded output",  "EXPAND",     NULL,             NULL,    NULL       },
    { UNIT_EXPAND,   0,             "compact output",   "COMPACT",    NULL,             NULL,    NULL,      },

/*    Entry Flags           Value  Print String  Match String  Validation   Display    Descriptor */
/*    --------------------  -----  ------------  ------------  ----------  ----------  ---------- */
    { MTAB_XDV | MTAB_NMO,    1,   "VFU",        NULL,         NULL,       &show_vfu,  NULL       },
    { MTAB_XDV | MTAB_NC,     0,   "VFU",        "VFU",        &set_vfu,   &show_vfu,  NULL       },

    { 0 }
    };


/* Debugging trace list */

static DEBTAB lpa_deb [] = {                    /* debugging trace list */
    { "CMD",   TRACE_CMD   },                   /*   trace printer and controller commands */
    { "INCO",  TRACE_INCO  },                   /*   trace printer and controller command initiations and completions */
    { "STATE", TRACE_STATE },                   /*   trace controller state changes */
    { "SERV",  TRACE_SERV  },                   /*   trace unit service scheduling calls and entries */
    { "XFER",  TRACE_XFER  },                   /*   trace print data transmissions */
    { "BUF",   TRACE_BUFFR },                   /*   trace data buffer filling and emptying */
    { NULL,    0           }
    };


/* Device descriptor */

#define DEVICE_FLAGS        (TO_CHANNEL (11) | TO_ADDRESS (4))

DEVICE lpa_dev = {
    "LPA",                                      /* device name */
    lpa_unit,                                   /* unit array */
    lpa_reg,                                    /* register array */
    lpa_mod,                                    /* modifier array */
    UNIT_COUNT,                                 /* number of units */
    10,                                         /* address radix */
    32,                                         /* address width = 4 GB */
    1,                                          /* address increment */
    8,                                          /* data radix */
    8,                                          /* data width */
    NULL,                                       /* examine routine */
    NULL,                                       /* deposit routine */
    &lpa_reset,                                 /* reset routine */
    NULL,                                       /* boot routine */
    &lpa_attach,                                /* attach routine */
    &lpa_detach,                                /* detach routine */
    &lpa_dib,                                   /* device information block pointer */
    DEVICE_FLAGS | DEV_DISABLE | DEV_DEBUG,     /* device flags */
    0,                                          /* debug control flags */
    lpa_deb,                                    /* debug flag name table */
    NULL,                                       /* memory size change routine */
    NULL,                                       /* logical device name */
    NULL,                                       /* (4.0 dummy) help routine */
    NULL,                                       /* (4.0 dummy) help attach routine */
    NULL,                                       /* (4.0 dummy) help context */
    NULL,                                       /* (4.0 dummy) description */
    lpa_adjust                                  /* the device adjustment function pointer */
    };


/* HP-IB translator local allocation state declarations */

static uint32 lpa_count = 1;                    /* the current count of LPA device instances */

static INSTANCE lpas [MAX_LPA_COUNT] = {        /* LPA bus instances, indexed by instance ID */
    { &lpa_dev, &lpa_state }                    /*   instance 0 (original) */
    };



/***********************************************
 *                                             *
 * HP-IB translator local SCP support routines *
 *                                             *
 ***********************************************/


/* Adjust the device instances.

   This routine is called by SCP to adjust device instances that are newly
   allocated or pending deletion.  On entry, "dvptr" points at the "sim_devices"
   element that contains a pointer to the first of potentially several device
   instances to be adjusted, and "delta" contains the number of instances just
   created (if positive) or to be deleted (if negative).

   When the "dvptr" parameter is NULL, SCP expects the return value to be
   formatted as follows:

      31  30  ... 25  24 |23  22  ..  17  16 |15  14 13  ...  2   1   0
     +---+---+---+---+---+---+---+---+---+---+---+--+---+---+---+---+---+
     | clone identifier  |  instance limit   |     sim_devices size     |
     +---+---+---+---+---+---+---+---+---+---+---+--+---+---+---+---+---+

   The fields have the following meanings:

     sim_devices size      The total number of elements defined in the
                           "sim_devices" array, including the NULL element at
                           the end.  The difference between this number and the
                           number of initialized elements gives the number of
                           new device instances that can be added by the user.

     instance limit        The maximum number of instances allowed.  The count
                           includes the original instance.

     clone identifier      A character used to produce unique names for new
                           device instances.  The ID will be appended to the
                           original device name and incremented by the device
                           instance number.

   When the adjust function is called with a non-NULL pointer parameter, the
   routine performs the actions are necessary to prepare the indicated device
   instances for use or destruction.  The return value is either SCPE_OK if all
   adjustments succeeded, or an SCPE error code indicating the nature of the
   failure.  In the latter case, any allocations made by the routine prior to
   the failure must be freed before returning.  If this routine returns success,
   the device's "reset" routine, if defined, will be called with the -P
   (power-on) switch to complete initialization of the new device.


   Implementation notes:

    1. The deallocation part of the routine is entered either for a deletion
       request or for an addition request when a required allocation fails.

    2. Passing NULL to the "free" routine is allowed by the C standard, so we do
       not need to check for this condition before calling the routine.
*/

static uint32 lpa_adjust (DEVICE **dvptr, int32 delta)
{
LPA_STATE *nsptr;
DEVICE    *dptr;
char      *nptr;
DIB       *dibptr;
REG       *rptr;
int32     count;
t_stat    status = SCPE_OK;

if (dvptr == NULL)                                      /* if this is a limits request */
    return TO_DWORD (TO_WORD ('A', MAX_LPA_COUNT),      /*   then return the clone ID and instance limit */
                     SIM_DEVICE_SIZE);                  /*     and device array size */

else if (delta > 0) {                                   /* otherwise if device clones are being added */
    for (count = 0; count < delta; count++) {           /*   then for each new device clone */
        dptr = dvptr [count];                           /*     get a pointer to the new device */

        status = SCPE_INCOMP;                           /* preset for name duplication error recovery */

        nptr = dptr->name;                              /* save the newly assigned name */
        dptr->name = "";                                /*   and temporarily remove it */

        if (find_dev (nptr) != NULL) {                      /* if the device name is duplicated */
            cprintf ("Duplicate device name %s\n", nptr);   /*   then report it to the user */

            dptr->name = nptr;                          /* restore the assigned name */
            break;                                      /*   and fail at this point */
            }

        else                                            /* otherwise the name is unique */
            dptr->name = nptr;                          /*   so restore it to the device */

        status = SCPE_MEM;                              /* preset for memory error recovery */

        dptr = dvptr [count];                           /* get a pointer to the new device */

        dibptr = malloc (sizeof (DIB));                 /* allocate a new DIB structure */

        if (dibptr == NULL)                             /* if the allocation failed */
            break;                                      /*   then clean up before exiting */
        else                                            /* otherwise */
            dptr->ctxt = dibptr;                        /*   save the new DIB pointer */

        memcpy (dibptr, &lpa_dib, sizeof (DIB));        /* copy the original DIB structure */
        dibptr->card_index = lpa_count + count;         /*   and set the new card index */

        nsptr = calloc (1, sizeof (LPA_STATE));         /* allocate and zero a new LPA state structure */

        if (nsptr == NULL)                              /* if the allocation failed */
            break;                                      /*   then clean up before exiting */

        else {                                          /* otherwise */
            lpas [lpa_count + count].sptr = nsptr;      /*   save the new state pointer */
            lpas [lpa_count + count].dptr = dptr;       /*     and device pointer in the table */
            }

        for (rptr = dptr->registers; rptr->name != NULL; rptr++)    /* loop through the register entries */
            rptr->loc =                                             /*   and rebase from the state structure */
              (char *) rptr->loc - (char *) &lpa_state + (char *) nsptr;

        status = SCPE_OK;                               /* this pass succeeded */
        }                                               /*   so process the next instance */

    lpa_count = lpa_count + delta;                      /* bump the count of LPA instances */
    }                                                   /*   and continue until all instances are adjusted */


if (delta < 0 || status != SCPE_OK) {                   /* if instances are being deallocated */
    if (delta < 0)                                      /*   then if existing entries are to be deleted */
        delta = - delta;                                /*     then make the deletion count positive */

    lpa_count = lpa_count - delta;                      /* drop the count of LPA instances */

    for (count = 0; count < delta; count++) {           /* for each instance to be deleted */
        dptr = dvptr [count];                           /*   get the device pointer */

        if (dptr != NULL)                               /* if it is defined */
            free (dptr->ctxt);                          /*   then free the device context */

        free (lpas [lpa_count + count].sptr);           /* free the LPA state if it was allocated */

        lpas [lpa_count + count].dptr = NULL;           /* clear the */
        lpas [lpa_count + count].sptr = NULL;           /*   table entries */
        }
    }

return (uint32) status;                                 /* return the adjustment status */
}


/* Service the line printer unit.

   The line printer service is called to load a character into the printer
   buffer or to process a format command that causes the buffered line to be
   printed with specified paper movement.

   On entry, the "empty" state field has been set to the number of the bus
   buffer currently being emptied and transferred to the printer buffer, and the
   corresponding "count" element has been set to the count of bytes in that bus
   buffer.  The service routine is rescheduled after each byte transferred to
   the print buffer and after a buffered line is printed and the paper has
   advanced.  The scheduling delay represents the time required for the physical
   movement of the print hammers and paper advance.

   If a SET LPA OFFLINE command or a DETACH LPA command simulating an
   out-of-paper condition is given, the printer will not honor the command
   immediately if data exists in the print buffer or the printer is currently
   printing a line.  In this case, the action is deferred until the service
   routine is entered to complete a print operation.  When the printer is placed
   back online, this routine is called if data is present in the bus buffer to
   resume transferring characters to the printer.

   The hardware printer uses only seven data signals, so an additional signal,
   designated PAPER INSTRUCTION, is provided to the printer to indicate that the
   data lines carry a format command rather than a print character.  In
   simulation, we use the eighth bit in each buffered byte to differentiate
   between a character (0) and a format command (1).  If there is room in the
   print buffer, the character is loaded.  If not, then depending on the model,
   the printer either discards the character or automatically prints the buffer
   contents, advances the paper one line, and stores the new character in the
   empty buffer.  If a control character is sent but the printer cannot print
   it, a space is loaded in its place.

   A format command causes the current buffer to be printed, and then the paper
   is advanced by a prescribed amount.  Two output modes are provided: compact
   and expanded.

   In compact mode, a printed line is terminated by a CR LF pair, but subsequent
   line spacing is performed by LFs alone.  Also, a top-of-form request will
   emit a FF character instead of the number of LFs required to reach the top of
   the next form, and overprinting is handled by emitting a lone CR at the end
   of the line.  This mode is used when the printer output file will be sent to
   a physical printer connected to the host.

   In expanded mode, paper advance is handled solely by emitting CR LF pairs.
   Overprinting is handled by merging characters in the buffer.  This mode is
   used where the printer output file will be saved or manipulated as a text
   file.

   The format commands recognized by the printer are:

     0 x x 0 0 0 0 -- slew 0 lines (suppress spacing) after printing
          ...
     0 x x 1 1 1 1 -- slew 15 lines after printing

   and:

     1 x x 0 0 0 0 -- slew to VFU channel 1 after printing
          ...
     1 x x 1 0 1 1 -- slew to VFU channel 12 after printing

   A command to slew to a VFU channel that is not punched or to a VFU channel
   other than those defined for the printer will cause a tape fault, and the
   printer will go offline; setting the printer back online will clear the
   fault.  Otherwise, LFs or a FF (compact mode) or CR LF pairs (expanded mode)
   will be added to the buffer to advance the paper the required number of
   lines.

   Not all printers can overprint.  A request to suppress spacing on a printer
   that cannot (e.g., the HP 2607) is treated as a request for single spacing.

   If the stream write fails, an error message is displayed on the simulation
   console, a printer alarm condition is set (which takes the printer offline),
   and SCPE_IOERR is returned to cause a simulation stop to give the user the
   opportunity to fix the problem.  Simulation may then be resumed, either with
   the printer set back online if the problem is fixed, or with the printer
   remaining offline if the problem is uncorrectable.


   Implementation notes:

    1. When a paper-out condition is detected, the 2607 printer goes offline
       only when the next top-of-form is reached.  The 2611/13/17/18/19 printers
       go offline as soon as the current line completes.

    2. Because attached files are opened in binary mode, newline translation
       (i.e., from LF to CR LF) is not performed by the host system.  Therefore,
       we write explicit CR LF pairs to end lines, even in compact mode, as
       required for fidelity to HP peripherals.  If bare LFs are used by the
       host system, the printer output file must be postprocessed to remove the
       CRs.

    3. Overprinting in expanded mode is simulated by merging the lines in the
       buffer.  A format command to suppress spacing resets the buffer index but
       saves the previous buffer length as a "high water mark" that will be
       extended if the overlaying line is longer.  This process may be repeated
       as many times as desired before issuing a format command that prints the
       buffer.

       When overlaying characters, if a space overlays a printing character, a
       printing character overlays a space, or a printing character overlays
       itself, then the printing character is retained.  Otherwise, an
       "overprint character," which defaults to DEL but can be changed by the
       user, replaces the character in the buffer.

    4. Printers that support 12-channel VFUs treat the VFU format command as
       modulo 16.  Printers that support 8-channel VFUs treat the command as
       modulo 8.

    5. Elements 1 through N of the VFU array contain the punch data for lines 1
       through N of the form.  Element 0 contains the cumulative punch data for
       all form lines, i.e., the logical OR of elements 1 through N.

    6. As a convenience to the user, the printer output file is flushed when a
       TOF operation is performed.  This permits inspection of the output file
       from the SCP command prompt while output is ongoing.

    7. The user may examine the TFAULT and PFAULT registers to determine why the
       printer went offline.

    8. The routine is not scheduled if the printer is not online, and the
       printer cannot be taken offline with data in the print buffer, so we do
       not need to check that condition here.  Also, as online implies paper
       loaded, we do not need to check that the unit is attached or that
       "uptr->fileref" is not NULL.

    9. Explicit tests for lowercase and control characters are much faster and
       are used rather than calls to "islower" and "iscntrl", which must
       consider the current locale.

   10. This routine can be entered when there is nothing to print if no Print
       Data commands follow a paper instruction.  In that case, the reentry
       delay represents the print action and paper motion time.
*/

static t_stat lp_service (UNIT *uptr)
{
const uint32       lpa_id  = uptr->Instance;            /* the instance identifier */
LPA_STATE *const   lpa     = lpas [lpa_id].sptr;        /* the pointer to the channel state */
DEVICE *const      dptr    = lpas [lpa_id].dptr;        /* the pointer to the controlling device */
const uint32       address = ADDRESS (dptr->flags);     /* the bus address of the device containing the unit */
const PRINTER_TYPE model   = MODEL (uptr->flags);       /* the printer model */
DELAY              event;
uint8              data_byte, format_byte;
uint32             channel, line_count, slew_count;

tpprintf (dptr, TRACE_SERV, "Device %u printer unit service entered\n",
          address);

if (lpa->count [lpa->empty] == 0)                           /* if the current buffer is empty */
    return SCPE_OK;                                         /*   then there is nothing to do */
else                                                        /* otherwise */
    data_byte = lpa->buffer [lpa->empty] [lpa->eindex++];   /*   get the next byte from the bus buffer */

if (data_byte & PAPER_CMD) {                            /* if this is a paper format command */
    format_byte = data_byte & FORMAT_MASK;              /*   then discard unused bits */

    tpprintf (dptr, TRACE_XFER, "Format code %03o sent to printer\n",
              format_byte);

    if (lpa->ovp_index > lpa->pindex)                   /* if the overprinted line is longer than the current line */
        lpa->pindex = lpa->ovp_index;                   /*   then extend the current buffer index */

    if (lpa->pindex > 0 && format_byte != FORMAT_SUPPRESS)  /* if printing will occur, then trace it */
        tpprintf (dptr, TRACE_CMD, "Printed %u character%s on line %u\n",
                  lpa->pindex, (lpa->pindex == 1 ? "" : "s"), lpa->current_line);

    if (format_byte == FORMAT_SUPPRESS                  /* if this is a "suppress space" request */
      && print_props [model].overprints) {              /*   and the printer is capable of overprinting */
        slew_count = 0;                                 /*     then do not slew after printing */

        if (uptr->flags & UNIT_EXPAND) {                /* if the printer is in expanded mode */
            if (lpa->pindex > lpa->ovp_index)           /*   then if the current line is longer than the overprinted line */
                lpa->ovp_index = lpa->pindex;           /*     then extend the overprinted line */

            lpa->pindex = 0;                            /* reset the buffer index to overprint the next line */
            }

        else                                            /* otherwise the printer is in compact mode */
            lpa->printer [lpa->pindex++] = CR;          /*   so overprint by emitting a CR without a LF */

        tpprintf (dptr, TRACE_CMD, "Printer commanded to suppress spacing on line %u\n",
                  lpa->current_line);
        }

    else if (format_byte & FORMAT_VFU) {                /* otherwise if this is a VFU command */
        if (print_props [model].vfu_channels == 8)      /*   then if it's an 8-channel VFU */
            format_byte &= FORMAT_VFU_8_MASK;           /*     then only three bits are significant */

        channel = VFU_CHANNEL_1 >> (format_byte - FORMAT_VFU_BIAS - 1); /* set the requested channel */

        tpprintf (dptr, TRACE_CMD, "Printer commanded to slew to VFU channel %u from line %u\n",
                  format_byte - FORMAT_VFU_BIAS, lpa->current_line);

        lpa->tape_fault = not (channel & lpa->VFU [0]); /* a tape fault occurs if there is no punch in this channel */

        slew_count = 0;                                 /* initialize the slew counter */

        do {                                            /* the VFU always slews at least one line */
            slew_count++;                               /* increment the slew counter */
            lpa->current_line++;                        /*   and the line counter */

            if (lpa->current_line > lpa->form_length)   /* if the current line is beyond the end of the form */
                lpa->current_line = 1;                  /*   then reset to the top of the next form */
            }
        while (not (lpa->tape_fault                     /* continue until a tape fault occurs */
          || channel & lpa->VFU [lpa->current_line]));  /*   or a punch is present */
        }

    else {                                              /* otherwise it must be a slew command */
        slew_count = format_byte;                       /*   so get the number of lines to slew */

        if (format_byte == FORMAT_SUPPRESS)             /* if the printer cannot overprint */
            slew_count = 1;                             /*   then the paper advances after printing */

        tpprintf (dptr, TRACE_CMD, "Printer commanded to slew %u line%s from line %u\n",
                  slew_count, (slew_count == 1 ? "" : "s"), lpa->current_line);

        lpa->current_line += slew_count;                /* move the current line */

        if (lpa->current_line > lpa->form_length)       /* if the current line is beyond the end of the form */
            lpa->current_line -= lpa->form_length;      /*   then it extends onto the next form */
        }

    if (format_byte == FORMAT_VFU_CHAN_1                /* if a TOF was requested */
      && not (uptr->flags & UNIT_EXPAND)                /*   and the printer is in compact mode */
      && slew_count > 1) {                              /*     and more than one line is needed to reach the TOF */
        if (lpa->pindex > 0) {                          /*       then if the buffer not empty */
            lpa->printer [lpa->pindex++] = CR;          /*         then print */
            lpa->printer [lpa->pindex++] = LF;          /*           the current line */
            }

        lpa->printer [lpa->pindex++] = FF;              /* emit a FF to move to the TOF */
        }

    else if (slew_count > 0) {                          /* otherwise a slew is needed */
        lpa->printer [lpa->pindex++] = CR;              /*   then emit a CR LF */
        lpa->printer [lpa->pindex++] = LF;              /*     to print the current line */

        line_count = slew_count;                        /* get the number of lines to slew */

        while (--line_count > 0) {                      /* while movement is needed */
            if (uptr->flags & UNIT_EXPAND)              /* if the printer is in expanded mode */
                lpa->printer [lpa->pindex++] = CR;      /*   then blank lines are CR LF pairs */

            lpa->printer [lpa->pindex++] = LF;          /* otherwise just LFs are used */
            }
        }

    if (lpa->pindex > 0) {                                  /* if the buffer is not empty */
        uptr->pos += sizeof (uint8)                         /*   then write the buffer to the printer file */
          * (t_addr) fwrite (lpa->printer, sizeof (uint8),  /*     and update the file position */
                             lpa->pindex, uptr->fileref);

        lpa->pindex = 0;                                /* clear the print buffer index */
        lpa->ovp_index = 0;                             /*   and any existing overprint index */

        if (lpa->paper_fault) {                         /* if an out-of-paper condition is pending */
            if (print_props [model].fault_at_eol        /*   then if the printer faults at the end of any line */
              || lpa->current_line == 1)                /*     or the printer is at the top of the form */
                return lpa_detach (uptr);               /*       then complete it now with the printer offline */
            }

        else if (lpa->tape_fault) {                     /* otherwise if a referenced VFU channel was not punched */
            tpprintf (dptr, TRACE_CMD, "Commanded VFU channel %u is not punched\n",
                      format_byte - FORMAT_VFU_BIAS);
            set_alarm (lpa_id, uptr, model);            /*   then take the printer offline */
            }

        else if (lpa->offline_pending)                  /* otherwise if a non-alarm offline request is pending */
            set_locality (lpa_id, uptr, Offline);       /*   then take the printer offline now */
        }

    if (print_props [model].vfu_channels > 8) {             /* if the printer VFU has more than 8 channels */
        if (lpa->VFU [lpa->current_line] & VFU_CHANNEL_9)   /*   then if channel 9 is punched for this line */
            lpa->status |= ST_VFU_9;                        /*     then report it in the device status */
        else                                                /*   otherwise */
            lpa->status &= ~ST_VFU_9;                       /*     clear the associated status */

        if (lpa->VFU [lpa->current_line] & VFU_CHANNEL_12)  /* if channel 12 is punched for this line */
            lpa->status |= ST_VFU_12;                       /*   then report it in the device status */
        else                                                /*   otherwise */
            lpa->status &= ~ST_VFU_12;                      /*     clear the associated status */
        }

    if (format_byte == FORMAT_VFU_CHAN_1)               /* if a TOF request was performed */
        fflush (uptr->fileref);                         /*   then flush the file buffer for inspection */

    event = Print_Delay;                                /* schedule the event */
    uptr->wait = lpa->dlyptr->print                     /*   for the printing delay */
                   + slew_count * lpa->dlyptr->advance; /*     plus the paper advance delay */

    if (slew_count > 0)                                 /* if the paper was advanced */
        tpprintf (dptr, TRACE_CMD, "Printer advanced %u line%s to line %u\n",
                  slew_count, (slew_count == 1 ? "" : "s"), lpa->current_line);
    }


else {                                                  /* otherwise it's a print character */
    data_byte = data_byte & DATA_MASK;                  /*   and only the lower 7 bits are used */

    if (data_byte > '_'                                 /* if the character is "lowercase" */
      && print_props [model].char_set == 64)            /*     but the printer doesn't support it */
        data_byte = data_byte - 040;                    /*       then shift it to "uppercase" */

    if ((data_byte < ' ' || data_byte == DEL)           /* if the character is a control character */
      && print_props [model].char_set != 128)           /*   but the printer doesn't support it */
        data_byte = ' ';                                /*     then substitute a space */

    if (lpa->pindex < print_props [model].line_length) {    /* if there is room in the buffer */
        if (lpa->ovp_index == 0                             /*   then if not overprinting */
          || lpa->pindex >= lpa->ovp_index                  /*     or past the current buffer limit */
          || lpa->printer [lpa->pindex] == ' ')             /*     or overprinting a blank */
            lpa->printer [lpa->pindex] = data_byte;         /*       then store the character */

        else if (data_byte != ' '                       /* otherwise if we're overprinting a character */
          && data_byte != lpa->printer [lpa->pindex])   /*   with a different character */
            lpa->printer [lpa->pindex] =                /*     then substitute the overprint character */
              (uint8) lpa->overprint_char;

        lpa->pindex++;                                  /* increment the buffer index */

        event = Load_Delay;                             /* schedule the event */
        uptr->wait = lpa->dlyptr->buffer_load;          /*   for the print buffer loading delay */

        tpprintf (dptr, TRACE_XFER, "Character %s sent to printer\n",
                  fmt_char (data_byte));
        }

    else if (print_props [model].autoprints) {          /* otherwise if a buffer overflow auto-prints */
        tpprintf (dptr, TRACE_CMD, "Buffer overflow printed %u characters on line %u\n",
                  lpa->pindex, lpa->current_line);

        lpa->printer [lpa->pindex++] = CR;              /*   then tie off */
        lpa->printer [lpa->pindex++] = LF;              /*     the current buffer */

        uptr->pos += sizeof (uint8)                         /* write the buffer to the printer file */
          * (t_addr) fwrite (lpa->printer, sizeof (uint8),  /*   and update the file position */
                             lpa->pindex, uptr->fileref);

        lpa->current_line = lpa->current_line + 1;      /* move the paper one line */

        if (lpa->current_line > lpa->form_length)       /* if the current line is beyond the end of the form */
            lpa->current_line = 1;                      /*   then reset to the top of the next form */

        tpprintf (dptr, TRACE_CMD, "Printer advanced 1 line to line %u\n",
                  lpa->current_line);

        lpa->ovp_index = 0;                             /* clear any accumulated overprint index */

        lpa->printer [0] = data_byte;                   /* store the overflowing character */
        lpa->pindex = 1;                                /*   in the empty buffer */

        event = Print_Delay;                            /* schedule the event */
        uptr->wait = lpa->dlyptr->print                 /*   for the printing delay */
                       + lpa->dlyptr->advance           /*     plus the paper advance delay */
                       + lpa->dlyptr->buffer_load;      /*       plus the buffer load delay */

        tpprintf (dptr, TRACE_XFER, "Character %s sent to printer\n",
                  fmt_char (data_byte));
        }

    else {                                              /* otherwise the printer does not auto-print */
        event = Load_Delay;                             /*   so discard the character and schedule the event */
        uptr->wait = lpa->dlyptr->buffer_load;          /*     for the print buffer loading delay */

        tpprintf (dptr, TRACE_CMD, "Buffer overflow discards character %s\n",
                  fmt_char (data_byte));
        }
    }

if (lpa->eindex >= lpa->count [lpa->empty]) {           /* if there are no more characters in the bus buffer */
    lpa->count [lpa->empty] = 0;                        /*   then free the current buffer */

    tpprintf (dptr, TRACE_BUFFR, "Buffer %u emptied\n", lpa->empty);

    lpa->empty = (lpa->empty + 1) % 8;                  /* move to the next bus buffer */
    lpa->eindex = 0;                                    /*   and reset the index */

    lpa->status |= ST_BUFFER_READY;                     /* set buffer-available status */

    if (lpa->status & ST_ONLINE)                        /* if the printer is still online */
        set_response (lpa_id, address, SET);            /*   then assert the poll response */
    }

if (ferror (uptr->fileref)) {                           /* if a host file system error occurred */
    report_error (uptr->fileref);                       /*   then report the error to the console */

    set_alarm (lpa_id, uptr, model);                    /* take the printer offline */
    return SCPE_IOERR;                                  /*   and stop the simulator */
    }

else                                                    /* otherwise the transfer succeeded */
    return activate_unit (lpa_id, uptr, event);         /*   so schedule the next transfer to the printer */
}


/* Service the HP-IB controller.

   This service routine is called to execute scheduled bus controller events.
   Events are scheduled to execute bus commands that do not complete
   immediately.  Examples of scheduled commands are Amigo Identify, Device
   Specified Jump, and Read and Write Loopback.  The actions to be taken upon
   service entry depend on the current state of the controller.

   The poll response is determined at command completion.  Poll response is
   asserted only when a free buffer exists, and either the printer is online or
   the DSJ value is non-zero.  The latter can occur when, e.g. ,the printer runs
   out of paper.  This condition takes the printer offline and sets DSJ = 1.  If
   the poll response were denied by the offline condition, then the status could
   not be interrogated.  However, once a Read Status command is done to obtain
   the offline status, DSJ is cleared, and the poll response is denied because
   the printer is offline.


   Implementation notes:

    1. The second identify byte must be returned to the GIC within 5
       microseconds or the channel program will fail.  This is because the CPP
       does not wait for the second byte to be returned once the first byte is
       available.  So both bytes must arrive in the GIC FIFO before the channel
       Wait instruction resumes.

    2. We must have a local "completed" status, rather than relying on
       "lpa->active" to indicate completion, because calling "hpib_source" (GIC)
       to deny NRFD calls "transfer_data" (GIC), which calls "lpa_hpib_accept"
       with the next byte, which calls "process_data", which sets "lpa->active"
       false when it receives the final byte.  But when this sequence unwinds
       back here, we find that "lpa->active" is now false for the next-to-last
       byte, so we exit the reception state without ever denying the NRFD signal
       that process_data asserts for the final byte. This results in a hung bus
       and a GIC DMA timeout.
*/

static t_stat bus_service (UNIT *uptr)
{
const uint32     lpa_id     = uptr->Instance;           /* the instance identifier */
LPA_STATE *const lpa        = lpas [lpa_id].sptr;       /* the pointer to the channel state */
DEVICE *const    dptr       = lpas [lpa_id].dptr;       /* the pointer to the controlling device */
const uint32     channel    = CHANNEL (dptr->flags);    /* the channel address of the bus controller */
const uint32     address    = ADDRESS (dptr->flags);    /* the bus address of the device containing the unit */
const BUS_CABLE  error_byte = Bus_DAV | Bus_EOI | 1;    /* the value returned for a transaction error */
BUS_CABLE        bus;
t_bool           completed   = FALSE;                   /* TRUE if the command has completed */
t_stat           result      = SCPE_OK;                 /* the result of the service call */
CN_STATE         entry_state = lpa->state;              /* the controller state at entry */

tpprintf (dptr, TRACE_SERV, "Device %u bus unit state %s service entered\n",
          address, state_names [entry_state]);

switch (entry_state) {                                  /* dispatch the controller state */

    case Command_Send:                                              /* send command data to the bus */
        if (not (lpa->bus & (Bus_ATN | Bus_NRFD))) {                /* if the bus is ready for data */
            bus = Bus_DAV | lpa->parameter [lpa->param_index++];    /*   then get the next byte to send */
            lpa->parameter_count--;                                 /*     and count it */

            if (lpa->parameter_count == 0)              /* if the transfer is complete */
                bus |= Bus_EOI;                         /*   then tag the last byte with EOI */

            lpa->bus = hpib_source (channel, address, bus); /* send the byte to the bus */

            if (lpa->parameter_count == 0)                  /* if all bytes have been sent */
                completed = TRUE;                           /*   then the command is complete */
            else                                            /* otherwise the transfer continues */
                activate_unit (lpa_id, uptr, Status_Delay); /*   so reschedule the event service */
            }
        break;


    case Command_Receive:                               /* receive command data */
        if (lpa->parameter_count == 0)                  /* if all bytes have been received */
            completed = TRUE;                           /*   then the command is complete */

        lpa->bus = hpib_source (channel, address,       /* deny NRFD to remove the handshake hold */
                                lpa->bus & ~Bus_NRFD);

        entry_state = lpa->state;                       /* pick up any state change from accepting the byte */
        break;


    case Execution_Receive:                             /* receiving data from the host */
        completed = not lpa->active;                    /* the command is complete if the transaction completed */

        lpa->bus = hpib_source (channel, address,       /* deny NRFD to remove the handshake hold */
                                lpa->bus & ~Bus_NRFD);

        entry_state = lpa->state;                        /* pick up any state change from accepting the byte */
        break;


    case Completion_Wait:                               /* wait for command completion */
        completed = TRUE;                               /* the command is now complete */
        break;


    case Error_Source:                                  /* send data after an error */
        if (not (lpa->bus & (Bus_ATN | Bus_NRFD))) {    /* if the bus is ready for data */
            lpa->bus = hpib_source (channel, address,   /*   then source a dummy byte to the bus */
                                    error_byte);

            lpa->status |= ST_PROTOCOL_ERROR;           /* register the error status */
            lpa->dsj = DSJ_Bus_Error;                   /*   and report a protocol error */
            completed = TRUE;                           /*     and terminate the command */
            }
        break;


    case Error_Sink:                                    /* absorb data after an error */
        if (not lpa->active) {                          /* if the transaction is complete */
            lpa->status |= ST_PROTOCOL_ERROR;           /*   then register the error status */
            lpa->dsj = DSJ_Bus_Error;                   /*     and report a protocol error */
            completed = TRUE;                           /*       and terminate the command */
            }

        lpa->bus = hpib_source (channel, address,       /* deny NRFD to remove the handshake hold */
                               lpa->bus & ~Bus_NRFD);
        break;


    default:                                            /* no other states schedule service */
        tpprintf (dptr, TRACE_STATE, "Device %u bus unit state %s not handled\n",
                  address, state_names [entry_state]);

        result = SCPE_IERR;                             /* signal an internal error */
        break;
    }                                                   /* end of interface state dispatch */


if (completed) {                                        /* if the command completed */
    tpprintf (dptr, TRACE_INCO, "Device %u %s completed with DSJ %u\n",
              address, command_names [lpa->command], lpa->dsj);

    lpa->command = Invalid;                             /*   then invalidate the command */
    lpa->state = Command_Ready;                         /*     and return to the ready state */
    lpa->active = FALSE;                                /*       with the bus transaction complete */

    if (lpa->status & ST_BUFFER_READY                   /* if a bus buffer is available */
      && (lpa->status & ST_ONLINE                       /*   and the printer is online */
      || lpa->dsj != DSJ_No_Error))                     /*   or an unreported error is present */
        set_response (lpa_id, address, SET);            /*     then enable the poll response */
    else                                                /* otherwise */
        set_response (lpa_id, address, CLEAR);          /*   disable the poll response */
    }

if (TRACINGP (dptr, TRACE_STATE)                            /* if tracing is active */
  && lpa->state != entry_state)                             /*   and the controller state has changed */
    trace_state (dptr, address, entry_state, lpa->state);   /*     then trace the state transition */

return result;                                          /* return the result of the service */
}


/* Device reset routine.

   This routine is called for a RESET or RESET LPA command.  With the addition
   of the -P switch, it is the simulation equivalent of the initial application
   of power to the translator PCA and the line printer.  Without the switch, the
   bus controller and printer are idled; this action has no direct hardware
   analog.  Clearing the print buffer ensures that a subsequent DETACH LPA will
   succeed.

   In addition, if a power-on reset (RESET -P) is done, the original FASTTIME
   settings are restored.
*/

static t_stat lpa_reset (DEVICE *dptr)
{
DIB *const       dibptr  = (DIB *) dptr->ctxt;          /* a pointer to the device context */
const uint32     lpa_id  = dibptr->card_index;          /* the instance ID of the protocol */
LPA_STATE *const lpa     = lpas [lpa_id].sptr;          /* the pointer to the channel state */
const uint32     address = ADDRESS (dptr->flags);       /* the bus address of the device */
t_stat           result  = SCPE_OK;

if (sim_switches & P_SWITCH) {                          /* if this is a power-on reset */
    hpib_register (lpa_registration);                   /*   then register our protocol handler */

    BUS_UPTR->Instance = lpa_id;                        /* set the instance ID */
    LP_UPTR->Instance  = lpa_id;                        /*   of the bus and printer units */

    memset (lpa, 0, sizeof (LPA_STATE));                /* clear the state structure */

    lpa->fast_times = initial_times;                    /* restore the original optimized timing values */

    if (dptr->flags & DEV_REALTIME)                         /* if the printer is in real-time mode */
        lpa->dlyptr = &real_times [MODEL (LP_UPTR->flags)]; /*   then point at the times for the current model */
    else                                                    /* otherwise */
        lpa->dlyptr = &lpa->fast_times;                     /*   point at the fast times */

    result = load_vfu (lpa_id, LP_UPTR, NULL);          /* load the standard VFU tape */

    lpa->punched_char   = 'O';                          /* restore the character displayed for a punched VFU channel */
    lpa->unpunched_char = '.';                          /*   and the character displayed for an unpunched channel */
    lpa->overprint_char = DEL;                          /*     and the character to use if overprinting */

    lpa->status = ST_POWER_FAIL;                        /* indicate power restoration */
    }

else {                                                  /* otherwise this is a normal reset */
    clear_controller (BUS_UPTR, lpa);                   /*   so clear just the bus controller */

    lpa->state   = Command_Ready;                       /* reset the controller state */
    lpa->command = Invalid;                             /*   and invalidate the current command */
    lpa->status &= ~ST_DYNAMIC;                         /*     and clear the dynamic status bits */
    }

lpa->status |= ST_BUFFER_READY;                         /* free buffers are now available */

if (LP_UPTR->flags & UNIT_ATT) {                        /* if paper is loaded */
    lpa->paper_fault = FALSE;                           /*   then there is no paper fault */

    if (LP_UPTR->flags & UNIT_OFFLINE)                  /* if the printer is offline */
        lpa->dsj = DSJ_Print_Error;                     /*   then record a printer error */

    else {                                              /* otherwise */
        lpa->status |= ST_ONLINE;                       /*   set online status */
        lpa->dsj = DSJ_No_Error;                        /*     and record an available printer */
        set_response (lpa_id, address, SET);            /*       and enable the parallel poll response */
        }
    }

else {                                                  /* otherwise the printer is out of paper */
    lpa->paper_fault = TRUE;                            /*   and a paper fault exists */
    lpa->status |= ST_PAPER_OUT;                        /*     so set paper-out status */
    lpa->dsj = DSJ_Print_Error;                         /*       and log a printer error */
    }

if (lpa->status & ST_BUS_ERROR)                         /* if a bus protocol or parity error exists */
    lpa->dsj = DSJ_Bus_Error;                           /*   then record a bus error */

else if (lpa->status & ST_POWER_FAIL) {                 /* otherwise if power has failed */
    lpa->dsj = DSJ_Power_Error;                         /*   then record a power failure */
    set_response (lpa_id, address, SET);                /*     and enable the parallel poll response */
    }

sim_cancel (BUS_UPTR);                                  /* cancel any in-progress bus controller action */
sim_cancel (LP_UPTR);                                   /*   and printer action */

lpa->pindex = lpa->ovp_index = 0;                       /* clear the print buffer indices to idle the printer */

lpa->listeners = BUS_UNADDRESSED;                       /* unaddress all current listeners */
lpa->talker    = BUS_UNADDRESSED;                       /*   and the current talker */

return result;
}


/* Run-time initialization.

   This bus routine is called during I/O initialization each time machine
   instruction execution is to begin.  The routine returns TRUE to permit
   execution to commence or FALSE to inhibit execution.  Initialization is
   optional for modules but is required in our case to set the presence bit
   corresponding to the assigned bus address.  We also ensure that any prior bus
   assignment is cleared from the set of devices checking command parity and
   responding to a parallel poll.


   Implementation notes:

    1. The parallel poll response set uses HP-IB bit order, where bit 0 is the
       MSB, while the bus device presence set uses C bit order, where bit 0 is
       the LSB.  Consequently, the response set cannot be masked by the present
       set to account for a bus address change but instead must be calculated
       separately from the bus address.
*/

static t_bool lpa_initialize (DEVICE *dptr, DIB *dibptr)
{
const uint32     lpa_id = dibptr->card_index;   /* the instance ID of the protocol */
LPA_STATE *const lpa    = lpas [lpa_id].sptr;   /* the pointer to the channel state */

lpa->present = BIT (ADDRESS (dptr->flags));             /* set the "device is present" bit */
lpa->parity &= lpa->present;                            /*   and ensure only present devices are checking parity */

lpa->response_set &= HPIB_BIT (ADDRESS (dptr->flags));  /* mask poll responses to just those devices present */

return TRUE;                                            /* initialization always succeeds */
}


/* Attach the printer image file.

   The specified file is attached to the indicated unit.  This is the simulation
   equivalent of loading paper into the printer and pressing the ONLINE button.
   The transition from offline to online asserts the parallel poll response if a
   bus buffer is available.

   A new image file may be requested by giving the "-N" switch to the ATTACH
   command.  If an existing file is specified with "-N", it will be cleared; if
   specified without "-N", printer output will be appended to the end of the
   existing file content.  In all cases, the paper is positioned at the top of
   the form.


   Implementation notes:

    1. If we are called during a RESTORE command to reattach a file previously
       attached when the simulation was SAVEd, the device status and file
       position are not altered.

    2. The pointer to the appropriate event delay times is set in case we are
       being called during a RESTORE command (the assignment is redundant
       otherwise).
*/

static t_stat lpa_attach (UNIT *uptr, char *cptr)
{
const     uint32 lpa_id = uptr->Instance;       /* the instance ID of the protocol */
DEVICE    *const dptr   = lpas [lpa_id].dptr;   /* the pointer to the controlling device */
LPA_STATE *const lpa    = lpas [lpa_id].sptr;   /* the pointer to the channel state */
t_stat           result;

result = attach_unit (uptr, cptr);                      /* attach the specified printer image file */

if (result == SCPE_OK                                   /* if the attach was successful */
  && not (sim_switches & SIM_SW_REST)) {                /*   and we are not being called during a RESTORE command */
    lpa->current_line = 1;                              /*     then reset the line counter to the top of the form */

    tpprintf (dptr, TRACE_CMD, "Printer paper loaded\n");

    set_locality (lpa_id, uptr, Online);                /* set the printer online */
    }

lpa->paper_fault = FALSE;                               /* clear any existing paper fault */

if (dptr->flags & DEV_REALTIME)                         /* if the printer is in real-time mode */
    lpa->dlyptr = &real_times [MODEL (uptr->flags)];    /*   then point at the times for the current model */
else                                                    /* otherwise */
    lpa->dlyptr = &lpa->fast_times;                     /*   point at the fast times */

return result;                                          /* return the result of the attach */
}


/* Detach the printer image file.

   The specified file is detached from the indicated unit.  This is the
   simulation equivalent of running out of paper or unloading the paper from the
   printer.  The out-of-paper condition cause a paper fault alarm, and the
   printer goes offline.  The transition from online to offline denies the
   parallel poll response.

   When the printer runs out of paper, it will not go offline until characters
   present in the buffer are printed and paper motion stops.  In addition, the
   2607 printer waits until the paper reaches the top-of-form position before
   going offline.

   In simulation, entering a DETACH LP command while the printer is busy will
   defer the file detach until print operations reach the top of the next form
   (2607) or until the current print operation completes (2611/13/17/18/19).  An
   immediate detach may be forced by adding the -F switch to the DETACH command.
   This simulates physically removing the paper from the printer and succeeds
   regardless of the current printer state.


   Implementation notes:

    1. During simulator shutdown, this routine is called for both units, not
       just the printer unit.  The printer must be detached, even if a detach
       has been deferred, to ensure that the file is closed properly.  We do
       this in response to a detach request with the SIM_SW_SHUT switch present.

    2. The DETACH ALL command will fail if any detach routine returns a status
       other than SCPE_OK.  Because a deferred detach is not fatal, we must
       return SCPE_OK, but we still want to print a warning to the user.

    3. Because the 2607 only paper faults at TOF, we must explicitly set the
       offline_pending flag, as set_alarm may not have been called.

    4. A forced detach is performed if the CPU is halted or the simulator is
       exiting, as either way the target operating system will not be sending
       any more output to the printer.
*/

static t_stat lpa_detach (UNIT *uptr)
{
const uint32       lpa_id = uptr->Instance;             /* the instance ID of the protocol */
DEVICE *const      dptr   = lpas [lpa_id].dptr;         /* the pointer to the controlling device */
LPA_STATE *const   lpa    = lpas [lpa_id].sptr;         /* the pointer to the channel state */
const PRINTER_TYPE model  = MODEL (uptr->flags);        /* the printer model number */

if (uptr->flags & UNIT_ATTABLE)                         /* if we're being called for the printer unit */
    if (not (uptr->flags & UNIT_ATT))                   /*   then if the unit is not currently attached */
        return SCPE_UNATT;                              /*     then report it */

    else {                                                      /* otherwise */
        if (sim_switches & (F_SWITCH | SIM_SW_SHUT)             /*   if this is a forced detach or shut down request */
          || not cpu_is_running ()) {                           /*     or the CPU is halted */
            if (lpa->pindex != 0) {                             /*       then if the buffer is not empty */
                if (lpa->printer [lpa->pindex - 1] != LF        /*         then if the last buffer character */
                  && lpa->printer [lpa->pindex - 1] != FF) {    /*           is not a line feed or form feed */
                    lpa->printer [lpa->pindex++] = CR;          /*             then tie off */
                    lpa->printer [lpa->pindex++] = LF;          /*               the current buffer */
                    }

                fwrite (lpa->printer, sizeof (uint8),   /* write the partial buffer to the printer file */
                        lpa->pindex, uptr->fileref);

                lpa->pindex = 0;                        /* clear the printer buffer index to permit detaching */
                }

            lpa->current_line = 1;                      /* reset the printer to TOF to enable detaching */
            sim_cancel (uptr);                          /*   and terminate any print action in progress */
            }

        if ((print_props [model].fault_at_eol           /* if the printer faults at the end of any line */
          || lpa->current_line == 1)                    /*   or the printer is at the top of the form */
          && set_alarm (lpa_id, uptr, model)) {         /*   and a paper alarm is accepted */
            lpa->paper_fault = TRUE;                    /*     then set the out-of-paper condition */

            tpprintf (dptr, TRACE_CMD, "Printer is out of paper\n");

            return detach_unit (uptr);                  /*     then detach the unit */
            }

        else {                                          /* otherwise the alarm was rejected at this time */
            lpa->paper_fault = TRUE;                    /*  then set the out-of-paper condition */
            lpa->offline_pending = TRUE;                /*    but defer the detach */

            tpprintf (dptr, TRACE_CMD, "Paper out request deferred until print completes\n");

            cprintf ("%s\n", sim_error_text (SCPE_INCOMP)); /* report that the actual detach must be deferred */
            return SCPE_OK;                                 /*   until the buffer has been printed */
            }
        }

else                                                    /* otherwise */
    return SCPE_UNATT;                                  /*   we've been called for the wrong unit */
}


/* Set the printer model.

   This validation routine is called to set the model of the printer.  The
   "value" parameter is one of the UNIT_26nn constants that indicates the new
   model.  Validation isn't necessary, except to detect a model change and alter
   the real-time delays accordingly.
*/

static t_stat set_model (UNIT *uptr, int32 value, char *cptr, void *desc)
{
const uint32     lpa_id = uptr->Instance;       /* the instance ID of the protocol */
DEVICE *const    dptr   = lpas [lpa_id].dptr;   /* the pointer to the controlling device */
LPA_STATE *const lpa    = lpas [lpa_id].sptr;   /* the pointer to the channel state */

if (dptr->flags & DEV_REALTIME)                         /* if the printer is in real-time mode */
    lpa->dlyptr = &real_times [MODEL (value)];          /*   then point at the times for the current model */
else                                                    /* otherwise */
    lpa->dlyptr = &lpa->fast_times;                     /*   point at the fast times */

return SCPE_OK;                                         /* allow the reassignment to proceed */
}


/* Set the printer online or offline.

   This validation routine is called to set the printer online or offline.  The
   "value" parameter is UNIT_OFFLINE if the printer is going offline and is zero
   if the printer is going online.  This simulates pressing the ON/OFFLINE
   button on the printer.  The unit must be attached (i.e., paper must be
   loaded), before the printer may be set online or offline.

   If the printer is being taken offline, the buffer is checked to see if any
   characters are present.  If they are, or if the printer unit is currently
   scheduled (i.e., executing a print operation), the offline request is
   deferred until printing completes, and the routine returns "Command not
   complete" status to inform the user.  Otherwise, the unit is set offline, and
   parallel poll response is denied to indicate that the printer is not ready.

   If the printer is being put online and paper is present, the unit is set
   online, and any paper or tape fault present is cleared.  If the currently
   emptying bus buffer is occupied, the printer service is scheduled to continue
   the transfer.

   As a special case, a detach (out-of-paper condition) or offline request that
   has been deferred until printing completes may be canceled by setting the
   printer online.  No other action is taken, because the printer has never
   transitioned to the offline state.


   Implementation notes:

    1. Although a deferred offline request is not fatal, we return SCPE_INCOMP
       to prevent "set_cmd" from setting the UNIT_OFFLINE bit in the unit flags
       before the printer actually goes offline.
*/

static t_stat set_on_offline (UNIT *uptr, int32 value, char *cptr, void *desc)
{
const uint32     lpa_id = uptr->Instance;       /* the instance ID of the protocol */
DEVICE *const    dptr   = lpas [lpa_id].dptr;   /* the pointer to the controlling device */
LPA_STATE *const lpa    = lpas [lpa_id].sptr;   /* the pointer to the channel state */

if (not (uptr->flags & UNIT_ATT))                       /* if the printer is detached */
    return SCPE_UNATT;                                  /*   then it can't be set online or offline */

else if (value == UNIT_ONLINE)                          /* otherwise if this is an online request */
    if (lpa->paper_fault && lpa->offline_pending) {     /*   then if an out-of-paper condition is deferred */
        lpa->paper_fault = FALSE;                       /*     then cancel the request */
        lpa->offline_pending = FALSE;                   /*       leaving the file attached */
        }

    else                                                /*   otherwise it's a normal online request */
        set_locality (lpa_id, uptr, Online);            /*     so set the printer online */

else if (set_locality (lpa_id, uptr, Offline) == FALSE) {   /* otherwise if it cannot be set offline now */
    tpprintf (dptr, TRACE_CMD, "Offline request deferred until print completes\n");
    return SCPE_INCOMP;                                     /*   then let the user know */
    }

return SCPE_OK;                                         /* return operation success */
}


/* Set the VFU tape.

   This validation routine is entered to set up the VFU on the printer.  It is
   invoked by one of two commands:

     SET LPA VFU
     SET LPA VFU=<filename>

   The first form loads the standard 66-line tape into the VFU.  The second form
   loads the VFU with the tape image specified by the filename.  The format of
   the tape image is described in the comments for the "lpa_load_vfu" routine.

   On entry, "uptr" points at the printer unit, "cptr" points to the first
   character after the "VFU" keyword, and the "value" and "desc" parameters are
   unused.  If "cptr" is NULL, then the first command form was given, and the
   "lpa_load_vfu" routine is called with a NULL file stream pointer to indicate
   that the standard VFU tape should be used.  Otherwise, the second command
   form was given, and "cptr" points to the supplied filename.  The file is
   opened, and the "lpa_load_vfu" routine is called with the stream pointer to
   load the VFU tape image contained therein.
*/

static t_stat set_vfu (UNIT *uptr, int32 value, char *cptr, void *desc)
{
const uint32 lpa_id = uptr->Instance;           /* the instance ID of the protocol */
FILE         *vfu_stream;
t_stat       result;

if (cptr == NULL)                                       /* if a VFU reset is requested */
    result = load_vfu (lpa_id, uptr, NULL);             /*   then reload the standard VFU tape */

else if (*cptr == '\0')                                 /* otherwise if the filename was omitted */
    return SCPE_MISVAL;                                 /*   then report the missing argument */

else {                                                  /* otherwise the filename was specified */
    vfu_stream = fopen (cptr, "r");                     /*   so attempt to open it */

    if (vfu_stream == NULL)                             /* if the open failed */
        return SCPE_OPENERR;                            /*   then report the error */

    result = load_vfu (lpa_id, uptr, vfu_stream);       /* load the VFU tape from the file */

    fclose (vfu_stream);                                /* close the file */
    }

return result;                                          /* return the result of the load */
}


/* Show the VFU tape.

   This display routine is called to show the content of the tape currently
   loaded in the printer's VFU.  The "value" parameter indicates how the routine
   was called.  It is 0 if a SHOW LPA command was given and 1 if a SHOW LPA VFU
   command was issued.  For the former, only the VFU title is displayed.  The
   latter displays the VFU title, followed by a header labelling each of the
   channel columns, and then one line for each line of the form consisting of
   punch and no-punch characters, according to the VFU definition.

   The output stream is passed in the "st" parameter, and the "uptr" and "desc"
   parameters are ignored.


   Implementation notes:

    1. Setting the string precision for the header lines trims them to the
       number of channels appropriate for the printer model selected.
*/

static t_stat show_vfu (FILE *st, UNIT *uptr, int32 value, void *desc)
{
static const char header_1 [] = " Ch 1 Ch 2 Ch 3 Ch 4 Ch 5 Ch 6 Ch 7 Ch 8 Ch 9 Ch10 Ch11 Ch12";
static const char header_2 [] = " ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ---- ----";

const uint32       lpa_id        = uptr->Instance;                      /* the instance ID of the protocol */
LPA_STATE *const   lpa           = lpas [lpa_id].sptr;                  /* the pointer to the channel state */
const PRINTER_TYPE model         = MODEL (uptr->flags);                 /* the printer model number */
const uint32       channel_count = print_props [model].vfu_channels;    /* the count of VFU channels */
uint32             chan, line, current_channel;

if (value == 0)                                         /* if we're called for a summary display */
    fputs (lpa->vfu_title, st);                         /*   then output only the VFU title */

else {                                                          /* otherwise the full VFU definition is requested */
    fprintf (st, "\n%s tape is loaded.\n\n", lpa->vfu_title);   /*   so start by displaying the VFU title */

    fprintf (st, "Line %.*s\n", channel_count * 5, header_1);   /* display the */
    fprintf (st, "---- %.*s\n", channel_count * 5, header_2);   /*   channel headers */

    for (line = 1; line <= lpa->form_length; line++) {  /* loop through the VFU array */
        fprintf (st, "%3d ", line);                     /*   and display the current form line number */

        current_channel = VFU_CHANNEL_1;                /* start with channel 1 */

        for (chan = 1; chan <= channel_count; chan++) { /* loop through the defined channels */
            fputs ("    ", st);                         /*  and add some space */

            if (lpa->VFU [line] & current_channel)      /* if the current channel is punched for this line */
                fputc (lpa->punched_char, st);          /*   then display a punched location */
            else                                        /* otherwise */
                fputc (lpa->unpunched_char, st);        /*   display an unpunched location */

            current_channel = current_channel >> 1;     /* move to the next channel */
            }

        fputc ('\n', st);                               /* end the line */
        }
    }

return SCPE_OK;                                         /* the display routine always succeeds */
}



/***************************************
 *                                     *
 * HP-IB translator local bus routines *
 *                                     *
 ***************************************/


/* Report the protocol configuration.

   This routine is called during run-time initialization to report the
   configuration of this GIC protocol handler.  On entry, "channel" contains the
   channel number for which the report is requested.  The routine returns a
   pointer to an array of structures, each containing a pointer to the device
   instance, the instance ID of the protocol handler, and a bitmap of the
   devices present on the bus and supported by the protocol.

   For a given channel, there may be multiple LPA instances that host units on
   that channel.  Each LPA instance supports a single printer occupying a single
   bus address.  If more than one printer is required, then additional LPA
   instances may be created and assigned to the same channel, as long as they
   have different bus address assignments.

   Therefore, the report array may contain more than one reporting structure,
   with the end of the valid elements followed by an element containing a NULL
   device pointer.  This array is defined as static so that the pointer returned
   remains valid while the caller unloads the information from the structure(s).


   Implementation notes:

    1. While there can be only eight possible devices on a given channel, the
       report array must be large enough to hold one element for each possible
       LPA instance, plus an additional one to hold the NULL element indicating
       the end of the report.  This is because a user might not have changed the
       channel assignments after increasing the count of instances, resulting in
       all instances on the same channel.  This error will be picked up when
       channel conflicts are checked, but that may occur after we are called,
       and so we must be prepared for the worst case.
*/

static BUS_REPORT *lpa_report (uint32 channel)
{
static BUS_REPORT report [MAX_LPA_COUNT + 1];
DEVICE            *dptr;
BUS_REPORT        *rptr;
uint32            lpa_id;

memset (report, 0, sizeof report);                      /* clear any previous report */
rptr = report;                                          /*   and restart at the first element */

for (lpa_id = 0; lpa_id < lpa_count; lpa_id++) {        /* search through the instance table */
    dptr = lpas [lpa_id].dptr;                          /*   and get the next device pointer */

    if (CHANNEL (dptr->flags) == channel                /* if the instance is configured for the channel */
      && not (dptr->flags & DEV_DIS)) {                 /*   and the device is enabled */
        rptr->dptr     = dptr;                          /*   then report the device pointer */
        rptr->instance = lpa_id;                        /*     and the instance ID */
        rptr->present  = BIT (ADDRESS (dptr->flags));   /*       and the assigned address bit */

        rptr++;                                         /* point at the next entry */
        }
    }

return report;                                          /* return the configuration report pointer */
}


/* Accept a command or data byte from the bus.

   This routine is called whenever a controller or talker has sourced a data
   byte to the bus, or whenever a controller, talker, or listener has changed
   the control state of the bus.  It is responsible for decoding commands and
   for calling routines to process secondaries, incoming data, and control state
   changes.

   When accepting a byte with DAV asserted, it corresponds in hardware to
   performing the IEEE-488 AH (Acceptor Handshake), T (Talker), L (Listener),
   RL (Remote Local), and DC (Device Clear) interface functions.  When changing
   the control signals without sourcing a byte, i.e., with DAV denied, it
   corresponds to performing the IEEE-488 SR (Service Request) function or as a
   participant in the SH (Source Handshake) function when asserting and denying
   NRFD (Not Ready for Data).

   If data is present, it decodes primary, secondary, talk, and listen commands,
   and receives data bytes if any of the attached devices are current listeners.
   For control changes, it responds to parallel polls and the IFC (Interface
   Clear), DCL (Device Clear), and SDC (Selected Device Clear) commands.

   On entry, the "lpa_id" parameter is the instance identifier of the LPA device
   connected to the bus, and the "bus" parameter is the set of eight control and
   eight data lines representing the HP-IB.

   If the DAV signal is not asserted, then a control signal change occurred.
   There are three conditions to which we must respond:

    1. A parallel poll is being conducted.  We record the poll state, so that if
       a poll response is subsequently enabled while a poll is still in
       progress, we can notify the controller.  Then we return the poll
       response from the current device.

    2. An Interface Clear is initiated.  IFC unaddresses all devices, so any
       data transactions in progress must be terminated as though an Untalk and
       Unlisten had been accepted from the bus.

    3. Attention or Not Ready for Data is denied.  A device addressed to talk
       must wait for ATN to deny before data may be sent.  Also, a listener that
       has asserted NRFD must deny it before a talker may send data.  If our
       talker is sending data and both ATN and NRFD are denied, then we reenter
       the service routine to send the next byte.

   If DAV is asserted, then a command or data byte is present on the bus.  If
   the ATN signal is asserted, parity is checked if enabled to ensure that it is
   correct.  If it is not, then a command parity error is logged, and the bus is
   frozen by asserting NDAC.  Otherwise, the command byte is decoded to
   determine the proper action.

   Primary commands placed on the bus may be universal (applying to all present
   devices) or addressed (applying only to those devices that are addressed to
   listen).  Address commands select the device that will talk and the device(s)
   that will listen.  If the addressed device is present, the byte is accepted;
   a talk or listen command directed to a device not present is ignored.

   Secondary commands and data bytes are processed only if one or more listeners
   are present.

   This routine is called once for each byte sourced and is responsible for
   notifying the bus controller.  The type of byte passed is determined by the
   state of the ATN signal and, if ATN is asserted, by the high-order bits of
   the value.

   If a Talk, Unlisten, or Untalk command is received while a data transaction
   is currently in progress, the transfer is ended.  This can happen in the
   middle of printer data write in burst mode; if the transfer is incomplete, it
   will be set up again once the status of the prior burst is checked.  A new
   Listen command does not end the transfer, because the current listener
   remains addressed.

   For secondaries, the Amigo Identify command sequence is decoded here, as it
   does not follow the standard HP-IB protocol.  Otherwise, secondaries are
   processed by the bus controller if it is addressed to listen or talk.

   If ATN is not asserted, a data byte has been sourced.  It is processed by
   the bus controller if it is addressed to listen.

   After processing is complete, the routine returns the set of bits
   corresponding to the addresses of the current listeners.


   Implementation notes:

    1. As this device hosts only a single bus controller on a single bus
       address, only one device will be present and therefore only be one
       listener bit will be active at a time.  However, the acceptor routine is
       written in a more general way in case this ever changes.

    2. The HP-IB Parity Checking command says that if checking is enabled and a
       channel command does not have odd parity, "...the device will NOT accept
       the command, i.e., Not Data Accepted (NDAC) will remain in the low state.
       This condition will remain until the host removes Data Valid (DAV) and
       corrects the channel command parity."  Because all devices on the bus
       drive the common NDAC line, any device holding it low will prevent the
       data handshake from completing.  Therefore, we need only test if the map
       of parity checkers is non-zero before declining acceptance of a byte with
       bad parity.  Before returning, we log a Parity Error in the bus
       controller.

    3. Identify is handled as a special case because of the odd bus protocol
       employed (it is an Untalk followed by a secondary containing a device
       address, and the device puts data on the bus without having been
       addressed as the talker).  If a secondary is immediately preceded by an
       Untalk, and the address contained in the secondary is present, then an
       Identify command has been received.  When the command is processed, the
       device is temporarily assigned as the talker for the duration of the
       command, even though it is preceded by an Untalk.  This allows us to know
       which SIMH unit to reschedule when the controller denies ATN to permit
       return of the identity bytes.

    4. The "process_data" routine may assert NRFD to hold off the talker, so we
       remove the DAV and EOI signals from the bus state before calling.  This
       prevents the assertion call from being misinterpreted as a data transfer.

    5. This protocol module maintains the current state of the channel, so that
       the routines know which devices are present and addressed.  This is more
       efficient than querying each bus address at each call to determine
       whether a device is present and, if so, whether it is presently addressed
       to talk or listen.  Consequently, it will be called only if one or more
       devices are present (for commands) or listening (for data).  If the
       routine is called inadvertently, it simply returns with no action taken.

    6. The manual gives no hint as to how the Universal Device Clear and
       Selected Device Clear commands affect the bus controller.  For other
       interfaces, these commands reset the associated controllers to their
       power-up states.  The two commands are equivalent and restart the bus
       controller firmware without affecting either the bus addressing state or
       the printer state.

    7. The "parity" state field is a bitmap of those devices on the bus that
       have enabled command parity checking.  However, the LPA device occupies
       only a single bus address, so the presence of any bit in the "parity"
       field indicates that the addressed device is checking parity.
*/

static uint32 lpa_hpib_accept (uint32 lpa_id, BUS_CABLE bus)
{
LPA_STATE *const lpa         = lpas [lpa_id].sptr;      /* the pointer to the channel state */
DEVICE *const    dptr        = lpas [lpa_id].dptr;      /* the pointer to the controlling device */
const uint32     channel     = CHANNEL (dptr->flags);   /* the channel address of the bus controller */
const uint32     bus_command = BUS_COMMAND (bus);       /* the command present on the bus */
const uint32     bus_address = BUS_ADDRESS (bus);       /* the address indicated for addressing commands */
const uint32     address_bit = BIT (bus_address);       /* the bit corresponding to the bus address */
const uint32     address     = ADDRESS (dptr->flags);   /* the bus address configured for the device */
uint32           dev_map;

lpa->bus = bus;                                         /* synchronize the bus */

lpa->poll_active = (bus & BUS_PPOLL) == BUS_PPOLL;      /* record the parallel poll state */

if (lpa->poll_active)                                   /* if a poll is currently active */
    return lpa->response_set;                           /*   then return the poll response set */

else if (not (bus & Bus_DAV)) {                         /* otherwise if this is a control change */
    if (bus & Bus_IFC) {                                /*   then if Interface Clear is asserted */
        if (lpa->active)                                /*     then if a data transfer is active */
            end_transaction (lpa_id, BUS_PCG);          /*       then abort the transaction */

        lpa->listeners = BUS_UNADDRESSED;               /* unaddress all listeners */
        lpa->talker    = BUS_UNADDRESSED;               /*   and unaddress the current talker */
        }

    else if (not (bus & (Bus_ATN | Bus_NRFD))           /* otherwise if the bus is ready for data */
      && lpa->present & lpa->talker                     /*   and a talker is addressed */
      && lpa->state == Error_Source                     /*   and the controller is waiting to send an error byte */
      && not sim_is_active (BUS_UPTR))                  /*   and the controller unit is not scheduled */
        bus_service (BUS_UPTR);                         /*     then call it directly to source the byte */

    return lpa->present & lpa->listeners;               /* return the current set of listeners */
    }

else if (lpa->present == BUS_NONE)                      /* otherwise if there are no devices present */
    return 0;                                           /*   then the listener bitmap is empty */

else if (bus & Bus_ATN)                                 /* otherwise if this is a command */
    if (lpa->parity && odd_parity [BUS_DATA (bus)]) {   /*   then if parity is enabled but wrong */
        lpa->status |= ST_PARITY_ERROR;                 /*     then log a command parity error */
        lpa->dsj = DSJ_Bus_Error;                       /*       and record a bus error */

        lpa->bus = hpib_source (channel, address,           /* assert NDAC to hang the bus */
                                bus & ~Bus_DAV | Bus_NDAC); /*   and clear DAV for a control change */
        }

    else switch (BUS_GROUP (bus)) {                     /* otherwise dispatch the bus group */

        case BUS_PCG:                                   /* Primary Command Group */
            if (bus_command == BUS_DCL)                 /* if the command is a Universal Device Clear */
                dev_map = lpa->present;                 /*   then get all devices present on the bus */

            else if (bus_command == BUS_SDC)            /* otherwise if it's a Selected Device Clear */
                dev_map = lpa->listeners;               /*   then get just the set of listeners */

            else                                        /* otherwise the command is unsupported */
                dev_map = 0;                            /*   and so is ignored */

            if (dev_map) {                              /* if the device is addressed or present */
                lpa->command = Device_Clear;            /*   then set up the clear command */
                execute_command (lpa_id, address, bus); /*     and execute it */
                }
            break;


        case BUS_LAG:                                   /* Listen Address Group */
            if (bus_address == BUS_UNADDRESS) {         /* if this is an Unlisten */
                if (lpa->active)                        /*   then if unlistening while the channel is active */
                    end_transaction (lpa_id, BUS_LAG);  /*     then end the transaction */

                lpa->listeners = BUS_UNADDRESSED;       /* unaddress all of the listeners */
                }

            else {                                      /* otherwise someone is being addressed */
                lpa->listeners &= ~BUS_UNADDRESSED;     /*   so clear the unaddressed bit */

                if (lpa->present & address_bit) {       /* if the addressed device is present */
                    lpa->listeners |= address_bit;      /*   then it becomes a listener */

                    if (lpa->talker == address_bit)     /* if it is also the current talker */
                        lpa->talker = BUS_UNADDRESSED;  /*   then untalk it */
                    }
                }
            break;


        case BUS_TAG:                                   /* Talk Address Group */
            if (lpa->active                             /* if readdressing while the channel is active */
              && lpa->present & lpa->talker             /*   and a talker is currently addressed */
              && address_bit != lpa->talker)            /*     and the same address is not being respecified */
                end_transaction (lpa_id, BUS_TAG);      /*       then end the transaction */

            if (bus_address == BUS_UNADDRESS)           /* if this is an Untalk */
                lpa->talker = BUS_UNADDRESSED;          /*   then clear the talker */

            else {                                      /* otherwise someone is being addressed */
                lpa->talker = BUS_NONE;                 /*   so clear the unaddressed bit */

                if (lpa->present & address_bit) {       /* if the addressed device is present */
                    lpa->talker = address_bit;          /*   then it becomes the current talker */

                    if (lpa->listeners & address_bit)   /* if the new talker had been listening */
                        lpa->listeners &= ~address_bit; /*   then clear its listener bit */
                    }
                }
            break;


        case BUS_SCG:                                   /* Secondary Command Group */
            if (lpa->talker == BUS_UNADDRESSED          /* if an Untalk preceded the secondary */
              && lpa->present & address_bit) {          /*   and the addressed device is present */
                lpa->command = Amigo_Identify;          /*     then this is an Identify sequence */

                lpa->active = TRUE;                     /* activate the transaction */
                lpa->talker = address_bit;              /*   and temporarily address the device to talk */
                execute_command (lpa_id, address, bus); /*     and start the command */
                }

            else if (lpa->present & (lpa->listeners | lpa->talker)) /* otherwise if the device is addressed */
                process_secondary (lpa_id, address, bus);           /*   then process the secondary */
            break;
        }                                               /* end of command dispatching */

else {                                                  /* otherwise it is bus data (ATN is denied) */
    lpa->bus &= ~(Bus_DAV | Bus_EOI);                   /*   so remove the handshake from the bus state */

    if (lpa->present & lpa->listeners)                  /* if the device is listening */
        process_data (lpa_id, address, bus);            /*   then process the data byte */
    }

return lpa->present & lpa->listeners;                   /* return the current set of listeners */
}



/*******************************************
 *                                         *
 * HP-IB translator local utility routines *
 *                                         *
 *******************************************/


/* Clear the bus controller.

   The bus controller is cleared without affecting the bus addressing state or
   the printer state.  This involves canceling any in-progress controller event
   and setting the bus controller state fields to their initial values.  If the
   clear is a result of a Device Clear command, the status bits associated with
   the controller and the DSJ value are also reset.  These are not done if the
   clear results from a RESET LPA command.

   The bus controller state and command code are not reset, as we want to
   complete the command normally if we are called for a Device Clear.
*/

static void clear_controller (UNIT *uptr, LPA_STATE *lpa)
{
sim_cancel (uptr);                                      /* cancel any in-progress event */
lpa->active = FALSE;                                    /*   and end any transaction */

lpa->param_index = 0;                                   /* clear the parameter buffer index */
lpa->parameter_count = 0;                               /*   and the buffer count */

memset (lpa->count, 0, sizeof lpa->count);              /* clear the bus buffer occupancy counters */
lpa->fill   = lpa->empty  = 0;                          /*   and the buffer selectors */
lpa->findex = lpa->eindex = 0;                          /*     and the buffer indices */

lpa->parser = Match_Disabled;                           /* disable escape recognition */

if (lpa->command == Device_Clear) {                     /* if this is a programmed clear */
    lpa->dsj = DSJ_No_Error;                            /*   then report that no error exists */
    lpa->status &= ~ST_STATIC;                          /*     and clear the static status errors */
    }

lpa->status |= ST_BUFFER_READY;                         /* set buffer available status */

return;                                                 /* return with the controller idle */
}


/* Process a secondary command.

   This routine is called to process a secondary command received from the bus
   for an addressed device.  Secondaries are used by the Amigo protocol to
   identify the various commands, specifically:

           Hex
     T/L  Value  Identifies
     ---  -----  ---------------------
      L    00    Print Data
      L    02    Set Escape
      L    10    Amigo Clear
      L    1E    Write Loopback
      L    1F    Initiate Self Test

      T    0E    Read Status
      T    10    Device Specified Jump
      T    1E    Read Loopback
      T    1F    Read Self Test Result

   On entry, the "lpa_id" parameter is the instance identifier of the LPA device
   connected to the bus, the "address" parameter contains the bus address of the
   device, which will be either a listener or talker, and the "bus" parameter
   contains the set of eight control and eight data lines representing the HP-IB
   state.

   The routine decodes the valid secondaries, checks that the bus controller is
   in the appropriate state for the supplied secondary, and establishes the new
   controller state as a result of receiving the secondary.

   Error recovery is as follows:

    - An unsupported talk secondary sends a single data byte tagged with EOI.

    - An unsupported listen secondary accepts and discards any accompanying data
      bytes until EOI is asserted or an Unlisten is received.


   Implementation notes:

    1. The HP-IB Translator Service Manual does not specify the action taken if
       an invalid secondary is received.  In simulation, this is reported as a
       Protocol Error and the command is invalidated.

    2. The bus controller enables its poll response to notify the host that a
       bus data buffer is available.  However, the host might not wait for a
       poll and might initiate the next command while the poll is still pending.
       In simulation, the poll response is scheduled as a service event that
       represents the controller overhead.  This event may still be pending when
       a secondary is received, and if so, the service event is canceled before
       proceeding with the next command.
*/

static void process_secondary (uint32 lpa_id, uint32 address, BUS_CABLE bus)
{
LPA_STATE *const lpa         = lpas [lpa_id].sptr;      /* the pointer to the channel state */
DEVICE *const    dptr        = lpas [lpa_id].dptr;      /* the pointer to the controlling device */
const uint32     secondary   = BUS_ADDRESS (bus);       /* the secondary address without parity */
const uint32     address_bit = BIT (address);           /* the bit corresponding to the bus address */
const CN_STATE   entry_state = lpa->state;              /* the controller state at entry */

if (lpa->state == Completion_Wait)                      /* if a previous poll response is still pending */
    sim_cancel (BUS_UPTR);                              /*   then cancel it to avoid confusion */

lpa->active = TRUE;                                     /* activate the transaction */

if (lpa->listeners & address_bit)                       /* if this device is listening */
    switch (secondary) {                                /*   then dispatch the listen secondary */

        case 0x00:                                      /* Print Data */
            lpa->command = Print_Data;                  /* set the command */
            break;


        case 0x02:                                      /* Set Escape */
            lpa->command = Set_Escape;                  /* set the command */
            break;


        case 0x10:                                      /* Amigo Clear */
            lpa->command = Amigo_Clear;                 /* set the command */
            break;


        case 0x1E:                                      /* Write Loopback */
            lpa->command = Write_Loopback;              /* set the command */
            break;


        case 0x1F:                                      /* Initiate Self-Test */
            lpa->command = Initiate_Self_Test;          /* set the command */
            break;


        default:                                        /* an unsupported listen secondary was received */
            lpa->command = Invalid;                     /*   so set an invalid command */
            lpa->state = Error_Sink;                    /*     and sink as many bytes as the host wants to send */
            break;
        }

else                                                    /* otherwise the device is talking */
    switch (secondary) {                                /*   so dispatch the talk secondary */

        case 0x0E:                                      /* Read Status */
            lpa->command = Read_Status;                 /* set the command */
            break;


        case 0x10:                                      /* Device Specified Jump */
            lpa->command = Device_Specified_Jump;       /* set the command */
            break;


        case 0x1E:                                      /* Read Loopback */
            lpa->command = Read_Loopback;               /* set the command */
            break;


        case 0x1F:                                      /* Read Self-Test Result */
            lpa->command = Read_Self_Test;              /* set the command */
            break;


        default:                                        /* an unsupported talk secondary was received */
            lpa->command = Invalid;                     /*   so set an invalid command */
            lpa->state = Error_Source;                  /*     and source an error byte tagged with EOI */

            activate_unit (lpa_id, BUS_UPTR, Data_Delay);   /* schedule the data transfer */
            break;
        }


if (lpa->command == Invalid)                            /* if a bad secondary was received */
    tpprintf (dptr, TRACE_CMD, "Device %u invalid %s secondary %02XH received\n",
              address, (lpa->listeners & address_bit ? "listen" : "talk"), secondary);

else {                                                  /* otherwise the command is accepted */
    lpa->parameter_count =                              /*   so set the number */
      cmd_props [lpa->command].parameter_count;         /*     of the expected parameters */

    if (lpa->parameter_count > 0)                       /* if the command has parameters */
        lpa->param_index = 0;                           /*   then reset the parameter index */

    lpa->state = cmd_props [lpa->command].next_state;   /* enter the next state */

    if (lpa->state != Parameter_Wait)                   /* if not waiting for a parameter */
        execute_command (lpa_id, address, bus);         /*   then start the command */
    }

if (TRACINGP (dptr, TRACE_STATE)                            /* if tracing is active */
  && lpa->state != entry_state)                             /*   and the controller state has changed */
    trace_state (dptr, address, entry_state, lpa->state);   /*     then trace the state transition */

return;
}


/* Process a data byte.

   This routine is called to process a data byte received from the bus for an
   addressed device.  Depending on the controller state, the byte represents
   either a command parameter or a data value.

   On entry, the "lpa_id" parameter is the instance identifier of the LPA device
   connected to the bus, the "address" parameter contains the bus address of the
   device, which will be a listener, and the "bus" parameter contains the set of
   eight control and eight data lines representing the HP-IB state.

   The routine interprets the data byte in the context of the current bus
   controller state, as follows:

     Parameter_Wait
     --------------
     The value is a parameter to the current command.  It is placed in the
     parameter buffer, the expected parameter count is decremented, and if no
     more are expected, the pending command is executed.

     Execution_Receive
     -----------------
     The value is a data byte received from the host during a data transfer to
     the bus buffer.  If escape parsing is enabled, the character is checked for
     the expected sequence; if the sequence is invalid, then error recovery
     consists of ignoring characters until the end of the sequence.  If the
     escape sequence completes and is valid, the corresponding paper instruction
     is loaded into the bus buffer.  If escape parsing is disabled, or the start
     of an escape sequence has not been detected, then the character is loaded
     into the bus buffer.

     If the buffer is full or the byte is tagged with EOI, the buffer is tied
     off and sent to the printer.  The NRFD signal is asserted on the HP-IB to
     hold off the talker and the bus controller is scheduled to process the
     byte.

     Command_Receive
     ---------------
     The value is a data byte received from the host for the bus controller.  It
     is placed in the parameter buffer, and the expected reception count is
     decremented.  If the byte is tagged with EOI, then the count is cleared to
     indicate that no more bytes are expected.  Otherwise, if the count is
     already zero, and the byte is not tagged with EOI, then an error has
     occurred, and the controller proceeds to sink as many bytes as the host
     wishes to send.  The NRFD signal is asserted on the HP-IB to hold off the
     talker, and the bus controller is scheduled to process the byte.

     Error_Sink
     ----------
     The value is a data byte that is discarded due to a previous error.  If the
     byte is tagged with EOI, the transaction is terminated.  Otherwise, the
     NRFD signal is asserted on the HP-IB to hold off the talker, and the bus
     controller is scheduled to complete the bus handshake after the data
     transfer time elapses.

   Entry in any other state results in a transition to the Error_Sink state to
   discard the unexpected data.  An example of this is entry in the
   Command_Ready state.  This can only occur if the host sends a Listen followed
   by data without an intervening secondary.  In this case (and the others), we
   discard the data; a Protocol Error will be reported when the final byte of
   the transaction is received.


   Implementation notes:

    1. The HP 2563 printer manual says that two-character escape sequences use
       the characters "0" though "~" (60 to 176 octal) as the second character,
       while parameter escape sequences use the characters "!" through "/" (41
       to 57 octal) as the second character.  Error recovery therefore is
       complete if the second character indicates a two-character escape,
       whereas parameter escape recovery consists of absorbing characters until
       an uppercase letter is seen.
*/

static void process_data (uint32 lpa_id, uint32 address, BUS_CABLE bus)
{
LPA_STATE *const   lpa         = lpas [lpa_id].sptr;        /* the pointer to the channel state */
DEVICE *const      dptr        = lpas [lpa_id].dptr;        /* the pointer to the controlling device */
const uint32       channel     = CHANNEL (dptr->flags);     /* the channel address of the bus controller */
const PRINTER_TYPE model       = MODEL (LP_UPTR->flags);    /* the printer model */
const uint8        data        = BUS_DATA (bus);            /* the data value from the bus */
const CN_STATE     entry_state = lpa->state;                /* the controller state at entry */
uint8              paper_command;

switch (lpa->state) {                                   /* dispatch the controller state */

    case Parameter_Wait:                                /* waiting for a parameter */
        lpa->parameter [lpa->param_index++] = data;     /* save the parameter in the buffer */
        lpa->parameter_count--;                         /*   and drop the remaining parameter count */

        if (lpa->parameter_count == 0)                  /* if no more are expected */
            execute_command (lpa_id, address, bus);     /*   then execute the command */
        break;


    case Execution_Receive:                             /* receiving data from the host */
        switch (lpa->parser) {                          /* dispatch the parser state */

            case Match_ESC:                             /* looking for an ESC character */
                if (data == ESC) {                      /* if the character is Escape */
                    lpa->value = 0;                     /*   then clear the escape parameter value */
                    lpa->parser = Match_Amp;            /*     and move to the next state */
                    break;
                    }

            /* fall through into Match_Disabled case */

            case Match_Disabled:                                /* escape detection is disabled */
                lpa->buffer [lpa->fill] [lpa->findex++] = data; /*   so just store the character in the buffer */
                break;


            case Match_Amp:                             /* look for an Ampersand */
                if (data == '&')                        /* if the expected character is present */
                    lpa->parser = Match_L;              /*   then move to the next state */

                else if (data >= '0' && data <= '~')    /* otherwise if this is a two-character escape */
                    lpa->parser = Match_ESC;            /*   then return to the initial parser state */

                else                                    /* otherwise */
                    lpa->parser = Match_Error;          /*   move to the error state */
                break;


            case Match_L:                               /* look for an uppercase L */
                if (data == 'l')                        /* if the expected character is present */
                    lpa->parser = Match_Number;         /*   then move to the next state */
                else                                    /* otherwise */
                    lpa->parser = Match_Error;          /*   move to the error state */
                break;


            case Match_Number:                                  /* look for a numeric value */
                if (data >= '0' && data <= '9')                 /* if the character is a digit */
                    lpa->value = lpa->value * 10 + data - '0';  /*   then convert and accumulate the value */

                else if (data == 'U') {                     /* otherwise if slewing to a count */
                    if (lpa->value > 63) {                  /*   then if the count is invalid */
                        lpa->dsj = DSJ_Bus_Error;           /*     then report a bus error */
                        lpa->status |= ST_PROTOCOL_ERROR;   /*       and set protocol error status */
                        }

                    else do {                                       /* otherwise slew the number of lines indicated */
                        if (lpa->value <= SLEW_MAX) {               /* if the printer can slew as requested */
                            paper_command = PAPER_CMD               /*   then form the slew instruction */
                                              | (uint8) lpa->value;
                            lpa->value = 0;                         /* slewing is complete */
                            }

                        else {                                      /* otherwise break up the job */
                            paper_command = PAPER_CMD | SLEW_MAX;   /*   and slew the maximum count possible */
                            lpa->value = lpa->value - SLEW_MAX;     /*     and drop the remaining line count */
                            }

                        lpa->buffer [lpa->fill] [lpa->findex++] = paper_command;    /* store the paper instruction */
                        }
                    while (lpa->value > 0);             /* continue until all requested lines are skipped */

                    lpa->parser = Match_ESC;            /* return to the initial parser state */
                    }

                else if (data == 'V') {                                     /* otherwise if slewing to a VFU channel */
                    if (lpa->value < 1                                      /*   then if the channel number */
                      || lpa->value > print_props [model].vfu_channels) {   /*     is invalid */
                        lpa->dsj = DSJ_Bus_Error;                           /*       then report a bus error */
                        lpa->status |= ST_PROTOCOL_ERROR;                   /*         and set protocol error status */
                        }

                    else                                                    /* otherwise the channel number is value */
                        lpa->buffer [lpa->fill] [lpa->findex++] =           /*   so form the channel instruction */
                          PAPER_CMD | FORMAT_VFU | (uint8) lpa->value - 1;  /*     biasing as required by the printer */

                    lpa->parser = Match_ESC;            /* return to the initial parser state */
                    }

                else if (data != ' ')                   /* otherwise if the character is not a blank */
                    lpa->parser = Match_Error;          /*   then move to the error state */
                break;


            case Match_Error:                           /* silently absorb characters */
                if (data >= 'A' && data <= 'Z')         /*   until an uppercase letter is seen */
                    lpa->parser = Match_ESC;            /*     and then return to the initial parser state */
                break;
            }                                           /* all cases are handled */

        if (bus & Bus_EOI || lpa->findex == 128)        /* if this is the last byte of the transaction */
            end_transaction (lpa_id, BUS_SCG);          /*   then tie off the bus buffer */

        lpa->bus = hpib_source (channel, address,       /* assert NRFD to hold off the talker */
                               lpa->bus | Bus_NRFD);

        activate_unit (lpa_id, BUS_UPTR, Data_Delay);   /* schedule the unit to accept the data */
        break;


    case Command_Receive:                               /* receiving command bytes */
        lpa->parameter [lpa->param_index++] = data;     /* save the byte in the parameter buffer */
        lpa->parameter_count--;                         /*   and drop the remaining count */

        if (bus & Bus_EOI)                              /* if this is the last byte from the host */
            lpa->parameter_count = 0;                   /*   then there are no more expected */

        else if (lpa->parameter_count == 0)             /* otherwise if more are coming but are not wanted */
            lpa->state = Error_Sink;                    /*   then sink as many bytes as the host wants to send */

        lpa->bus = hpib_source (channel, address,       /* assert NRFD to hold off the talker */
                               lpa->bus | Bus_NRFD);

        activate_unit (lpa_id, BUS_UPTR, Data_Delay);   /* schedule the unit to accept the data */
        break;


    default:                                            /* data was received in the wrong state */
        lpa->state = Error_Sink;                        /* sink all data arriving in the wrong state */

    /* fall through into the Error_Sink state */

    case Error_Sink:                                    /* sinking data bytes after an error */
        if (bus & Bus_EOI)                              /* if the transfer is complete */
            lpa->active = FALSE;                        /*   then end the transaction */

        else                                            /* otherwise */
            lpa->bus = hpib_source (channel, address,   /*   assert NRFD to hold off the talker */
                                   lpa->bus | Bus_NRFD);

        activate_unit (lpa_id, BUS_UPTR, Data_Delay);   /* schedule the unit to accept the data */
        break;
    }                                                   /* end of state dispatch */

if (TRACINGP (dptr, TRACE_STATE)                            /* if tracing is active */
  && lpa->state != entry_state)                             /*   and the controller state has changed */
    trace_state (dptr, address, entry_state, lpa->state);   /*     then trace the state transition */

return;
}


/* Execute an Amigo line printer command.

   This routine is called to execute bus commands.  On entry, the "lpa_id"
   parameter is the instance identifier of the LPA device connected to the bus,
   the "address" parameter contains the bus address of the device, which will be
   a listener, and the "bus" parameter contains the set of eight control and
   eight data lines representing the HP-IB state.  The command to execute is
   present in the "command" field of the device state structure, and any
   parameter required by the command has already been received and is present in
   the state structure "parameter" array.

   This routine implements the following commands:

     Bus  Sec  Operation
     ---  ---  -------------------------------
     UNT  MSA  Identify
      14   -   Universal Device Clear
      04   -   Selected Device Clear

     MLA   00  Print data
     MLA   02  Enable/Disable Escape Sequences
     MLA   10  Amigo Clear
     MLA   1E  Write Loopback
     MLA   1F  Initiate Self-Test

     MTA   0E  Read Status
     MTA   10  Device-Specified Jump
     MTA   1E  Read Loopback
     MTA   1F  Read Self-Test Result

   After command-specific processing is done, a bus controller service event
   will be scheduled if it is required to advance or complete the command.


   Implementation notes:

    1. The Device Clear command must call "bus_service" directly, rather than
       scheduling an event, as otherwise the GIC may start the next command
       before this one completes.

    2. The Print Data command disables the poll response to ensure that the
       window between bus buffer fill completion and the check for another
       available buffer is closed.  Otherwise, a parallel poll conducted after
       the last byte for the last buffer was sent and before the bus controller
       had disabled the poll response for buffer unavailability could get the
       wrong response and therefore lose print data.

       A similar window exists for the Read Status command.  If DSJ > 0, the
       poll response assertion is maintained until the status is read.  We
       disable the poll response at the start of the command to ensure that a
       poll conducted after the status is sent but before the command completes
       is not misinterpreted by the host.

    3. For the Set Escape command parameter, the MPE printer driver "HIOLPRT2"
       sends 200 octal to enable escape recognition and 202 octal to disable
       recognition.  There is no indication in the driver or the service manual
       of how the bus controller reacts if a different parameter byte is
       supplied.  We elect to report a Protocol Error for this case.

    4. An Amigo Clear command performs no action itself, other than configuring
       bus parity checking.  The Selected Device Clear that follows in the
       protocol sequence performs the actual bus controller clear.

    5. The Write Loopback command accepts up to 256 bytes from the host and
       saves them in the parameter buffer, where they reside until being sent
       back to the host for a Read Loopback command.  To avoid corrupting the
       buffer by an intervening Device Specified Jump command, the DSJ byte is
       always stored at buffer index 256 after the last Write Loopback byte.

    6. The Initiate Self-Test command is accepted but is otherwise ignored.  The
       Read Self-Test Result command returns a dummy value that indicates that
       the test completed successfully.
*/

static void execute_command (uint32 lpa_id, uint32 address, BUS_CABLE bus)
{
LPA_STATE *const lpa         = lpas [lpa_id].sptr;      /* the pointer to the channel state */
DEVICE *const    dptr        = lpas [lpa_id].dptr;      /* the pointer to the controlling device */
t_bool           unsupported = FALSE;                   /* TRUE if the command or utility is unsupported */

tpprintf (dptr, TRACE_INCO, "Device %u %s started\n",
          address, command_names [lpa->command]);

switch (lpa->command) {                                 /* dispatch the command opcode */

    case Device_Clear:
        clear_controller (BUS_UPTR, lpa);               /* clear the bus controller */

        if (TRACINGP (dptr, TRACE_STATE))               /* if tracing is active */
            trace_state (dptr, address, lpa->state,     /*   then trace the coming state transition */
                         Completion_Wait);

        lpa->state = Completion_Wait;                   /* set up for command completion */
        bus_service (BUS_UPTR);                         /*   and complete immediately */
        break;


    case Print_Data:
        set_response (lpa_id, address, CLEAR);          /* disable the poll response and wait for the data */
        break;


    case Set_Escape:
        lpa->state = Completion_Wait;                   /* set up the command completion */

        if (lpa->parameter [0] == ENABLE_ESC            /* if enabling escape sequences */)
            lpa->parser = Match_ESC;                    /*   then set the initial parser state */

        else if (lpa->parameter [0] == DISABLE_ESC)     /* otherwise if disabling escape sequences */
            lpa->parser = Match_Disabled;               /*   then inhibit parsing */

        else {                                          /* otherwise the option is invalid */
            lpa->status |= ST_PROTOCOL_ERROR;           /*   to set protocol error status */
            lpa->dsj = DSJ_Bus_Error;                   /*     and report a bus error */

            tpprintf (dptr, TRACE_CMD, "Device %u Set Escape invalid parameter %03o\n",
                      address, lpa->parameter [0]);

            unsupported = TRUE;                         /* the command is unsupported */
            }

        if (not unsupported)                            /* if the command is valid */
            tpprintf (dptr, TRACE_CMD, "Device %u Set Escape %s executing\n",
                      address, condition_names [lpa->parameter [0] == ENABLE_ESC]);
        break;


    case Amigo_Clear:
        if (lpa->parameter [0] & 1)                     /* if the enable bit is set */
            lpa->parity |= BIT (address);               /*   then indicate that our device checks parity */
        else                                            /* otherwise */
            lpa->parity &= ~BIT (address);              /*   disable parity checking for our device */

        tpprintf (dptr, TRACE_INCO, "Device %u Amigo Clear command parity %s\n",
                  address, condition_names [lpa->parameter [0] & 1]);
        break;


    case Write_Loopback:
        break;                                          /* no special action is needed */


    case Initiate_Self_Test:
        tpprintf (dptr, TRACE_INCO, "Device %u Initiate Self-Test%s%.0u executing\n",
                  address, (lpa->parameter [0] ? " subtest " : ""), lpa->parameter [0]);

        lpa->state = Completion_Wait;                   /* the command completes immediately */
        break;


    case Read_Status:
        set_response (lpa_id, address, CLEAR);          /* disable the poll response before sending the status */

        tpprintf (dptr, TRACE_INCO, "Device %u DSJ %u status is %s\n", address,
        lpa->dsj, fmt_bitset (lpa->status, status_format));

        lpa->dsj = DSJ_No_Error;                        /* clear the DSJ after reporting it */

        lpa->parameter [0] = lpa->status;               /* return the status word */
        lpa->status &= ~ST_STATIC;                      /*   and then clear the static error bits */
        break;


    case Device_Specified_Jump:
        lpa->parameter [256] = (uint8) lpa->dsj;        /* save the DSJ at the end of the buffer */
        lpa->param_index = 256;                         /*   to preserve the Write Loopback data */

        tpprintf (dptr, TRACE_CMD, "Device %u Device Specified Jump %u executing\n",
                  address, lpa->dsj);
        break;


    case Read_Loopback:
        lpa->parameter_count = lpa->param_index;        /* use the buffer count */
        lpa->param_index = 0;                           /*   given by the Write Loopback command */
        break;


    case Read_Self_Test:
        lpa->parameter [0] = 0;                         /* return a dummy code indicating a successful test */
        break;


    case Amigo_Identify:
        lpa->parameter [0] = 0x20;                      /* set the printer ID */
        lpa->parameter [1] = 0x0A;                      /*   to indicate the HP-IB translator PCA */

        lpa->parameter_count = 2;                       /* set the count and index here */
        lpa->param_index = 0;                           /*   as process_secondary is not called for Identify */

        tpprintf (dptr, TRACE_CMD, "Device %u Amigo Identify %02X%02XH executing\n",
                  address, lpa->parameter [0], lpa->parameter [1]);

        if (TRACINGP (dptr, TRACE_STATE))               /* if tracing is active */
            trace_state (dptr, address, lpa->state,     /*   then trace the state transition */
                         Command_Send);

        lpa->state = Command_Send;                      /* set up the next state */
        break;


    default:                                            /* commands not implemented */
        unsupported = TRUE;                             /*   are ignored */
        break;
    }


if (cmd_props [lpa->command].use_controller)            /* if the command uses the bus controller */
    activate_unit (lpa_id, BUS_UPTR, Activation_Delay); /*   then schedule the event service */

if (unsupported)
    tpprintf (dptr, TRACE_CMD, "Device %u %s command not implemented\n",
              address, command_names [lpa->command]);

else if (not cmd_props [lpa->command].reported)
    tpprintf (dptr, TRACE_CMD, "Device %u %s executing\n",
              address, command_names [lpa->command]);

return;                                                 /* return with the command initiated or completed */
}


/* End an active channel transaction.

   This routine is called to end a channel transaction in progress.  Termination
   occurs when a device is unaddressed in the middle of a data transmission
   during a command.  This usually represents an error, except in the case of
   burst transfers, which are expected to terminate at the end of the burst.

   On entry, the "lpa_id" parameter is the instance identifier of the LPA device
   connected to the bus, and the "bus_group" parameter contains the HP-IB
   command group from the current bus state.

   If the Primary Command Group is passed, then the termination results from an
   IFC (interface clear) assertion, and so represents a channel abort.  If the
   Secondary Command Group is passed, then this termination is the normal end
   after receiving a byte tagged with EOI.  If the Listen Address Group or Talk
   Address Group, then the termination is the (likely) result of a burst
   transfer ending without an EOI byte and so is expected.  In all cases, if the
   current bus data buffer is occupied, it is tied off, buffer-ready status is
   updated, and the print buffer load service is scheduled if it is not active.
   If an EOI did not terminate the transaction, the bus event service routine is
   called to complete the command.
*/

static void end_transaction (uint32 lpa_id, BUS_DATA bus_group)
{
LPA_STATE *const lpa     = lpas [lpa_id].sptr;          /* the pointer to the channel state */
DEVICE *const    dptr    = lpas [lpa_id].dptr;          /* the pointer to the controlling device */
const uint32     address = ADDRESS (dptr->flags);       /* the bus address of the device containing the unit */

lpa->active = FALSE;                                    /* end the channel transaction */

if (lpa->parser != Match_Disabled)                      /* if escape sequence parsing is enabled */
    lpa->parser = Match_ESC;                            /*   then reset the parser state */

if (lpa->findex > 0) {                                  /* if the current buffer is occupied */
    lpa->count [lpa->fill] = lpa->findex;               /*   then set the byte count */

    tpprintf (dptr, TRACE_BUFFR, "Buffer %u filled (%u/128 bytes)\n",
              lpa->fill, lpa->findex);

    lpa->findex = 0;                                    /* reset the filling index */
    lpa->fill = (lpa->fill + 1) % 8;                    /*   and advance to the next buffer */

    if (lpa->count [lpa->fill] == 0)                    /* if another buffer is available */
        lpa->status |= ST_BUFFER_READY;                 /*   then set buffer available status */
    else                                                /* otherwise */
        lpa->status &= ~ST_BUFFER_READY;                /*   clear buffer available status */

    if (lpa->eindex == 0 && lpa->status & ST_ONLINE) {  /* if the printer is idle and online */
        LP_UPTR->wait = lpa->dlyptr->buffer_load;       /*   then schedule the buffer load delay */
        activate_unit (lpa_id, LP_UPTR, Load_Delay);    /*     to start printing the data */
        }
    }

if (bus_group != BUS_SCG) {
    if (bus_group == BUS_PCG)                                   /* if the end is unexpected */
        tpprintf (dptr, TRACE_CMD, "Device %u channel abort\n", /*   then report a channel abort */
                  address);

    if (TRACINGP (dptr, TRACE_STATE))                               /* if tracing is active */
        trace_state (dptr, address, lpa->state, Completion_Wait);   /*   then trace the state transition */

    lpa->state = Completion_Wait;                       /* set the completion state */
    bus_service (BUS_UPTR);                             /*   and call the event routine to end the transaction */
    }

return;
}


/* Enable or disable a device's parallel poll response.

   This routine changes the poll response for a specified bus device.  It
   corresponds in hardware to a device's IEEE-488 PPE (parallel poll enable) or
   PPD (parallel poll disable) state.  When the controller conducts a parallel
   poll, the responses of all eight bus devices are placed on the data bus for
   retrieval by the controller.  An enabled poll response indicates to the host
   that a device is requesting service.

   On entry, the "lpa_id" parameter is the instance identifier of the LPA device
   connected to the bus, the "address" parameter is the bus address of the
   device whose poll response is changing, and the "response" parameter
   indicates whether the response should be set (enabled) or cleared (disabled).
   The poll response bit corresponding to the bus address is included or
   excluded from the set of poll responses.

   If a parallel poll is in progress when a poll response is set, the poll is
   conducted again to reflect the new response.


   Implementation notes:

    1. The set of poll responses is maintained by the channel to improve
       performance, so that we do not need to query each device during a poll to
       ascertain its response.
*/

static void set_response (uint32 lpa_id, uint32 address, FLIP_FLOP response)
{
LPA_STATE *const lpa           = lpas [lpa_id].sptr;    /* the pointer to the channel state */
const BUS_DATA   old_responses = lpa->response_set;     /* the current set of poll responses */
DEVICE *const    dptr          = lpas [lpa_id].dptr;    /* the pointer to the controlling device */
const uint32     channel       = CHANNEL (dptr->flags); /* the channel address of the bus controller */

if (response == SET)                                    /* if the poll response is set */
    lpa->response_set |= HPIB_BIT (address);            /*   then include the poll bit in the response set */
else                                                    /* otherwise we are clearing the poll response */
    lpa->response_set &= ~HPIB_BIT (address);           /*   so remove the poll bit */

if (lpa->response_set != old_responses) {               /* if the poll response has changed, report it */
    tpprintf (dptr, TRACE_STATE, "Device %u parallel poll response %s\n",
              address, condition_names [response]);

    if (lpa->poll_active && response == SET)                    /* if a parallel poll is in progress */
        lpa->bus = hpib_source (channel, address, lpa->bus);    /*   then conduct it again with the new response */
    }

return;
}


/* Activate the bus controller or printer unit.

   This routine activates the SIMH unit belonging to the instance of the LPA
   device specified by the "lpa_id" parameter and pointed to by the "uptr"
   parameter with a delay specified by the "delay" parameter.  For all but
   printer delays, the "delay" is an index into the "bus_delays" array, whose
   fields specify the delay scheduling time and the reason for the delay.  For
   printer delays, the delay has been calculated and set into the unit wait
   field before the routine is called.  If tracing is enabled, the activation is
   logged to the debug file.

   The status of the activation is returned to the caller.


   Implementation notes:

    1. We want to avoid printing trace lines when the unit is already scheduled
       because the time reported will be ignored in favor of the remaining event
       scheduling time.  However, there is no acceptable substitute for calling
       "sim_is_active" to determine if the unit is currently scheduled, as
       "sim_activate" returns SCPE_OK regardless of whether or not the unit is
       already in the queue.  Checking "uptr->next" doesn't work if the unit is
       last in the queue, and checking "unit->time" doesn't work because
       although the "time" field is reset to zero when the unit is removed from
       the queue, time = 0 is a legal "sim_activate" parameter.  "sim_activate"
       calls "sim_is_active" on entry, but there doesn't seem to be a way of
       avoiding this double call for most activation requests.
*/

static t_stat activate_unit (uint32 lpa_id, UNIT *uptr, DELAY delay)
{
LPA_STATE *const lpa  = lpas [lpa_id].sptr;             /* the pointer to the channel state */
DEVICE *const    dptr = lpas [lpa_id].dptr;             /* the pointer to the controlling device */
const uint32     unit = (uint32) (uptr - dptr->units);  /* the unit number */
int32            time;

if (sim_is_active (uptr))                               /* if the unit is already scheduled */
    return SCPE_OK;                                     /*   then return with nothing to do */

else if (bus_delays [delay].ticks == 0) {               /* otherwise if a calculated delay is specified */
    time = uptr->wait;                                  /*   then get it from the unit wait field */

    tpprintf (dptr, TRACE_SERV, "Device %u %s unit %s delay %d service scheduled\n",
              ADDRESS (dptr->flags), unit_names [unit],
              bus_delays [delay].name, time);
    }

else {                                                  /* otherwise */
    time = bus_delays [delay].ticks;                    /*   get the indicated delay */

    tpprintf (dptr, TRACE_SERV, "Device %u %s unit state %s %s delay %d service scheduled\n",
              ADDRESS (dptr->flags), unit_names [unit],
              state_names [lpa->state], bus_delays [delay].name, time);
    }

return sim_activate (uptr, time);                       /* activate the unit and return the result */
}


/* Set an alarm condition.

   This routine is called when an alarm condition exists.  An alarm occurs when
   paper is out (paper fault) or a VFU command addresses a channel that does not
   contain a punch (tape fault).  In response, the printer goes offline and, for
   all models except the 2607, becomes not-ready.

   On entry, the routine attempts to set the printer offline.  If this succeeds,
   and the printer has a separate not-ready indication, then it is set
   not-ready.  If it fails, for reasons explained in the comments for the
   "set_on_offline" routine, it will be set offline and not-ready when printing
   completes.


   Implementation notes:

    1. The 26069A translator reports both paper faults and tape faults as "paper
       out" in the status byte.
*/

static t_bool set_alarm (uint32 lpa_id, UNIT *uptr, PRINTER_TYPE model)
{
LPA_STATE *const lpa = lpas [lpa_id].sptr;      /* the pointer to the channel state */

if (set_locality (lpa_id, uptr, Offline)) {             /* if the printer went offline */
    if (print_props [model].not_ready)                  /*   then if the printer reports ready status separately */
        lpa->status |= ST_PAPER_OUT;                    /*     then set the printer not-ready */

    return TRUE;                                        /* return success status */
    }

else                                                    /* otherwise the offline request is pending */
    return FALSE;                                       /*   so return deferral status */
}


/* Set the printer locality.

   This routine is called to set the printer online or offline, and it returns
   TRUE if the request succeeded or FALSE if it was deferred.  An online request
   always succeeds, so it is up to the caller to ensure that going online is
   permissible (e.g., that paper is loaded into the printer).  An offline
   request succeeds only if the printer is idle.  If characters are present in
   the print buffer, or if the printer is printing or slewing, then the request
   is deferred until the current line is complete.
*/

static t_bool set_locality (uint32 lpa_id, UNIT *uptr, LOCALITY printer_state)
{
LPA_STATE *const lpa     = lpas [lpa_id].sptr;          /* the pointer to the channel state */
DEVICE *const    dptr    = lpas [lpa_id].dptr;          /* the pointer to the controlling device */
const uint32     address = ADDRESS (dptr->flags);       /* the bus address of the device containing the unit */

if (printer_state == Offline) {                         /* if the printer is going offline */
    if (lpa->pindex == 0                                /*   then if the print buffer is empty */
      && not sim_is_active (uptr)) {                    /*     and the printer is idle */
        uptr->flags |= UNIT_OFFLINE;                    /*       then set the printer offline now */
        lpa->status &= ~ST_ONLINE;                      /*         and update the printer status */
        lpa->dsj = DSJ_Print_Error;                     /*           and indicate a printer problem */
        }

    else {                                              /*   otherwise the request must wait */
        lpa->offline_pending = TRUE;                    /*     until the line is printed */
        return FALSE;                                   /*       and the command is not complete */
        }
    }

else {                                                      /* otherwise the printer is going online */
    uptr->flags &= ~UNIT_OFFLINE;                           /*   so clear the unit flag */
    lpa->status = lpa->status & ~ST_PAPER_OUT | ST_ONLINE;  /*     and indicate that the printer is online */

    lpa->paper_fault = FALSE;                           /* clear any paper fault */
    lpa->tape_fault  = FALSE;                           /*   and any tape fault */

    if (lpa->dsj == DSJ_Print_Error)                    /* if the printer was out of paper but otherwise OK */
        lpa->dsj = DSJ_No_Error;                        /*   then clear the error */

    if (lpa->status & ST_BUFFER_READY)                  /* if a bus buffer is available */
        set_response (lpa_id, address, SET);            /*   then assert the poll response */
    else                                                /* otherwise */
        set_response (lpa_id, address, CLEAR);          /*   deny the poll response */

    if (lpa->count [lpa->empty] > 0) {                  /* if the current buffer is occupied */
        uptr->wait = lpa->dlyptr->buffer_load;          /*   then schedule the buffer load delay */
        activate_unit (lpa_id, uptr, Load_Delay);       /*     to restart the transfer to the printer */
        }
    }

tpprintf (dptr, TRACE_CMD, "Printer set %s\n",
          locality_names [printer_state]);

lpa->offline_pending = FALSE;                           /* the operation completed */
return TRUE;                                            /*   so return success */
}


/* Load the VFU.

   The printer VFU is loaded either with a custom tape image stored in the file
   associated with the stream "vf" or with the standard 66-line tape if the
   stream parameter is NULL.  The "lpa_id" parameter is the instance identifier
   of the LPA device connected to the bus, and the "uptr" parameter points to
   the printer unit.

   The standard VFU tape (02607-80024 for the 8-channel HP 2607 and 02613-80001
   for the 12-channel HP 2611, 2613, 2617, 2618, and 2619) defines the channels
   as:

     Chan  Description
     ----  --------------
       1   Top of form
       2   Bottom of form
       3   Single space
       4   Double space
       5   Triple space
       6   Half page
       7   Quarter page
       8   Sixth page
       9   Bottom of form

   ...with channels 10-12 uncommitted.

   A custom tape file starts with a VFU definition line and then contains one
   channel-definition line for each line of the form.  The number of lines
   establishes the form length.  Channel 1 must be dedicated to the top-of-form,
   but the other channels may be defined as desired.

   A semicolon appearing anywhere on a line begins a comment, and the semicolon
   and all following characters are ignored.  Zero-length lines, including lines
   beginning with a semicolon, are ignored.

   Note that a line containing one or more blanks is not a zero-length line, so,
   for example, the line " ; a comment starting in column 2" is not ignored.

   The first (non-ignored) line in the file is a VFU definition line of this
   exact form:

     VFU=<punch characters>,<no-punch character>{,<title>}

   ...where:

     Parameter           Description
     ------------------  -------------------------------------------------------
     punch characters    a string of one or more characters used interchangeably
                         to represent a punched location

     no-punch character  a single character representing a non-punched location

     title               an optional descriptive string printed by the SHOW LP
                         VFU command ("Custom VFU" is used by default)

   If the "VFU" line is missing or not of the correct form, then "Format error"
   status is returned, and the VFU tape is not changed.

   The remaining (non-ignored) lines define the channels punched for each line
   of the printed form.  The line format consists of a sequence of punch,
   no-punch, and "other" characters in channel order.  Each punch or no-punch
   character defines a channel state, starting with channel 1 and proceeding
   left-to-right until all channels for the VFU are defined; if the line
   terminates before all channels are defined, the remaining channels are set to
   the no-punch state.  Any "other" characters (i.e., neither a punch character
   nor a no-punch character) are ignored and may be used freely to delineate the
   tape channels.

   Examples using the standard 66-line tape definition for an 8-channel VFU:


     ; the VFU definition                 |   VFU=1234567,  ; no-punch is a ' '
     VFU=1,0,a binary tape image          |
                                          |   1             ; top of form
     ; the channel definitions            |     345         ; form line 1
                                          |     3           ; form line 2
     10111111   ; top of form             |     34          ; form line 3
     00100000   ; single space            |     3 5         ; form line 4
     0011       ; channels 5-8 no-punch   |     34          ; form line 5
                                          |
     -------------------------------------+-------------------------------------
                                          |
     VFU=X,-,blanks are "others"          |   VFU=TO,.,brackets are "others"
                                          |   ; 1   2   3   4   5   6   7   8
     X  -  X  X  X  X  X  X   ; line 1    |   ;--- --- --- --- --- --- --- ---
     -  -  X  -  -  -  -  -   ; line 2    |    [T]  .  [O] [O] [O] [O] [O] [O]
     -  -  X  X  -  -  -  -   ; line 3    |     .   .  [O]  .   .   .   .   .
                                          |     .   .  [O] [O]  .   .   .   .


   On entry, the "vf" parameter determines whether the standard tape or a custom
   tape is to be loaded.  If "vf" is NULL, a standard 66-line tape is generated
   and stored in the tape buffer.  Otherwise, a custom tape file is read and
   parsed, and the assembled VFU entries are stored in the tape buffer.  After
   generation or a successful tape load, the buffer is copied to the VFU array,
   the form length is set, the current line is reset to the top-of-form, and the
   state of VFU channels 9 and 12 are set into the device status.


   Implementation notes:

    1. VFU array elements 1-n correspond to form print lines 1-n.  Element 0 is
       the logical OR of all of the other elements and is used during VFU format
       command processing to determine if a punch is present somewhere in a
       given channel.
*/

static t_stat load_vfu (uint32 lpa_id, UNIT *uptr, FILE *vf)
{
LPA_STATE *const   lpa   = lpas [lpa_id].sptr;          /* the pointer to the channel state */
const PRINTER_TYPE model = MODEL (uptr->flags);         /* the printer model number */
uint32             line, channel;
int32              len;
char               buffer [LINE_SIZE], punch [LINE_SIZE], no_punch;
char               *bptr, *tptr;
uint32             tape [VFU_SIZE] = { 0 };

if (vf == NULL) {                                       /* if the standard VFU is requested */
    tape [ 1] = VFU_CHANNEL_1;                          /*   then punch channel 1 for the top of form */
    tape [60] = VFU_CHANNEL_2 | VFU_CHANNEL_9;          /*     and channels 2 and 9 for the bottom of form */

    for (line = 1; line <= 60; line++) {                            /* load each of the 60 printable lines */
        tape [line] |= VFU_CHANNEL_3                                /* punch channel 3 for single space */
                         | (line %  2 ==  1 ? VFU_CHANNEL_4 : 0)    /* punch channel 4 for double space */
                         | (line %  3 ==  1 ? VFU_CHANNEL_5 : 0)    /* punch channel 5 for triple space */
                         | (line % 30 ==  1 ? VFU_CHANNEL_6 : 0)    /* punch channel 6 for next half page */
                         | (line % 15 ==  1 ? VFU_CHANNEL_7 : 0)    /* punch channel 7 for next quarter page */
                         | (line % 10 ==  1 ? VFU_CHANNEL_8 : 0);   /* punch channel 8 for next sixth page */

        tape [0] |= tape [line];                        /* accumulate the channel punches */
        }

    lpa->form_length = 66;                              /* set the form length */
    strcpy (lpa->vfu_title, "Standard VFU");            /*   and set the title */
    }

else {                                                  /* otherwise load a custom VFU from the file */
    len = read_vfu_line (vf, buffer, sizeof buffer);    /*  starting with the first line */

    if (len <= 0                                            /* if there isn't one */
      || strncmp (buffer, "VFU=", strlen ("VFU=")) != 0) {  /*   or it's not a VFU definition statement */
        cputs ("Missing VFU definition line\n");            /*     then complain to the console */
        return SCPE_FMT;                                    /*       and fail with a format error */
        }

    bptr = buffer + strlen ("VFU=");                    /* point at the first control argument */

    tptr = strtok (bptr, ",");                          /* parse the punch token */

    if (tptr == NULL) {                                             /* if it's missing */
        cputs ("Missing punch field in the VFU control line\n");    /*   then complain to the console */
        return SCPE_FMT;                                            /*     and fail with a format error */
        }

    strcpy (punch, tptr);                               /* save the set of punch characters */

    tptr = strtok (NULL, ",");                          /* parse the no-punch token */

    if (tptr == NULL){                                              /* if it's missing */
        cputs ("Missing no-punch field in the VFU control line\n"); /*   then complain to the console */
        return SCPE_FMT;                                            /*     and fail with a format error */
        }

    no_punch = *tptr;                                   /* save the no-punch character */

    tptr = strtok (NULL, ",");                          /* parse the optional title */

    if (tptr != NULL)                                   /* if it's present */
        strcpy (lpa->vfu_title, tptr);                  /*   then save the user's title */
    else                                                /* otherwise */
        strcpy (lpa->vfu_title, "Custom VFU");          /*   use a generic title */


    for (line = 1; line <= VFU_MAX; line++) {               /* load up to the maximum VFU tape length */
        len = read_vfu_line (vf, buffer, sizeof buffer);    /* read a tape definition line */

        if (len <= 0)                                   /* if at the EOF or an error occurred */
            break;                                      /*   then the load is complete */

        tptr = buffer;                                  /* point at the first character of the line */

        channel = VFU_CHANNEL_1;                        /* set the channel 1 indicator */

        while (channel != 0 && *tptr != '\0') {         /* loop until the channel or definition is exhausted */
            if (strchr (punch, *tptr) != NULL) {        /* if the character is in the punch set */
                tape [line] |= channel;                 /*   then punch the current channel */
                channel = channel >> 1;                 /*     and move to the next one */
                }

            else if (*tptr == no_punch)                 /* otherwise if the character is the no-punch character */
                channel = channel >> 1;                 /*   then move to the next channel */

            tptr++;                                     /* otherwise the character is neither, so ignore it */
            }

        tape [0] |= tape [line];                        /* accumulate the channel punches */
        }


    if (not (tape [1] & VFU_CHANNEL_1)) {                   /* if there is no channel 1 punch in the first line */
        cputs ("Missing punch in channel 1 of line 1\n");   /*   then complain to the console */
        return SCPE_FMT;                                    /*     and fail with a format error */
        }

    lpa->form_length = line - 1;                        /* set the form length */
    }

memcpy (lpa->VFU, tape, sizeof lpa->VFU);               /* copy the tape buffer to the VFU array */

lpa->current_line = 1;                                  /* reset the line counter to the top of the form */

if (print_props [model].vfu_channels > 8) {             /* if the printer VFU has more than 8 channels */
    if (lpa->VFU [lpa->current_line] & VFU_CHANNEL_9)   /*   then if channel 9 is punched for this line */
        lpa->status |= ST_VFU_9;                        /*     then report it in the device status */
    else                                                /*   otherwise */
        lpa->status &= ~ST_VFU_9;                       /*     remove it from the reported status */

    if (lpa->VFU [lpa->current_line] & VFU_CHANNEL_12)  /* if channel 12 is punched for this line */
        lpa->status |= ST_VFU_12;                       /*   then report it in the device status */
    else                                                /* otherwise */
        lpa->status &= ~ST_VFU_12;                      /*   remove it from the reported status */
    }

return SCPE_OK;                                         /* the VFU was successfully loaded */
}


/* Read a line from the VFU file.

   This routine reads a line from the VFU tape image file designated by the file
   stream parameter "vf", stores the data in the string buffer pointed to by
   "line" and whose size is given by "size", and returns the length of that
   string.  Comments are stripped from lines that are read, and the routine
   continues to read until a non-zero-length line is found.  If the end of the
   file was reached, the return value is 0.  If a file error occurred, the
   return value is -1.


   Implementation notes:

    1. The routine assumes that the file was opened in text mode, so that
       automatic CRLF-to-LF conversion is done if needed.  This simplifies
       the end-of-line removal.
*/

static int32 read_vfu_line (FILE *vf, char *line, uint32 size)
{
char  *result;
int32 len = 0;

while (len == 0) {
    result = fgets (line, size, vf);                    /* get the next line from the file */

    if (result == NULL)                                 /* if an error occurred */
        if (feof (vf))                                  /*   then if the end of file was seen */
            return 0;                                   /*     then return an EOF indication */

        else {                                          /*   otherwise */
            report_error (vf);                          /*     report the error to the console */
            return -1;                                  /*       and return an error indication */
            }

    len = strlen (line);                                /* get the current line length */

    if (len > 0 && line [len - 1] == '\n')              /* if the last character is a newline */
        line [--len] = '\0';                            /*   then remove it and decrease the length */

    result = strchr (line, ';');                        /* search for a comment indicator */

    if (result != NULL) {                               /* if one was found */
        *result = '\0';                                 /*   then truncate the line at that point */
        len = (int32) (result - line);                  /*     and recalculate the line length */
        }
    }

return len;
}


/* Report a stream I/O error to the console.

   If a stream I/O error has been detected, this routine will print an error
   message to the simulation console and clear the stream's error indicator.
*/

static void report_error (FILE *stream)
{
cprintf ("%s simulator printer I/O error: %s\n",        /* report the error to the console */
         sim_name, strerror (errno));

clearerr (stream);                                      /* clear the error */

return;
}


/* Trace a state transition.

   This routine is called to trace a bus controller state transition.  On entry,
   "dptr" points at the device, "address" is the device bus address, and
   "entry_state" and "exit_state" are the starting and ending states,
   respectively.

   Tracing must be active when this routine is called.
*/

static void trace_state (DEVICE *dptr, uint32 address, CN_STATE entry_state, CN_STATE exit_state)
{
hp_trace (dptr, TRACE_STATE, "Device %u transitioned from %s state to %s state\n",
          address, state_names [entry_state], state_names [exit_state]);
return;
}
