/* hp3000_gic_ma.c: HP 3000 7970E "Amigo" Magnetic Tape Drive simulator

   Copyright (c) 2024, J. David Bryan

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
   AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
   ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
   WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of the author shall not be used
   in advertising or otherwise to promote the sale, use or other dealings in
   this Software without prior written authorization from the author.

   MA           7970E "Amigo" Magnetic Tape Drive simulator

   31-May-24    JDB     Corrected poll response masking during initialization
   16-Apr-24    JDB     First release
   22-Sep-23    JDB     Passes the D7970Sxx Magnetic Tape Diagnostic
   22-Aug-23    JDB     Created

   References:
     - 7970E HP-IB Interface Manual Programming and Service
         (07970-90919, November 1982)
     - Magnetic Tape Diagnostic Manual
         (30070-90015, March 1982)
     - IEEE Standard Digital Interface for Programmable Instrumentation
         (IEEE-488A-1980, September 1979)


   The HP-IB interface for the HP 7970E 1/2-inch magnetic tape drive permits the
   connection of one to four drives to the HP Interface Bus.  This is the
   required connection method when using the drive with IMB-based HP 3000
   computer systems.  The interface installs in the chassis of the master drive
   and replaces the HP 30215 tape controller and parallel interface that is used
   with SIO-based 3000s such as the Series II and III.  The master drive could
   be ordered with an HP-IB interface by specifying Option 436, or a
   parallel-interface drive could be retrofitted with an HP-IB interface by
   ordering HP part number 26072A.

   Designed originally for connection to the HP 300 ("Amigo") computer, the
   command set was expanded to include several commands related to HP-IB
   operation.  They include the Amigo Identify and Amigo Clear sequences, Read
   and Write Loopback channel tests, and Device Specified Jump commands.  The
   interface also includes a 128-byte FIFO buffer between the drives and the bus
   to enable more efficient use of the General I/O Channel (GIC) to which it is
   attached.  The interface occupies a single bus address, with each drive
   having a user-selectable unit number from 0-3.

   The interface responds to the following commands.  In the following table,
   the bus addressing state is indicated by U [untalk], L [listen], and T
   [talk], and the secondary and opcode numeric values are in hex.

   The commands that are implemented are:

     Bus  Sec  Opc  Operation
     ---  ---  ---  ------------------------------------------------
      U   MSA   -   Amigo Identify
      U    -   14   Universal Device Clear (primary command)
      L    -   04   Selected Device Clear (primary command)

      L   00    -   Write Data
      L   01    -   Tape Command
      L   01   01   Select Unit 0
      L   01   02   Select Unit 1
      L   01   03   Select Unit 2
      L   01   04   Select Unit 3
      L   01   05   Write Record
      L   01   06   Write File Mark
      L   01   07   Write Gap
      L   01   08   Read Record
      L   01   09   Forward Space Record
      L   01   0A   Backspace Record
      L   01   0B   Forward Space File
      L   01   0C   Backspace File
      L   01   0D   Rewind
      L   01   0E   Rewind/Offline
      L   01   0F   Read Record Backward

      L   07    -   End Command
      L   07   01   Clear Poll Response
      L   07   02   Inhibit Poll Response for Read Bursts
      L   07   04   Enable Rewind Completion Response or Auto-Select
      L   07   10   Clear DSJ
      L   07   20   Abort Current Tape Operation
      L   10   **   Amigo Clear
      L   1E    -   Write Loopback

      T   00    -   Read Data
      T   01    -   Read Status
      T   02    -   Read Byte Count
      T   10    -   Device Specified Jump
      T   1E    -   Read Loopback

   The Universal Device Clear and the Selected Device Clear initialize the
   interface to its power-up state.  The Amigo Clear command and its parameter
   are accepted but are otherwise ignored.

   Tape read and write record commands supply or accept data in 64-byte bursts
   and respond with parallel polls after each burst.  If the host performs the
   read or write in record mode, it should ignore the per-burst poll responses
   that are generated.  For a record longer than 64 bytes, the End command with
   the Clear Poll Response option should follow a record-mode transfer, as the
   interface will generate one additional parallel poll response when the
   command is complete.

   The HP-IB bus transaction sequence for each of the commands is listed below:

     Identify
     --------

       ATN  UNT     Untalk
       ATN  MSA     My secondary address
            DAB     ID data byte 01H
       EOI  DAB     ID data byte 83H
       ATN  OTA     Talk 30


     Universal Device Clear
     ----------------------

       ATN  PCG     Primary command 14H
            ppe     Parallel poll enabled when clear completes


     Selected Device Clear
     ---------------------

       ATN  MLA     My listen address
       ATN  PCG     Primary command 04H
            ppe     Parallel poll enabled when clear completes
       ATN  UNL     Unlisten


     Amigo Clear
     -----------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 10H (Amigo Clear)
            ppd     Parallel poll disabled
       EOI  DAB     Unused data byte
       ATN  SDC     Selected device clear
            ppe     Parallel poll enabled when clear completes
       ATN  UNL     Unlisten


     Select Unit
     -----------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 01H
            ppd     Parallel poll disabled
       EOI  DAB     Opcode 01H-04H
       ATN  UNL     Unlisten
            ppe     Parallel poll enabled


     Write File Mark, Write Gap
     --------------------------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 01H
            ppd     Parallel poll disabled
       EOI  DAB     Opcode 06H, 07H
       ATN  UNL     Unlisten
            ...     (writing)
            ppe     Parallel poll enabled


     Forward Space Record, Backspace Record, Forward Space File, Backspace File
     --------------------------------------------------------------------------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 01H
            ppd     Parallel poll disabled
       EOI  DAB     Opcode 09H-0CH
       ATN  UNL     Unlisten
            ...     (spacing)
            ppe     Parallel poll enabled


     Rewind
     ------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 01H
            ppd     Parallel poll disabled
       EOI  DAB     Opcode 0DH
       ATN  UNL     Unlisten
            ...     (rewind started)
            ppe     Parallel poll enabled


     Rewind and Go Offline
     ---------------------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 01H
            ppd     Parallel poll disabled
       EOI  DAB     Opcode 0EH
       ATN  UNL     Unlisten
            ...     (rewind started)


     Read Record, Read Record Backward
     ---------------------------------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 01H
            ppd     Parallel poll disabled
       EOI  DAB     Opcode 08H, 0FH
       ATN  UNL     Unlisten
            ppe     Parallel poll enabled

       ATN  MTA     My talk address                                    <--+
       ATN  SCG     Secondary command 10H (DSJ)                           |
            ppd     Parallel poll disabled                                |
       EOI  DAB     Reporting byte                                        |
       ATN  UNT     Untalk                                                |
                                                                          |
       ATN  MTA     My talk address                                       |
       ATN  SCG     Secondary command 00H (Read Data)                     |
            DAB     Read data byte #1                                     |
            ...     (more data bytes)                                     |
       EOI  DAB     Read data byte #n                                     |
            ppe     Parallel poll enabled                                 |
       ATN  UNT     Untalk                                                |
                                                                          |
       [ burst mode delivers <= 64 bytes with EOI only on last burst ] ---+

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 07H (End command)
       EOI  DAB     Opcode 01H (Clear Poll Responses)
            ppd     Parallel poll disabled
       ATN  UNL     Unlisten
            ...     (tape postamble read)
            ppe     Parallel poll enabled

       ATN  MTA     My talk address
       ATN  SCG     Secondary command 10H (DSJ)
            ppd     Parallel poll disabled
       EOI  DAB     Reporting byte
       ATN  UNT     Untalk


     Write Record
     ------------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 01H
            ppd     Parallel poll disabled
       EOI  DAB     Opcode 05H
       ATN  UNL     Unlisten
            ppe     Parallel poll enabled

       ATN  MTA     My talk address                                    <--+
       ATN  SCG     Secondary command 10H (DSJ)                           |
            ppd     Parallel poll disabled                                |
       EOI  DAB     Reporting byte                                        |
       ATN  UNT     Untalk                                                |
                                                                          |
       ATN  MLA     My listen address                                     |
       ATN  SCG     Secondary command 00H (Write Data)                    |
            DAB     Write data byte #1                                    |
            ...     (more data bytes)                                     |
       EOI  DAB     Write data byte #n                                    |
            ppe     Parallel poll enabled                                 |
       ATN  UNL     Unlisten                                              |
                                                                          |
       [ burst mode delivers <= 64 bytes with EOI only on last burst ] ---+

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 07H (End command)
       EOI  DAB     Opcode 01H (Clear Poll Responses)
            ppd     Parallel poll disabled
       ATN  UNL     Unlisten
            ...     (tape postamble written)
            ppe     Parallel poll enabled

       ATN  MTA     My talk address
       ATN  SCG     Secondary command 10H (DSJ)
            ppd     Parallel poll disabled
       EOI  DAB     Reporting byte
       ATN  UNT     Untalk


     End Command
     -----------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 07H
       EOI  DAB     Opcode byte
            ppd     Parallel poll disabled
       ATN  UNL     Unlisten


     Device Specified Jump
     ---------------------

       ATN  MTA     My talk address
       ATN  SCG     Secondary command 10H
            ppd     Parallel poll disabled
       EOI  DAB     Reporting byte
       ATN  UNT     Untalk


     Write Loopback and Read Loopback
     --------------------------------

       ATN  MLA     My listen address
       ATN  SCG     Secondary command 1EH
            ppd     Parallel poll disabled
            DAB     Loopback data byte #1
            ...
       EOI  DAB     Loopback data byte #256
            ppe     Parallel poll enabled
       ATN  UNL     Unlisten

       ATN  MTA     My talk address
       ATN  SCG     Secondary command 1EH
            ppd     Parallel poll disabled
            DAB     Loopback data byte #1
            ...
       EOI  DAB     Loopback data byte #129
            ppe     Parallel poll enabled
       ATN  UNT     Untalk

   For Write Loopback, the interface accepts 256 bytes but retains only the
   first 128 bytes.  The Read Loopback will return the first 128 bytes and then
   return a byte of value 177 octal as the 129th byte.


   The Read Status command returns three status bytes in the following format:

       0   1 | 2   3   4 | 5   6   7
     +---+---+---+---+---+---+---+---+
     | F | L | T | S | C | N | M | O | Byte 1
     +---+---+---+---+---+---+---+---+
     | - | unit  | D | R | W | U | I | Byte 2
     +---+---+---+---+---+---+---+---+
     | - | - | P | E |    online     | Byte 3
     +---+---+---+---+---+---+---+---+

   Where:

     F = End of File
     L = Load Point
     T = End of Tape
     S = Single-Track Error
     C = Command Rejected
     N = No Write Ring
     M = Multiple-Track Error
     O = Online

     D = Data Error
     R = Tape Runaway
     W = Rewinding
     U = Tape Unit Busy
     I = Interface Busy

     P = Power Restored
     E = Command Parity Error

   Unit:

     00 = reporting unit 0
     01 = reporting unit 1
     10 = reporting unit 2
     11 = reporting unit 3

   Online:

     ---1 = unit 0 placed online
     --1- = unit 1 placed online
     -1-- = unit 2 placed online
     1--- = unit 3 placed online

   In simulation, each interface instance provides up to four connected tape
   drives assigned to a given GIC channel number and responding to a
   user-selectable bus address from 0-7.  Multiple instances may reside on the
   same GIC if they are assigned different addresses.

   This module simulates the bus controller.  The tape controller and drive
   simulations are provided by the HP magnetic tape controller simulator library
   (hp_tapelib).  Rather than simulating the signal interaction specific to
   these two cards, the HP tape library simulates an abstract controller having
   an electrical interface modelled on the HP 13037 disc controller.  The bus
   interface and tape controller interact via 16-bit data, flag, and function
   buses.  Commands, status, and data are exchanged across the data bus, with
   the flag bus providing indications of the state of the interface and the
   function bus indicating what actions the interface must take in response to
   command processing by the controller.

   While the interface and controller are idle, a drive unit that changes from
   Not Ready to Ready status will assert a parallel poll response.  This occurs
   when an offline drive is put online (e.g., after mounting a tape) and when a
   rewinding drive completes the action and is repositioned at the load point.

   The simulator provides five SIMH units for use by the tape controller
   library.  Units 0-3 correspond to the four tape drives, while unit 4
   corresponds to the tape controller.  A sixth unit, unit 5, simulates the
   action of the bus controller.

   By default, each tape drive unit's address corresponds to its SIMH unit
   number.  Disabling a unit is equivalent to disconnecting that drive's cables
   from the interface.

   All commands except Tape Command are executed by the bus controller.  The
   controller has nine possible states, as follows:

     Command_Ready     -- Waiting for a new command to be received.  This is the
                          idle state for the bus controller.

     Parameter_Wait    -- Waiting for a command parameter to be received.  For
                          example, Tape Command is followed by a parameter byte
                          identifying the specific command, and End Command is
                          followed by an option byte specifying the actions to
                          take.

     Command_Send      -- Waiting to send a command byte, for instance the
                          status bytes for the Read Status command.  The bus ATN
                          and NRFD signals must be denied before a byte is
                          sourced.

     Command_Receive   -- Waiting to receive a command byte, for example the
                          bytes of a Write Loopback command.  When a data byte
                          is accepted, the simulator asserts NRFD to hold off
                          the talker until the bus service is entered in this
                          state.

     Execution_Send    -- Waiting to send a data byte for a Read Data tape
                          command.  The same bus conditions listed in the
                          Command_Send state are required before data is
                          sourced.

     Execution_Receive -- Waiting to receive a data byte for a Write Data tape
                          command.  As in the Command_Receive state, NRFD is
                          asserted to hold off the talker until the bus event is
                          serviced.

     Completion_Wait   -- Waiting for a command to complete before optionally
                          asserting a parallel poll response.  The bus
                          controller enters this state as soon as the tape
                          controller has accepted a tape command; this frees the
                          bus controller to accept another command while the
                          tape operation continues.  Commands that must wait for
                          the conclusion of tape operations defer their poll
                          responses until the tape commands complete.

     Error_Source      -- Waiting to send an error byte in response to an
                          invalid or undefined talk secondary.  If an illegal
                          secondary is received, such as a Read Data secondary
                          following a Write Data command, entry in the bus
                          controller event service sources a single byte of
                          value 1 with EOI to satisfy the inbound transfer.

     Error_Sink        -- Waiting to receive an error byte in response to an
                          invalid or undefined listen secondary.  If such a
                          secondary is received, the bus controller event
                          service will receive and discard as many bytes as the
                          host sends until one tagged with EOI is received.

   The bus controller initiates all commands.  For bus commands such as Amigo
   Identify or Device Specified Jump, only the bus controller unit is scheduled.
   For tape commands that do not involve data transfers, such as Rewind or
   Forward Space File, the bus controller passes the command to the tape
   controller, which then schedules the selected tape unit to execute the
   command.

   Read Data and Write Data commands are passed from the bus controller to the
   tape controller, which schedules the initial processing wait on the tape
   controller unit and initial tape motion on the selected tape unit.  Then
   these two units operate in parallel to transfer data between the tape image
   and the HP-IB.  For a read operation, the tape drive unit is scheduled to
   fill the FIFO buffer, while the tape controller unit is scheduled to empty
   the buffer and pass data bytes back to the bus.  For a write operation, the
   roles are reversed; the tape controller unit fills the FIFO, while the tape
   drive unit empties it.  Once the data transfer completes, the tape drive unit
   handles the cessation of tape motion and the completion of the tape command.

   Because the command protocol separates the tape command, such as Write
   Record, from the data transfer, and because each involves the receipt of its
   own command secondary, the potential exists for the secondaries to be
   mismatched.  For instance, a Write Record command could be followed by a Read
   Data secondary, or vice versa.  Or it could be followed by some unknown talk
   or listen secondary.  The controller must be able to recover from these in a
   way that allows the channel program to complete -- presumably to follow with
   a DSJ = 1 return and a Request Status command that will show a Command Reject
   error.

   If a secondary error occurs, the tape controller continues to operate,
   loading the FIFO with dummy data or unloading and discarding tape data, while
   the bus controller is scheduled to source or sink the bytes that accompany
   the invalid secondary.

   The simulation provides both realistic and optimized ("fast") timing modes.
   The setting determines the scheduling delays for the SIMH units representing
   the tape drives as well as the tape and bus controllers.


   Implementation notes:

    1. The HP-IB controller exists is both "old" and "new" versions.  An earlier
       revision of the interface manual dated September 1978 does not show the
       Abort Current Tape Operation bit (bit 6) option of the End command.  The
       tape diagnostic performs a test to determine which controller is being
       used.  If it detects an old controller, it reports, "Note: This
       controller is suitable for Series 30/33 mainframes only" and skips steps
       221 (Read Record Backward and detect Tape Runaway), 229 (Read Record and
       detect Tape Runaway), and 305 (End with Abort Current Tape Operation).

       The test writes a record of 200 bytes using bursts of 63 bytes.  It
       expects an old controller to succeed with DSJ = 0 and a new controller to
       fail with DSJ = 1 and a Data Timing Error.  The simulator implements the
       new controller behavior.

    2. The intent of Burst Mode is to free up the bus after a relatively fast
       transfer between a device buffer and main memory while the buffer empties
       or refills relatively slowly.  Burst mode is provided by the 7970E HP-IB
       tape drive and was used by the MPE driver (module 35, HIOTAPE0).  Other
       tape drives, as well as all disc drives, use Record Mode exclusively.
       However, an update to the 7970 driver changed it to use Record Mode as
       well, with this comment: "This was done to prevent tape-runaway problems
       due to CPP going away for periods of time longer than the tape could
       stand."  Although the simulator implements Burst Mode handling, it is
       only used by the tape diagnostic.
*/



#include "hp3000_defs.h"
#include "hp3000_io.h"
#include "hp3000_imb.h"                         /* must be included after hp3000_io.h */
#include "hp3000_gic.h"
#include "hp_tapelib.h"

#include "sim_tape.h"



/* Program constants */

#define CHANNEL_COUNT       (CHAN_MAX + 1)      /* count of IMB channels */

#define MAX_MA_COUNT        (CHANNEL_COUNT - 2) /* count of potential MA instances */

#define MA_AUXUNITS         1                   /* the number of HP-IB controller units */

#define DRIVE_COUNT         (TL_MAXDRIVE + 1)           /* the number of tape drive units */
#define TAPE_COUNT          (DRIVE_COUNT + TL_AUXUNITS) /* the total number of tape library units */
#define UNIT_COUNT          (TAPE_COUNT + MA_AUXUNITS)  /* the total number of units */

#define TAPE_CNTLR_UPTR     (&dptr->units [TL_CNTLR_UNIT])  /* the tape controller unit pointer */
#define HPIB_CNTLR_UPTR     (&dptr->units [TL_HPIB_UNIT])   /* the HP-IB controller unit pointer */

#define P_SWITCH            SWMASK ('P')        /* perform a power-on reset */


/* Device flags and accessors.

   DEV_V_UF + 15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
             +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
             | - | - | - | - | - | - | - | - |  address  |    channel    | R |
             +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     R = realtime mode is enabled


#define DEV_REALTIME        DF_BIT (0)          * defined by the HP tape library *
*/

#define DEV_ADDRESS         DF_BITS (7, 5)      /* controller HP-IB address */
#define DEV_CHANNEL         DF_BITS (4, 1)      /* GIC channel address */

#define ADDRESS(w)          DF_BITS_TO (7, 5, w)

#define CHANNEL(w)          DF_BITS_TO (4, 1, w)
#define TO_CHANNEL(w)       TO_DF_BITS (4, 1, w)


/* Unit flags and accessors.

   UNIT_V_UF + 15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
   MTUF_V_UF + 10   9   8   7   6   5   4   3   2   1   0
              +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
              | - | - | - | R | O |   unit    |  model    |    (reserved)     |
              +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     R     = the unit is rewinding
     O     = the unit is online
     unit  = the unit selector setting
     model = the DRIVE_TYPE enumeration constant for the simulated drive

   The lower flag bits are reserved for use by the SIMH magnetic tape library.


#define UNIT_MODEL          MT_BITS (2, 0)      * defined by the HP tape library *
#define UNIT_UNIT           MT_BITS (5, 3)      * defined by the HP tape library *
#define UNIT_ONLINE         MT_BIT  (6)         * defined by the HP tape library *
#define UNIT_REWINDING      MT_BIT  (7)         * defined by the HP tape library *
*/

#define Instance            buf                 /* unit field for the instance ID of the device */


/* HP-IB interface commands */

typedef enum {                                  /* HP-IB interface commands */
    Invalid = 0,                                /*   invalid command (default for reset) */
    Write_Data,                                 /*   MLA 00 */
    Tape_Command,                               /*   MLA 01 */
    End_Command,                                /*   MLA 07 */
    Amigo_Clear,                                /*   MLA 10 */
    Write_Loopback,                             /*   MLA 1E */
    Read_Data,                                  /*   MTA 00 */
    Read_Status,                                /*   MTA 01 */
    Read_Byte_Count,                            /*   MTA 02 */
    Device_Specified_Jump,                      /*   MTA 10 */
    Read_Loopback,                              /*   MTA 1E */
    Amigo_Identify                              /*   UNT MSA */
    } IF_COMMAND;

/* HP-IB interface command names */

static const char * const if_command_names [] = {   /* command names, indexed by IF_COMMAND */
    "invalid",
    "Write Data",
    "Tape",
    "End",
    "Amigo Clear",
    "Write Loopback",
    "Read Data",
    "Read Status",
    "Read Byte Count",
    "Device Specified Jump",
    "Read Loopback",
    "Amigo Identify"
    };


/* HP-IB interface state declarations.

   The bus controller states determine the actions performed by the bus event
   service routine.  All commands begin and end in the Command_Ready state.  The
   intermediate state transitions for each command are as follows:

     Command                    State Transitions
     -------------------------  ------------------------------------
     Amigo Identify             Command_Send
     Universal Device Clear     (bus controller not scheduled)
     Selected Device Clear      (bus controller not scheduled)
     Amigo Clear                (bus controller not scheduled)

     Tape Command               Parameter_Wait ->
       Select Unit 0              Completion_Wait
       Select Unit 1              Completion_Wait
       Select Unit 2              Completion_Wait
       Select Unit 3              Completion_Wait
       Write Record               Completion_Wait
       Write File Mark            Completion_Wait
       Write Gap                  Completion_Wait
       Read Record                Completion_Wait
       Forward Space Record       Completion_Wait
       Backspace Record           Completion_Wait
       Forward Space File         Completion_Wait
       Backspace File             Completion_Wait
       Rewind                     Completion_Wait
       Rewind/Offline             Completion_Wait
       Read Record Backward       Completion_Wait
     End Command                Parameter_Wait ->
       Clear Poll Response        Completion_Wait
       Inhibit Poll Response      Completion_Wait
       Enable Rew Compl/Select    Completion_Wait
       Clear DSJ                  Completion_Wait
       Abort Current Tape Op      Completion_Wait

     Read Data                  Execution_Send -> Completion_Wait
     Write Data                 Execution_Receive -> Completion_Wait
     Read Status                Command_Send -> Completion_Wait
     Read Byte Count            Command_Send -> Completion_Wait
     Device Specified Jump      Command_Send -> Completion_Wait
     Read Loopback              Command_Send -> Completion_Wait
     Write Loopback             Command_Receive -> Completion_Wait
*/

typedef enum {                                  /* controller state */
    Command_Ready = 0,                          /*   waiting for a command message secondary */
    Parameter_Wait,                             /*   waiting for a parameter reception */
    Command_Send,                               /*   sending command status */
    Command_Receive,                            /*   receiving command data */
    Execution_Send,                             /*   sending read data or status */
    Execution_Receive,                          /*   receiving write data */
    Completion_Wait,                            /*   waiting for command completion */
    Error_Source,                               /*   sending bytes for error recovery */
    Error_Sink                                  /*   receiving bytes for error recovery */
    } IF_STATE;

/* HP-IB interface state names */

static const char * const if_state_names [] = { /* state names, indexed by IF_STATE */
    "Command Ready",
    "Parameter Wait",
    "Command Send",
    "Command Receive",
    "Execution Send",
    "Execution Receive",
    "Completion Wait",
    "Error Source",
    "Error Sink"
    };


/* HP-IB interface command properties */

typedef struct {                                /* command properties */
    IF_STATE  next_state;                       /*   the next command state */
    t_bool    completion_poll;                  /*   TRUE if the command ends with a parallel poll response */
    t_bool    use_controller;                   /*   TRUE if the command uses the controller unit */
    } CMD_PROPERTIES;

static const CMD_PROPERTIES cmd_props [] = {    /* command properties, indexed by IF_COMMAND */
/*                        Compl    Use   */
/*    Next State           Poll   Cntlr  */
/*    ------------------  ------  ------ */
    { Command_Ready,      TRUE,   FALSE  },     /*   invalid command */
    { Execution_Receive,  FALSE,  FALSE  },     /*   Write Data */
    { Parameter_Wait,     FALSE,  TRUE   },     /*   Tape Command */
    { Parameter_Wait,     FALSE,  TRUE   },     /*   End Command */
    { Parameter_Wait,     TRUE,   FALSE  },     /*   Amigo Clear */
    { Command_Receive,    TRUE,   TRUE   },     /*   Write Loopback */
    { Execution_Send,     FALSE,  FALSE  },     /*   Read Data */
    { Command_Send,       FALSE,  TRUE   },     /*   Read Status */
    { Command_Send,       FALSE,  TRUE   },     /*   Read Byte Count */
    { Command_Send,       FALSE,  TRUE   },     /*   Device Specified Jump */
    { Command_Send,       TRUE,   TRUE   },     /*   Read Loopback */
    { Command_Send,       FALSE,  TRUE   }      /*   Amigo Identify */
    };


/* End command accessors.

       7   6 | 5   4   3 | 2   1   0
     +---+---+---+---+---+---+---+---+
     | - | - | A | D | - | R | I | P | Parameter
     +---+---+---+---+---+---+---+---+

   Where:

     A = Abort Current Tape Operation
     D = Clear DSJ
     R = Enable Rewind Completion Response or Auto-Select
     I = Inhibit Poll Response for Read Bursts
     P = Clear Poll Response
*/

typedef enum {                                  /* End command options */
    End_Abort           = BIT (5),              /*   (A) Abort Current Tape Operation */
    End_Clear_DSJ       = BIT (4),              /*   (D) Clear DSJ */
    End_Enable_Response = BIT (2),              /*   (R) Enable Rewind Completion Response or Auto-Select */
    End_Inhibit_Poll    = BIT (1),              /*   (I) Inhibit Poll Response for Read Bursts */
    End_Clear_Poll      = BIT (0)               /*   (P) Clear Poll Response */
    } EN_COMMANDS;

static const BITSET_NAME end_names [] = {       /* End command option names */
    "abort",                                    /*   bit  5 */
    "clear DSJ",                                /*   bit  4 */
    NULL,                                       /*   bit  3 */
    "enable rewind/auto-select",                /*   bit  2 */
    "inhibit poll",                             /*   bit  1 */
    "clear poll"                                /*   bit  0 */
    };

static const BITSET_FORMAT end_format =         /* names, offset, direction, alternates, bar */
    { FMT_INIT (end_names, 0, msb_first, no_alt, no_bar) };


/* Status byte accessors.

       7   6 | 5   4   3 | 2   1   0
     +---+---+---+---+---+---+---+---+
     | F | L | T | S | C | N | M | O | Byte 1
     +---+---+---+---+---+---+---+---+
     | - | unit  | D | R | W | U | I | Byte 2
     +---+---+---+---+---+---+---+---+
     | - | - | P | E |    online     | Byte 3
     +---+---+---+---+---+---+---+---+

   Where:

     F = End of File
     L = Load Point
     T = End of Tape
     S = Single-Track Error
     C = Command Rejected
     N = No Write Ring
     M = Multiple-Track Error
     O = Online

     D = Data Timing Error
     R = Tape Runaway
     W = Rewinding
     U = Tape Unit Busy
     I = Interface Busy

     P = Power Restored
     E = Command Parity Error

   Unit:

     00 = reporting unit 0
     01 = reporting unit 1
     10 = reporting unit 2
     11 = reporting unit 3

   Online:

     ---1 = unit 0 placed online
     --1- = unit 1 placed online
     -1-- = unit 2 placed online
     1--- = unit 3 placed online
*/

#define S1_TAPE_MARK        BIT (7)             /* (F) tape mark (supplied by hp_tapelib) */
#define S1_LOAD_POINT       BIT (6)             /* (L) load point (supplied by hp_tapelib) */
#define S1_EOT              BIT (5)             /* (T) end of tape (supplied by hp_tapelib) */
#define S1_PARITY           BIT (4)             /* (S) single-track error (supplied by hp_tapelib) */
#define S1_REJECT           BIT (3)             /* (C) command reject (supplied by hp_tapelib) */
#define S1_PROTECTED        BIT (2)             /* (N) no write ring (supplied by hp_tapelib) */
#define S1_MTE              BIT (1)             /* (M) multiple-track error */
#define S1_READY            BIT (0)             /* (O) unit online (supplied by hp_tapelib) */

#define S2_TIMING           BIT (4)             /* (D) data timing error (supplied by hp_tapelib) */
#define S2_RUNAWAY          BIT (3)             /* (R) tape runaway (supplied by hp_tapelib) */
#define S2_REWINDING        BIT (2)             /* (W) unit is rewinding (supplied by hp_tapelib) */
#define S2_TAPE_BUSY        BIT (1)             /* (U) unit is busy (supplied by hp_tapelib) */
#define S2_IF_BUSY          BIT (0)             /* (I) interface is busy (supplied by hp_tapelib) */

#define S2_UNIT(w)          BITS_TO (6, 5, w)

#define S3_POWER            BIT (5)             /* (P) power restored */
#define S3_PARITY           BIT (4)             /* (E) command parity error */
#define S3_UNIT_3_ONLINE    BIT (3)             /* unit 3 placed online */
#define S3_UNIT_2_ONLINE    BIT (2)             /* unit 2 placed online */
#define S3_UNIT_1_ONLINE    BIT (1)             /* unit 1 placed online */
#define S3_UNIT_0_ONLINE    BIT (0)             /* unit 0 placed online */

/* Status byte formatter declarations */

static const BITSET_NAME status_1_names [] = {  /* Status byte 1 names */
    "tape mark",                                /*   bit  7 */
    "load point",                               /*   bit  6 */
    "end of tape",                              /*   bit  5 */
    "single-track error",                       /*   bit  4 */
    "command rejected",                         /*   bit  3 */
    "no write ring",                            /*   bit  2 */
    "multiple-track error",                     /*   bit  1 */
    "online"                                    /*   bit  0 */
    };

static const BITSET_FORMAT status_1_format =    /* names, offset, direction, alternates, bar */
    { FMT_INIT (status_1_names, 0, msb_first, no_alt, no_bar) };


static const BITSET_NAME status_2_names [] = {  /* Status byte 2 names */
    "data error",                               /*   bit  4 */
    "tape runaway",                             /*   bit  3 */
    "rewinding",                                /*   bit  2 */
    "tape busy",                                /*   bit  1 */
    "interface busy"                            /*   bit  0 */
    };

static const BITSET_FORMAT status_2_format =    /* names, offset, direction, alternates, bar */
    { FMT_INIT (status_2_names, 0, msb_first, no_alt, append_bar) };


static const BITSET_NAME status_3_names [] = {  /* Status byte 3 names */
    "power restored",                           /*   bit  5 */
    "command parity error",                     /*   bit  4 */
    "unit 3 placed online",                     /*   bit  3 */
    "unit 2 placed online",                     /*   bit  2 */
    "unit 1 placed online",                     /*   bit  1 */
    "unit 0 placed online"                      /*   bit  0 */
    };

static const BITSET_FORMAT status_3_format =    /* names, offset, direction, alternates, bar */
    { FMT_INIT (status_3_names, 0, msb_first, no_alt, no_bar) };


/* Tape command code to tape controller opcode translation table */

static const CNTLR_OPCODE to_opcode [] = {      /* opcode translation table (fully decoded) */
    Invalid_Opcode,                             /*   000 --- = reserved */
    Select_Unit_0,                              /*   001 --- = Select Unit 0 */
    Select_Unit_1,                              /*   002 --- = Select Unit 1 */
    Select_Unit_2,                              /*   003 --- = Select Unit 2 */
    Select_Unit_3,                              /*   004 --- = Select Unit 3 */
    Write_Record,                               /*   005 WRR = Write Record */
    Write_File_Mark,                            /*   006 WFM = Write File Mark */
    Write_Gap,                                  /*   007 GAP = Write Gap */
    Read_Record,                                /*   010 RDR = Read Record */
    Forward_Space_Record,                       /*   011 FSR = Forward Space Record */
    Backspace_Record,                           /*   012 BSR = Backspace Record */
    Forward_Space_File,                         /*   013 FSF = Forward Space File */
    Backspace_File,                             /*   014 BSF = Backspace File */
    Rewind,                                     /*   015 REW = Rewind */
    Rewind_Offline,                             /*   016 RST = Rewind and Reset */
    Read_Record_Backward                        /*   017 RRB = Read Record Backward */
    };

#define LAST_OPCODE         017                 /* last valid Tape Command opcode */


/* Bus controller delay properties */

typedef enum {                                  /* bus controller delays */
    Activation_Delay = 0,                       /*   the delay before command activation begins */
    Status_Delay,                               /*   the delay between returned status bytes */
    Data_Delay                                  /*   the delay between returned data bytes */
    } DELAY;

typedef struct {                                /* bus delay structure */
    const char  *name;                          /*   delay name */
    int32       ticks;                          /*   delay in event ticks */
    } BUS_DELAY;

static const BUS_DELAY bus_delays [3] = {       /* bus delays, indexed by DELAY */
    { "activation", uS (4.0) },
    { "status",     uS (2.0) },
    { "data",       uS (1.2) }
    };


/* Tape controller initial optimized delays */

static const DELAY_PROPS initial_times = {
/*     rewind    rewind     rewind       BOT         IR      per-byte     cntlr   */
/*     start    time/inch    stop       time        time     data xfer   overhead */
/*    --------  ---------  --------  -----------  ---------  ----------  -------- */
      uS (10),   uS (1),   uS (10),   uS (10),    uS (200),   uS (5),    uS (10)
      };


/* HP-IB channel state declarations.

   The channel state structure contains the tape library controller, the state
   of the bus and current bus command, a buffer for loopback data and command
   parameters, a bitmap of drive attention requests, and current tape operation
   values.


   Implementation notes:

    1. The tape controller and tape units maintain their own state in the
       CNTLR_VARS structure and unit structure fields, respectively.  These are
       managed by the tape library.

    2. Unit attention occurs for two reasons: a rewind operation has completed,
       or the unit has been placed online.  The HP-IB controller saves these
       separately for each of the four units.  If the Enable Rewind Completion
       Response/Auto-Select command is issued, then a parallel poll response is
       generated for each unit that has either completed rewinding or has been
       placed online.  Multiple attention requests are prioritized, recognizing
       rewind completions for units 0-3 before online transitions for units 0-3.
       In simulation, the attention requests are encoded in the "attention"
       state field as follows:

          15  14  13  12  11  10   9   8   7   6   5   4   3   2   1   0
         +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
         | 0   0   0   0   0   0   0   0 | A | A | A | A | R | R | R | R |
         +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
                                         | 3   2   1   0 | 3   2   1   0 |
                                         |    online     | rewind compl. |

       Applying the LSB_NUMBER macro then obtains the correctly prioritized
       attention request.

    3. The bus interface maintains the power-restored and command-parity-error
       status bits in the "interface_state" field.  These are merged with the
       other bits maintained by the tape library to form the third status byte.

    4. The talker field contains the bit corresponding to the talker address.
       This is analogous to the listeners field, which contains the bits
       corresponding to the addresses of the listeners.  Untalk and Unlisten are
       represented by bit 31 in the fields.

    5. The buffer field is used to hold the 128 data bytes received for a Write
       Loopback command plus the one extra byte added for a Read Loopback
       command.  The last byte is also used to hold the DSJ value, as DSJ is
       executed between the Write Loopback and Read Loopback commands,
*/

#define ONLINE_ATTN(w)      BITS_TO (7, 4, w)

#define TO_ONLINE_ATTN(w)   TO_BITS (7, 4, w)
#define TO_REWIND_ATTN(w)   TO_BITS (3, 0, w)

#define CLEAR_ATTN(u)       (~((BIT (0) | BIT (4)) << (u))) /* bits to clear a unit's attention */

typedef struct {                                /* channel state */
    CNTLR_VARS      controller;                 /*   the tape controller */
    BUS_CABLE       bus;                        /*   the current HP-IB state */
    uint32          present;                    /*   bus address bitmap of devices present on the bus */
    uint32          parity;                     /*   bus address bitmap of parity checking devices */
    uint32          listeners;                  /*   bus address bitmap of the listeners */
    uint32          talker;                     /*   bus address bit of the talker */
    t_bool          active;                     /*   TRUE if a channel transaction is active */
    t_bool          poll_active;                /*   TRUE if a parallel poll is in progress */
    BUS_DATA        response_set;               /*   set of parallel poll responses */
    IF_STATE        state;                      /*   interface state */
    IF_COMMAND      command;                    /*   interface command */
    uint32          parameter;                  /*   interface command parameter */
    uint32          byte_count;                 /*   count of bytes read or written */
    uint32          interface_status;           /*   interface status (power and parity) */
    uint32          dsj;                        /*   the Device Specified Jump value */
    uint32          attention;                  /*   the drive attention bitmap */
    t_bool          auto_select;                /*   TRUE if rewind completion or auto selection is enabled */
    uint8           buffer [129];               /*   the command data-return buffer */
    uint32          index;                      /*   current index into the buffer */
    uint32          count;                      /*   remaining byte count */
    DELAY_PROPS     fast_times;                 /*   optimized times, indexed by DELAY */
    CNTLR_OPCODE    opcode;                     /*   the tape controller opcode */
    CNTLR_IBUS      data_bus;                   /*   the tape controller data bus */
    CNTLR_FLAG_SET  flag_bus;                   /*   the tape controller flag bus */
    CNTLR_CLASS     command_class;              /*   the current command classification */
    TL_BUFFER       record [TL_BUFSIZE];        /*   the tape controller record buffer */
    } MA_STATE;


/* Multiple instance structure declaration */

typedef struct {                                /* MA instance structure */
    DEVICE    *dptr;                            /*   a pointer to the device instance */
    MA_STATE  *sptr;                            /*   a pointer to the interface state instance */
    } INSTANCE;


/* HP-IB interface local state declarations.

   The local state consists of the bus controller and tape controller state
   structures.  The original device's bus state structure is used as a template
   to create a new state instance for each new device instance.  The tape state
   structure is reinitialized to its initial value for new instances.
*/

DEVICE ma_dev;                                  /* incomplete device structure */

static MA_STATE ma_state = {                    /* bus controller state */
    { CNTLR_INIT (HP_IB, ma_dev, ma_state.record, ma_state.fast_times) }
    };


/* HP-IB interface local SCP support routine declarations */

static t_stat     tape_service      (UNIT   *uptr);
static t_stat     hpib_service      (UNIT   *uptr);
static t_stat     ma_reset          (DEVICE *dptr);
static t_stat     ma_boot           (int32  unit_number, DEVICE *dptr);
static t_bool     ma_initialize     (DEVICE *dptr, DIB *dibptr);
static CNTLR_VARS ma_set_controller (DEVICE *dptr, MA_STATE *mptr);
static uint32     ma_adjust         (DEVICE **dvptr, int32 delta);
static t_stat     ma_attach         (UNIT   *uptr, char *cptr);
static t_stat     ma_detach         (UNIT   *uptr);
static t_stat     ma_onoffline      (UNIT   *uptr, int32 value, char *cptr, void *desc);


/* HP-IB interface local bus routine declarations */

static ACCEPTOR ma_hpib_accept;
static REPORTER ma_report;


/* HP-IB interface local utility routine declarations */

static void   process_secondary (uint32 ma_id, uint32 address, BUS_CABLE bus);
static void   process_data      (uint32 ma_id, uint32 address, BUS_CABLE bus);
static void   execute_command   (uint32 ma_id, uint32 address, BUS_CABLE bus);
static void   end_transfer      (uint32 ma_id, BUS_DATA bus_group);
static void   set_response      (uint32 ma_id, uint32 address, FLIP_FLOP response);
static t_stat activate_unit     (uint32 ma_id, UNIT *uptr, DELAY delay);
static t_stat call_controller   (uint32 ma_id, UNIT *uptr);
static void   check_attention   (uint32 ma_id, uint32 address);


/* HP-IB registration declaration */

static const BUS_REGISTER ma_registration = {   /* bus registration */
    MA_Index,                                   /*   the dispatcher table index */
    { ma_report,                                /*   a pointer to the protocol reporting routine */
      ma_hpib_accept }                          /*   a pointer to the bus acceptor routine */
    };


/* HP-IB interface SCP data structure declarations */


/* Device information block */

static DIB ma_dib = {
    NULL,                                       /* the device's IOP interface function pointer */
    ma_initialize,                              /* the execution initialization function pointer */
    DEVNO_UNUSED,                               /* the device number */
    SRNO_UNUSED,                                /* the service request number */
    INTPRI_UNUSED,                              /* the interrupt priority */
    INTMASK_UNUSED,                             /* the interrupt mask */
    0,                                          /* the card index */
    CLEAR,                                      /* the interrupt request flip-flop */
    CLEAR,                                      /* the interrupt active flip-flop */
    FALSE,                                      /* TRUE if service has been requested */
    NULL,                                       /* the device's IMB interface function pointer */
    CHAN_UNUSED,                                /* the IMB channel number */
    SLOT_UNUSED,                                /* the IMB slot number */
    CPU_IS_IMB | SYS_SF                         /* the supported system configurations */
    };


/* Unit list.

   The first four SIMH units correspond to the four tape drives (unit numbers
   0-3).  The fifth unit is the tape controller, and the sixth unit is the bus
   controller.  The tape drives and controller are operated and maintained by
   the tape library.
*/

#define UNIT_FLAGS          (UNIT_7970E | UNIT_ATTABLE | UNIT_ROABLE | UNIT_DISABLE)

static UNIT ma_unit [UNIT_COUNT] = {
    { UDATA (tape_service, UNIT_FLAGS, 0) },    /* drive unit 0 */
    { UDATA (tape_service, UNIT_FLAGS, 0) },    /* drive unit 1 */
    { UDATA (tape_service, UNIT_FLAGS, 0) },    /* drive unit 2 */
    { UDATA (tape_service, UNIT_FLAGS, 0) },    /* drive unit 3 */
    { UDATA (tape_service, UNIT_DIS,   0) },    /* tape controller unit */
    { UDATA (hpib_service, UNIT_DIS,   0) }     /* HP-IB controller unit */
    };


/* Register list.

   We provide register access to the optimized timing table to allow the user to
   alter these if desired.  The HP-IB controller hardware does not have
   externally accessible registers, but we provide a few "pseudo-registers" to
   provide a view into the controller state.


   Implementation notes:

    1. All of the register locations are based on the state structure, except
       for those that have depth = DRIVE_COUNT, which are based on the unit
       array.  If this condition changes, the "ma_adjust" routine will have to
       be changed accordingly.
*/

static REG ma_reg [] = {
/*    Macro   Name     Location               Width   Offset       Flags       */
/*    ------  -------  ---------------------  ------  ------  ---------------- */
    { YRDATA (DEVS,    ma_state.present,        8),           PV_RZRO | REG_RO },
    { FLDATA (XFER,    ma_state.active,                  0),  PV_RZRO | REG_RO },
    { FLDATA (PP,      ma_state.poll_active,             0),  PV_LEFT | REG_RO },
    { YRDATA (PPR,     ma_state.response_set,   8),           PV_RZRO          },
    { DRDATA (ISTATE,  ma_state.state,          4),           PV_LEFT          },
    { DRDATA (ICMD,    ma_state.command,        4),           PV_LEFT          },
    { DRDATA (COUNT,   ma_state.byte_count,    16),           PV_LEFT          },
    { DRDATA (DSJ,     ma_state.dsj,            2),           PV_LEFT          },

      TL_REGS (ma_state.controller, ma_unit, DRIVE_COUNT, ma_state.record, ma_state.fast_times),

    { SVDATA (STATE,   ma_state)                                               },

    { NULL }
    };


/* Modifier list */

#define ETAB_CHANNEL        MTAB_EDV | MTAB_ERX (10) | MTAB_EMN (2)
#define ETAB_ADDRESS        MTAB_EDV | MTAB_ERX (10)

static MTAB ma_mod [] = {

    TL_MODS (ma_state.controller, TL_NONE, TL_FIXED, ma_onoffline),

/*    Mask Value   Match/Max Print String  Match String  Valid  Display  Descr  4.x   Expansion Flags */
/*    -----------  ---------  -----------  ------------  -----  -------  -----  ----  --------------- */
    { DEV_CHANNEL,    15,     "channel ",  "CHANNEL",    NULL,  NULL,    NULL,  NULL, ETAB_CHANNEL    },
    { DEV_ADDRESS,     7,     "bus=",      "BUS",        NULL,  NULL,    NULL,  NULL, ETAB_ADDRESS    },

    { 0 }
    };


/* Debugging trace list */

static DEBTAB ma_deb [] = {                     /* debugging trace list */
    { "CMD",   TRACE_CMD   },                   /*   trace interface and controller commands */
    { "INCO",  TRACE_INCO  },                   /*   trace interface and controller command initiations and completions */
    { "STATE", TRACE_STATE },                   /*   trace state changes */
    { "SERV",  TRACE_SERV  },                   /*   trace unit service scheduling calls and entries */
    { "XFER",  TRACE_XFER  },                   /*   trace data receptions and transmissions */
    { "IOBUS", TRACE_IOBUS },                   /*   trace tape controller bus signals and data words */
    { NULL,    0           }
    };


/* Device descriptor.

   Defining the device as having a 16-bit data width with an address increment
   of 2 means that attached file addresses are byte addresses but that files may
   be examined and deposited as either bytes or words.  Calling the HP word
   examine and deposit routines ensures that words are processed in big-endian
   (MSB first) format, regardless of the host platform convention.
*/

DEVICE ma_dev = {
    "MA",                                       /* device name */
    ma_unit,                                    /* unit array */
    ma_reg,                                     /* register array */
    ma_mod,                                     /* modifier array */
    UNIT_COUNT,                                 /* number of units */
    10,                                         /* address radix */
    32,                                         /* address width = 4 GB */
    2,                                          /* address increment */
    8,                                          /* data radix */
    16,                                         /* data width */
    &hp_word_examine,                           /* examine routine */
    &hp_word_deposit,                           /* deposit routine */
    &ma_reset,                                  /* reset routine */
    &ma_boot,                                   /* boot routine */
    &ma_attach,                                 /* attach routine */
    &ma_detach,                                 /* detach routine */
    &ma_dib,                                    /* device information block pointer */
    DEV_DISABLE | DEV_DEBUG | TO_CHANNEL (9),   /* device flags */
    0,                                          /* debug control flags */
    ma_deb,                                     /* debug flag name table */
    NULL,                                       /* memory size change routine */
    NULL,                                       /* logical device name */
    NULL,                                       /* (4.0 dummy) help routine */
    NULL,                                       /* (4.0 dummy) help attach routine */
    NULL,                                       /* (4.0 dummy) help context */
    NULL,                                       /* (4.0 dummy) description */
    ma_adjust                                   /* the device adjustment function pointer */
    };


/* HP-IB interface local allocation state declarations */

static uint32 ma_count = 1;                     /* the current count of MA device instances */

static INSTANCE mas [MAX_MA_COUNT] = {          /* MA bus instances, indexed by instance ID */
    { &ma_dev, &ma_state }                      /*   instance 0 (original) */
    };



/****************************************
 *                                      *
 * Interface local SCP support routines *
 *                                      *
 ****************************************/


/* Service the tape controller or a drive unit.

   This service routine is called to execute scheduled controller command phases
   for the specified unit.  The actions to be taken depend on the current state
   of the controller and the drive unit.

   This routine is entered when a tape unit or the tape controller unit is ready
   to execute the next command phase.  Generally, the controller library handles
   all of the tape operations.  All that is necessary is to notify the
   controller, which will process the next phase of command execution.  Because
   the controller can overlap operations, in particular scheduling rewinds on
   several drive units simultaneously, each drive unit carries its own current
   operation code and execution phase.  The controller uses these to determine
   what to do next.
*/

static t_stat tape_service (UNIT *uptr)
{
const  uint32 ma_id   = uptr->Instance;                 /* the instance identifier */
DEVICE *const dptr    = mas [ma_id].dptr;               /* the pointer to the controlling device */
const  uint32 address = ADDRESS (dptr->flags);          /* the bus address of the device containing the unit */
const  uint32 unit    = UNIT (uptr->flags);             /* the unit number */

tpprintf (dptr, TRACE_SERV, "Device %u %s service entered\n",
          address, tl_unit_name (unit));

return call_controller (ma_id, uptr);                   /* return the result of the controller call */
}


/* Service the HP-IB controller.

   This service routine is called to execute scheduled bus controller events.
   Events are scheduled to execute bus commands that do not involve the tape
   units.  Examples are Amigo Identify, Device Specified Jump, and Read and
   Write Loopback.  The actions to be taken upon service entry depend on the
   current state of the controller.


   Implementation notes:

    1. The second identify byte must be returned to the GIC within 5
       microseconds or the channel program will fail.  This is because the CPP
       does not wait for the second byte to be returned once the first byte is
       available.  So both bytes must arrive in the GIC FIFO before the channel
       Wait instruction resumes.
*/

static t_stat hpib_service (UNIT *uptr)
{
const uint32    ma_id      = uptr->Instance;            /* the instance identifier */
MA_STATE *const ma         = mas [ma_id].sptr;          /* the pointer to the channel state */
DEVICE   *const dptr       = mas [ma_id].dptr;          /* the pointer to the controlling device */
const uint32    channel    = CHANNEL (dptr->flags);     /* the channel address of the bus controller */
const uint32    address    = ADDRESS (dptr->flags);     /* the bus address of the device containing the unit */
const uint32    unit       = UNIT (uptr->flags);        /* the unit number */
const BUS_CABLE error_byte = Bus_DAV | Bus_EOI | 1;     /* the value returned for a transaction error */
BUS_CABLE       bus;
t_bool          completed   = FALSE;                    /* TRUE if the command has completed */
t_stat          result      = SCPE_OK;                  /* the result of the service call */
IF_STATE        entry_state = ma->state;                /* the controller state at entry */

tpprintf (dptr, TRACE_SERV, "Device %u %s state %s service entered\n",
          address, tl_unit_name (unit), if_state_names [entry_state]);

switch (entry_state) {                                  /* dispatch the controller state */

    case Command_Send:                                  /* send command data to the bus */
        if (not (ma->bus & (Bus_ATN | Bus_NRFD))) {     /* if the bus is ready for data */
            bus = Bus_DAV | ma->buffer [ma->index++];   /*   then get the next byte to send */
            ma->count--;                                /*     and count it */

            if (ma->count == 0)                         /* if the transfer is complete */
                bus |= Bus_EOI;                         /*   then tag the last byte with EOI */

            ma->bus = hpib_source (channel, address, bus);  /* send the byte to the bus */

            if (ma->count == 0)                             /* if all bytes have been sent */
                completed = TRUE;                           /*   then terminate the command */
            else                                            /* otherwise the execution phase continues */
                activate_unit (ma_id, uptr, Status_Delay);  /*   so reschedule the next transfer */
            }
        break;


    case Command_Receive:                               /* receive command data */
        if (ma->count == 0 || ma->bus & Bus_EOI)        /* if the transfer is complete */
            completed = TRUE;                           /*   then terminate the command */

        ma->bus = hpib_source (channel, address, ma->bus & ~Bus_NRFD);  /* deny NRFD to remove the handshake hold */

        entry_state = ma->state;                        /* pick up any state change from accepting the byte */
        break;


    case Completion_Wait:                               /* wait for command completion */
        completed = TRUE;                               /* indicate that a poll response may be generated */
        break;


    case Error_Source:                                  /* send data after an error */
        if (not (ma->bus & (Bus_ATN | Bus_NRFD))) {     /* if the bus is ready for data */
            ma->bus = hpib_source (channel, address,    /*   then source a dummy byte to the bus */
                                   error_byte);

            ma->flag_bus |= EOD;                        /* stop any ongoing tape read or write */
            completed = TRUE;                           /*   and terminate the command */
            }
        break;


    case Error_Sink:                                    /* absorb data after an error */
        if (ma->count == 0) {                           /* if the transfer is complete */
            ma->flag_bus |= EOD;                        /*   then stop any ongoing tape read or write */

            completed = TRUE;                           /* terminate the command */
            }

        ma->bus = hpib_source (channel, address,        /* deny NRFD to remove the handshake hold */
                               ma->bus & ~Bus_NRFD);
        break;


    default:                                            /* no other states schedule service */
        tpprintf (dptr, TRACE_STATE, "Device %u %s state %s not handled\n",
                  address, tl_unit_name (unit), if_state_names [entry_state]);

        result = SCPE_IERR;                             /* signal an internal error */
        break;
    }                                                   /* end of interface state dispatch */


if (completed) {                                        /* if the command completed */
    ma->active = FALSE;                                 /*   then end any data transfer */
    ma->state = Command_Ready;                          /*     and enable the next command */

    if (cmd_props [ma->command].completion_poll)        /* if a completion poll is needed */
        set_response (ma_id, address, SET);             /*   then enable the poll response */

    if (ma->command != Tape_Command)
        tpprintf (dptr, TRACE_INCO, "Device %u %s completed with DSJ %u\n",
                  address, if_command_names [ma->command], ma->dsj);

    else if (ma->opcode != Invalid_Opcode)
        tpprintf (dptr, TRACE_INCO, "Device %u Tape Command %s %s with DSJ %u\n",
                  address, tl_opcode_name (ma->opcode),
                  (ma->opcode <= Select_Unit_3 ? "completed" : "continuing"), ma->dsj);

    ma->command = Invalid;                              /* invalidate the command after completion */
    }

if (ma->state != entry_state)                           /* if the controller state has changed */
    tpprintf (dptr, TRACE_STATE, "Device %u %s transitioned from %s state to %s state\n",
              address, tl_unit_name (unit),
              if_state_names [entry_state], if_state_names [ma->state]);

return result;                                          /* return the result of the service */
}


/* Device reset routine.

   This routine is called for a RESET, RESET MA, or BOOT MA command.  With the
   addition of the -P switch, it is the simulation equivalent of applying or
   cycling power on the master tape drive.  Without the switch, all of the tape
   units are idled; this action has no direct hardware analog.

   In addition, if a power-on reset (RESET -P) is done, the original FASTTIME
   settings are restored.
*/

static t_stat ma_reset (DEVICE *dptr)
{
DIB *const      dibptr = (DIB *) dptr->ctxt;    /* a pointer to the device context */
const uint32    ma_id  = dibptr->card_index;    /* the instance ID of the protocol */
MA_STATE *const ma     = mas [ma_id].sptr;      /* the pointer to the channel state */
UNIT            *uptr;
uint32          unit;

if (sim_switches & P_SWITCH) {                          /* if this is a power-on reset */
    hpib_register (ma_registration);                    /*   then register our protocol handler */

    memset (ma, 0, sizeof (MA_STATE));                  /* clear the state structure */
    ma->controller = ma_set_controller (dptr, ma);      /*  and reset the initial controller configuration */

    ma->dsj = 1;                                        /* set DSJ to 1 to request a status read */
    ma->interface_status = S3_POWER;                    /*   and indicate power restoration */

    set_response (ma_id, ADDRESS (dptr->flags), SET);   /* enable the parallel poll response */
    }

ma->flag_bus = INTOK;                                   /* enable drive attention reporting */

ma->fast_times = initial_times;                         /* restore the original optimized timing values */

ma->listeners = BUS_UNADDRESSED;                        /* unaddress all current listeners */
ma->talker    = BUS_UNADDRESSED;                        /*   and the current talker */

for (unit = 0; unit < UNIT_COUNT; unit++) {             /* loop through the SIMH units */
    uptr = &dptr->units [unit];                         /*   and get the unit pointer */

    uptr->Instance = ma_id;                             /* set the ID of the controlling instance */

    if (not (uptr->flags & UNIT_DIS))                   /* if the unit is not disabled */
        sim_cancel (uptr);                              /*   then cancel any current activation */
    }

return tl_reset (&ma->controller);                      /* reset the controller and return the result */
}


/* Device boot routine.

   This routine is called for the BOOT MA command to initiate the system cold
   load procedure for the tape.  It is the simulation equivalent to setting the
   Load thumbwheel switches to the appropriate channel and device numbers and
   then pressing the LOAD system control panel switch.

   For the Starfish, the switch register is set to the device number of the IMB
   adapter in the lower byte and the device number (channel/device) of the boot
   device in the upper byte.

   The cold load microcode does a Select Unit 0 command, so we verify that the
   user has specified unit 0.
*/

static t_stat ma_boot (int32 unit_number, DEVICE *dptr)
{
const uint32 channel = CHANNEL (dptr->flags);   /* the channel address of device */
const uint32 device  = ADDRESS (dptr->flags);   /* the bus address of the device containing the unit */
HP_WORD      load;

if (unit_number != 0)                                   /* if a unit other than 0 is specified */
    return SCPE_NOFNC;                                  /*   then fail with a Command not allowed error */

else {                                                  /* otherwise */
    if (sys_configuration & CPU_IS_IMB)                 /*   if this is an IMB CPU */
        load = TO_DEVNO (channel, device);              /*     then get the target device number */
    else                                                /*   otherwise */
        load = TO_WORD (TO_DEVNO (channel, device),     /*     form the target and IMBA device numbers */
                        IMBA_DEVNO);                    /*       for the Starfish */

    cpu_front_panel (load, Cold_Load);                  /* set up the cold load */

    return SCPE_OK;                                     /* return to run the bootstrap */
    }
}


/* Run-time initialization.

   This bus routine is called during I/O initialization each time machine
   instruction execution is to begin.  The routine returns TRUE to permit
   execution to commence or FALSE to inhibit execution.  Initialization is
   optional for modules but is required in our case to set the presence bit
   corresponding to the assigned bus address.  We also ensure that any prior bus
   assignment is cleared from the set of devices checking command parity and
   responding to a parallel poll.


   Implementation notes:

    1. The parallel poll response set uses HP-IB bit order, where bit 0 is the
       MSB, while the bus device presence set uses C bit order, where bit 0 is
       the LSB.  Consequently, the response set cannot be masked by the present
       set to account for a bus address change but instead must be calculated
       separately from the bus address.
*/

static t_bool ma_initialize (DEVICE *dptr, DIB *dibptr)
{
const uint32      ma_id = dibptr->card_index;   /* the instance ID of the protocol */
MA_STATE *const   ma    = mas [ma_id].sptr;     /* the pointer to the channel state */
CNTLR_VARS *const cvptr = &ma->controller;      /* the pointer to the tape controller */

ma->present |= BIT (ADDRESS (dptr->flags));             /* set the "device is present" bit */
ma->parity &= ma->present;                              /*   and ensure only present devices are checking parity */

ma->response_set &= HPIB_BIT (ADDRESS (dptr->flags));   /* mask poll responses to just those devices present */

return tl_initialize (cvptr);                           /* return the result of the tape initializer */
}


/* Initialize the tape controller structure.

   This routine initializes a newly allocated or newly reset tape controller.
   It is called by the instance adjuster to set up the correct values in the
   fields of the tape controller structure and the reset routine to restore the
   controller parameters after a power-on reset clears the instance state
   structure.


   Implementation notes:

    1. This routine is needed because Microsoft Visual C++ 2008 does not support
       postfix structure initializer operator expressions from the C99 standard.
*/

static CNTLR_VARS ma_set_controller (DEVICE *dptr, MA_STATE *mptr)
{
CNTLR_VARS initial = { CNTLR_INIT (HP_IB, *dptr, mptr->record, mptr->fast_times) };

return initial;                                         /* return the initialized structure */
}


/* Adjust the device instances.

   This routine is called by SCP to adjust device instances that are newly
   allocated or pending deletion.  On entry, "dvptr" points at the "sim_devices"
   element that contains a pointer to the first of potentially several device
   instances to be adjusted, and "delta" contains the number of instances just
   created (if positive) or to be deleted (if negative).

   When the "dvptr" parameter is NULL, SCP expects the return value to be
   formatted as follows:

      31  30  ... 25  24 |23  22  ..  17  16 |15  14 13  ...  2   1   0
     +---+---+---+---+---+---+---+---+---+---+---+--+---+---+---+---+---+
     | clone identifier  |  instance limit   |     sim_devices size     |
     +---+---+---+---+---+---+---+---+---+---+---+--+---+---+---+---+---+

   The fields have the following meanings:

     sim_devices size      The total number of elements defined in the
                           "sim_devices" array, including the NULL element at
                           the end.  The difference between this number and the
                           number of initialized elements gives the number of
                           new device instances that can be added by the user.

     instance limit        The maximum number of instances allowed.  The count
                           includes the original instance.

     clone identifier      A character used to produce unique names for new
                           device instances.  The ID will be appended to the
                           original device name and incremented by the device
                           instance number.

   When the adjust function is called with a non-NULL pointer parameter, the
   routine performs the actions are necessary to prepare the indicated device
   instances for use or destruction.  The return value is either SCPE_OK if all
   adjustments succeeded, or an SCPE error code indicating the nature of the
   failure.  In the latter case, any allocations made by the routine prior to
   the failure must be freed before returning.  If this routine returns success,
   the device's "reset" routine, if defined, will be called with the -P
   (power-on) switch to complete initialization of the new device.


   Implementation notes:

    1. The deallocation part of the routine is entered either for a deletion
       request or for an addition request when a required allocation fails.

    2. Register locations that have depth = DRIVE_COUNT refer to the unit array.
       All other non-NULL register locations refer to the state structure.  If a
       third reference becomes necessary, indicators should be added in the form
       of user flags instead of expanding this ad-hoc determination.
*/

static uint32 ma_adjust (DEVICE **dvptr, int32 delta)
{
MA_STATE *nsptr;
DEVICE   *dptr;
char     *nptr;
DIB      *dibptr;
UNIT     *uptr;
REG      *rptr;
int32    count;
t_stat   status = SCPE_OK;

if (dvptr == NULL)                                      /* if this is a limits request */
    return TO_DWORD (TO_WORD ('A', MAX_MA_COUNT),       /*   then return the clone ID and instance limit */
                     SIM_DEVICE_SIZE);                  /*     and device array size */

else if (delta > 0) {                                   /* otherwise if device clones are being added */
    for (count = 0; count < delta; count++) {           /*   then for each new device clone */
        dptr = dvptr [count];                           /*     get a pointer to the new device */

        status = SCPE_INCOMP;                           /* preset for name duplication error recovery */

        nptr = dptr->name;                              /* save the newly assigned name */
        dptr->name = "";                                /*   and temporarily remove it */

        if (find_dev (nptr) != NULL) {                      /* if the device name is duplicated */
            cprintf ("Duplicate device name %s\n", nptr);   /*   then report it to the user */

            dptr->name = nptr;                          /* restore the assigned name */
            break;                                      /*   and fail at this point */
            }

        else                                            /* otherwise the name is unique */
            dptr->name = nptr;                          /*   so restore it to the device */

        status = SCPE_MEM;                              /* preset for memory error recovery */

        dptr = dvptr [count];                           /* get a pointer to the new device */

        dibptr = malloc (sizeof (DIB));                 /* allocate a new DIB structure */

        if (dibptr == NULL)                             /* if the allocation failed */
            break;                                      /*   then clean up before exiting */
        else                                            /* otherwise */
            dptr->ctxt = dibptr;                        /*   save the new DIB pointer */

        memcpy (dibptr, &ma_dib, sizeof (DIB));         /* copy the original DIB structure */
        dibptr->card_index = ma_count + count;          /*   and set the new card index */

        nsptr = calloc (1, sizeof (MA_STATE));          /* allocate and zero a new MA state structure */

        if (nsptr == NULL)                                          /* if the allocation failed */
            break;                                                  /*   then clean up before exiting */
        else                                                        /* otherwise */
            nsptr->controller = ma_set_controller (dptr, nsptr);    /*   initialize the new tape controller structure */

        mas [ma_count + count].sptr = nsptr;            /* save the new state pointer */
        mas [ma_count + count].dptr = dptr;             /*   and device pointer in the table */

        uptr = dptr->units;                             /* point at the new unit array */

        for (rptr = dptr->registers; rptr->name != NULL; rptr++)    /* loop through the register entries */
            if (rptr->depth == DRIVE_COUNT)                         /* if the register is a unit reference */
                rptr->loc =                                         /*   then rebase from the unit array */
                  (char *) rptr->loc - (char *) ma_unit + (char *) uptr;
            else                                                    /* otherwise */
                rptr->loc =                                         /*   rebase from the state structure */
                  (char *) rptr->loc - (char *) &ma_state + (char *) nsptr;

        status = SCPE_OK;                               /* this pass succeeded */
        }                                               /*   so process the next instance */

    ma_count = ma_count + delta;                        /* bump the count MA instances */
    }


if (delta < 0 || status != SCPE_OK) {                   /* if device clones are being deallocated */
    if (delta < 0)                                      /*   then if existing entries are to be deleted */
        delta = - delta;                                /*     then make the deletion count positive */

    ma_count = ma_count - delta;                        /* drop the count of MA instances */

    for (count = 0; count < delta; count++) {           /* for each device clone to be deleted */
        dptr = dvptr [count];                           /*   get the device pointer */

        if (dptr != NULL)                               /* if it is defined */
            free (dptr->ctxt);                          /*   then free the device context */

        if (mas [ma_count + count].sptr != NULL)        /* if the MA state was allocated */
            free (mas [ma_count + count].sptr);         /*   then free it */

        mas [ma_count + count].dptr = NULL;             /* clear the */
        mas [ma_count + count].sptr = NULL;             /*   table entries */
        }
    }

return (uint32) status;                                 /* return the adjustment status */
}


/* Attach a tape image file to a drive unit.

   The tape drive specified by the "uptr" parameter is attached to the tape
   image file whose name is specified by the "cptr" parameter.  This is the
   simulation equivalent of mounting a tape reel on the drive and pressing the
   LOAD and ONLINE buttons.  The transition from offline to online causes a Unit
   Attention interrupt.

   The controller library routine handles command validation and setting the
   appropriate drive unit status.  It will return an error code if the command
   fails.  Otherwise, it will return SCPE_INCOMP if the command must be
   completed with a controller call, or SCPE_OK if the command is complete.  If
   the controller is idle, a call will be needed to poll the drives for
   attention; otherwise, the drives will be polled the next time the controller
   becomes idle.


   Implementation notes:

    1. If we are called during a RESTORE command to reattach a file previously
       attached when the simulation was SAVEd, the unit status will not be
       changed by the controller, so the unit will not request attention.
*/

static t_stat ma_attach (UNIT *uptr, char *cptr)
{
const uint32      ma_id = uptr->Instance;                   /* the instance identifier */
CNTLR_VARS *const cvptr = &mas [ma_id].sptr->controller;    /* the pointer to the tape controller */
t_stat            result;

result = tl_attach (cvptr, uptr, cptr);                 /* attach the drive */

if (result == SCPE_INCOMP) {                            /* if the controller must be called before returning */
    call_controller (ma_id, NULL);                      /*   then let it know to poll the drives */
    return SCPE_OK;                                     /*     and return with success */
    }

else                                                    /* otherwise the attach failed */
    return result;                                      /*   so return the status of the attach */
}


/* Detach a tape image file from a unit.

   The tape drive specified by the "uptr" parameter is unloaded, and the
   attached file, if any, is detached.  Unloading a tape leaves the drive
   offline.
*/

static t_stat ma_detach (UNIT *uptr)
{
const uint32      ma_id = uptr->Instance;                   /* the instance identifier */
CNTLR_VARS *const cvptr = &mas [ma_id].sptr->controller;    /* the pointer to the tape controller */

return tl_detach (cvptr, uptr);                         /* return the result of the detach */
}


/* Set a drive online or offline.

   The SET MAn OFFLINE command simulates pressing the RESET button, and the SET
   MAn ONLINE command simulates pressing the ONLINE button.  The transition from
   offline to online causes a Unit Attention interrupt.  The SET request fails
   if there is no tape mounted on the drive, i.e., if the unit is not attached
   to a tape image file.

   The controller library routine handles command validation and setting the
   appropriate drive unit status.  It will return an error code if the command
   fails.  Otherwise, it will return SCPE_INCOMP if the command must be
   completed with a controller call, or SCPE_OK if the command is complete.  If
   the controller is idle, a call will be needed to poll the drives for
   attention; otherwise, the drives will be polled the next time the controller
   becomes idle.
*/

static t_stat ma_onoffline (UNIT *uptr, int32 value, char *cptr, void *desc)
{
const uint32      ma_id  = uptr->Instance;                  /* the instance identifier */
CNTLR_VARS *const cvptr  = &mas [ma_id].sptr->controller;   /* the pointer to the tape controller */
const t_bool      online = (value == UNIT_ONLINE);          /* TRUE if the drive is being put online */
t_stat            result;

result = tl_onoffline (cvptr, uptr, online);            /* set the drive online or offline */

if (result == SCPE_INCOMP) {                            /* if the controller must be called before returning */
    call_controller (ma_id, NULL);                      /*   then let it know to poll the drives */
    return SCPE_OK;                                     /*     before returning with success */
    }

else                                                    /* otherwise */
    return result;                                      /*   return the status of the load or unload */
}



/**************************************
 *                                    *
 * HP-IB interface local bus routines *
 *                                    *
 **************************************/


/* Accept a command or data byte from the bus.

   This routine is called whenever a controller or talker has sourced a data
   byte to the bus, or whenever a controller, talker, or listener has changed
   the control state of the bus.  It is responsible for decoding commands and
   for calling routines to process secondaries, incoming data, and control state
   changes.

   When accepting a byte with DAV asserted, it corresponds in hardware to
   performing the IEEE-488 AH (Acceptor Handshake), T (Talker), L (Listener),
   RL (Remote Local), and DC (Device Clear) interface functions.  When changing
   the control signals without sourcing a byte, i.e., with DAV denied, it
   corresponds to performing the IEEE-488 SR (Service Request) function or as a
   participant in the SH (Source Handshake) function when asserting and denying
   NRFD (Not Ready for Data).

   If data is present, it decodes primary, secondary, talk, and listen commands,
   and receives data bytes if any of the attached devices are current listeners.
   For control changes, it responds to parallel polls and the IFC (Interface
   Clear), DCL (Device Clear), and SDC (Selected Device Clear) commands.

   On entry, the "ma_id" parameter is the instance identifier of the MA device
   connected to the bus, and the "bus" parameter is the set of eight control and
   eight data lines representing the HP-IB.

   If the DAV signal is not asserted, then a control signal change occurred.
   There are three conditions to which we must respond:

    1. A parallel poll is being conducted.  We record the poll state, so that if
       a poll response is subsequently enabled while a poll is still in
       progress, we can notify the controller.  Then we return the poll
       response from the current device.

    2. An Interface Clear is initiated.  IFC unaddresses all units, so any data
       transactions in progress must be terminated as though an Untalk and
       Unlisten had been accepted from the bus.

    3. Attention or Not Ready for Data is denied.  A device addressed to talk
       must wait for ATN to deny before data may be sent.  Also, a listener that
       has asserted NRFD must deny it before a talker may send data.  If a
       talker is sending data and both ATN and NRFD are denied, then we reenter
       the service routine to send the next byte.

   If DAV is asserted, then a command or data byte is present on the bus.  If
   the ATN signal is asserted, parity is checked to ensure that it is correct.
   If it is not, then a command parity error is logged.  Otherwise, the command
   byte is decoded to determine the proper action.

   Primary commands placed on the bus may be universal (applying to all present
   devices) or addressed (applying only to those devices that are addressed to
   listen).  Address commands select the device that will talk and the device(s)
   that will listen.  If the addressed device is present, the byte is accepted;
   a talk or listen command directed to a device not present is ignored.

   Secondary commands and data bytes are processed only if one or more listeners
   are present.

   This routine is called once for each byte sourced and is responsible for
   notifying the bus controller.  The type of byte passed is determined by the
   state of the ATN signal and, if ATN is asserted, by the high-order bits of
   the value.

   If a Talk, Unlisten, or Untalk command is received while a data transaction
   is currently in progress, the transfer is ended.  This can happen in the
   middle of a tape read or write in burst mode; if the transfer is incomplete,
   it will be set up again once the status of the prior burst is checked.  A new
   Listen command does not end the transfer, because the current listener
   remains addressed.

   For secondaries, the Amigo Identify command sequence is decoded here, as it
   does not follow the standard HP-IB protocol.  Otherwise, secondaries are
   processed by the bus controller if it is addressed to listen or talk.

   If ATN is not asserted, a data byte has been sourced.  It is processed by
   the bus controller if it is addressed to listen.

   After processing is complete, the routine returns the set of bits
   corresponding to the addresses of the current listeners.


   Implementation notes:

    1. As this device hosts only a single set of four drives on a single bus
       address, only one bus device will be present and therefore only be one
       listener bit will be active at a time.  However, the acceptor routine is
       written in a more general way in case this ever changes.

    2. The HP-IB Parity Checking command says that if checking is enabled and a
       channel command does not have odd parity, "...the device will NOT accept
       the command, i.e., Not Data Accepted (NDAC) will remain in the low state.
       This condition will remain until the host removes Data Valid (DAV) and
       corrects the channel command parity."  Because all devices on the bus
       drive the common NDAC line, any device holding it low will prevent the
       data handshake from completing.  Therefore, we need only test if the map
       of parity checkers is non-zero before declining acceptance of a byte with
       bad parity.  Before returning, we log a Command Parity Error in the bus
       controller.

    3. Identify is handled as a special case because of the odd bus protocol
       employed (it is an Untalk followed by a secondary containing a device
       address, and the device puts data on the bus without having been
       addressed as the talker).  If a secondary is immediately preceded by an
       Untalk, and the address contained in the secondary is present, then an
       Identify command has been received.  When the command is processed, the
       device is temporarily assigned as the talker for the duration of the
       command, even though it is preceded by an Untalk.  This allows us to know
       which SIMH unit to reschedule when the controller denies ATN to permit
       return of the identity bytes.

    4. The "process_data" routine may assert NRFD to hold off the talker, so we
       remove the DAV and EOI signals from the bus state before calling.  This
       prevents the assertion call from being misinterpreted as a data transfer.

    5. This protocol module maintains the current state of the channel, so that
       the routines know which devices are present and addressed.  This is more
       efficient than querying each bus address at each call to determine
       whether a device is present and, if so, whether it is presently addressed
       to talk or listen.  Consequently, it will be called only if one or more
       devices are present (for commands) or listening (for data).  If the
       routine is called inadvertently, it simply returns with no action taken.

    6. The manual states that a Universal Device Clear and a Selected Device
       Clear initialize the interface to its power-up state.  The two commands
       are equivalent and reset the hardware and firmware.  They do not affect
       the bus addressing state.

       Although not documented, Device Clear sets the "tape-unit-placed-online"
       bit in the status word for each drive that is online at the time the
       command is executed.  The diagnostic relies on this in its CLEARDEV
       function.  This would also occur at any time when the HP-IB interface
       received the "SL" signal from a drive, indicating that the user pushed
       the ONLINE button on the drive.  But the bit will not be set for a drive
       that has been taken offline either by the user via the RESET button on
       the drive or remotely via the "CL" command signal, nor will the Device
       Clear place the drive online again.
*/

static uint32 ma_hpib_accept (uint32 ma_id, BUS_CABLE bus)
{
MA_STATE *const   ma          = mas [ma_id].sptr;       /* the pointer to the channel state */
DEVICE   *const   dptr        = mas [ma_id].dptr;       /* the pointer to the controlling device */
CNTLR_VARS *const cvptr       = &ma->controller;        /* the pointer to the tape controller */
const uint32      address     = BUS_ADDRESS (bus);      /* the address indicated for addressing commands */
const uint32      address_bit = BIT (address);          /* the bit corresponding to the address */
UNIT              *uptr;
IF_STATE          entry_state;
uint32            dev_map;

ma->bus = bus;                                          /* synchronize the bus */

ma->poll_active = (bus & BUS_PPOLL) == BUS_PPOLL;       /* record the parallel poll state */

if (ma->poll_active)                                    /* if a poll is currently active */
    return ma->response_set;                            /*   then return the poll response set */

else if (not (bus & Bus_DAV)) {                         /* otherwise if this is a control change */
    if (bus & Bus_IFC) {                                /*   then if interface clear is asserted */
        if (ma->active)                                 /*     then if a data transfer is active */
            end_transfer (ma_id, BUS_PCG);              /*       then abort the transaction */

        ma->listeners = BUS_UNADDRESSED;                /* unaddress all listeners */
        ma->talker    = BUS_UNADDRESSED;                /*   and unaddress the current talker */
        }

    else if (not (bus & (Bus_ATN | Bus_NRFD))           /* otherwise if the bus is ready for data */
      && ma->present & ma->talker) {                    /*   and a talker is addressed */
        if (ma->command == Read_Data)                   /*     then if this is a Read Data command */
            uptr = TAPE_CNTLR_UPTR;                     /*       then point at the tape controller */
        else                                            /*     otherwise */
            uptr = HPIB_CNTLR_UPTR;                     /*       point at the bus controller */

        if ((ma->state == Execution_Send                /* if the device is waiting to send data */
          || ma->state == Error_Source)                 /*   or source an error byte */
          && not sim_is_active (uptr))                  /*   and the controller unit is not scheduled */
            uptr->action (uptr);                        /*     then call it directly to source the byte */
        }

    return ma->present & ma->listeners;                 /* return the current set of listeners */
    }

else if (ma->present == BUS_NONE)                       /* otherwise if there are no devices present */
    return 0;                                           /*   then the listener bitmap is empty */

else if (bus & Bus_ATN) {                               /* otherwise if this is a command */
    if (ma->parity && odd_parity [BUS_DATA (bus)])      /*   then if parity is enabled but wrong */
        ma->interface_status |= S3_PARITY;              /*     then log a command parity error */

    else switch (BUS_GROUP (bus)) {                     /* otherwise dispatch the bus group */

        case BUS_PCG:                                   /* Primary Command Group */
            switch (address) {                          /* dispatch the primary command */

                case 0x04:                              /* Selected Device Clear or Amigo Clear */
                case 0x14:                              /* Universal Device Clear */
                    dev_map = ma->present;              /* get all devices present on the bus */

                    if (address == 0x04)                /* if clearing only the selected devices */
                        dev_map &= ma->listeners;       /*   then mask to the set of devices listening */

                    if (dev_map) {                      /* if the device is addressed or present */
                        tl_reset (cvptr);               /*   then reset the controller to stop the drives */

                        ma->interface_status = 0;       /* clear the interface status */

                        ma->attention = TO_ONLINE_ATTN (tl_status (cvptr)); /* reset the online drive status */

                        set_response (ma_id, ADDRESS (dptr->flags), SET);   /* enable the poll response */
                        }

                    ma->active = FALSE;                 /* end any active data transaction */
                    break;


                default:                                /* unsupported universal commands */
                    break;                              /*   are ignored */
                }
            break;


        case BUS_LAG:                                   /* Listen Address Group */
            if (address == BUS_UNADDRESS) {             /* if this is an Unlisten */
                if (ma->active)                         /*   then if unlistening while the channel is active */
                    end_transfer (ma_id, BUS_LAG);      /*     then end the transaction */

                ma->listeners = BUS_UNADDRESSED;        /* unaddress all of the listeners */
                }

            else {                                      /* otherwise someone is being addressed */
                ma->listeners &= ~BUS_UNADDRESSED;      /*   so clear the unaddressed bit */

                if (ma->present & address_bit) {        /* if the addressed device is present */
                    ma->listeners |= address_bit;       /*   then it becomes a listener */

                    if (ma->talker == address_bit)      /* if it is also the current talker */
                        ma->talker = BUS_UNADDRESSED;   /*   then untalk it */
                    }
                }
            break;


        case BUS_TAG:                                   /* Talk Address Group */
            if (ma->active                              /* if readdressing while the channel is active */
              && ma->present & ma->talker               /*   and a talker is currently addressed */
              && address_bit != ma->talker)             /*     and the same address is not being respecified */
                end_transfer (ma_id, BUS_TAG);          /*       then end the transaction */

            if (address == BUS_UNADDRESS)               /* if this is an Untalk */
                ma->talker = BUS_UNADDRESSED;           /*   then clear the talker */

            else {                                      /* otherwise someone is being addressed */
                ma->talker = BUS_NONE;                  /*   so clear the unaddressed bit */

                if (ma->present & address_bit) {        /* if the addressed device is present */
                    ma->talker = address_bit;           /*   then it becomes the current talker */

                    if (ma->listeners & address_bit)    /* if the new talker had been listening */
                        ma->listeners &= ~address_bit;  /*   then clear its listener bit */
                    }
                }
            break;


        case BUS_SCG:                                   /* Secondary Command Group */
            if (ma->talker == BUS_UNADDRESSED           /* if an Untalk preceded the secondary */
              && ma->present & address_bit) {           /*   and the addressed device is present */
                ma->command = Amigo_Identify;           /*     then this is an Identify sequence */

                entry_state = ma->state;                /* save the entry state */

                ma->talker = address_bit;               /* temporarily address the device to talk */
                execute_command (ma_id, address, bus);  /*   and start the command */

                tpprintf (dptr, TRACE_STATE, "Device %u transitioned from %s state to %s state\n",
                          address, if_state_names [entry_state], if_state_names [ma->state]);
                }

            else if (ma->present & (ma->listeners | ma->talker))        /* otherwise if the device is addressed */
                process_secondary (ma_id, ADDRESS (dptr->flags), bus);  /*   then process the secondary */
            break;
        }                                               /* end of command dispatching */
    }

else {                                                  /* otherwise it is bus data (ATN is denied) */
    ma->bus &= ~(Bus_DAV | Bus_EOI);                    /*   so remove the handshake from the bus state */

    if (ma->present & ma->listeners)                        /* if the device is listening */
        process_data (ma_id, ADDRESS (dptr->flags), bus);   /*   then process the data byte */
    }

return ma->present & ma->listeners;                     /* return the current set of listeners */
}


/* Report the protocol configuration.

   This routine is called during run-time initialization to report the
   configuration of this GIC protocol handler.  On entry, "channel" contains the
   channel number for which the report is requested.  The routine returns a
   pointer to an array of structures, each containing a pointer to the device
   instance, the instance ID of the protocol handler, and a bitmap of the
   devices present on the bus and supported by the protocol.

   For a given channel, there may be multiple MA instances that host units on
   that channel.  Each MA instance supports four tape drives occupying a single
   bus address.  If more than four drives are required, or if concurrent data
   transfers are desired, then a second MA instance may be created and assigned
   to the same channel and a different bus address so that it can contribute the
   remainder of the bus devices.

   Therefore, the report array may contain more than one reporting structure,
   with the end of the valid elements followed by an element containing a NULL
   device pointer.  This array is defined as static so that the pointer returned
   remains valid while the caller unloads the information from the structure(s).


   Implementation notes:

    1. While there can be only eight possible devices on a given channel, the
       report array must be large enough to hold one element for each possible
       MA instance, plus an additional one to hold the NULL element indicating
       the end of the report.  This is because a user might not have changed the
       channel assignments after increasing the count of instances, resulting in
       all instances on the same channel.  This error will be picked up when
       channel conflicts are checked, but that may occur after we are called,
       and so we must be prepared for the worst case.
*/

static BUS_REPORT *ma_report (uint32 channel)
{
static BUS_REPORT report [MAX_MA_COUNT + 1];
DEVICE            *dptr;
BUS_REPORT        *rptr;
uint32            ma_id;

memset (report, 0, sizeof report);                      /* clear any previous report */
rptr = report;                                          /*   and restart at the first element */

for (ma_id = 0; ma_id < ma_count; ma_id++) {            /* search through the instance table */
    dptr = mas [ma_id].dptr;                            /*   and get the next device pointer */

    if (CHANNEL (dptr->flags) == channel                /* if the instance is configured for the channel */
      && not (dptr->flags & DEV_DIS)) {                 /*   and the device is enabled */
        rptr->dptr     = dptr;                          /*   then report the device pointer */
        rptr->instance = ma_id;                         /*     and the instance ID */
        rptr->present  = BIT (ADDRESS (dptr->flags));   /*       and the assigned address bit */

        rptr++;                                         /* point at the next entry */
        }
    }

return report;                                          /* return the configuration report pointer */
}



/******************************************
 *                                        *
 * HP-IB interface local utility routines *
 *                                        *
 ******************************************/


/* Process a secondary command.

   This routine is called to process a secondary command received from the bus
   for an addressed device.  Secondaries are used by the Amigo protocol to
   identify the various commands, specifically:

           Hex
     T/L  Value  Identifies
     ---  -----  ---------------------
      L    00    Write Data
      L    01    Tape Command
      L    07    End Command
      L    10    Amigo Clear
      L    1E    Write Loopback

      T    00    Read Data
      T    01    Read Status
      T    02    Read Byte Count
      T    10    Device Specified Jump
      T    1E    Read Loopback

   On entry, the "ma_id" parameter is the instance identifier of the MA device
   connected to the bus, the "address" parameter contains the bus address of the
   device, which will be either a listener or talker, and the "bus" parameter
   contains the set of eight control and eight data lines representing the HP-IB
   state.

   The routine decodes the valid secondaries, checks that the bus controller is
   in the appropriate state for the supplied secondary, and establishes the new
   controller state as a result of receiving the secondary.

   Bus controller commands operate independently of tape controller commands,
   except when a data transfer from a Read Record or Write Record command is in
   progress.  Because the protocol uses separate secondaries for the command and
   data, the Read Data and Write Data secondaries verify that the controller is
   executing a read or write command, respectively.  If it is not, then an error
   condition exists.

   Error recovery is as follows:

    - An unsupported talk secondary sends a single data byte tagged with EOI.

    - An unsupported listen secondary accepts and discards any accompanying data
      bytes until EOI is asserted or an Unlisten is received.

    - A Read Data secondary that is not permitted sends a single data byte
      tagged with EOI.

    - An Write Data secondary that is not permitted accepts and discards data
      bytes until EOI is asserted or an Unlisten is received.


   Implementation notes:

    1. The HP-IB Interface Manual does not specify the action taken if an
       invalid secondary is received.  In simulation, this is reported as a
       Command Reject error and the command is invalidated.  Also XFRNG is added
       to the flag bus so that the tape controller will terminate any
       in-progress data transfer gracefully.

    2. The bus controller enables its poll response to notify the host that a
       command has completed.  However, the host might not wait for a poll and
       might initiate the next command while the poll is still pending.  In
       simulation, the poll response is scheduled as a service event that
       represents the controller overhead.  This event may still be pending when
       the secondaries are received, and if so, the service event is canceled
       before proceeding with the next command.  This is to ensure that the poll
       response from the preceding command is not confused with that of the new
       command.
*/

static void process_secondary (uint32 ma_id, uint32 address, BUS_CABLE bus)
{
MA_STATE   *const ma          = mas [ma_id].sptr;       /* the pointer to the channel state */
DEVICE     *const dptr        = mas [ma_id].dptr;       /* the pointer to the controlling device */
CNTLR_VARS *const cvptr       = &ma->controller;        /* the pointer to the tape controller */
const uint32      secondary   = BUS_ADDRESS (bus);      /* the secondary address without parity */
const IF_STATE    entry_state = ma->state;              /* the controller state at entry */
t_bool            abort       = FALSE;

if (ma->state == Completion_Wait)                       /* if a previous poll response is still pending */
    sim_cancel (HPIB_CNTLR_UPTR);                       /*   then cancel it to avoid confusion */

if (ma->listeners & BIT (address))                      /* if this device is listening */
    switch (secondary) {                                /*   then dispatch the listen secondary */

        case 0x00:                                      /* Receive Write Data */
            ma->command = Write_Data;                   /*   then set the command */

            if (ma->command_class == Class_Write)       /* if a write is in progress */
                ma->active = TRUE;                      /*   then (re)start the transaction */

            else {                                      /* otherwise */
                ma->count = D32_UMAX;                   /*   sink as many bytes as the host wants to send */
                ma->state = Error_Sink;                 /*     until one is tagged with EOI */
                abort = TRUE;
                }
            break;


        case 0x01:                                      /* Tape command */
            if (ma->state == Command_Ready)             /* if we are ready for a command */
                ma->command = Tape_Command;             /*   then set the command and wait for the parameter */

            else {                                      /* otherwise */
                ma->count = D32_UMAX;                   /*   sink as many bytes as the host wants to send */
                ma->state = Error_Sink;                 /*     until one is tagged with EOI */
                abort = TRUE;                           /*       to discard the new command */
                }
            break;


        case 0x07:                                      /* End command */
            ma->command = End_Command;                  /* set the command and wait */
            break;


        case 0x10:                                      /* Amigo Clear */
            ma->command = Amigo_Clear;                  /* set the command and wait */
            break;


        case 0x1E:                                      /* Write Loopback */
            ma->command = Write_Loopback;               /* set the command and wait */
            break;


        default:                                        /* an unsupported listen secondary was received */
            ma->command = Invalid;                      /* set an invalid command */

            ma->count = D32_UMAX;                       /* sink as many bytes as the host wants to send */
            ma->state = Error_Sink;                     /*   until one is tagged with EOI */
            abort = TRUE;
            break;
        }

else                                                    /* otherwise the device is talking */
    switch (secondary) {                                /*   so dispatch the talk secondary */

        case 0x00:                                      /* Send Read Data */
            ma->command = Read_Data;                    /*   then set the command */

            if (ma->command_class == Class_Read)        /* if a read is in progress */
                ma->active = TRUE;                      /*   then (re)start the transaction */

            else {                                      /* otherwise */
                ma->state = Error_Source;               /*   source an error byte tagged with EOI */
                abort = TRUE;
                }
            break;


        case 0x01:                                      /* Read Status */
            ma->command = Read_Status;                  /* set the command */
            break;


        case 0x02:                                      /* Read Byte Count */
            ma->command = Read_Byte_Count;              /* set the command */
            break;


        case 0x10:                                      /* Device Specified Jump */
            ma->command = Device_Specified_Jump;        /* set up the reporting command */
            break;


        case 0x1E:                                      /* Read Loopback */
            ma->command = Read_Loopback;                /* set the command */
            break;


        default:                                        /* an unsupported talk secondary was received */
            ma->command = Invalid;                      /* set an invalid command */

            ma->state = Error_Source;                   /* source an error byte tagged with EOI */
            abort = TRUE;
            break;
        }


if (abort) {                                            /* if a bad secondary was received */
    if (ma->command == Invalid)                         /*   then if the secondary is undefined */
        tpprintf (dptr, TRACE_CMD, "Device %u invalid %s secondary %02XH rejected\n",
                  address, (ma->listeners & BIT (address) ? "listen" : "talk"), secondary);

    else {                                              /* otherwise the secondary mode is wrong */
        tpprintf (dptr, TRACE_CMD, "Device %u %s command rejected\n",
                  address, if_command_names [ma->command]);

        ma->command = Invalid;                          /* invalidate the data transfer command */
        }

    if (cvptr->state == Idle_State) {                   /* if the tape controller is idle */
        ma->data_bus = Invalid_Opcode;                  /*   then force a command reject */
        ma->flag_bus |= CMRDY | XFRNG;                  /*     for the invalid */
        call_controller (ma_id, NULL);                  /*       command secondary */
        }

    else                                                /* otherwise the controller is busy */
        ma->flag_bus |= XFRNG;                          /*  so mark the current data transfer as no good */
    }

else {                                                  /* otherwise the command is accepted */
    ma->state = cmd_props [ma->command].next_state;     /*   so enter the next state */

    set_response (ma_id, address, CLEAR);               /* disable the poll response */

    if (ma->state != Parameter_Wait)                    /* if not waiting for a parameter */
        execute_command (ma_id, address, bus);          /*   then start the command */
    }

if (ma->state != entry_state)
    tpprintf (dptr, TRACE_STATE, "Device %u transitioned from %s state to %s state\n",
              address, if_state_names [entry_state], if_state_names [ma->state]);

return;
}


/* Process a data byte.

   This routine is called to process a data byte received from the bus for an
   addressed device.  Depending on the controller state, the byte represents
   either a command opcode, a command parameter, or a data value.

   On entry, the "ma_id" parameter is the instance identifier of the MA device
   connected to the bus, the "address" parameter contains the bus address of the
   device, which will be a listener, and the "bus" parameter contains the set of
   eight control and eight data lines representing the HP-IB state.

   The routine interprets the data byte in the context of the current bus
   controller state, as follows:

     Parameter_Wait
     --------------
     The value is a parameter to the current command.  It is placed in the
     parameter buffer, the default next state (Completion_Wait) is set, and the
     pending command is executed.

     Execution_Receive
     -----------------
     The value is a data byte received from the host during a data transfer to
     the tape controller.  It is placed on the controller's data bus, DTRDY is
     asserted on the flag bus, and the byte counts for the data buffer and the
     transaction are incremented.  If the byte is tagged with EOI, EOD is also
     asserted on the flag bus to indicate the end of the transfer.  The NRFD
     signal is asserted on the HP-IB to hold off the talker and the tape
     controller is scheduled to receive the byte.

     Command_Receive
     ---------------
     The value is a data byte received from the host for the bus controller.
     The preceding command secondary has indicated the number of bytes to
     expect.  If the count is not exhausted, the byte is added to the buffer,
     the buffer index is incremented, and the count is decremented.  If the byte
     is tagged with EOI, then the count is cleared to indicate that no more
     bytes are expected.  Otherwise, if the count is already zero, and the byte
     is not tagged with EOI, then the controller proceeds to sink as many bytes
     as the host wishes to send.  The NRFD signal is asserted on the HP-IB to
     hold off the talker, and the bus controller is scheduled to process the
     byte.

     Error_Sink
     ----------
     The value is a data byte that is discarded due to a previous error.  If the
     byte is tagged with EOI, the remaining transaction count is set to zero to
     end the transaction.  The NRFD signal is asserted on the HP-IB to hold off
     the talker, and the bus controller is scheduled to complete the bus
     handshake after the data transfer time elapses.

   Entry in any other state results in a transition to the Error_Sink state to
   discard the unexpected data.  An example of this is entry in the
   Command_Ready state.  This can only occur if the host sends a Listen followed
   by data without an intervening secondary.  In this case (and the others), we
   discard the data and report a Command Reject error.
*/

static void process_data (uint32 ma_id, uint32 address, BUS_CABLE bus)
{
MA_STATE *const ma          = mas [ma_id].sptr;         /* the pointer to the channel state */
DEVICE   *const dptr        = mas [ma_id].dptr;         /* the pointer to the controlling device */
const uint32    channel     = CHANNEL (dptr->flags);    /* the channel address of the bus controller */
const uint8     data        = BUS_DATA (bus);           /* the data value from the bus */
const IF_STATE  entry_state = ma->state;                /* the controller state at entry */

switch (ma->state) {                                    /* dispatch the controller state */

    case Parameter_Wait:                                /* waiting for a parameter */
        ma->parameter = data;                           /* save the parameter */

        ma->state = Completion_Wait;                    /* preset the completion state */
        execute_command (ma_id, address, bus);          /*   and execute the command */
        break;


    case Execution_Receive:                             /* receiving data from the host */
        ma->data_bus = data;                            /* save the data byte */
        ma->flag_bus |= DTRDY;                          /*   and set the data ready flag */

        ma->count++;                                    /* count the buffer byte */
        ma->byte_count++;                               /*   and the record length */

        if (bus & Bus_EOI)                              /* if this is the last byte from the host */
            ma->flag_bus |= EOD;                        /*   then add an EOD response from the controller */

        ma->bus = hpib_source (channel, address,        /* assert NRFD to hold off the talker */
                               ma->bus | Bus_NRFD);

        activate_unit (ma_id, TAPE_CNTLR_UPTR, Data_Delay); /* activate the unit to write the data */
        break;


    case Command_Receive:                               /* receiving command bytes */
        if (ma->count > 0) {                            /* if there are more to receive */
            ma->buffer [ma->index++] = data;            /*   then add the byte to the buffer */
            ma->count--;                                /*     and drop the remaining count */
            }

        if (bus & Bus_EOI)                              /* if this is the last byte from the host */
            ma->count = 0;                              /*   then there are no more expected */

        else if (ma->count == 0) {                      /* otherwise if more are coming but are not wanted */
            ma->count = D32_UMAX;                       /*   then sink as many bytes as the host wants to send */
            ma->state = Error_Sink;                     /*     until one is tagged with EOI */
            }

        ma->bus = hpib_source (channel, address,        /* assert NRFD to hold off the talker */
                               ma->bus | Bus_NRFD);

        activate_unit (ma_id, HPIB_CNTLR_UPTR, Data_Delay); /* schedule the next reception */
        break;


    default:                                            /* data was received in the wrong state */
        ma->state = Error_Sink;                         /* sink all data arriving in the wrong state */

    /* fall through into the Error_Sink state */

    case Error_Sink:                                    /* sinking data bytes after an error */
        if (bus & Bus_EOI)                              /* if this is the last byte from the host */
            ma->count = 0;                              /*   then terminate the transfer with this byte */

        ma->bus = hpib_source (channel, address,        /* assert NRFD to hold off the talker */
                               ma->bus | Bus_NRFD);

        activate_unit (ma_id, HPIB_CNTLR_UPTR, Data_Delay); /* schedule the next reception */
        break;
    }                                                   /* end of state dispatch */

if (ma->state != entry_state)
    tpprintf (dptr, TRACE_STATE, "Device %u transitioned from %s state to %s state\n",
              address, if_state_names [entry_state], if_state_names [ma->state]);

return;
}


/* Execute an Amigo tape command.

   This routine is called to execute tape and bus commands.  On entry, the
   "ma_id" parameter is the instance identifier of the MA device connected to
   the bus, the "address" parameter contains the bus address of the device,
   which will be a listener, and the "bus" parameter contains the set of eight
   control and eight data lines representing the HP-IB state.  The command to
   execute is present in the "opcode" field of the device state structure, and
   any parameter required by the command has already been received and is
   present in the "parameter" state field.

   This routine implements the following Amigo tape commands:

     Bus  Sec  Operation
     ---  ---  ---------------------
      L   00   Write Data
      L   01   Tape Command
      L   07   End Command
      L   10   Amigo Clear
      L   1E   Write Loopback

      T   00   Read Data
      T   01   Read Status
      T   02   Read Byte Count
      T   10   Device Specified Jump
      T   1E   Read Loopback

      U   MSA  Amigo Identify


   Implementation notes:

    1. The Write Loopback command accepts 128 bytes from the host and saves them
       in the first 128 bytes of the state buffer.  They are sent back to the
       host for a Read Loopback command, along with a 129th byte of value 177
       octal.  To avoid corrupting the buffer by an intervening Device Specified
       Jump command, the DSJ byte is always stored at buffer index 128 after the
       last Write Loopback byte.

    2. An Amigo Clear command is ignored by the HP-IB controller.  However, the
       Selected Device Clear that follows in the protocol sequence does clear
       the controller.
*/

static void execute_command (uint32 ma_id, uint32 address, BUS_CABLE bus)
{
MA_STATE   *const ma    = mas [ma_id].sptr;     /* the pointer to the channel state */
DEVICE     *const dptr  = mas [ma_id].dptr;     /* the pointer to the controlling device */
CNTLR_VARS *const cvptr = &ma->controller;      /* the pointer to the tape controller */
uint32            tape_status;
t_bool            reported = FALSE;             /* TRUE if the command trace has been reported */
t_bool            unsupported = FALSE;          /* TRUE if the command or utility is unsupported */

if (ma->command == Tape_Command) {                      /* if this is a tape command */
    if (ma->parameter > LAST_OPCODE)                    /*   then if the parameter value is invalid */
        ma->opcode = Invalid_Opcode;                    /*     then tag it */
    else                                                /*   otherwise */
        ma->opcode = to_opcode [ma->parameter];         /*     convert to the controller opcode */

    tpprintf (dptr, TRACE_INCO, "Device %u Tape Command %s started\n",
              address, tl_opcode_name (ma->opcode));
    }

else
    tpprintf (dptr, TRACE_INCO, "Device %u %s started\n",
              address, if_command_names [ma->command]);

ma->index = 0;                                          /* reset the buffer index for the new command */

switch (ma->command) {                                  /* dispatch the command opcode */

    case Write_Data:
        ma->count = 0;                                  /* reset the buffer byte counter */
        break;


    case Tape_Command:
        if (ma->opcode >= Select_Unit_0 && ma->opcode <= Select_Unit_3) {   /* if it's a unit select command */
            ma->attention &= CLEAR_ATTN (ma->opcode);                       /*   then clear the attention bits */
            ma->auto_select = FALSE;                                        /*     and disable auto-selection */
            }

        tpprintf (dptr, TRACE_CMD, "Device %u Tape Command %s executing\n",
                  address, tl_opcode_name (ma->opcode));

        reported = TRUE;                                /* the command has been reported */

        ma->data_bus = (CNTLR_IBUS) ma->opcode;         /* put the opcode on the data bus */
        ma->flag_bus |= CMRDY | CMXEQ;                  /*   and assert command ready and execute on the flag bus */

        call_controller (ma_id, NULL);                  /* tell the controller to start the command */

        ma->byte_count = 0;                             /* clear the read/write byte counter */
        break;


    case End_Command:
        tpprintf (dptr, TRACE_CMD, "Device %u End | %s executing\n",
                  address, fmt_bitset (ma->parameter, end_format));

        reported = TRUE;                                /* the command has been reported */

        if (ma->parameter & End_Abort)                  /* if the Abort Current Tape Operation bit is set */
            tl_clear (cvptr);                           /*   then clear the controller */

        if (ma->parameter & End_Clear_DSJ)              /* if the Clear DSJ bit is set */
            ma->dsj = 0;                                /*   then clear the DSJ response */

        if (ma->parameter & End_Clear_Poll)             /* if the Clear Poll Response bit is set */
            set_response (ma_id, address, CLEAR);       /*   then disable the poll response */

        if (ma->parameter & End_Inhibit_Poll)           /* if the Inhibit Poll Response for Read Bursts bit is set */
            ma->flag_bus |= EOD;                        /*   then set the EOD indication */

        if (ma->parameter & End_Enable_Response) {      /* if the Enable Rewind Completion or Auto-Select bit is set */
            ma->auto_select = TRUE;                     /*   then enable auto-selection */
            check_attention (ma_id, address);           /*     and check for a pending drive attention */
            }
        break;


    case Amigo_Clear:
        tpprintf (dptr, TRACE_INCO, "Device %u Amigo Clear ignored\n",
                  address);

        reported = TRUE;                                /* suppress the "command is executing" report */

        ma->state = Command_Ready;                      /* an Amigo Clear is ignored by the controller */
        break;


    case Write_Loopback:
        ma->count = 128;                                /* set the count of bytes to accept */
        break;


    case Read_Data:
        ma->count = 0;                                  /* reset the buffer byte counter */
        break;


    case Read_Status:
        tape_status = tl_status (cvptr);                /* get the current tape status */

        ma->buffer [0] = LOWER_BYTE (UPPER_WORD (tape_status)); /* break the status word */
        ma->buffer [1] = UPPER_BYTE (tape_status);              /*   into its constituent bytes */
        ma->buffer [2] = LOWER_BYTE (tape_status)               /*     for return to the host */
                           | (uint8) ma->interface_status;      /*       and merge in the interface state */

        ma->count = 3;                                  /* set the buffer byte count */
        ma->interface_status = 0;                       /*   and clear the power restored and parity error bits */

        tpprintf (dptr, TRACE_INCO, "Device %u status is %s, %sunit %u, %s\n",
                  address, fmt_bitset (ma->buffer [0], status_1_format),
                  fmt_bitset (ma->buffer [1], status_2_format), S2_UNIT (ma->buffer [1]),
                  fmt_bitset (ma->buffer [2], status_3_format));
        break;


    case Read_Byte_Count:
        ma->buffer [0] = UPPER_BYTE (ma->byte_count);   /* return the 16-bit byte count */
        ma->buffer [1] = LOWER_BYTE (ma->byte_count);   /*   from the last read or write operation */

        ma->count = 2;                                  /* set the buffer byte count */

        tpprintf (dptr, TRACE_CMD, "Device %u Read Byte Count %u executing\n",
                  address, ma->byte_count);

        reported = TRUE;                                /* the command has been reported */
        break;


    case Device_Specified_Jump:
        ma->buffer [128] = (uint8) ma->dsj;             /* save the DSJ in the buffer */

        ma->index = 128;                                /* set the starting buffer index */
        ma->count = 1;                                  /*   and the buffer count */

        tpprintf (dptr, TRACE_CMD, "Device %u Device Specified Jump %u executing\n",
                  address, ma->dsj);

        reported = TRUE;                                /* the command has been reported */

        ma->dsj = 0;                                    /* clear the DSJ after reporting it */
        set_response (ma_id, address, CLEAR);           /*   and disable the poll response */
        break;


    case Read_Loopback:
        ma->count = 129;                                /* set the (preloaded) buffer count */
        ma->buffer [128] = 0177;                        /*   and add the prescribed final byte */
        break;


    case Amigo_Identify:                                /* Identify */
        ma->buffer [0] = 0x01;                          /* set the code */
        ma->buffer [1] = 0x83;                          /*   to identify the HP-IB 7970E */

        ma->index = 0;                                  /* set the starting buffer index */
        ma->count = 2;                                  /*   and the count of bytes to send */

        tpprintf (dptr, TRACE_CMD, "Device %u Amigo Identify %02X%02XH executing\n",
                  address, ma->buffer [0], ma->buffer [1]);

        reported = TRUE;                                /* the command has been reported */

        ma->state = Command_Send;                       /* set up the next state */
        break;


    default:                                            /* commands not implemented */
        unsupported = TRUE;                             /*   or unsupported opcodes are ignored */
        break;
    }


if (cmd_props [ma->command].use_controller)                     /* if the command uses the bus controller */
    activate_unit (ma_id, HPIB_CNTLR_UPTR, Activation_Delay);   /*   then schedule the event service */

if (unsupported)
    tpprintf (dptr, TRACE_CMD, "Device %u %s command not implemented\n",
              address, if_command_names [ma->command]);

else if (not reported)
    tpprintf (dptr, TRACE_CMD, "Device %u %s executing\n",
              address, if_command_names [ma->command]);

return;                                                 /* return with the command initiated or completed */
}


/* End an active channel transaction.

   This routine is called to end a channel transaction in progress.  Termination
   occurs when a device is unaddressed in the middle of a data transmission
   during a command.  This usually represents an error, except in the case of
   burst transfers, which are expected to terminate at the end of the burst.

   On entry, the "ma_id" parameter is the instance identifier of the MA device
   connected to the bus, and the "bus_group" parameter contains the command
   group from the current bus state.

   If the Primary Command Group is passed, then the termination results from an
   IFC (interface clear) assertion, and so represents a channel abort.  If the
   Listen Address Group or Talk Address Group, then the termination is the
   (likely) result of a burst transfer and so is expected.  In either case, the
   bus controller state is changed to wait for the completion of the current
   transfer.


   Implementation notes:

    1. The HP-IB controller exists is both "old" and "new" versions.  A burst
       write of less than 64 bytes that does not end with EOI fails with DSJ = 1
       and a Data Timing Error with the new controller.  This succeeds with the
       old controller.  We implement the former behavior, and the tape
       diagnostic tests for this condition, although this difference is not
       mentioned in the interface manual.
*/

static void end_transfer (uint32 ma_id, BUS_DATA bus_group)
{
MA_STATE *const ma      = mas [ma_id].sptr;             /* the pointer to the channel state */
DEVICE *const   dptr    = mas [ma_id].dptr;             /* the pointer to the controlling device */
const  uint32   address = ADDRESS (dptr->flags);        /* the bus address of the device containing the unit */

if (bus_group == BUS_PCG)                                               /* if the end is unexpected */
    tpprintf (dptr, TRACE_CMD, "Device %u channel abort\n", address);   /*   then report an abort */

tpprintf (dptr, TRACE_STATE, "Device %u %s transitioned from %s state to %s state\n",
          address, tl_unit_name (ma->active ? TL_CNTLR_UNIT : TL_HPIB_UNIT),
          if_state_names [ma->state], if_state_names [Completion_Wait]);

ma->state = Completion_Wait;                            /* set the completion state */
hpib_service (HPIB_CNTLR_UPTR);                         /*   and call the event routine to end the transaction */

if (bus_group == BUS_LAG                                /* if the controller is listening */
  && not (ma->flag_bus & EOD) && ma->count < 64)        /*   and a burst < 64 bytes did not end with EOI */
    ma->flag_bus |= OVRUN;                              /*     then signal a Data Timing Error */

return;
}


/* Enable or disable a device's parallel poll response.

   This routine changes the poll response for a specified bus device.  It
   corresponds in hardware to a device's IEEE-488 PPE (parallel poll enable) or
   PPD (parallel poll disable) state.  When the controller conducts a parallel
   poll, the responses of all eight bus devices are placed on the data bus for
   retrieval by the controller.  An enabled poll response indicates to the host
   that a device is requesting service.

   On entry, the "ma_id" parameter is the instance identifier of the MA device
   connected to the bus, the "address" parameter is the bus address of the
   device whose poll response is changing, and the "response" parameter
   indicates whether the response should be set (enabled) or cleared (disabled).
   The poll response bit corresponding to the bus address is included or
   excluded from the set of poll responses.

   If a parallel poll is in progress when a poll response is set, the poll is
   conducted again to reflect the new response.


   Implementation notes:

    1. The set of poll responses is maintained by the channel to improve
       performance, so that we do not need to query each device during a poll to
       ascertain its response.
*/

static void set_response (uint32 ma_id, uint32 address, FLIP_FLOP response)
{
static const char *response_name [] = {         /* poll response names, indexed by Boolean response */
    "disabled",                                 /*   FALSE = poll response is disabled */
    "enabled"                                   /*   TRUE  = poll response is enabled */
    };
MA_STATE *const ma            = mas [ma_id].sptr;        /* the pointer to the channel state */
const BUS_DATA  old_responses = ma->response_set;        /* the current set of poll responses */
DEVICE   *const dptr          = mas [ma_id].dptr;        /* the pointer to the controlling device */
const uint32    channel       = CHANNEL (dptr->flags);   /* the channel address of the bus controller */

if (response == SET)                                    /* if the poll response is set */
    ma->response_set |= HPIB_BIT (address);             /*   then include the poll bit in the response */
else                                                    /* otherwise we are clearing the poll response */
    ma->response_set &= ~HPIB_BIT (address);            /*   so remove the poll bit */

if (ma->response_set != old_responses) {                /* if the poll response has changed */
    tpprintf (dptr, TRACE_STATE, "Device %d parallel poll response %s\n",
              address, response_name [response]);

    if (ma->poll_active && response == SET)                 /* if a parallel poll is in progress */
        ma->bus = hpib_source (channel, address, ma->bus);  /*   then conduct it again with the new response */
    }
return;
}


/* Activate a controller unit.

   This routine activates the SIMH unit belonging to the instance of the MA
   device specified by the "ma_id" parameter and pointed to by the "uptr"
   parameter with a delay specified by the "delay" parameter.  The "delay" is an
   index into the "bus_delays" array, whose fields specify the delay scheduling
   time and the reason for the delay.  If tracing is enabled, the activation is
   logged to the debug file.

   The status of the activation is returned to the caller.
*/

static t_stat activate_unit (uint32 ma_id, UNIT *uptr, DELAY delay)
{
MA_STATE *const ma   = mas [ma_id].sptr;        /* the pointer to the channel state */
DEVICE   *const dptr = mas [ma_id].dptr;        /* the pointer to the controlling device */

if (uptr->next != NULL)                                 /* if the unit is already scheduled */
    return SCPE_OK;                                     /*   then return with nothing to do */

tpprintf (dptr, TRACE_SERV, "Device %u %s state %s %s delay %d service scheduled\n",
          ADDRESS (dptr->flags), tl_unit_name (UNIT (uptr->flags)),
          if_state_names [ma->state], bus_delays [delay].name,
          bus_delays [delay].ticks);

return sim_activate (uptr, bus_delays [delay].ticks);   /* activate the unit and return the result */
}


/* Call the tape controller.

   The abstract tape controller connects to the HP-IB interface via 16-bit data,
   flag, and function buses.  The controller monitors the flag bus and reacts to
   the interface changing the flag states by placing or accepting data on the
   data bus and issuing commands to the interface via the function bus.

   In simulation, a call to the tl_controller routine informs the controller of
   a (potential) change in flag state.  The current set of flags and the data
   bus value are supplied, and the controller returns a combined set of
   functions and a data bus value.

   The controller must be called any time there is a change in the state of the
   interface or the drive units.  Generally, the cases that require notification
   are when the interface:

     - has a new command to execute
     - has detected the channel starting, ending, or aborting the transfer
     - has a new data byte available to send
     - has obtained the last data byte received
     - has received a unit service event notification
     - has detected the mounting of the tape reel on a drive
     - has detected a drive being placed online or offline

   The set of returned functions is processed sequentially, updating the
   interface state as indicated.  The functions have the following significance:

    IFGTC -- Interface Get Command.  The supplied command has been processed,
             and the command classification is available on the function bus.
             The classification will be Class_Invalid if the command was not
             executed.

    IFIN  -- Interface In.  A data byte is unloaded from the FIFO and is
             available on the function bus.

    DVEND -- Device End.  The last byte of the record has been returned.

    STDFL -- Set Data Flag.  Another byte is available to be read from the FIFO.

    IFOUT -- Interface Out.  The data byte supplied has been loaded into the
             FIFO, and there is space for another byte.

    RQSRV -- Request Service.  The FIFO level is more than half full for a read,
             or is less than half full for a write, or the command is ready to
             proceed past a Wait instruction in the channel program.  A parallel
             poll response should be asserted to advance the channel program.

    STINT -- Set Interrupt.  A data transfer timing error has occurred, and DSJ
             should be set to 1 to indicate an error condition.

    STCFL -- Set Control Flag.  The command is either invalid and has been
             rejected, or it is valid and has completed.  A parallel poll
             response should be asserted to inform the channel program
             processor.

    DATTN -- Drive Attention.  A tape drive unit is requesting attention, and
             the unit number is available on the function bus.  Unit attention
             is only recognized when the controller is idle.

    SCPE  -- SCP Error Status.  A fatal error was returned from a tape library
             call and the SCPE error code is present on the function bus.  If
             the error was recoverable, SCPE_OK will be returned, although
             status should be checked for the nature of the error.

   Because the tape is a synchronous device, overrun or underrun can occur if
   the bus controller is not ready when the drive must transfer data.  There are
   four conditions that lead to an overrun or underrun:

    1. The drive is ready with a tape read byte, but the interface FIFO is full.

    2. The drive needs a tape write byte, but the interface FIFO is empty.

    3. The GIC attempts to read a byte, but the FIFO is empty.

    4. The GIC attempts to write a byte, but the FIFO is full.

   The tape controller detects the first two conditions and returns the STINT
   function, which sets the data overrun flag.  The hardware design of the
   interface prevents the last two conditions, as the GIC will wait until the
   interface asserts DAV and sources a byte, or the interface asserts NRFD to
   hold off the GIC transfer.


   Implementation notes:

    1. In hardware, data overrun and underrun are detected as each byte is moved
       between the tape unit and the FIFO.  In simulation, OVRUN will not be
       asserted when the controller is called with the full or empty buffer;
       instead, it will be asserted for the next controller call after
       receiving the STINT function.  Because the controller will be called for
       the tape stop phase, and because OVRUN isn't checked until that point,
       this "late" assertion does not affect overrun or underrun detection.

    2. When the controller returns IFIN, STDFL, and a data byte, it can be
       sourced to the bus without checking for NRFD.  This is because the first
       transfer occurred as a result of ATN and NRFD denying, and subsequent
       transfers are not initiated by the STDFL handler if NRFD is asserted.  In
       that case, denial of NRFD will resume the transfer.

    3. The HP-IB controller does not set DSJ = 1 when a tape mark is seen as a
       result of a Forward Space File or a Backspace File, even though the End
       of File bit in status byte 1 is set.
*/

static t_stat call_controller (uint32 ma_id, UNIT *uptr)
{
MA_STATE   *const ma      = mas [ma_id].sptr;           /* the pointer to the channel state */
DEVICE     *const dptr    = mas [ma_id].dptr;           /* the pointer to the controlling device */
CNTLR_VARS *const cvptr   = &ma->controller;            /* the pointer to the tape controller */
const uint32      channel = CHANNEL (dptr->flags);      /* the channel address of the bus controller */
const uint32      address = ADDRESS (dptr->flags);      /* the bus address of the device containing the unit */
CNTLR_IFN_IBUS    result;
CNTLR_IFN_SET     command_set;
CNTLR_IFN         command;
uint32            unit_bit;
t_stat            status = SCPE_OK;

result = tl_controller (cvptr, uptr,                    /* call the controller to start or continue a command */
                        ma->flag_bus, ma->data_bus);

command_set = TLIFN (result);                           /* get the set of interface commands to execute */

while (command_set) {                                   /* process the set of returned interface commands */
    command = TLNEXTIFN (command_set);                  /* isolate the next command */

    switch (command) {                                  /* dispatch an interface command */

        case IFIN:                                      /* Interface In */
            ma->bus = Bus_DAV | TLIBUS (result);        /* set up the data byte */

            if (TLIFN (result) & DVEND)                 /* if the transfer is complete */
                ma->bus |= Bus_EOI;                     /*   then tag the last byte with EOI */

            ma->bus = hpib_source (channel, address, ma->bus);  /* send the byte to the bus */

            ma->byte_count++;                           /* count the transfer */
            break;


        case IFOUT:                                     /* Interface Out */
            ma->bus = hpib_source (channel, address,    /* deny NRFD to remove the handshake hold */
                                   ma->bus & ~Bus_NRFD);

            ma->flag_bus &= ~DTRDY;                     /* clear the data ready flag */
            set_response (ma_id, address, CLEAR);       /*   and disable the poll response */
            break;


        case IFGTC:                                     /* Interface Get Command */
            ma->flag_bus &= ~(CMRDY | CMXEQ);           /* clear the command flags */

            ma->command_class = (CNTLR_CLASS) TLIBUS (result);  /* save the command classification */
            break;


        case STDFL:                                                 /* Set Data Flag */
            if (ma->command == Read_Data                            /* if data is being read */
              && not (ma->bus & (Bus_ATN | Bus_NRFD)))              /*   and the bus is ready for it */
                activate_unit (ma_id, TAPE_CNTLR_UPTR, Data_Delay); /*     then activate the unit to return the data */
            break;


        case STINT:                                     /* Set Interrupt */
            ma->dsj = 1;                                /* set the error state */
            ma->flag_bus |= OVRUN;                      /*   and flag a data overrun */
            break;


        case STCFL:                                     /* Set Command Flag */
            ma->flag_bus = INTOK;                       /* reset the interface flags */

            tpprintf (dptr, TRACE_INCO, "Device %u Tape Command %s completed with DSJ %u\n",
                      address, tl_opcode_name (ma->opcode), ma->dsj);

        /* fall through into the RQSRV case */

        case RQSRV:                                     /* Request Service */
            set_response (ma_id, address, SET);         /* enable the poll response */
            break;


        case DVEND:                                     /* Device End */
            set_response (ma_id, address, CLEAR);       /* disable the poll response */
            break;


        case DATTN:                                         /* Drive Attention */
            unit_bit = BIT (DATTN_UNIT (TLIBUS (result)));  /* get the unit bit of the drive requesting attention */

            if (TLIBUS (result) & DATTN_ONLINE)             /* if attention is for an online transition */
                ma->attention |= TO_ONLINE_ATTN (unit_bit); /*   then set the bit in the online field */
            else                                            /* otherwise attention is for rewind completion */
                ma->attention |= TO_REWIND_ATTN (unit_bit); /*   so set the bit in the rewind field */

            if (ma->auto_select)                        /* if rewind completion/auto-selection is enabled */
                check_attention (ma_id, address);       /*   then process the attention bit */
            break;


        case SCPE:                                      /* SCP Error Status */
            if (cvptr->call_status != MTSE_TMK          /* if the status is not "tape mark seen" */
              || ma->opcode != Forward_Space_File       /*   or the command is not Forward Space File */
              && ma->opcode != Backspace_File)          /*   and not Backspace File */
                ma->dsj = 1;                            /*     then indicate an error condition */

            status = TLIBUS (result);                   /* get the status code */
            break;
        }                                               /* all cases are handled */

    command_set &= ~command;                            /* remove the current command from the set */
    }                                                   /*   and continue with the remaining commands */

return status;                                          /* return the result of the call */
}


/* Check for drive attention.

   This routine is called when rewind completion or online auto-selection is
   enabled, and the tape controller returns the DATTN ("Drive Attention")
   function as a result of a drive going online or a rewind completing.  It is
   also called when an End - Enable Rewind Completion or Auto-Select command is
   received, in case a prior DATTN had seen while this option was disabled.

   On entry, the "ma_id" parameter is the instance identifier of the MA device
   connected to the bus, and the "address" parameter is the bus address of the
   device to check.  If an attention bit is set, the routine obtains the
   highest-priority condition from the set.  A rewind condition will return a
   bit number from 0-3, corresponding to units 0-3, and DSJ is set to 0.  An
   online condition will return a bit number from 4-7, also corresponding to
   units 0-3. In this case, DSJ is set to 1, and a Unit Select command for the
   unit number is sent to the tape controller.  In both cases, parallel poll
   response is asserted, and the serviced condition bit is cleared.


   Implementation notes:

    1. It is safe to call the tape controller with the Unit Select command
       because drive attention only occurs when the controller is idle.
*/

static void check_attention (uint32 ma_id, uint32 address)
{
MA_STATE *const ma    = mas [ma_id].sptr;       /* a pointer to the channel state */
DEVICE   *const dptr  = mas [ma_id].dptr;       /* a pointer to the controlling device */
uint32          unit, bit_number;

if (ma->attention) {                                    /* if a unit is requesting attention */
    bit_number = LSB_NUMBER (ma->attention);            /*   then get the bit number of the lowest unit */

    if (bit_number < DRIVE_COUNT) {                     /* if this is a rewind completion notice */
        ma->dsj = 0;                                    /*   then set DSJ to 0 */

        tpprintf (dptr, TRACE_CMD, "Device %u %s rewind completion response\n",
                  address, tl_unit_name (bit_number));
        }

    else {                                              /* otherwise this is an online transition notice */
        ma->dsj = 1;                                    /*   so set DSJ to 1 to indicate */

        unit = LSB_NUMBER (ONLINE_ATTN (ma->attention));    /* get the number of the lowest unit */

        ma->opcode = Select_Unit_0 + unit;              /* set up the appropriate Select Unit command */
        ma->data_bus = (CNTLR_IBUS) ma->opcode;         /*   and place it on the data bus */
        ma->flag_bus = CMRDY | CMXEQ | INTOK;           /*     and assert command ready and execute on the flag bus */

        call_controller (ma_id, NULL);                  /* tell the controller to start the command */

        tpprintf (dptr, TRACE_CMD, "Device %u %s auto-selected\n",
                  address, tl_unit_name (unit));
        }

    set_response (ma_id, address, SET);                 /* enable the parallel poll response */
    ma->attention &= ~BIT (bit_number);                 /*   and clear the attention bit just serviced */
    }

return;
}
