/* hp3000_cpu.c: HP 3000 Central Processing Unit simulator

   Copyright (c) 2016-2024, J. David Bryan

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
   AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
   ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
   WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   Except as contained in this notice, the name of the author shall not be used
   in advertising or otherwise to promote the sale, use or other dealings in
   this Software without prior written authorization from the author.

   CPU          HP 3000 Series III/58 Central Processing Unit

   08-Nov-24    JDB     Corrected DUMP modifier entry order
   22-Jul-24    JDB     Added the "cpu_is_running" global routine
   01-Jul-24    JDB     Added CSMAP register
                        Added CST Expansion firmware option
   15-Apr-24    JDB     Adjusted postlude status validity checks
   23-Mar-24    JDB     Added system halt 5 for CST length = 0
   22-Mar-24    JDB     Moved PFARS from unit flags to system configuration
                        Removed "set_pfars" routine
                        Use "mem_size" global as replacement for MEMSIZE macro
   21-Mar-24    JDB     Moved "cpu_interrupt_pending" from the base instruction module
                        Changed "cpu_run_mode_interrupt" from global to local
   18-Mar-24    JDB     The SHOW CPU SPEED command is now CPU-specific
   14-Mar-24    JDB     Moved "cpu_exit_procedure" to the base instruction module
   09-Mar-24    JDB     Added "cpu_bank_mask" global
   29-Feb-24    JDB     Now shows base registers before a trap is taken
   04-Jan-24    JDB     Added a cold-load simulator stop option
   29-Dec-23    JDB     "imb_initialize" now returns a device number for pending IRQ
   28-Dec-23    JDB     Moved "cpu_start_dispatcher" to the base instruction module
                        Moved "hp_device_conflict" from the system module
   26-Oct-23    JDB     Added IRQ tracing
   25-Oct-23    JDB     pclk_interval now set in cpu_reset
   16-Oct-23    JDB     First successful run of MPE V/E through startup and shutdown
   08-Oct-23    JDB     First successful run of DUS Version 3.06
   07-Oct-23    JDB     Added Amigo clock declarations
   30-Sep-23    JDB     Added "cpu_imb_interface" routine to provide IMB memory
                        Added TRACE_IMBUS flag to trace IMB memory accesses
   03-Aug-23    JDB     Now synchronizes the PCLK service with the keyboard poll unit
                        Added "hp_reset_poll" call to "cpu_reset"
   26-Mar-23    JDB     Cleaned up unused macro
   01-Mar-23    JDB     Added calls to imb_initialize and cpp_service in sim_instr
   06-Jan-23    JDB     Traced stops are now always reported as instruction traces
   25-Nov-22    JDB     Added Starfish cold loading support
   07-Sep-22    JDB     Added check for "iop_initialize" failure
   21-Jan-22    JDB     Set SIM_SW_HIDE if sim_instr status may be ignored if quiet
   26-Oct-21    JDB     Renamed DEB_xxx trace flags to TRACE_xxx
   07-Apr-21    JDB     Added "Series III+" machine
                        Expanded bank register widths from 4 to 6 bits
   01-Apr-21    JDB     Removed REG_FIT from register definitions (obsolete)
   11-Oct-20    JDB     Moved UNIT_OPTS to hp3000_cpu.h for updating ease
   24-Sep-20    JDB     Traps now trace the register set
   05-Sep-20    JDB     Enabled EIS option
   26-Aug-20    JDB     Corrected line ends for trace to stdout
   04-Jul-20    JDB     Postlude trace now calls "sim_error_text" unconditionally
   09-Dec-19    JDB     Replaced debugging macros with tracing macros
   20-Jun-19    JDB     Added wait time to process clock trace
   08-Apr-19    JDB     Suppress stop messages for step and breakpoints in DO files
   07-Apr-19    JDB     Moved breakpoint test before run/halt-mode test
   21-Feb-19    JDB     Resuming into PAUS with pending IRQ now sets P correctly
                        Moved "debug_save" from global to "sim_instr" local
   06-Feb-19    JDB     Corrected trace report for simulation stop
   27-Dec-18    JDB     Revised fall through comments to comply with gcc 7
   05-Sep-17    JDB     Removed the -B (binary display) option; use -2 instead
                        Changed REG_A (permit any symbolic override) to REG_X
   19-Jan-17    JDB     Added comments describing the OPND and EXEC trace options
   29_Dec-16    JDB     Changed the status mnemonic flag from REG_S to REG_T
   07-Nov-16    JDB     Renamed cpu_byte_to_word_ea to cpu_byte_ea
   03-Nov-16    JDB     Added zero offsets to the cpu_call_procedure calls
   01-Nov-16    JDB     Added per-instruction trace capability
   24-Oct-16    JDB     Renamed SEXT macro to SEXT16
   22-Sep-16    JDB     Moved byte_to_word_address from hp3000_cpu_base.c
   21-Sep-16    JDB     Added CIS/NOCIS option
   01-Sep-16    JDB     Add power fail/power restore support
   23-Aug-16    JDB     Add module interrupt support
   14-Jul-16    JDB     Implemented the cold dump process
                        Changed "loading" EXEC_STATE to "waiting"
   11-Jul-16    JDB     Change "cpu_unit" from a UNIT to an array of one UNIT
   08-Jun-16    JDB     Corrected %d format to %u for unsigned values
   16-May-16    JDB     ACCESS_PROPERTIES.name is now a pointer-to-constant
   13-May-16    JDB     Modified for revised SCP API function parameter types
   24-Mar-16    JDB     Changed memory word type from uint16 to MEMORY_WORD
   21-Mar-16    JDB     Changed cpu_ccb_table type from uint16 to HP_WORD
   11-Mar-16    JDB     Fixed byte EA calculations with negative indexes
   22-Dec-15    JDB     First release version
   01-Apr-15    JDB     First successful run of MPE V/R through account login
   11-Dec-12    JDB     Created

   References:
     - HP 3000 Series II System Microprogram Listing
         (30000-90023, August 1976)
     - HP 3000 Series II/III System Reference Manual
         (30000-90020, July 1978)
     - Machine Instruction Set Reference Manual
         (30000-90022, June 1984)
     - HP 3000 Series 40/44 Computer Systems Reference/Training Manual
         (30090-90001, July 1981)
     - Series 64/68/70 Computer Systems Microcode Manual
         (30140-90045, October 1986)


   The HP 3000 is a family of general-purpose business computers that were sold
   by Hewlett-Packard from 1972 through 2001.  There are two major divisions
   within this family: the "classic" 16-bit, stack-oriented CISC machines, and
   the "Precision Architecture" 32-bit, register-oriented RISC machines that
   succeeded them.  All machines run versions of MPE, the "Multiprogramming
   Executive" operating system.

   Within the "classic" division, there are two additional subdivisions, based
   on the method used for peripheral connections: the original "SIO" machines,
   and the later "HP-IB" machines.  The I/O interfacing hardware differs between
   the two types of machines, as do the privileged I/O machine instructions.
   The user instruction sets are identical, as are the register sets visible to
   the programmer.  The I/O drivers are different to account for the hardware
   differences, and therefore they run slightly different versions of MPE.

   This implementation simulates two "classic" machines: the Series III and the
   Series 58.

   The Series III represents the high end of the SIO machines , which consist of
   the 3000 CX, the Series I, the Series II, and the Series III.  These machines
   use I/O interfaces based on the I/O Processor (IOP) bus.

   The CX and Series I, which is a repackaged CX, are essentially subsets of the
   Series II/III -- a smaller instruction set, limited memory size, and
   lower-precision floating-point instructions.  Simulation of these machines
   may be added in the future.

   The CX and Series I support 64K 16-bit words of memory.  The Series II
   supports up to 256K, divided into four banks of 64K words each, and the
   Series III extends this to 1024K words using 16 banks (although MPE V/R can
   accommodate bank numbers up to six bits in length).

   The Series 58 is in the upper-mid range of the HP-IB machines, which consist
   of some 19 models from the low-end Series 30 to the high-end Series 70.  The
   I/O interfaces for these machines connect to an Intermodule Bus (IMB).  The
   various models differ mainly in memory and I/O capacity and processor speed,
   with maximum memory sizes varying from 512K for the Series 30 to 8192K for
   the Series 70.  The Series 58 supports 4096K.

   Memory is divided into variable-length code and data segments, with the
   latter containing a program's global data area and stack.  Memory protection
   is accomplished by checking program, data, and stack accesses against segment
   base and limit registers, which can be set only by MPE.  Bounds violations
   cause automatic hardware traps to a handler routine within MPE.  Some
   violations may be permitted; for example, a Stack Overflow trap may cause MPE
   to allocate a larger stack and then restart the interrupted instruction.
   Almost all memory references are position-independent, so moving segments to
   accommodate expansion requires only resetting of the segment registers to
   point at the new locations.  Code segments are fully reentrant and shareable,
   and both code and data are virtual, as the hardware supports absent code and
   data segment traps.

   The classic 3000s are stack machines.  Most of the instructions operate on
   the value on the top of the stack (TOS) or on the TOS and the next-to-the-top
   of the stack (NOS).  To improve execution speed, the 3000 has a set of
   hardware registers that are accessed as the first four locations at the top
   of the stack, while the remainder of the stack locations reside in main
   memory.  A hardware register renamer provides fast stack pushes and pops
   without physically copying values between registers.

   In hardware, the stack registers are referenced internally as TR0-TR3 and
   externally as RA-RD.  An access to the RA (TOS) register is translated by the
   renamer to access TR0, TR1, etc. depending on which internal register is
   designated as the current top of the stack.  For example, assume that RA
   corresponds to TR0.  To push a new value onto the top of the stack, the
   renamer is adjusted so that RA corresponds to TR1, and the new value is
   stored there.  In this state, RB corresponds to TR0, the previous TOS (and
   current NOS) value.  Additional pushes rename RA as TR2 and then TR3, with RB
   being renamed to TR1 and then TR2, and similarly for RC and RD.  The number
   of valid TOS registers is given by the value in the SR register, and the
   first stack location in memory is given by the value in the SM register.
   Pops reverse the sequence: a pop with RA corresponding to TR3 renames RA to
   TR2, RB to TR1, etc.  When all four stack registers are in use, a push will
   copy the value in RD to the top of the memory stack (SM + 1) before the
   registers are renamed and the new value stored into RA.  Similarly, when all
   four stack registers are empty, a pop simply decrements the SM register,
   effectively deleting the top of the memory stack.

   Because of the renamer, the microcode can always use RA to refer to the top
   of the stack, regardless of which internal TR register is being used, and
   similarly for RB-RD.  Execution of each instruction begins with a stack
   preadjustment, if needed, that loads the TOS registers that will be used by
   the instruction from the top of the memory stack.  For example, if only RA
   contains a value (i.e., the SR register value is 1), the ADD instruction,
   which adds the values in RA and RB, will load RB with the value on the top of
   the memory stack, the SR count will be incremented, and the SM register will
   be decremented.  On the other hand, if both RA and RB contained values (SR >=
   2), then no preadjustment would be required before the ADD instruction
   microcode manipulated RA and RB.

   In simulation, the renamer is implemented by physically copying the values
   between registers, as this is much faster than mapping via array index
   values, as the hardware does.  A set of functions provides the support to
   implement the hardware stack operations:

     cpu_push       - empty the TOS register (SR++, caller stores into RA)
     cpu_pop        - delete the TOS register (SR--)
     cpu_queue_up   - move from memory to the lowest TOS register (SR++, SM--)
     cpu_queue_down - move from the lowest TOS register to memory (SR--, SM++)
     cpu_flush      - move all stack registers to memory (SR = 0 on return)
     cpu_adjust_sr  - queue up until the required SR count is reached
     cpu_mark_stack - write a stack marker to memory

   The renamer is described in US Patent 3,737,871 (Katzman, "Stack Register
   Renamer", June 1973).

   The MPE operating system is supported by special microcode features that
   perform code and data segment mapping, segment access bounds checking,
   privilege checking, etc.  The layout of certain in-memory tables is known to
   both the OS and the microcode and is used to validate execution of
   instructions.  For instance, every stack instruction is checked for a valid
   access within the stack segment boundaries, which are set up by the OS
   before program dispatch.  For this reason, the 3000 cannot be operated as a
   "bare" machine, because these tables would not have been initialized.
   Similarly, the "cold load" process by which the OS is loaded from storage
   media into memory is entirely in microcode, as machine instructions cannot be
   executed until the required tables are present in memory.

   This OS/microcode integration means that the microcode may detect conditions
   that make continued execution impossible.  An example would be a not-present
   segment fault for the segment containing the disc I/O driver.  If such a
   condition is detected, the CPU does a "system halt."  This fatal microcode
   error, distinct from a regular programmed halt, causes operation to cease
   until the CPU is reset.

   The CPU hardware for the SIO includes a free-running 16-bit process clock
   that increments once per millisecond whenever a user program is executing.
   MPE uses the process clock to charge CPU usage to programs, and thereby to
   users, groups, and accounts.  Instructions are provided to read (RCLK) and
   set (SCLK) the process clock.  A separate hardware clock present on an I/O
   interface provides the 100 msec. timing interval for MPE to maintain the
   time-of-day.

   The HP-IB machines do not have a dedicated hardware counter for the process
   clock, nor is there a dedicated hardware timer with a programmable interrupt
   interval for the time-of-day clock.  Instead, there is a simple 1 msec. CPU
   counter that provides periodic interrupts that the firmware intercepts and
   uses to provide a software process clock and a software programmable timer.


   The data types supported by the instruction set are:

     - 8-bit unsigned byte
     - 16-bit unsigned integer ("logical" format)
     - 16-bit two's-complement integer
     - 32-bit two's-complement integer
     - 32-bit sign-magnitude floating point
     - 64-bit sign-magnitude floating point

   Multi-word values are stored in memory with the most-significant words in the
   lowest addresses.  The stack is organized in memory from lower to higher
   addresses, so a value on the stack has its least-significant word on the top
   of the stack.

   Machine instructions are initially decoded by a sub-opcode contained in the
   four highest bits, as follows (note that the HP 3000 numbers bits from
   left-to-right, i.e., bit 0 is the MSB and bit 15 is the LSB):

       0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   0   0 |   1st stack opcode    |   2nd stack opcode    |  Stack
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+


       0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   0   1 | X |   shift opcode    |      shift count      |  Shift
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   0   1 | I |   branch opcode   |+/-|  P displacement   |  Branch
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   0   1 | X |  bit test opcode  |     bit position      |  Bit Test
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+


       0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   1   0 | 0   0   0   0 |  move op  | opts/S decrement  |  Move
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   1   0 | 0   0   0   0 |  special op   | 0   0 | sp op |  Special
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   1   0 | 0   0   0   1 |  option set   | option subset |  Firmware
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   1   0 |  imm opcode   |       immediate operand       |  Immediate
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   1   0 | field opcode  |    J field    |    K field    |  Field
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   1   0 |  register op  | SK| DB| DL| Z |STA| X | Q | S |  Register
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+


       0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   1   1 | 0   0   0   0 |  I/O opcode   |    K field    |  I/O (IOP)
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   1   1 | 0   0   0   0 |  cntl opcode  | 0   0 | cn op |  Control
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   1   1 |  program op   |            N field            |  Program
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   1   1 | immediate op  |       immediate operand       |  Immediate
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   1   1 |   memory op   |        P displacement         |  Memory
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     The memory, loop, and branch instructions occupy the remainder of the
     sub-opcodes (04-17):

       0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |   memory op   | X | I |         mode and displacement         |  Memory
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
                             | 0 | 0 |     P+ displacement 0-255     |
                             +---+---+---+---+---+---+---+---+---+---+
                             | 0 | 1 |     P- displacement 0-255     |
                             +---+---+---+---+---+---+---+---+---+---+
                             | 1 | 0 |    DB+ displacement 0-255     |
                             +---+---+---+---+---+---+---+---+---+---+
                             | 1 | 1 | 0 |   Q+ displacement 0-127   |
                             +---+---+---+---+---+---+---+---+---+---+
                             | 1 | 1 | 1 | 0 | Q- displacement 0-63  |
                             +---+---+---+---+---+---+---+---+---+---+
                             | 1 | 1 | 1 | 1 | S- displacement 0-63  |
                             +---+---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |   memory op   | X | I | s |     mode and displacement         |   Memory
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
                                 | 0 |    DB+ displacement 0-255     |
                                 +---+---+---+---+---+---+---+---+---+
                                 | 1 | 0 |   Q+ displacement 0-127   |
                                 +---+---+---+---+---+---+---+---+---+
                                 | 1 | 1 | 0 | Q- displacement 0-63  |
                                 +---+---+---+---+---+---+---+---+---+
                                 | 1 | 1 | 1 | S- displacement 0-63  |
                                 +---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   1   0   1 |loop op| 0 |+/-|    P-relative displacement    |  Loop
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 1   1   0   0 | I | 0   1 | > | = | < | P+- displacement 0-31 |  Branch
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   The HP-IB machines implement their model-specific instructions using two
   words, as follows:

       0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   1   0 | 0   0   0   0 | 0   1   0 | 0   0   1   0   0 |  Clock
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   0   0   0   0   0   0   0   0   0   0 |   operation   |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   1   0 | 0   0   0   0 | 1   1   0   0   0 | 0   1   0 |  I/O (IMB)
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   0   0   0   0   0   0   0   0   0   0 | I/O operation |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Optional firmware extension instruction sets occupy instruction codes
   020400-020777, except for the DMUL (020570) and DDIV (020571) instructions
   that are part of the base set, as follows:

       0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   1   0 | 0   0   0   1 | 0   1   1   1 | 1   0   0 | x |  DMUL/DDIV
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   1   0 | 0   0   0   1 | 0   0   0   0 | 1 | ext fp op |  Extended FP
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   1   0 | 0   0   0   1 | 0   0   0   1 |    APL op     |  APL
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   1   0 | 0   0   0   1 | 0   0   1   1 |   COBOL op    |  COBOL
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | 0   0   1   0 | 0   0   0   1 | 1 |  options  |  decimal op   |  Decimal
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Most instructions are defined by unique 16-bit codes that specify the
   operations, operands, addressing modes, shift counts, etc.  For each of these
   instructions, there is only one canonical form that encodes a given
   instruction (or instruction pair, in the case of stack instructions).  For
   example, the octal value 140003 is the only value that encodes the "BR P+3"
   instruction.

   There are also instruction codes that contain one or more bits designated as
   "reserved" in the Machine Instruction Set Reference Manual.  For some of
   these instructions, the reserved bits do not affect instruction decoding.
   Each instruction of this type has a defined canonical form -- typically with
   the reserved bits set to zero -- but will execute identically if one of the
   undefined forms is used.  For example, the "MOVE" instructions define bits
   12-13 as 00, but the bits are not decoded, so values of 01, 10, and 11 for
   these bits will result in identical execution.  Specifically, "MOVE 0" is
   encoded canonically as octal value 020020, but the undefined codes 020024,
   020030, and 020034 execute "MOVE 0" as well.

   For the rest of these instructions, the reserved bits are decoded and will
   affect the instruction interpretation.  An example of this is the "IXIT"
   instruction.  It also defines bits 12-13 as 00 (canonical encoding 020360),
   but in this case the bits must be 00 for the instruction to execute; any
   other value (e.g., undefined codes 020364, 020370, or 020374) executes as a
   "PCN" instruction, whose canonical encoding is 020362.

   Finally, some codes are not assigned to any instructions, or they are
   assigned to instructions that are supplied by optional firmware that is not
   present in the machine.  These instruction codes are designated as
   "unimplemented" and will cause Unimplemented Instruction traps if they are
   executed.  Examples are the stack operation code 072 in either the left-hand
   or right-hand position (i.e., instruction codes 0072xx and 00xx72), codes
   020410-020415 if the Extended Instruction Set firmware option is not
   installed, and codes 036000-036777.

   When the simulator examines the bit patterns of instructions to execute, each
   will fall into one of four categories:

    1. Defined (canonical) instruction encodings.

    2. Undefined (non-canonical) instruction encodings, where reserved fields
       are "don't care" bits (e.g., MOVE).

    3. Undefined (non-canonical) instruction encodings, where reserved fields
       are decoded (e.g., IXIT).

    4. Unimplemented instruction encodings (e.g., stack opcode 072 or EADD with
       no EIS installed).

   When examining memory or register values in instruction-mnemonic form, the
   names of the canonical instructions in category 1 are displayed in uppercase,
   as are the names of the non-canonical instructions in category 2.  The
   non-canonical instruction names in category 3 are displayed in lowercase.
   This is to indicate to the user that the instructions that will be executed
   may not be the ones expected from the decoding.  Instruction names in
   category 4 that correspond to supported firmware options are displayed in
   uppercase, regardless of whether or not the option is enabled.  Category 4
   encodings that do not correspond to instructions are displayed in octal.

   With the exception of the model-specific two-word instructions of the HP-IB
   machines, all of the base set instructions are one word in length.  Most of
   the firmware extension instructions occupy one word as well, although some
   are two words long.  If the first word of a two-word instruction is valid but
   the second is not, an Unimplemented Instruction trap will occur, with P
   pointing to the location after the second word.

   The simulator provides five stop conditions related to instruction execution
   that may be enabled with a SET CPU STOP=<stop> command:

     <stop>    Action
     --------  ------------------------------------
     LOOP      stop on an infinite loop
     PAUSE     stop on a PAUS instruction
     UNDEF     stop on an undefined instruction
     UNIMPL    stop on an unimplemented instruction
     COLDLOAD  stop when a cold-load trap is taken

   If an enabled stop condition is detected, execution ceases with the
   instruction pending, and control returns to the SCP prompt.  When simulation
   stops, execution may be resumed in two ways.  With the exception of the
   COLDLOAD stop, if the cause of the stop has not been remedied and the stop
   has not been disabled, resuming execution with CONTINUE, STEP, GO, or RUN
   will cause the stop to occur again.  Alternately, specifying the "-B" switch
   with any of the preceding commands will resume execution while bypassing the
   stop for the current instruction.  Resuming after a COLDLOAD stop continues
   with the cold-load trap; no bypass switch is needed.

   The LOOP option stops the simulator if it attempts to execute an instruction
   that enters an infinite loop (e.g., BR P+0).  The branch instructions TBA,
   TBX, BCC, BR, BCY, BNCY, BOV, and BNOV result in an infinite loop if the
   branch displacement is zero and the branch condition is true.  The remaining
   branch instructions cannot result in an infinite loop, as they all modify the
   CPU state and so eventually reach a point where they drop out of the loop.

   The PAUSE option stops the simulator if execution of a PAUS instruction is
   attempted.  This instruction normally suspends instruction fetching until an
   interrupt occurs.  Clearing the stop and resuming execution suspends the
   fetch/execute process until an external interrupt occurs.  Resuming with the
   stop bypassed continues execution with the instruction following the PAUS;
   this is the same action that occurs when pressing the HALT button and then
   the RUN button in hardware.

   The UNDEF option stops the simulator if execution of a non-canonical
   instruction from decoding category 3 (i.e., an instruction containing a
   decoded reserved bit pattern other than that defined in the Machine
   Instruction Set manual) is attempted.  The intent is to catch instructions
   containing reserved fields with values that change the meaning of those
   instructions.  Bypassing the stop will decode and execute the instruction in
   the same manner as the HP 3000 microcode.

   The UNIMPL option stops the simulator if execution of an instruction from
   decoding category 4 is attempted.  Bypassing the stop will cause an
   Unimplemented Instruction trap.  Instructions that depend on the presence of
   firmware options are implemented if the option is present and unimplemented
   if the option is absent.  For example, instruction code 020410 executes as
   the EADD instruction if the Extended Instruction Set is enabled but is
   unimplemented if the EIS is disabled.  It is displayed as EADD in either
   case.

   The COLDLOAD option stops the simulator when a cold-load trap is taken.  This
   occurs after a LOAD, START, or DUMP command has loaded an operating system
   into memory and is about to commence execution.  The stop allows simulator
   command files to perform any necessary configuration or set up any scripted
   replies before system startup begins.  If the stop is not set, HP-IB systems
   will proceed directly into OS execution, while SIO systems will stop and wait
   for a RUN, GO, or CONTINUE command before proceeding.

   The instructions in decoding category 2 whose non-canonical forms do not
   cause an UNDEF stop are:

           Canonical  Reserved
     Inst  Encoding     Bits    Defined As    Decoded As
     ----  ---------  --------  -----------   -----------
     SCAN   010600     10-15    0 0 0 0 0 0   x x x x x x
     TNSL   011600     10-15    0 0 0 0 0 0   x x x x x x
     MOVE   020000     12-13        0 0           x x
     MVB    020040     12-13        0 0           x x
     MVBL   020100        13          0             x
     SCW    020120        13          0             x
     MVLB   020140        13          0             x
     SCU    020160        13          0             x
     CMPB   020240     12-13        0 0           x x
     RSW    020300     12-14        0 0 0         x x x
     LLSH   020301     12-14        0 0 0         x x x
     PLDA   020320     12-14        0 0 0         x x x
     PSTA   020321     12-14        0 0 0         x x x
     LSEA   020340     12-13        0 0           x x
     SSEA   020341     12-13        0 0           x x
     LDEA   020342     12-13        0 0           x x
     SDEA   020343     12-13        0 0           x x
     PAUS   030020     12-15        0 0 0 0       x x x x

   The instructions in category 3 whose non-canonical forms cause an UNDEF stop
   are:

           Canonical  Reserved
     Inst  Encoding     Bits    Defined As  Decoded As
     ----  ---------  --------  ----------  ----------
     IXIT   020360     12-15     0 0 0 0     0 0 0 0
     LOCK   020361     12-15     0 0 0 1     n n 0 1
     PCN    020362     12-15     0 0 1 0     n n n 0
     UNLK   020363     12-15     0 0 1 1     n n 1 1

     SED    030040     12-15     0 0 0 x     n n n x

     XCHD   030060     12-15     0 0 0 0     0 0 0 0
     PSDB   030061     12-15     0 0 0 1     n n 0 1
     DISP   030062     12-15     0 0 1 0     n n n 0
     PSEB   030063     12-15     0 0 1 1     n n 1 1

     SMSK   030100     12-15     0 0 0 0     0 0 0 0
     SCLK   030101     12-15     0 0 0 1     n n n n
     RMSK   030120     12-15     0 0 0 0     0 0 0 0
     RCLK   030121     12-15     0 0 0 1     n n n n

   Where:

     x = 0 or 1
     n = any collective value other than 0

   In hardware, the SED instruction works correctly only if opcodes 030040 and
   030041 are used.  Opcodes 030042-030057 also decode as SED, but the status
   register is set improperly (the I bit is cleared, bits 12-15 are rotated
   right twice and then ORed into the status register).  In simulation, opcodes
   030042-030057 work correctly but will cause an UNDEF simulation stop if
   enabled.


   The CPU simulator provides extensive tracing capabilities that may be enabled
   with the SET CONSOLE DEBUG=<filename> and SET CPU DEBUG=<trace> commands.
   The trace options that may be specified are:

     Trace  Action
     -----  -------------------------------------------
     INSTR  trace instructions executed
     DATA   trace memory data accesses
     FETCH  trace memory instruction fetches
     REG    trace registers
     OPND   trace memory operands
     EXEC   trace matching instruction execution states
     IRQ    trace interrupt and trap requests
     PSERV  trace process clock service events
     IMBUS  trace IMB signals and data words

   A section of an example trace is:

     >>CPU fetch: 00.010342  020320    instruction fetch
     >>CPU instr: 00.010341  000300  ZROX,NOP
     >>CPU   reg: 00.006500  000000    X 000000, M i t r o c CCG
     >>CPU fetch: 00.010343  041100    instruction fetch
     >>CPU instr: 00.010342  020320  PLDA
     >>CPU  data: 00.000000  001340    absolute read
     >>CPU   reg: 00.006500  000001    A 001340, X 000000, M i t r o c CCG
     >>CPU fetch: 00.010344  037777    instruction fetch
     >>CPU instr: 00.010343  041100  LOAD DB+100
     >>CPU  data: 00.002100  123003    data read
     >>CPU   reg: 00.006500  000002    A 123003, B 001340, X 000000, M i t r o c CCL
     >>CPU fetch: 00.010345  023404    instruction fetch
     >>CPU instr: 00.010344  037777  ANDI 377
     >>CPU   reg: 00.006500  000002    A 000003, B 001340, X 000000, M i t r o c CCG
     >>CPU fetch: 00.010346  002043    instruction fetch
     >>CPU instr: 00.010345  023404  MPYI 4
     >>CPU   reg: 00.006500  000002    A 000014, B 001340, X 000000, M i t r o c CCG
     >>CPU fetch: 00.010347  020320    instruction fetch

   The INSTR option traces instruction executions.  Each instruction is printed
   before it is executed.  The two opcodes of a stack instruction are printed
   together before the left-hand opcode is executed.  If the right-hand opcode
   is not NOP, it is reprinted before execution, with dashes replacing the
   just-executed left-hand opcode.

   The DATA option traces reads from and writes to memory.  Each access is
   classified by the memory bank register that is paired with the specified
   offset; they are: dma, absolute, program, data, and stack.  DMA accesses
   derive their bank addresses from the banks specified in Set Bank I/O program
   orders.  Absolute accesses always use bank 0.  Program, data, and stack
   accesses use the bank addresses in the PBANK, DBANK, and SBANK registers,
   respectively.

   The FETCH option traces instruction fetches from memory.  These accesses are
   separated from those traced by the DATA option because fetches usually are of
   little interest except when debugging the fetch/execute sequence.  Because
   the HP 3000 has a two-stage pipeline, fetches load the NIR (Next Instruction
   Register) with the instruction after the instruction to be executed from the
   CIR (Current Instruction Register).

   The REG option traces register values.  Two sets of registers are printed.
   After executing each instruction, the currently active TOS registers, the
   index register, and the status register are printed.  After executing an
   instruction that may alter the base registers, the program, data, and stack
   segment base registers are printed.

   The OPND option traces memory byte operand values.  Some instructions take
   memory and register operands that are difficult to decode from DATA or REG
   traces.  This option presents these operands in a higher-level format.  The
   memory bank and address values are always those of the operands.  The operand
   data and values printed are specific to the instruction.  For example, the
   ALGN instruction prints its source and target operands, digit counts, and
   fraction counts, and the EDIT instruction displays its subprogram operations.

   The EXEC option traces the execution of instructions that match
   user-specified criteria.  When a match occurs, all CPU trace options are
   turned on for the duration of the execution of the matched instruction.  The
   prior trace settings are restored when a match fails.  This option allows
   detailed tracing of specified instructions while minimizing the log file size
   compared to a full instruction trace.

   The IRQ option traces hardware interrupts and microcode traps.  Dispatcher
   entries and interrupt returns are also traced.  The traces include the
   program counter at the time of the interrupt or trap.

   The PSERV option traces process clock event service entries.  Each trace
   reports whether or not the CPU was executing on the Interrupt Control Stack
   when the process clock ticked.  Execution on the ICS implies that the
   operating system is executing.  As the process clock ticks every millisecond,
   enabling PSERV tracing can quickly produce a large number of trace lines.

   The IMBUS option traces the IMB signals and data words sent to the memory
   controller of HP-IB systems.  For these systems, the controller resides on
   the Intermodule Bus and in simulation is hosted in this CPU module.

   The various trace formats are interpreted as follows:

     >>CPU instr: 00.010341  000300  ZROX,NOP
                  ~~ ~~~~~~  ~~~~~~  ~~~~~~~~
                  |    |       |       |
                  |    |       |       +-- instruction mnemonic(s)
                  |    |       +---------- octal data (instruction opcode)
                  |    +------------------ octal address (P)
                  +----------------------- octal bank (PBANK)


     >>CPU  data: 00.000000  001340    absolute read
     >>CPU  data: 00.002100  123003    data read
                  ~~ ~~~~~~  ~~~~~~    ~~~~~~~~~~~~~
                  |    |       |         |
                  |    |       |         +-- memory access classification
                  |    |       +------------ octal data (memory contents)
                  |    +-------------------- octal address (effective address)
                  +------------------------- octal bank (PBANK, DBANK, or SBANK)


     >>CPU fetch: 00.010342  020320    instruction fetch
                  ~~ ~~~~~~  ~~~~~~    ~~~~~~~~~~~~~~~~~
                  |    |       |         |
                  |    |       |         +-- memory access classification
                  |    |       +------------ octal data (instruction opcode)
                  |    +-------------------- octal address (P + 1)
                  +------------------------- octal bank (PBANK)


     >>CPU   reg: 00.006500  000002    A 123003, B 001340, X 000000, M i t r o c CCL
                  ~~ ~~~~~~  ~~~~~~    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                  |    |       |         |
                  |    |       |         +-- register values (from 0-4 TOS registers, X, STA)
                  |    |       +------------ octal stack register count (SR)
                  |    +-------------------- octal stack memory address (SM)
                  +------------------------- octal bank (SBANK)

     >>CPU   reg: 00.000000  000001    PB 010000, PL 025227, DL 001770, DB 002000, Q 006510, Z 007000
                  ~~ ~~~~~~  ~~~~~~    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                  |    |       |         |
                  |    |       |         +-- base register values
                  |    |       +------------ current code segment number (from STA)
                  |    +-------------------- zero
                  +------------------------- octal bank (DBANK)


     >>CPU  opnd: 00.135771  000000    DFLC '+','!'
     >>CPU  opnd: 00.045071  000252    target fraction 3 length 6,"002222"
                  ~~ ~~~~~~  ~~~~~~    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                  |    |       |         |
                  |    |       |         +-- operand-specific value
                  |    |       +------------ operand-specific data
                  |    +-------------------- octal address (effective address)
                  +------------------------- octal bank (PBANK, DBANK, or SBANK)


     >>CPU   irq: 00.001240  000006  external interrupt
     >>CPU   irq: 00.023736  000000  unimplemented instruction trap
                  ~~ ~~~~~~  ~~~~~~  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
                  |    |       |       |
                  |    |       |       +-- interrupt classification
                  |    |       +---------- parameter
                  |    +------------------ octal address (P) at interrupt
                  +----------------------- octal bank (PBANK) at interrupt


     >>CPU pserv: Process clock delay 3970 service entered on the user stack

    >>CPU  imbus: Memory controller received opcode Memory Write address 03015470 data 041407
    >>CPU  imbus: Memory controller returned data 000000 with signals ADN | DDN | PRO

   The process clock offers a user-selectable choice of calibrated or realistic
   timing.  Calibrated timing adjusts the clock to match actual elapsed time
   (i.e., wall-clock time).  Realistic timing bases the process-clock interval
   on machine instructions executed, using a mean instruction time of 2.5
   microseconds.  Running on a typical host platform, the simulator is one or
   two orders of magnitude faster than a real HP 3000, so the number of machine
   instructions executed for a given calibrated time interval will be
   correspondingly greater.  When the process clock is calibrated, the current
   simulation speed, expressed as a multiple of the speed of a real HP 3000
   Series III, may be obtained with the SHOW CPU SPEED command.  The speed
   reported will not be representative if the machine was executing a PAUS
   instruction when the simulator was stopped.

   When enabled by a SET CPU IDLE command, execution of a PAUS instruction will
   idle the simulator.  While idle, the simulator does not use any host system
   processor time.  Idle detection requires that the process clock and system
   clock be set to calibrated timing.  Idling is disabled by default.


   Implementation notes:

    1. Four timing sources in the simulator may be calibrated to wall-clock
       time.  These are the process clock, the system clock, the ATC poll timer,
       and the ADCC poll timer.  The process clock is always enabled and
       running, although the PCLK register only increments if the CPU is not
       executing on the ICS.  The system clock and poll timers run continuously
       if their respective devices are enabled.

       The four sources must be synchronized to allow efficient simulator
       idling.  This is accomplished by designating a separate keyboard poll
       timer as the master device.  This device runs continuously and calls the
       SCP timer calibration routines to track the host platform's wall clock.
       Each of the other sources set their wait times to the keyboard poll wait.
       All four devices will then enter their respective service routines
       concurrently.

    2. In hardware, the process clock period is fixed at one millisecond, and
       the system clock period, while potentially variable for the SIO systems,
       is set by MPE to one millisecond with an interrupt every 100 ticks.
       These periods are too short to allow the simulator to idle, as the host
       OS clock resolution is typically also one millisecond.

       Therefore, the process and system clock simulators are scheduled with
       ten-millisecond service times, and the PCLK and counter registers are
       incremented by ten for each event service.  To present the correct values
       when the registers are read, the counts are incremented by amounts
       proportional to the fractions of the service intervals that have elapsed
       when the reads occur.

    3. In simulation, the TOS renamer is implemented by permanently assigning
       the register names RA-RD to TR [0]-TR [3], respectively, and copying the
       contents between registers to pop and push values.  An alternate
       implementation approach is to use a renaming register, RN, that tracks
       the correspondence between registers and array entries, and to assign the
       register names dynamically using modular indices, e.g., RA is TR [RN], RB
       is TR [(RN + 1) & 3], etc.  In lieu of copying, incrementing and
       decrementing RN is done to pop and push values.

       Both implementations were mocked up and timing measurements made.  The
       results were that copying values is much faster than mapping via array
       index values, so this is the implementation chosen.

    4. When CST expansion firmware is installed, the physical/logical mapping
       state of the current code segment is kept in the CSMAP register, with bit
       1 set for physical mapping and clear for logical mapping.  This is the
       same format needed when writing stack markers and corresponds to the
       STMK_M value.  When the firmware is not installed, the CSMAP register
       value is irrelevant, except that it must not be written as part of a
       stack marker; if it is, then the CPU diagnostics fail.
*/



#include "hp3000_defs.h"
#include "hp3000_cpu.h"
#include "hp3000_cpu_io.h"
#include "hp3000_mem.h"
#include "hp3000_io.h"



/* Program constants */

#define PCLK_PERIOD         mS (1)              /* 1 millisecond process clock period */
#define PCLK_MULTIPLIER     10                  /* number of hardware process clock ticks per service */

#define PCLK_5X_PERIOD      (uint32) (PCLK_PERIOD * 2.83 + 0.5) /* process clock period for the Series 52/58 */

#define CPU_IO_RESET        0                   /* reset CPU and all I/O devices */
#define IO_RESET            1                   /* reset just the I/O devices */

#define CTL_MASK            0177400u            /* cold dump control byte mask */
#define DEV_MASK            0000377u            /* cold dump device byte mask */

#define DUMP_BANK_MASK      0017                /* IOP cold dump supports only 4-bit bank widths */

#define SIO_JUMP            0000000u            /* Jump unconditionally */
#define SIO_SBANK           0014000u            /* Set bank */
#define SIO_ENDIN           0034000u            /* End with interrupt */
#define SIO_CNTL            0040000u            /* Control */
#define SIO_WRITE           0060000u            /* Write 4096 words */
#define SIO_READ            0077760u            /* Read 16 words */

#define MS_CN_GAP           0000005u            /* Write Gap */
#define MS_CN_RST           0000011u            /* Rewind and Reset */
#define MS_CN_BSR           0000012u            /* Backspace Record */
#define MS_CN_WFM           0000015u            /* Write File Mark */

#define MS_ST_PROTECTED     0001000u            /* write protected */
#define MS_ST_READY         0000400u            /* unit ready */

#define MS_ST_MASK          (MS_ST_PROTECTED | MS_ST_READY)


/* CPU global SCP data definitions */

DEVICE cpu_dev;                                 /* incomplete device structure */

REG *sim_PC = NULL;                             /* the pointer to the P register */


/* CPU global data structures */

/* CPU registers */

HP_WORD CIR    = 0;                             /* current instruction register */
HP_WORD NIR    = 0;                             /* next instruction register */
HP_WORD PB     = 0;                             /* program base register */
HP_WORD P      = 0;                             /* program counter register */
HP_WORD PL     = 0;                             /* program limit register */
HP_WORD PBANK  = 0;                             /* program segment bank register */
HP_WORD DL     = 0;                             /* data limit register */
HP_WORD DB     = 0;                             /* data base register */
HP_WORD DBANK  = 0;                             /* data segment bank register */
HP_WORD Q      = 0;                             /* stack marker register */
HP_WORD SM     = 0;                             /* stack memory register */
HP_WORD SR     = 0;                             /* stack register counter */
HP_WORD Z      = 0;                             /* stack limit register */
HP_WORD SBANK  = 0;                             /* stack segment bank register */
HP_WORD TR [4] = { 0, 0, 0, 0 };                /* top of stack registers */
HP_WORD X      = 0;                             /* index register */
HP_WORD STA    = 0;                             /* status register */
HP_WORD SWCH   = 0;                             /* switch register */
HP_WORD CPX1   = 0;                             /* run-mode interrupt flags register */
HP_WORD CPX2   = 0;                             /* halt-mode interrupt flags register */
HP_WORD MOD    = 0;                             /* module control register */
HP_WORD PCLK   = 0;                             /* process clock register */
HP_WORD CNTR   = 0;                             /* microcode counter */
HP_WORD SP0    = 0;                             /* scratch pad register 0 */
HP_WORD CR     = 0;                             /* timer count register */
HP_WORD SIR    = 0;                             /* status and interrupt register */
HP_WORD CSMAP  = 0;                             /* code segment mapping register */


/* Condition Code B lookup table.

   Byte-oriented instructions set the condition code in the status word using
   Pattern B (CCB).  For this encoding:

     CCG = ASCII numeric character
     CCE = ASCII alphabetic character
     CCL = ASCII special character

   The simplest implementation of this pattern is a 256-way lookup table using
   disjoint condition code flags.  The SET_CCB macro uses this table to set the
   condition code appropriate for the supplied operand into the status word.
*/

const HP_WORD cpu_ccb_table [256] = {
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* NUL SOH STX ETX EOT ENQ ACK BEL */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* BS  HT  LF  VT  FF  CR  SO  SI  */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* DLE DC1 DC2 DC3 DC4 NAK SYN ETB */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* CAN EM  SUB ESC FS  GS  RS  US  */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* spa  !   "   #   $   %   &   '  */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /*  (   )   *   +   ,   -   .   /  */
    CFG, CFG, CFG, CFG, CFG, CFG, CFG, CFG,     /*  0   1   2   3   4   5   6   7  */
    CFG, CFG, CFL, CFL, CFL, CFL, CFL, CFL,     /*  8   9   :   ;   <   =   >   ?  */
    CFL, CFE, CFE, CFE, CFE, CFE, CFE, CFE,     /*  @   A   B   C   D   E   F   G  */
    CFE, CFE, CFE, CFE, CFE, CFE, CFE, CFE,     /*  H   I   J   K   L   M   N   O  */
    CFE, CFE, CFE, CFE, CFE, CFE, CFE, CFE,     /*  P   Q   R   S   T   U   V   W  */
    CFE, CFE, CFE, CFL, CFL, CFL, CFL, CFL,     /*  X   Y   Z   [   \   ]   ^   _  */
    CFL, CFE, CFE, CFE, CFE, CFE, CFE, CFE,     /*  `   a   b   c   d   e   f   g  */
    CFE, CFE, CFE, CFE, CFE, CFE, CFE, CFE,     /*  h   i   j   k   l   m   n   o  */
    CFE, CFE, CFE, CFE, CFE, CFE, CFE, CFE,     /*  p   q   r   s   t   u   v   w  */
    CFE, CFE, CFE, CFL, CFL, CFL, CFL, CFL,     /*  x   y   z   {   |   }   ~  DEL */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* 200 - 207 */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* 210 - 217 */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* 220 - 227 */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* 230 - 237 */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* 240 - 247 */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* 250 - 257 */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* 260 - 267 */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* 270 - 277 */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* 300 - 307 */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* 310 - 317 */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* 320 - 327 */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* 330 - 337 */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* 340 - 347 */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* 350 - 357 */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL,     /* 360 - 367 */
    CFL, CFL, CFL, CFL, CFL, CFL, CFL, CFL      /* 370 - 377 */
    };


/* Pending interrupt invocation state */

typedef enum {
    Initialize,
    Main_Loop,
    Executor
    } INVOCATION_STATE;


/* CPU global state */

jmp_buf     cpu_save_env;                       /* the saved environment for microcode aborts */
uint32      cpu_stop_flags;                     /* the simulation stop flag set */
HP_WORD     cpu_bank_mask;                      /* mask for the bank field of CST and DST */
HP_WORD     cpu_sir_irq_set;                    /* mask for the set of SIR interrupt sources */

POWER_STATE cpu_power_state  = Power_On;        /* the power supply state */
EXEC_STATE  cpu_micro_state  = Halted;          /* the microcode execution state */
t_bool      cpu_base_changed = FALSE;           /* TRUE if any base register is changed */


/* CPU local state */

static uint32  sim_stops      = 0;              /* the current simulation stop flag settings */
static uint32  cpu_speed      = 1;              /* the CPU speed, expressed as a multiplier of a real machine */
static uint32  pclk_increment = 1;              /* the process clock increment per event service */
static uint32  dump_control   = 0002006u;       /* the cold dump control word (default CNTL = 4, DEVNO = 6 */
static HP_WORD exec_mask      = 0;              /* the current instruction execution trace mask */
static HP_WORD exec_match     = D16_UMAX;       /* the current instruction execution trace matching value */

static HP_WORD start_device = TO_DEVNO (11, 0); /* the default START thumbwheel setting */
static HP_WORD dump_device  = TO_DEVNO (11, 0); /* the default DUMP thumbwheel setting */
static HP_WORD load_device  = TO_DEVNO ( 9, 0); /* the default LOAD thumbwheel setting */


/* CPU local data structures */

/* Interrupt classification names */

static const char *const interrupt_name [] = {  /* class names, indexed by IRQ_CLASS */
    "integer overflow",                         /*   000 irq_Integer_Overflow */
    "bounds violation",                         /*   001 irq_Bounds_Violation */
    "illegal memory address error",             /*   002 irq_Illegal_Address  */
    "non-responding module error",              /*   003 irq_Timeout          */
    "system parity error",                      /*   004 irq_System_Parity    */
    "address parity error",                     /*   005 irq_Address_Parity   */
    "data parity error",                        /*   006 irq_Data_Parity      */
    "module",                                   /*   007 irq_Module           */
    "external",                                 /*   010 irq_External         */
    "power fail",                               /*   011 irq_Power_Fail       */
    NULL,                                       /*   012 (not used)           */
    NULL,                                       /*   013 (not used)           */
    "system clock",                             /*   014 irq_System_Clock     */
    "ICS trap",                                 /*   015 irq_Trap             */
    "dispatch",                                 /*   016 irq_Dispatch         */
    "IXIT external"                             /*   017 irq_IXIT             */
    };


/* Trap classification names */

static const char *const trap_name [] = {       /* trap names, indexed by TRAP_CLASS */
    "no",                                       /*   000 trap_None                */
    "bounds violation",                         /*   001 trap_Bounds_Violation    */
    NULL,                                       /*   002 (unused)                 */
    NULL,                                       /*   003 (unused)                 */
    NULL,                                       /*   004 (unused)                 */
    NULL,                                       /*   005 (unused)                 */
    NULL,                                       /*   006 (unused)                 */
    NULL,                                       /*   007 (unused)                 */
    NULL,                                       /*   010 (unused)                 */
    NULL,                                       /*   011 (unused)                 */
    NULL,                                       /*   012 (unused)                 */
    NULL,                                       /*   013 (unused)                 */
    NULL,                                       /*   014 (unused)                 */
    NULL,                                       /*   015 (unused)                 */
    NULL,                                       /*   016 (unused)                 */
    NULL,                                       /*   017 (unused)                 */
    "unimplemented instruction",                /*   020 trap_Unimplemented       */
    "STT violation",                            /*   021 trap_STT_Violation       */
    "CST violation",                            /*   022 trap_CST_Violation       */
    "DST violation",                            /*   023 trap_DST_Violation       */
    "stack underflow",                          /*   024 trap_Stack_Underflow     */
    "privileged mode violation",                /*   025 trap_Privilege_Violation */
    NULL,                                       /*   026 (unused)                 */
    NULL,                                       /*   027 (unused)                 */
    "stack overflow",                           /*   030 trap_Stack_Overflow      */
    "user",                                     /*   031 trap_User                */
    NULL,                                       /*   032 (unused)                 */
    NULL,                                       /*   033 (unused)                 */
    NULL,                                       /*   034 (unused)                 */
    NULL,                                       /*   035 (unused)                 */
    NULL,                                       /*   036 (unused)                 */
    "absent code segment",                      /*   037 trap_CS_Absent           */
    "trace",                                    /*   040 trap_Trace               */
    "STT entry uncallable",                     /*   041 trap_Uncallable          */
    "absent data segment",                      /*   042 trap_DS_Absent           */
    "power on",                                 /*   043 trap_Power_On            */
    "cold load",                                /*   044 trap_Cold_Load           */
    "system halt"                               /*   045 trap_System_Halt         */
    };


/* CPU features table.

   The feature table is used to validate CPU feature changes within the subset
   of features supported by a given CPU.  Features in the standard list are
   enabled when the CPU model is selected.  If a feature appears in the standard
   list but NOT in the optional list, then it is standard equipment and cannot
   be disabled.  If a feature appears in the optional list, then it may be
   enabled or disabled as desired by the user.


   Implementation notes:

    1. The EIS was standard equipment for the Series II and III, although it
       was an option on the 3000 CX and Series I.  EIS can be disabled,
       equivalent in hardware to removing the address jumpers on the ROM PCA.

    2. The "Series III+" did not exist.  It is provided as an option to exercise
       the full memory range of MPE V/R, which can accommodate 6-bit bank
       numbers.

    3. The "Series 58+" did not exist either.  It is provided as an option to
       exercise the full memory range of MPE V/E, which can accommodate 8-bit
       bank numbers.

    4. The CST expansion firmware was standard on the Series 58, and no way was
       provided to disable it.  Consequently, the Series 58 could run only MPE
       V/E, as none of the earlier MPEs will run with the firmware enabled.  In
       simulation, we provide the SET CPU NOCSTX command to disable the
       firmware; this will allow earlier MPEs for the HP-IB systems, such as MPE
       V/P and MPE IV to be run if desired.
*/

typedef struct {
    uint32  std;                                /* list of standard features */
    uint32  opt;                                /* list of optional features */
    uint32  minmem;                             /* minimum configurable memory in 16-bit words */
    uint32  maxmem;                             /* maximum configurable memory in 16-bit words */
    uint32  bank_mask;                          /* memory bank register width mask */
    } FEATURE_TABLE;

static const FEATURE_TABLE cpu_features [] = {   /* features indexed by CPU_MODEL */
/*                Standard                         Optional              Minimum        Maximum     Bank  */
/*                Features                         Features               Memory        Memory      Mask  */
/*  ------------------------------------  --------------------------   ------------  -------------  ----  */
  { SYS_ARS | FW_EIS,                     SYS_ARS | FW_EIS,             64 * 1024,    256 * 1024,  0003 },  /* Series_II */
  { SYS_ARS | FW_EIS,                     SYS_ARS | FW_EIS | FW_CIS,   128 * 1024,   1024 * 1024,  0017 },  /* Series_III */
  { SYS_ARS | FW_EIS,                     SYS_ARS | FW_EIS | FW_CIS,   128 * 1024,   4096 * 1024,  0077 },  /* Series_IIIx */
  { SYS_ARS | FW_EIS | FW_CIS | FW_CSTX,  FW_CSTX,                    2048 * 1024,   4096 * 1024,  0077 },  /* Series_58 */
  { SYS_ARS | FW_EIS | FW_CIS | FW_CSTX,  FW_CSTX,                    2048 * 1024,  16384 * 1024,  0377 }   /* Series_58x */
  };


/* CPU local SCP support routine declarations */

static t_stat cpu_service (UNIT   *uptr);
static t_stat cpu_reset   (DEVICE *dptr);

static t_stat set_stops      (UNIT *uptr, int32 option,     char *cptr, void *desc);
static t_stat set_exec       (UNIT *uptr, int32 option,     char *cptr, void *desc);
static t_stat set_dump       (UNIT *uptr, int32 value,      char *cptr, void *desc);
static t_stat set_firmware   (UNIT *uptr, int32 option,     char *cptr, void *desc);
static t_stat set_ars        (UNIT *uptr, int32 option,     char *cptr, void *desc);
static t_stat set_thumbwheel (UNIT *uptr, int32 thumbwheel, char *cptr, void *desc);
static t_stat set_size       (UNIT *uptr, int32 new_size,   char *cptr, void *desc);
static t_stat set_model      (UNIT *uptr, int32 new_model,  char *cptr, void *desc);

static t_stat show_stops (FILE *st, UNIT *uptr, int32 value, void *desc);
static t_stat show_exec  (FILE *st, UNIT *uptr, int32 value, void *desc);
static t_stat show_dump  (FILE *st, UNIT *uptr, int32 value, void *desc);
static t_stat show_panel (FILE *st, UNIT *uptr, int32 value, void *desc);
static t_stat show_speed (FILE *st, UNIT *uptr, int32 value, void *desc);


/* CPU local utility routine declarations */

static t_bool        device_conflicts    (void);
static t_bool        interrupt_pending   (INVOCATION_STATE invocation, t_stat *status);
static void          run_mode_interrupt  (HP_WORD device_number);
static t_stat        halt_mode_interrupt (HP_WORD device_number);
static t_stat        machine_instruction (void);
static IMB_INTERFACE cpu_imb_interface;
static void          update_SIR          (void);
static void          trace_registers     (void);


/* CPU SCP data structures */

/* Device information block */

static DIB cpu_dib = {
    NULL,                                       /* the device's IOP interface function pointer */
    NULL,                                       /* the execution initialization function pointer */
    DEVNO_UNUSED,                               /* the device number */
    SRNO_UNUSED,                                /* the service request number */
    INTPRI_UNUSED,                              /* the interrupt priority */
    INTMASK_UNUSED,                             /* the interrupt mask */
    0,                                          /* the card index */
    CLEAR,                                      /* the interrupt request flip-flop */
    CLEAR,                                      /* the interrupt active flip-flop */
    FALSE,                                      /* TRUE if service has been requested */
    &cpu_imb_interface,                         /* the device's IMB interface function pointer */
    CHAN_UNUSED,                                /* the IMB channel number */
    SLOT_UNUSED,                                /* the IMB slot number */
    CPU_IS_IOP | CPU_IS_IMB                     /* the supported system configurations */
    };


/* Unit list.

   The CPU unit is marked with UNIT_FIX and the "capac" field is set to the
   configured memory size to support examining memory locations.  It also is
   used to schedule the process clock events.
*/

static UNIT cpu_unit [] = {
    { UDATA (&cpu_service, UNIT_CALTIME | UNIT_FIX | UNIT_BINK | UNIT_IDLE, 0) }
    };


/* Register list.

   The CPU register list exposes the machine registers for user inspection and
   modification.  User flags describe the permitted and default display formats,
   as follows:

     - REG_X  permits any symbolic display
     - REG_M  defaults to CPU instruction mnemonic display
     - REG_T  defaults to CPU status mnemonic display


   Implementation notes:

    1. The CNTR register is set to the value of the SR register when the
       micromachine halts or pauses.  This allows the SR value to be accessed by
       the diagnostics.  The top-of-stack registers are flushed to main memory
       when the machine halts or pauses, which alters SR.
*/

static REG cpu_reg [] = {
/*    Macro   Name     Location      Radix  Width  Offset       Flags      */
/*    ------  -------  ------------  -----  -----  ------  --------------- */
    { ORDATA (CIR,     CIR,                  16),          REG_M | REG_RO  },  /* current instruction register */
    { ORDATA (NIR,     NIR,                  16),          REG_M | REG_RO  },  /* next instruction register */
    { ORDATA (PB,      PB,                   16),                          },  /* program base register */
    { ORDATA (P,       P,                    16),                          },  /* program counter register */
    { ORDATA (PL,      PL,                   16),                          },  /* program limit register */
    { ORDATA (PBANK,   PBANK,                 6),                          },  /* 6-bit program segment bank register */
    { ORDATA (DL,      DL,                   16),                          },  /* data limit register */
    { ORDATA (DB,      DB,                   16),                          },  /* data base register */
    { ORDATA (DBANK,   DBANK,                 6),                          },  /* 6-bit data segment bank register */
    { ORDATA (Q,       Q,                    16),                          },  /* stack marker register */
    { ORDATA (SM,      SM,                   16),                          },  /* stack memory register */
    { ORDATA (SR,      SR,                    3),                          },  /* stack register counter */
    { ORDATA (Z,       Z,                    16),                          },  /* stack limit register */
    { ORDATA (SBANK,   SBANK,                 6),                          },  /* 6-bit stack segment bank register */
    { ORDATA (RA,      TR [0],               16),          REG_X           },  /* top of stack register */
    { ORDATA (RB,      TR [1],               16),          REG_X           },  /* top of stack - 1 register */
    { ORDATA (RC,      TR [2],               16),          REG_X           },  /* top of stack - 2 register */
    { ORDATA (RD,      TR [3],               16),          REG_X           },  /* top of stack - 3 register */
    { ORDATA (X,       X,                    16),          REG_X           },  /* index register */
    { ORDATA (STA,     STA,                  16),          REG_T           },  /* status register */
    { ORDATA (SWCH,    SWCH,                 16),          REG_X           },  /* switch register */
    { ORDATA (CPX1,    CPX1,                 16),                          },  /* run-mode interrupt flags */
    { ORDATA (CPX2,    CPX2,                 16),                          },  /* halt-mode interrupt flags */
    { ORDATA (PCLK,    PCLK,                 16),                          },  /* process clock register */
    { ORDATA (CR,      CR,                   16),                          },  /* timer count register */
    { ORDATA (SIR,     SIR,                  16),                          },  /* status and interrupt register */
    { ORDATA (CNTR,    CNTR,                  6),                 REG_HRO  },  /* microcode counter */
    { ORDATA (MOD,     MOD,                  16),                 REG_HRO  },  /* module control register */
    { ORDATA (SP0,     SP0,                  16),                 REG_HRO  },  /* scratch pad register 0 */
    { ORDATA (CSMAP,   CSMAP,                16),                 REG_HRO  },  /* code segment mapping register */

    { DRDATA (POWER,   cpu_power_state,       2),                 REG_HRO  },  /* system power supply state */
    { ORDATA (CONFIG,  sys_configuration,    32),                 REG_HRO  },  /* system configuration */

    { ORDATA (WRU,     sim_int_char,          8),                 REG_HRO  },  /* SCP interrupt character */
    { ORDATA (BRK,     sim_brk_char,          8),                 REG_HRO  },  /* SCP break character */
    { ORDATA (DEL,     sim_del_char,          8),                 REG_HRO  },  /* SCP delete character */

    { NULL }
    };


/* Modifier list */

#define ETAB_SET            (MTAB_EDV | MTAB_EGL)                       /* device | global */
#define ETAB_FW             (MTAB_EDV | MTAB_EGL | MTAB_EMK)            /* device | global | mask field */
#define ETAB_DEV            (ETAB_SET | MTAB_ERX (10) | MTAB_EMN (3))   /* device | global | radix 10 | min 3 */
#define ETAB_CTL            (ETAB_SET | MTAB_ERX (8))                   /* device | global | radix 8 */

static MTAB cpu_mod [] = {
/*      Entry           Print     Match       Validation        Display      Descriptor */
/*      Flags    Value  String    String       Routine          Routine       Pointer   */
/*    ---------  -----  -------  ---------  --------------  ---------------  ---------- */
    { MTAB_XDV,    0,   "IDLE",  "IDLE",    &sim_set_idle,  &sim_show_idle,  NULL       },
    { MTAB_XDV,    0,   NULL,    "NOIDLE",  &sim_clr_idle,  NULL,            NULL       },

/*      Mask       Match         Print         Match       Validation    Display      Descriptor       4.x   Expansion */
/*      Value      Value         String        String       Routine      Routine        Pointer       Dummy    Flags   */
/*    ---------  --------  -----------------  ---------  --------------  -------  ------------------  -----  --------- */
    { CPU_II,    CPU_II,   "Series II",       NULL,      &set_model,     NULL,    &sys_configuration, NULL,  ETAB_SET  },
    { CPU_III,   CPU_III,  "Series III",      "III",     &set_model,     NULL,    &sys_configuration, NULL,  ETAB_SET  },
    { CPU_IIIx,  CPU_IIIx, "Series III+",     "III+",    &set_model,     NULL,    &sys_configuration, NULL,  ETAB_SET  },
    { CPU_58,    CPU_58,   "Series 58",       "S58",     &set_model,     NULL,    &sys_configuration, NULL,  ETAB_SET  },
    { CPU_58x,   CPU_58x,  "Series 58+",      "S58+",    &set_model,     NULL,    &sys_configuration, NULL,  ETAB_SET  },

    { FW_EIS,    FW_EIS,   "EIS",             "EIS",     &set_firmware,  NULL,    &sys_configuration, NULL,  ETAB_FW   },
    { FW_EIS,    0,        "no EIS",          "NOEIS",   &set_firmware,  NULL,    &sys_configuration, NULL,  ETAB_FW   },

    { FW_CIS,    FW_CIS,   "CIS",             "CIS",     &set_firmware,  NULL,    &sys_configuration, NULL,  ETAB_FW   },
    { FW_CIS,    0,        "no CIS",          "NOCIS",   &set_firmware,  NULL,    &sys_configuration, NULL,  ETAB_FW   },

    { FW_CSTX,   FW_CSTX,  "CSTX",            "CSTX",    &set_firmware,  NULL,    &sys_configuration, NULL,  ETAB_FW   },
    { FW_CSTX,   0,        "no CSTX",         "NOCSTX",  &set_firmware,  NULL,    &sys_configuration, NULL,  ETAB_FW   },

    { SYS_ARS,   SYS_ARS,  "auto-restart",    "ARS",     &set_ars,       NULL,    &sys_configuration, NULL,  ETAB_SET  },
    { SYS_ARS,   0,        "no auto-restart", "NOARS",   &set_ars,       NULL,    &sys_configuration, NULL,  ETAB_SET  },


/*        Mask           Match             Print            Match      Validation  Display  Descriptor  */
/*        Value          Value             String           String       Routine   Routine    Pointer   */
/*    -------------  -------------  --------------------  -----------  ----------  -------  ----------- */
    { UNIT_CALTIME,  UNIT_CALTIME,  "calibrated timing",  "CALTIME",   NULL,       NULL,    NULL        },
    { UNIT_CALTIME,  0,             "realistic timing",   "REALTIME",  NULL,       NULL,    NULL        },

/*           Entry                         Print      Match         Validation       Display     Descriptor */
/*           Flags              Value      String     String         Routine         Routine      Pointer   */
/*    --------------------  -------------  --------  ----------  ----------------  ------------  ---------- */
    { MTAB_XDV,               128 * 1024,  NULL,       "128K",   &set_size,        NULL,         NULL       },
    { MTAB_XDV,               256 * 1024,  NULL,       "256K",   &set_size,        NULL,         NULL       },
    { MTAB_XDV,               384 * 1024,  NULL,       "384K",   &set_size,        NULL,         NULL       },
    { MTAB_XDV,               512 * 1024,  NULL,       "512K",   &set_size,        NULL,         NULL       },
    { MTAB_XDV,               768 * 1024,  NULL,       "768K",   &set_size,        NULL,         NULL       },
    { MTAB_XDV,              1024 * 1024,  NULL,      "1024K",   &set_size,        NULL,         NULL       },
    { MTAB_XDV,              2048 * 1024,  NULL,      "2048K",   &set_size,        NULL,         NULL       },
    { MTAB_XDV,              4096 * 1024,  NULL,      "4096K",   &set_size,        NULL,         NULL       },
    { MTAB_XDV,              8192 * 1024,  NULL,      "8192K",   &set_size,        NULL,         NULL       },
    { MTAB_XDV,             16384 * 1024,  NULL,     "16384K",   &set_size,        NULL,         NULL       },

    { MTAB_XDV,               Cold_Start,  NULL,     "START",    &set_thumbwheel,  NULL,         NULL       },
    { MTAB_XDV,               Cold_Dump,   NULL,     "DUMP",     &set_thumbwheel,  NULL,         NULL       },
    { MTAB_XDV,               Cold_Load,   NULL,     "LOAD",     &set_thumbwheel,  NULL,         NULL       },
    { MTAB_XDV | MTAB_NMO,        0,       "PANEL",  NULL,       NULL,             &show_panel,  NULL       },

    { MTAB_XDV | MTAB_NMO,        1,       "STOPS",  "STOP",     &set_stops,       &show_stops,  NULL       },
    { MTAB_XDV,                   0,       NULL,     "NOSTOP",   &set_stops,       NULL,         NULL       },

    { MTAB_XDV | MTAB_NMO,        1,       "EXEC",   "EXEC",     &set_exec,        &show_exec,   NULL       },
    { MTAB_XDV,                   0,       NULL,     "NOEXEC",   &set_exec,        NULL,         NULL       },

    { MTAB_XDV | MTAB_NMO,        0,       "DUMP",   NULL,       NULL,             &show_dump,   NULL       },
    { MTAB_XDV | MTAB_NMO,        0,       "SPEED",  NULL,       NULL,             &show_speed,  NULL       },

/*      Mask     Match  Print     Match     Validation  Display    Descriptor     4.x   Expansion */
/*      Value    Value  String    String    Routine     Routine     Pointer      Dummy    Flags   */
/*    ---------  -----  ------  ----------  ----------  -------  --------------  -----  --------- */
    { DEV_MASK,  127,   NULL,   "DUMPDEV",  &set_dump,   NULL,   &dump_control,   NULL,  ETAB_DEV  },
    { CTL_MASK,  255,   NULL,   "DUMPCTL",  &set_dump,   NULL,   &dump_control,   NULL,  ETAB_CTL  },

    { 0 }
    };


/* Debugging trace list */

static DEBTAB cpu_deb [] = {
    { "INSTR", TRACE_INSTR },                   /* trace instructions */
    { "DATA",  TRACE_DATA  },                   /* trace memory data accesses */
    { "FETCH", TRACE_FETCH },                   /* trace memory instruction fetches */
    { "REG",   TRACE_REG   },                   /* trace register values */
    { "OPND",  TRACE_OPND  },                   /* trace memory operand values */
    { "EXEC",  TRACE_EXEC  },                   /* trace instruction execution states */
    { "IRQ",   TRACE_IRQ   },                   /* trace interrupt and trap requests */
    { "PSERV", TRACE_PSERV },                   /* trace process clock service events */
    { "IMBUS", TRACE_IMBUS },                   /* trace IMB signals and data words (indirect) */
    { NULL,    0           }
    };

/* Debugging stop list */

static DEBTAB cpu_stop [] = {
    { "LOOP",     SS_LOOP     },                /* stop on an infinite loop */
    { "PAUSE",    SS_PAUSE    },                /* stop on a PAUS instruction */
    { "UNDEF",    SS_UNDEF    },                /* stop on an undefined instruction */
    { "UNIMPL",   SS_UNIMPL   },                /* stop on an unimplemented instruction */
    { "COLDLOAD", SS_COLDLOAD },                /* stop on a cold-load trap */
    { NULL,       0           }
    };


/* Device descriptor */

DEVICE cpu_dev = {
    "CPU",                                      /* device name */
    cpu_unit,                                   /* unit array */
    cpu_reg,                                    /* register array */
    cpu_mod,                                    /* modifier array */
    1,                                          /* number of units */
    8,                                          /* address radix */
    PA_WIDTH,                                   /* address width */
    1,                                          /* address increment */
    8,                                          /* data radix */
    16,                                         /* data width */
    &mem_examine,                               /* examine routine */
    &mem_deposit,                               /* deposit routine */
    &cpu_reset,                                 /* reset routine */
    NULL,                                       /* boot routine */
    NULL,                                       /* attach routine */
    NULL,                                       /* detach routine */
    &cpu_dib,                                   /* device information block pointer */
    DEV_DEBUG,                                  /* device flags */
    0,                                          /* debug control flags */
    cpu_deb,                                    /* debug flag name array */
    NULL,                                       /* memory size change routine */
    NULL                                        /* logical device name */
    };



/***********************************
 *                                 *
 * CPU global SCP support routines *
 *                                 *
 ***********************************/


/* Execute CPU instructions.

   This is the instruction decode routine for the HP 3000.  It is called from
   the simulator control program to execute instructions in simulated memory,
   starting at the simulated program counter.  It runs until the status to be
   returned is set to a value other than SCPE_OK.

   On entry, P points to the instruction to execute, and the "sim_switches"
   global contains any command-line switches included with the run command.  On
   exit, P points at the next instruction to execute (or the current
   instruction, in the case of a simulator stop during a PAUS instruction or
   after the first of two stack operations).

   Execution is divided into four phases.

   First, the instruction prelude configures the simulation state to resume
   execution.  This involves verifying that system power is on and there are no
   device conflicts (e.g., two devices with the same device number or two
   channels with the same channel number), initializing the I/O processor and
   channels, and setting the RUN switch if no other front panel switches are
   pressed.  These actions accommodate reconfiguration of the I/O device
   settings and program counter while the simulator is stopped.  The prelude
   also responds to one command-line switch: if "-B" is specified, the current
   set of simulation stop conditions is bypassed for the first instruction
   executed.  This allows, e.g., a PAUS instruction to be bypassed or an
   unimplemented instruction trap to be taken.

   Second, the microcode abort mechanism is set up.  Microcode aborts utilize
   the "setjmp/longjmp" mechanism to transfer control out of the instruction
   executors without returning through the call stack.  This allows an
   instruction to be aborted part-way through execution when continuation is
   impossible, e.g., due to a memory access violation.  It corresponds to direct
   microcode jumps out of the execution sequence and to the appropriate trap
   handlers.

   Third, the instruction execution loop decodes instructions and calls the
   individual executors in turn until a condition occurs that prevents further
   execution.  Examples of such conditions includes execution of a HALT
   instruction, a user stop request (CTRL+E) from the simulation console, a
   recoverable device error (such as an improperly formatted tape image), a
   user-specified breakpoint, and a simulation stop condition (such as execution
   of an unimplemented instruction).  The execution loop also polls for I/O
   events and device interrupts, and runs I/O channel cycles.  During
   instruction execution, the CIR register contains the currently executing
   instruction, the NIR register contains the next instruction to execute, and
   the P register points to the memory location two instructions ahead of the
   current instruction.

   Fourth, the instruction postlude updates the simulation state in preparation
   for returning to the SCP command prompt.  Devices that maintain an internal
   state different from their external state, such as the CPU process clock, are
   updated so that their internal and external states are fully consistent.
   This ensures that the state visible to the user during the simulation stop is
   correct.  It also ensures that the program counter points correctly at the
   next instruction to execute upon resumption.

   If enabled, the simulator is idled when a PAUS instruction has been executed
   and no service requests for the multiplexer, selector, or IMB channels are
   active.  Execution of a PAUS instruction suspends the fetch-and-execute
   process until an interrupt occurs or the simulator is stopped and then
   resumed with a GO -B or RUN -B command.

   The HP 3000 is a microcoded machine.  In hardware, the micromachine is always
   executing microinstructions, even when the CPU is "halted."  The halt/run
   state is simply a flip-flop setting, reflected in bit 15 of the CPX2 or SIR
   register and the RUN light on the front panel, that determines whether the
   "halt-mode" or "run-mode" microprogram is currently executing.

   In simulation, the "cpu_micro_state" variable indicates the state of the
   micromachine, i.e., which section of the microcode it is executing, while
   CPX2 or SIR bit 15 indicates whether the macromachine is halted or running.
   The micromachine may be in one of four states:

     - Running : the run-mode fetch-and-execute microcode is executing
     - Paused  : the run-mode PAUS instruction microcode is executing
     - Waiting : the halt-mode cold load or dump microcode is executing
     - Halted  : the halt-mode front panel microcode is executing

   Simulation provides a variety of stop conditions that break instruction
   execution and return to the SCP prompt with the CPU still in run mode.  These
   have no analog in hardware; the only way to stop the CPU is to press the HALT
   button on the front panel, which shifts the micromachine into halt-mode
   microcode execution.  When any of these conditions occur, the micromachine
   state is set to "halted," but the CPX2 or SIR run flag is remains set unless
   the stop was caused by execution of a HALT instruction.  Resuming execution
   with a STEP, CONTINUE, GO, or RUN command proceeds as though the hardware RUN
   switch was pressed after a programmed halt.  This provides the proper
   semantics for seamlessly stopping and restarting instruction execution.

   A microcode abort is performed by executing a "longjmp" to the abort handler,
   which is outside of and precedes the instruction execution loop.  The value
   passed to "longjmp" is a 32-bit integer containing the Segment Transfer Table
   index of the trap handler in the lower word and an optional parameter in the
   upper word.  Aborts are invoked by the MICRO_ABORT macro, which takes as its
   parameter a trap classification value, e.g.:

     MICRO_ABORT (trap_Privilege_Violation);
     MICRO_ABORT (trap_Integer_Zero_Divide);

   Some aborts require an additional parameter and must be invoked by the
   MICRO_ABORTP macro, which takes a trap classification value and a
   trap-specific value as parameters.  The traps that require additional
   parameters are:

     MICRO_ABORTP (trap_CST_Violation, segment_number);
     MICRO_ABORTP (trap_STT_Violation, segment_number);
     MICRO_ABORTP (trap_CS_Absent,     label/n/0);
     MICRO_ABORTP (trap_DS_Absent,     DST_number);
     MICRO_ABORTP (trap_Uncallable,    label);
     MICRO_ABORTP (trap_Trace,         label/n/0);
     MICRO_ABORTP (trap_User,          trap_number);

   trap_User is not usually called explicitly via MICRO_ABORTP; rather,
   MICRO_ABORT is used with one of the specific user-trap identifiers, e.g.,
   trap_Integer_Overflow, trap_Float_Overflow, trap_Decimal_Overflow, etc., that
   supplies both the trap classification and the trap parameter value.

   In addition, user traps must be enabled by setting the T-bit in the status
   word.  If the T bit is not set, a user trap sets the O-bit (overflow) in the
   status word and resumes execution with the next instruction instead of
   invoking the user trap handler.

   When an abort occurs, an equivalent PCAL to the appropriate STT entry is set
   up.  Then execution drops into the instruction loop to execute the first
   instruction of the trap handler.

   When the instruction loop is exited, the CPU process clock and system clock
   registers are updated, the micromachine is halted, and control returns to
   SCP.  Upon return, P points at the next instruction to execute, i.e., the
   instruction that will execute when the instruction loop is reentered.

   If the micromachine is paused, then P is reset to point to the PAUS
   instruction, which will be reexecuted when the routine is reentered.  If it
   is running, then P is reset to point to the current instruction if the stop
   allows it to be rerun, or at the next instruction.


   Implementation notes:

    1. While the Microsoft VC++ "setjmp" documentation says, "All variables
       (except register variables) accessible to the routine receiving control
       contain the values they had when longjmp was called," the ISO C99
       standard says, "All accessible objects have values...as of the time the
       longjmp function was called, except that the values of objects of
       automatic storage duration that are local to the function containing the
       invocation of the corresponding setjmp macro that do not have
       volatile-qualified type and have been changed between the setjmp
       invocation and longjmp call are indeterminate."

       Therefore, the "trace_save" variable is marked volatile to ensure that it
       is reloaded after a longjmp caused by a microcode abort.  None of the
       other automatic variables are changed or require their values to be
       maintained between the setjmp and a longjmp.

    2. In hardware, the NEXT microcode order present at the end of each
       instruction transfers the NIR content to the CIR, reads the memory word
       at P into the NIR, and increments P.  However, if an interrupt is
       present, then this action is omitted, and a microcode jump is performed
       to control store location 3, which then jumps to the microcoded interrupt
       handler.  In simulation, the CIR/NIR/P update is performed before the
       next instruction is executed, rather than after the last instruction
       completes, so that interrupts are handled before updating.

       In addition, the NEXT action is modified in hardware if the NIR contains
       a stack instruction with a non-NOP B stackop.  In this case, NEXT
       transfers the NIR content to the CIR, reads the memory word at P into the
       NIR, but does not increment P.  Instead, the R bit of the status register
       is set to indicate that a B stackop is pending.  When the NEXT at the
       completion of the A stackop is executed, the NIR and CIR are untouched,
       but P is incremented, and the R bit is cleared.  This ensures that if an
       interrupt or trap occurs between the stackops, P will point correctly at
       the next instruction to be executed.

       In simulation, following the hardware would require testing the NIR for a
       non-NOP B stackop at every pass through the instruction execution loop.
       To avoid this, the NEXT simulation unilaterally increments P, rather than
       only when a B stackop is not present, and the stack instruction executor
       tests for the B stackop and sets the R bit there.  However, by that time,
       P has already been incremented, so we decrement it there to return it to
       the correct value.

    3. The System Halt trap has no handler.  Instead, the simulator is halted,
       and control returns to the SCP prompt.

    4. The trace display for a trap reports the parameter value supplied with
       the microcode abort.  This is not necessarily the same as the parameter
       that is pushed on the stack for the trap handler.  As some traps, e.g.,
       trap_CST_Violation, can cause a System Halt, the placement of the trace
       call is dictated by the desire to report both the original trap and the
       System Halt trap, even though the placement results in the display of the
       incoming parameter value, rather than the stacked parameter value.

    5. A system halt occurs for a violation, absence, or trace trap while
       executing physical segment 1.  This condition is indicated by a segment
       number of 1 in the supplied parameter and, if the CST expansion firmware
       is installed, the physical mapping mode value in the CSMAP register.

    6. The execution trace (TRACE_EXEC) match test is performed in two parts to
       display the register values both before and after the instruction
       execution.  Consequently, the enable test is done before the register
       trace, and the disable test is done after.

    7. The execution test (exec_test) is set FALSE even though execution tracing
       is not specified.  This is done solely to reassure the compiler that the
       value is not clobbered by a longjmp call.

    8. The set of debug trace flags is restored in the postlude (and therefore
       also saved in the instruction prelude) to ensure that a simulation stop
       that occurs while an execution trace is in progress does not exit with
       the flags set improperly.

    9. The breakpoint test must be executed even when the CPU pauses to permit
       string breakpoint delays to expire.

   10. Before returning, the "hide" switch is set in "sim_switches" if the exit
       is a result of single-step expiration or a triggered breakpoint.  This
       indicates that the stop message may be omitted if operating within a
       quietly executing DO file.  The caller cannot determine this condition
       from the returned status because breakpoint stop status is a VM-specific
       value.

   11. The SIR is implemented only on IMB systems.  Because many bits duplicate
       those in the Series III CPX1 and CPX2 registers, we elect to work off of
       those registers and use the SIR only for the unique bits (IMB IRQ and
       CSRQ, and the system clock tick).  The other bits are set just before
       exiting the routine when the user might examine the register.
*/

t_stat sim_instr (void)
{
volatile uint32 trace_save;
int             abortval;
HP_WORD         label, parameter;
TRAP_CLASS      trap;
PLABEL          plab;
t_bool          exec_test;
uint32          device = 0;
t_stat          status = SCPE_OK;


/* Instruction prelude */

trace_save = cpu_dev.dctrl;                             /* save the current set of trace flags for later restoration */

if (sim_switches & SWMASK ('B'))                        /* if a simulation stop bypass was requested */
    cpu_stop_flags = SS_BYPASSED;                       /*   then clear the stop flags for the first instruction */
else                                                    /* otherwise */
    cpu_stop_flags = sim_stops;                         /*   set the stops as indicated */

if (cpu_power_state == Power_Off)                       /* if system power is off */
    status = STOP_POWER;                                /*   then execution is not possible until restoration */

else if (device_conflicts ())                           /* otherwise if the device assignments are inconsistent */
    status = SCPE_STOP;                                 /*   then inhibit execution */

else {                                                  /* otherwise */
    if (sys_configuration & (CPU_IS_IMB | SYS_SF)) {    /*   if this is an IMB or Starfish system */
        device = imb_initialize ();                     /*     then initialize the IMB */

        if (device == DEVNO_UNUSED)                     /* if IMB initialization failed */
            status = SCPE_STOP;                         /*   then inhibit execution */
        }

    if (sys_configuration & CPU_IS_IOP) {               /* if this is an IOP system */
        device = iop_initialize ();                     /*   then initialize the IOP */

        if (device == DEVNO_UNUSED)                     /* if I/O initialization failed */
            status = SCPE_STOP;                         /*   then inhibit execution */

        else {                                          /* otherwise */
            mpx_initialize ();                          /*   initialize the multiplexer channel */
            sel_initialize ();                          /*     and the selector channel */
            }
        }

    if (status == SCPE_OK                               /* if initialization succeeded */
      && not (CPX2 & CPX2_IRQ_SET))                     /*   and no halt-mode interrupts are present */
        CPX2 |= cpx2_RUNSWCH;                           /*     then assume a RUN command via STEP/CONT */

    interrupt_pending (Initialize, (t_stat *) &device); /* initialize with any "leftover" interrupt */
    }


/* Microcode abort processor */

abortval = setjmp (cpu_save_env);                       /* set the microcode abort handler */

if (abortval) {                                         /* if a microcode abort occurred */
    trap = TRAP (abortval);                             /*   then get the trap classification */
    parameter = PARAM (abortval);                       /*     and the optional parameter */

    label = TO_ISR_LABEL (trap);                        /* form the label from the STT number */

    if (TRACING (cpu_dev, TRACE_REG)) {                 /* if register tracing is enabled */
        cpu_base_changed = TRUE;                        /*   then show the full set of registers */
        trace_registers ();                             /*     just before the trap */
        }

    tprintf (cpu_dev, TRACE_INSTR | TRACE_IRQ, BOV_FORMAT "%s trap%s\n",
             PBANK, P - 1 & R_MASK, parameter, trap_name [trap],
             (trap == trap_User && not (STA & STATUS_T) ? " (disabled)" : ""));

    switch (trap) {                                     /* dispatch on the trap classification */

        case trap_None:                                 /* trap_None should never occur */
        case trap_System_Halt:
            CNTR = SR;                                  /* copy the stack register to the counter */
            cpu_flush ();                               /*   and flush the TOS registers to memory */

            SP0 = parameter;                            /* set SP0 to the parameter (system halt condition) */

            CPX2 = CPX2 & ~cpx2_RUN | cpx2_SYSHALT;     /* halt the CPU and set the system halt flag */
            status = STOP_SYSHALT;                      /*   and report the system halt condition */

            label = 0;                                  /* there is no trap handler for a system halt */
            break;


        case trap_CST_Violation:
            if (LABEL_SEGN (parameter) <= ISR_SEGN      /* if the trap occurred in segment 1 */
              && (CSMAP == STMK_M                       /*   and the segment is physically mapped */
              || not (sys_configuration & FW_CSTX)))    /*   or the CST firmware is not installed */
                MICRO_ABORT (trap_SysHalt_CSTV_1);      /*     then the failure is fatal */

        /* fall through into the next trap handler */

        case trap_STT_Violation:
            if (LABEL_SEGN (parameter) <= ISR_SEGN      /* if the trap occurred in segment 1 */
              && (CSMAP == STMK_M                       /*   and the segment is physically mapped */
              || not (sys_configuration & FW_CSTX)))    /*   or the CST firmware is not installed */
                MICRO_ABORT (trap_SysHalt_STTV_1);      /*     then the failure is fatal */

        /* fall through into the next trap handler */

        case trap_Unimplemented:
        case trap_DST_Violation:
        case trap_Stack_Underflow:
        case trap_Privilege_Violation:
        case trap_Bounds_Violation:
            parameter = label;                          /* the label is the parameter for these traps */

        /* fall through into the next trap handler */

        case trap_DS_Absent:
            cpu_flush ();                               /* flush the TOS registers to memory */
            cpu_mark_stack ();                          /*   and then write a stack marker */

        /* fall through into the next trap handler */

        case trap_Uncallable:
            break;                                      /* set up the trap handler */


        case trap_User:
            if (STA & STATUS_T) {                       /* if user traps are enabled */
                STA &= ~STATUS_O;                       /*   then clear overflow status */
                cpu_flush ();                           /*     and flush the TOS registers to memory */
                cpu_mark_stack ();                      /*       and write a stack marker */
                }                                       /*         and set up the trap handler */

            else {                                      /* otherwise in lieu of trapping */
                STA |= STATUS_O;                        /*   set overflow status */
                label = 0;                              /*     and continue execution with the next instruction */
                }
            break;


        case trap_CS_Absent:
            if (CPX1 & cpx1_ICSFLAG)                    /* if the trap occurred while on the ICS */
                MICRO_ABORT (trap_SysHalt_Absent_ICS);  /*   then the failure is fatal */

            else if (STATUS_CS (STA) <= ISR_SEGN        /* otherwise if the trap occurred in segment 1 */
              && (CSMAP == STMK_M                       /*   and the segment is physically mapped */
              || not (sys_configuration & FW_CSTX)))    /*   or the CST firmware is not installed */
                MICRO_ABORT (trap_SysHalt_Absent_1);    /*     then the failure is fatal */
            break;                                      /* otherwise set up the trap handler */


        case trap_Trace:
            if (STATUS_CS (STA) <= ISR_SEGN             /* if the trap occurred in segment 1 */
              && (CSMAP == STMK_M                       /*   and the segment is physically mapped */
              || not (sys_configuration & FW_CSTX)))    /*   or the CST firmware is not installed */
                MICRO_ABORT (trap_SysHalt_Trace_1);     /*     then the failure is fatal */
            break;                                      /* otherwise set up the trap handler */


        case trap_Cold_Load:                            /* this trap executes on the ICS */
            if (sys_configuration & CPU_IS_IOP          /* if this is an IOP system */
              || cpu_stop_flags & SS_COLDLOAD)          /*   or the user requested a cold-load stop */
                status = STOP_CLOAD;                    /*     then report that the cold load is complete */

            cpu_setup_ics_irq (irq_Trap, trap);         /* set up the ICS */
            break;                                      /*   and then the trap handler */


        case trap_Stack_Overflow:                           /* this trap executes on the ICS */
            if (CPX1 & cpx1_ICSFLAG)                        /*   so if the trap occurred while on the ICS */
                MICRO_ABORT (trap_SysHalt_Overflow_ICS);    /*     then the failure is fatal */

            cpu_setup_ics_irq (irq_Trap, trap);             /* otherwise, set up the ICS */
            break;                                          /*   and then the trap handler */


        case trap_Power_On:                             /* this trap executes on the ICS */
            cpu_setup_ics_irq (irq_Trap, trap);         /*   so set it up */
            cpu_power_state = Power_On;                 /*     and return to the power-on state */

            if (CPX2 & cpx2_INHPFARS)                   /* if auto-restart is inhibited */
                status = STOP_ARSINH;                   /*   then exit and wait for a manual restart */
            break;
        }                                               /* all cases are handled */


    if (label != 0) {                                   /* if the trap handler is to be called */
        STA = STATUS_M;                                 /*   then clear the status and enter privileged mode */

        SM = SM + 1 & R_MASK;                           /* increment the stack pointer */
        cpu_write_memory (stack, SM, parameter);        /*   and push the parameter on the stack */

        X = CIR;                                        /* save the current instruction for restarting */

        plab.label = label;                             /* create an external label descriptor */
        plab.mode  = External;                          /*   for the trap handler */

        cpu_call_procedure (plab, 0);                   /* set up to call the interrupt handling procedure */
        }

    sim_interval = sim_interval - 1;                    /* count the execution cycle that aborted */
    }


/* Instruction loop */

while (status == SCPE_OK) {                             /* execute until simulator status prevents continuation */

    if (interrupt_pending (Main_Loop, &status)) {       /* check for a pending interrupt and set up the handler */
        if (status != SCPE_OK)                          /*   but if an event service failed */
            break;                                      /*     then abort execution and report the failure */
        }

    else if (cpu_micro_state == Running) {                  /* otherwise if the micromachine is running */
        if (TRACING (cpu_dev, TRACE_EXEC | TRACE_REG)) {    /*   then if execution or register tracing is enabled */
            if (cpu_dev.dctrl & TRACE_EXEC)                 /*     then if tracing execution */
                if (STA & STATUS_R)                         /*       then if the right-hand stack op is pending */
                    exec_test =                             /*         then the execution test succeeds if */
                       (CIR << STACKOP_A_SHIFT              /*           the right-hand stack op */
                         & STACKOP_A_MASK                   /*             matches the test criteria */
                         & exec_mask) == exec_match;        /*               for the left-hand stack op value */
                else                                        /*       otherwise */
                    exec_test =                             /*         then the execution test succeeds if the next */
                       (NIR & exec_mask) == exec_match;     /*           instruction matches the test criteria */
            else                                            /*     otherwise */
                exec_test = FALSE;                          /*       there is no execution test */

            if (cpu_dev.dctrl & TRACE_EXEC              /* if execution tracing is enabled */
              && cpu_dev.dctrl != TRACE_ALL             /*   and is currently inactive */
              && exec_test) {                           /*     and the matching test succeeds */
                trace_save = cpu_dev.dctrl;             /*       then save the current trace flag set */
                cpu_dev.dctrl = TRACE_ALL;              /*         and turn on full tracing */
                }

            if (cpu_dev.dctrl & TRACE_REG)              /* if register tracing is enabled */
                trace_registers ();                     /*   then print the registers */

            if (cpu_dev.dctrl & TRACE_EXEC              /* if execution tracing is enabled */
              && cpu_dev.dctrl == TRACE_ALL             /*   and is currently active */
              && not exec_test) {                       /*     and the matching test fails */
                cpu_dev.dctrl = trace_save;             /*       then restore the saved trace flag set */
                hp_trace (&cpu_dev, TRACE_EXEC,         /*         and add a separator to the trace log */
                          "*****************\n");
                }
            }

        if (not (STA & STATUS_R)) {                     /* (NEXT) if the right-hand stack op is not pending */
            CIR = NIR;                                  /*   then update the current instruction */
            cpu_read_memory (fetch, P, &NIR);           /*     and load the next instruction */
            }

        P = P + 1 & R_MASK;                             /* point to the following instruction */

        if (TRACING (cpu_dev, TRACE_INSTR)) {           /* if instruction tracing is enabled */
            sim_eval [0] = CIR;                         /*   then save the instruction that will be executed */
            sim_eval [1] = NIR;                         /*     and the following word for evaluation */

            hp_trace (&cpu_dev, TRACE_INSTR, BOV_FORMAT,    /* print the address and the instruction opcode */
                      PBANK, P - 2 & R_MASK, CIR);          /*   as an octal value */

            if (fprint_cpu (sim_deb, sim_eval, 0, SIM_SW_STOP) == SCPE_ARG) /* print the mnemonic; if that fails */
                fprint_val (sim_deb, sim_eval [0], cpu_dev.dradix,          /*   then print the numeric */
                            cpu_dev.dwidth, PV_RZRO);                       /*     value again */

            if (sim_deb == stdout)                      /* if debug output is to the (raw) console */
                fputc ('\r', sim_deb);                  /*   then insert a carriage return */

            fputc ('\n', sim_deb);                      /* end the trace with a newline */
            }

        status = machine_instruction ();                /* execute one machine instruction */

        cpu_stop_flags = sim_stops;                     /* reset the stop flags as indicated */
        }

    else if (sim_idle_enab                              /* otherwise if idling is enabled while paused */
      && not (SIR & sir_IMBCSRQ)                        /*   and no IMB channel request is pending */
      && not sel_request && mpx_request_set == 0)       /*   and no IOP channel requests are pending */
        sim_idle (TMR_POLL, FALSE);                     /*     then idle the simulator */

    sim_interval = sim_interval - 1;                    /* count the execution cycle */
    }                                                   /*   and continue with the instruction loop */


/* Instruction postlude */

cpu_dev.dctrl = trace_save;                             /* restore the flag set */

cpu_update_pclk ();                                     /* update the process clock */
clk_update_counter ();                                  /*   and system clock counters */

if (sys_configuration & (CPU_IS_IMB | SYS_SF))          /* if this is an IMB or Starfish system */
    update_SIR ();                                      /*   then update the Status and Interrupt Register */

if (cpu_micro_state == Paused)                          /* if the micromachine is paused */
    P = P - 2 & R_MASK;                                 /*   then set P to point to the PAUS instruction */

else if (cpu_micro_state == Running)                    /* otherwise if it is running */
    if (status >= STOP_BASE && status <= STOP_RERUN)    /*   then if the instruction will be rerun when resumed */
        P = P - 2 & R_MASK;                             /*     then set P to point to it */
    else                                                /*   otherwise */
        P = P - 1 & R_MASK;                             /*     set P to point to the next instruction */

cpu_micro_state = Halted;                               /* halt the micromachine */

if (cpu_power_state == Power_Failing                    /* if power is failing */
  && status == STOP_HALT)                               /*   and we executed a HALT instruction */
    cpu_power_state = Power_Off;                        /*     then power will be off when we return */

if (status >= SCPE_OK)                                  /* if the status is valid */
    tprintf (cpu_dev, cpu_dev.dctrl,                    /*   then output the simulation stop reason */
             BOV_FORMAT "simulation stop: %s\n",
             PBANK, P, STA, sim_error_text (status));

if (status == SCPE_STEP || status == STOP_BRKPNT)       /* if stopping for a step or breakpoint */
    sim_switches |= SIM_SW_HIDE;                        /*   then permit omission if operating quietly */

return status;                                          /* return the status code that caused the stop */
}


/* Execute the LOAD and DUMP commands.

   This command processing routine implements the cold load and cold dump
   commands.  The syntax for the IOP machines is:

     LOAD { <control/devno> }
     DUMP { <control/devno> }

   The <control/devno> is a number that is deposited into the SWCH register
   before invoking the CPU cold load or cold dump facility.  The CPU radix is
   used to interpret the number; it defaults to octal.  If the number is
   omitted, the SWCH register value is not altered before loading or dumping.

   The syntax for the IMB machines is:

     START { <channel/device> }
     DUMP  { <channel/device> }
     LOAD  { <channel/device> }

   The <channel/device> is a number that is deposited into the SWCH register
   before invoking the CPU cold start, dump, or load facility.  The CPU radix is
   used to interpret the number; it defaults to octal.  If the number is
   omitted, the SWCH register value is set from the corresponding thumbwheel
   switch on the system control panel.

   On entry, the "arg" parameter is "Cold_Start" for a START command,
   "Cold_Load" for a LOAD command, or "Cold_Dump" for a DUMP command, and "buf"
   points at the remainder of the command line.  If characters exist on the
   command line, they are parsed, converted to a numeric value, and stored in
   the SWCH register.  Then the CPU's cold load/dump routine is called to set up
   the CPU state.  Finally, the CPU is started to begin the requested action.


   Implementation notes:

    1. The RUN command uses the RU_CONT argument instead of RU_RUN so that the
       run_cmd SCP routine will not reset all devices before entering the
       instruction executor.  The halt mode interrupt handlers for cold load and
       cold dump reset the simulator as appropriate for their commands (i.e.,
       the CPU and all I/O devices, or just the I/O devices, respectively).
*/

t_stat cpu_cold_cmd (int32 arg, char *buf)
{
const char *cptr;
char       gbuf [CBUFSIZE];
t_value    max;
t_stat     status;
HP_WORD    value;

if (*buf != '\0') {                                     /* if more command line characters exist */
    cptr = get_glyph (buf, gbuf, 0);                    /*   then get the next glyph */

    if (*cptr != '\0')                                  /* if that does not exhaust the input */
        return SCPE_2MARG;                              /*   then report that there are too many arguments */

    if (sys_configuration & CPU_IS_IMB)                 /* if this is an IMB system */
        max = DEVNO_MAX;                                /*   then the parameter is a device number */

    else if (arg == Cold_Start)                         /* otherwise if a START was done for an SIO CPU */
        return SCPE_NOFNC;                              /*   then report that the command is not allowed */

    else                                                /* otherwise */
        max = D16_UMAX;                                 /*   the parameter is a 16-bit value */

    value = (HP_WORD) get_uint (gbuf, cpu_dev.dradix,   /* get the parameter value */
                                max, &status);

    if (status == SCPE_OK)                              /* if a valid number was present */
        SWCH = value;                                   /*   then set it into the switch register */
    else                                                /* otherwise */
        return status;                                  /*   return the error status */
    }

else if (sys_configuration & CPU_IS_IMB)                /* otherwise if this is an IMB system */
    switch ((PANEL_SWITCH) arg) {                       /*   then dispatch on the command */

        case Cold_Start:                                /* START */
            SWCH = start_device;                        /* set the default start device */
            break;

        case Cold_Dump:                                 /* DUMP */
            SWCH = dump_device;                         /* set the default dump device */
            break;

        case Cold_Load:                                 /* LOAD */
            SWCH = load_device;                         /* set the default load device */
            break;

        case Run:                                       /* a run request */
            return SCPE_IERR;                           /*   should not be possible */
        }                                               /* all cases are handled */

else if (arg == Cold_Dump)                              /* otherwise if no IOP dump value was given */
    SWCH = dump_control;                                /*   then use the system control panel presets */

else if (arg == Cold_Start)                             /* otherwise if a START was done for an IOP system */
    return SCPE_NOFNC;                                  /*   then report that the command is not allowed */

cpu_front_panel (SWCH, (PANEL_SWITCH) arg);             /* set up the cold load or dump microcode */

return run_cmd (RU_CONT, buf);                          /* execute the halt-mode routine */
}


/* Execute the POWER commands.

   This command processing routine is called to initiate a power failure or
   power restoration.  The "cptr" parameter points to the power option keyword;
   the "arg" parameter is not used.

   The routine processes commands of the form:

     POWER { FAIL | OFF | DOWN }
     POWER { RESTORE | ON | UP }

   In simulation, the "cpu_power_state" global variable indicates the current
   state of system power.  The simulator starts in the Power_On state.  The
   POWER FAIL command moves from the Power_On to the Power_Failing state if the
   CPU is running, or to the Power_Off state if it is not.  Execution of a HALT
   in the Power_Failing state moves to the Power_Off state.  The POWER RESTORE
   command moves from the Power_Off to the Power_Returning state if the CPU is
   running, or to the Power_On state if it is not.  Execution of the power-on
   trap moves from the Power_Returning to the Power_On state.

   The POWER FAIL and POWER RESTORE commands are only valid in the Power_On and
   Power_Off states, respectively; otherwise, they print "Command not allowed."

   The four enumeration values model the states of the PON (power on) and PFW
   (power-fail warning) hardware signals, as follows:

     PON  PFW  State            Simulator Action
     ---  ---  ---------------  ----------------------------
      1    0   power on         executing normally
      1    1   power failing    executing with cpx1_PFINTR
      0    1   power off        will not execute
      0    0   power returning  executing with trap_Power_On

   In microcode, the power-fail routine updates to the word following the last
   word of the ICS to indicate the machine status.  For IOP machines, the
   current value of the CPX2 register is stored.  For IMB machines, the word is
   modified to set bit 0 if the CPU is running and to set bit 15 to indicate a
   power failure.  In either case, this value is used by the power-on routine to
   decide if the CPU was running (cpx2_RUN bit is set) or halted at the time of
   the power failure.  A power failure is indicated by setting the cpx1_PFINTR
   bit in the CPX1 register; this causes an interrupt to the power-failure
   routine in the operating system, which performs an orderly shutdown followed
   by a programmed HALT to wait for power to die.

   When power is restored, the power-on trap (trap_Power_On) is set up, and
   then, if the PF/ARS switch is in the "enable" position or the CPU is an IMB
   machine, the trap is taken, which restarts the operating system and any I/O
   that was in progress when power failed.  If the switch is in the "disable"
   position, the CPU remains halted, and the trap is taken when the RUN button
   is pressed.

   The POWER commands are entered at the SCP prompt.  If the machine was running
   at the time of power failure, execution is resumed automatically to execute
   the power-fail or power-restore OS routines.  If the machine was halted when
   the POWER commands were entered, the machine remains halted -- just the power
   state changes.


   Implementation notes:

    1. In hardware, when the power fail interrupt is serviced, the microcode
       sets the PWR INH flip-flop, which locks out the RUN switch and inhibits
       all other halt-mode (CPX2) and run-mode (CPX1) interrupts until the CPU
       is reset.  This ensures that the software power-fail interrupt handler
       executes unimpeded.  In simulation, a POWER FAIL command with the CPU
       running sets the power-fail interrupt bit in the CPX1 register and
       resumes execution in the Power_On state.  When the interrupt is detected
       by the "run_mode_interrupt" routine, the state is changed to
       Power_Failing.  In this state, run-mode interrupts are not recognized.
       Halt-mode interrupts need no special handling, as the power state is
       changed to Power_Off when the CPU halts.  Therefore, the CPU is never in
       a "halted-and-waiting-for-power-to-fade-away" state.

    2. The run command call uses the RU_CONT argument instead of RU_RUN so that
       the run_cmd SCP routine will not reset all devices before entering the
       instruction executor.  The halt-mode interrupt handlers for cold load and
       cold dump reset the simulator as appropriate for their commands (i.e.,
       the CPU and all I/O devices, or just the I/O devices, respectively).

    3. In order to set up the power-on trap, this routine presses the RUN button
       and continues the simulation.  After the trap is set up in the
       "sim_instr" routine, the halt-mode interrupt handler checks the PF/ARS
       switch and stops simulation if auto-restart is disabled.

    4. At power restoration on IMB CPUs, the word above the ICS stack limit
       (i.e., the word at location ZI + 1) has bit 0 set if the CPU was running
       and bit 15 set unilaterally.  If the CPU was halted, we clear bit 15,
       which corresponds to the CPX2 run flag.  This allows us to use the same
       "was running" test.
*/

t_stat cpu_power_cmd (int32 arg, char *cptr)
{
static CTAB options [] = {
    { "FAIL",    NULL, Power_Failing   },
    { "RESTORE", NULL, Power_Returning },
    { "OFF",     NULL, Power_Failing   },
    { "ON",      NULL, Power_Returning },
    { "DOWN",    NULL, Power_Failing   },
    { "UP",      NULL, Power_Returning },
    { NULL }
    };

char    gbuf [CBUFSIZE];
CTAB    *ctptr;
HP_WORD zi, zi_p1;
t_stat  status;

if (cptr == NULL || *cptr == '\0')                      /* if there is no option word */
    return SCPE_2FARG;                                  /*   then report a missing argument */

cptr  = get_glyph (cptr, gbuf, 0);                      /* parse (and upshift) the option specified */
ctptr = find_ctab (options, gbuf);                      /*   and look it up in the option table */

if (ctptr == NULL)                                      /* if the option is not valid */
    status = SCPE_ARG;                                  /*   then report a bad argument */

else if (*cptr != '\0')                                 /* otherwise if something follows the option */
    return SCPE_2MARG;                                  /*   then report too many arguments */

else if (ctptr->arg == Power_Failing)                   /* otherwise if a power-fail option was given */
    if (cpu_power_state != Power_On)                    /*   but the CPU power is not on */
        status = SCPE_NOFNC;                            /*     then the command is not allowed */

    else {                                              /* otherwise the failure is valid */
        if (sys_configuration & CPU_IS_IMB) {           /*   so if this is an IMB system */
            imb_assert_PFW ();                          /*     then assert a power-fail warning to all channels */

            cpu_read_memory (absolute, ICS_Z, &zi);     /* get the ICS stack limit */
            cpu_read_memory (absolute, zi + 1, &zi_p1); /*   and then get the word above the limit */

            zi_p1 = zi_p1 & ~HP_BIT (0) | HP_BIT (15);  /* clear bit 0 and set bit 15 to indicate a power failure */

            if (CPX2 & cpx2_RUN)                        /* if the CPU is running */
                zi_p1 |= HP_BIT (0);                    /*   then set bit 0 to indicate */

            cpu_write_memory (absolute, zi + 1, zi_p1); /* replace the word above the ICS limit */
            }

        else {                                          /* otherwise for IOP machines */
            iop_assert_PFWARN ();                       /*   send a power-fail warning to all devices */

            cpu_read_memory (absolute, ICS_Z, &zi);     /* get the ICS stack limit */
            cpu_write_memory (absolute, zi + 1, CPX2);  /*   and save the CPX2 value in the following word */
            }

        if (CPX2 & cpx2_RUN) {                          /* if the CPU is currently running */
            CPX1 |= cpx1_PFINTR;                        /*   then set the power-fail interrupt */
            CPX2 |= cpx2_RUNSWCH;                       /*     and assume a RUN command */
            status = run_cmd (RU_CONT, cptr);           /*       and continue execution */
            }

        else {                                          /* otherwise the CPU is currently halted */
            cpu_power_state = Power_Off;                /*   so remain halted in the "power is off" state */
            status = SCPE_OK;                           /*     and return command success */
            }
        }

else if (ctptr->arg == Power_Returning)                 /* otherwise if a power-restoration option was given */
    if (cpu_power_state != Power_Off)                   /*   but the CPU power is not off */
        status = SCPE_NOFNC;                            /*     then the command is not allowed */

    else {                                              /* otherwise the restoration is valid */
        reset_all_p (0);                                /*   so reset all devices to their power on states */

        cpu_read_memory (absolute, ICS_Z, &zi);         /* get the ICS stack limit */
        cpu_read_memory (absolute, zi + 1, &zi_p1);     /*   and then get the word above the limit */

        if (sys_configuration & CPU_IS_IOP)             /* if this is an IOP system */
            cpu_write_memory (absolute, zi + 1, CPX2);  /*   then replace it with the current CPX2 value */

        else if (not (zi_p1 & HP_BIT (0)))              /* otherwise if the IMB CPU was halted when power failed */
            zi_p1 = 0;                                  /*   then clear the run flag */

        if (zi_p1 & cpx2_RUN) {                         /* if the CPU was running at the time of power failure */
            cpu_power_state = Power_Returning;          /*   then move to the "power is returning" state */
            CPX2 |= cpx2_RUNSWCH;                       /*     and assume a RUN command */
            status = run_cmd (RU_CONT, cptr);           /*       and continue execution */
            }

        else {                                          /* otherwise the CPU was halted when power failed */
            cpu_power_state = Power_On;                 /*   so remain halted in the "power is on" state */
            status = SCPE_OK;                           /*     and return command success */
            }
        }

else                                                    /* otherwise a valid option has no handler */
    status = SCPE_IERR;                                 /*   so report an internal error */

return status;                                          /* return the operation status */
}



/*******************************
 *                             *
 * CPU global utility routines *
 *                             *
 *******************************/


/* Test if the CPU is running.

   This routine allows an I/O device to determine whether the CPU is running or
   halted.  Some I/O operations, such as requesting a release from the target
   operating system before unloading a tape cartridge during a DETACH, do not
   make sense when the operating system is not running.  Such conditional
   actions can be performed unconditionally if the CPU is not executing
   instructions.

   Note that this test does not return FALSE when execution is suspended by a
   breakpoint or the STEP command.  The CPU halts only when it executes a HALT
   instructionm, a System Halt occurs, or power is restored with the
   auto-restart switch set to the NOARS position.


   Implementation notes:

    1. We take advantage of the assignment of the RUN flip-flop to bit 15 of
       CPX2 as returning 0 or 1 for FALSE or TRUE.  Consequently, we do not need
       to test the masked result explicitly.
*/

t_bool cpu_is_running (void)
{
return (t_bool) (CPX2 & cpx2_RUN);                      /* return TRUE if the CPU is running */
}


/* Test for a pending interrupt.

   This routine is called from within instruction executors for interruptible
   instructions to test for a pending interrupt.  It counts an event tick and
   returns TRUE if the instruction should yield, either for an interrupt or for
   an event error, or FALSE if the instruction should continue.

   Instructions that potentially take a long time (e.g., MOVE, SCU, LLSH) test
   for pending interrupts after each word or byte moved or scanned.  The design
   of these instructions is such that an interrupt may be serviced and the
   instruction resumed without disruption.  For example, the MOVE instruction
   updates the source and target addresses and word count on the stack after
   each word moved.  If the instruction is interrupted, the values on the CPU
   stack indicate where to resume after the interrupt handler completes.


   Implementation notes:

    1. While simply calling "interrupt_pending" here looks inefficient, it
       actually turns out better than if the instruction executors called that
       routine directly.  This is because the routine is marked "inline," and
       all calls pass constants for the "invocation" parameter.  This causes the
       function call below to be replaced with a copy of the "interrupt_pending"
       code, but with all of the invocation-conditional tests removed.
       Effectively, the copy is customized for the instruction executors, while
       still having a single routine for simpler code maintenance.
*/

t_bool cpu_interrupt_pending (t_stat *status)
{
return interrupt_pending (Executor, status);            /* call the common interrupt tester */
}


/* Set up a front panel operation.

   This routine sets the SWCH register to the supplied value and then sets the
   appropriate bit in the CPX2 register.  This will cause a halt-mode interrupt
   when simulated execution is resumed.


   Implementation notes:

    1. We do this here to avoid having to export the registers and CPX values
       globally.
*/

void cpu_front_panel (HP_WORD switch_reg, PANEL_SWITCH request)
{
SWCH = switch_reg;                                      /* set the SWCH register value */

switch (request) {                                      /* dispatch on the request type */

    case Run:                                           /* a run request */
        CPX2 |= cpx2_RUNSWCH;                           /* set the RUN switch */
        break;

    case Cold_Start:                                    /* a cold start request */
    case Cold_Load:                                     /* a cold load request */
        CPX2 |= cpx2_LOADSWCH;                          /* set the LOAD switch */
        break;

    case Cold_Dump:                                     /* a cold dump request */
        CPX2 |= cpx2_DUMPSWCH;                          /* set the DUMP switch */
        break;
    }                                                   /* all cases are handled */

return;
}


/* Update the process clock and counter register.

   If the process clock is currently calibrated, then the service interval is
   actually ten times the hardware period of 1 millisecond.  This provides
   sufficient event service call spacing to allow idling to work.

   To present the correct value when the process clock is read, this routine is
   called to increment the count by an amount proportional to the fraction of
   the service interval that has elapsed.  In addition, it is called by the CPU
   instruction postlude, so that the PCLK register will have the correct value
   if it is examined from the SCP command prompt.

   A similar action is performed for the counter register.


   Implementation notes:

    1. The CPU unit pointer cannot be a "const" pointer because the parameter to
       "sim_activate_time" is not declared constant.

    2. The process clock is updated only if the CPU is not executing on the
       Interrupt Control Stack.  The decision to update or not is based on the
       execution state at the end of the interval.  This may not be correct if
       the state changed during the interval.  However, it should average out
       properly, given a large enough sample size, and assuming that ICS entry
       and exit are relatively random compared to the timing of calls to this
       routine.
*/

void cpu_update_pclk (void)
{
int32 elapsed, ticks;
UNIT *uptr = &cpu_unit [0];                     /* a pointer to the CPU process clock unit */

if (uptr->flags & UNIT_CALTIME) {                       /* if the process clock is calibrated */
    elapsed = uptr->wait - sim_activate_time (uptr);    /*   then elapsed time is original time less remaining time */

    ticks = (elapsed * PCLK_MULTIPLIER) / uptr->wait    /* the adjustment is the elapsed fraction of the multiplier */
              - (PCLK_MULTIPLIER - pclk_increment);     /*   less the amount of any adjustment already made */

    if (not (CPX1 & cpx1_ICSFLAG))                      /* if not executing on the ICS */
        PCLK = PCLK + ticks & R_MASK;                   /*   then update the process clock counter with rollover */

    CR = CR + ticks & R_MASK;                           /* update the counter register with rollover */
    pclk_increment = pclk_increment - ticks;            /*   and reduce the amount remaining to add at service */
    }

return;
}



/*********************************************
 *                                           *
 * CPU global instruction execution routines *
 *                                           *
 *********************************************/


/* Push the stack down.

   This routine implements the PUSH micro-order to create space on the stack for
   a new value.  On return, the new value may be stored in the RA register.

   If the SR register indicates that all of the TOS registers are in use, then
   the RD register is freed by performing a queue down.  Then the values in the
   TOS registers are shifted down, freeing the RA register.  Finally, SR is
   incremented in preparation for the store.
*/

void cpu_push (void)
{
if (SR == 4)                                            /* if all TOS registers are full */
    cpu_queue_down ();                                  /*   then move the RD value to memory */

RD = RC;                                                /* shift */
RC = RB;                                                /*   the register */
RB = RA;                                                /*     values down */

SR = SR + 1;                                            /* increment the register-in-use count */

return;
}


/* Pop the stack up.

   This routine implements the POP micro-order to delete the top-of-stack value.

   On entry, if the SR register indicates that all of the TOS registers are
   empty, then if decrementing the SM register would move it below the DB
   register value (for IOP machines) or Q register value (for IMB machines), and
   the CPU is not in privileged mode, then a Stack Underflow trap is taken.
   Otherwise, the stack memory pointer is decremented.

   If one or more values exist in the TOS registers, the values are shifted up,
   deleting the previous value in the RA register, and SR is decremented.
*/

void cpu_pop (void)
{
const uint32 is_imb = sys_configuration & CPU_IS_IMB;   /* non-zero if the CPU is an IMB system */

if (SR == 0) {                                          /* if the TOS registers are empty */
    if (SM <= (is_imb ? Q : DB) && NPRV)                /*   then if SM is not above Q/DB and the mode is non-privileged */
        MICRO_ABORT (trap_Stack_Underflow);             /*     then trap with a Stack Underflow */

    SM = SM - 1 & R_MASK;                               /* decrement the stack memory register */
    }

else {                                                  /* otherwise at least one TOS register is occupied */
    RA = RB;                                            /*   so shift */
    RB = RC;                                            /*     the register */
    RC = RD;                                            /*       values up */

    SR = SR - 1;                                        /* decrement the register-in-use count */
    }

return;
}


/* Queue a value from memory up to the register file.

   This routine implements the QUP micro-order to move the value at the top of
   the memory stack into the bottom of the TOS register file.  There must be a
   free TOS register when this routine is called.

   On entry, if decrementing the SM register would move it below the DB register
   value (for IOP machines) or Q register value (for IMB machines), and the CPU
   is not in privileged mode, then a Stack Underflow trap is taken.  Otherwise,
   the value pointed to by SM is read into the first unused TOS register, SM is
   decremented to account for the removed value, and SR is incremented to
   account for the new TOS register in use.


   Implementation notes:

    1. SR must be less than 4 on entry, so that TR [SR] is the first unused TOS
       register.

    2. SM and SR must not be modified within the call to cpu_read_memory.  For
       example, SR++ cannot be passed as a parameter.
*/

void cpu_queue_up (void)
{
const uint32 is_imb = sys_configuration & CPU_IS_IMB;   /* non-zero if the CPU is an IMB system */

if (SM <= (is_imb ? Q : DB) && NPRV)                    /* if SM is not above Q/DB and the mode is non-privileged */
    MICRO_ABORT (trap_Stack_Underflow);                 /*   then trap with a Stack Underflow */

else {                                                  /* otherwise */
    cpu_read_memory (stack, SM, &TR [SR]);              /*   read the value from memory into a TOS register */

    SM = SM - 1 & R_MASK;                               /* decrement the stack memory register */
    SR = SR + 1;                                        /*   and increment the register-in-use count */
    }

return;
}


/* Queue a value from the register file down to memory.

   This routine implements the QDWN micro-order to move the value at the bottom
   of the TOS register file into the top of the memory stack.  There must be a
   TOS register in use when this routine is called.

   On entry, if incrementing the SM register would move it above the Z register
   value, then a Stack Overflow trap is taken.  Otherwise, SM is incremented to
   account for the new value written, SR is decremented to account for the TOS
   register removed from use, and the value in that TOS register is written into
   memory at the top of the memory stack.


   Implementation notes:

    1. SR must be greater than 0 on entry, so that TR [SR - 1] is the last TOS
       register in use.

    2. SM and SR must not be modified within the call to cpu_write_memory.  For
       example, SR-- cannot be passed as a parameter.
*/

void cpu_queue_down (void)
{
if (SM >= Z)                                            /* if SM isn't below Z */
    MICRO_ABORT (trap_Stack_Overflow);                  /*   then trap with a Stack Overflow */

SM = SM + 1 & R_MASK;                                   /* increment the stack memory register */
SR = SR - 1;                                            /*   and decrement the register-in-use count */

cpu_write_memory (stack, SM, TR [SR]);                  /* write the value from a TOS register to memory */

return;
}


/* Flush the register file.

   This routine implements the PSHA microcode subroutine that writes the values
   of all TOS registers in use to the memory stack.  As each value is written,
   the SM register is incremented and the SR register is decremented.  On
   return, the SR register value will be zero.

   The routine does not check for stack overflow.
*/

void cpu_flush (void)
{
while (SR > 0) {                                        /* while one or more registers are in use */
    SM = SM + 1 & R_MASK;                               /*   increment the stack memory register */
    SR = SR - 1;                                        /*     and decrement the register-in-use count */

    cpu_write_memory (stack, SM, TR [SR]);              /* write the value from a TOS register to memory */
    }

return;
}


/* Adjust SR until it reaches a specified value.

   This routine implements the SRP1-SRP4 microcode subroutines that adjust the
   stack until the prescribed number (1-4) of TOS registers are occupied.  It
   performs queue ups, i.e., moves values from the top of the memory stack to
   the bottom of the register file, until the specified SR value is reached.
   Stack underflow is checked after the all of the values have been moved.  The
   routine assumes that at least one value must be moved.


   Implementation notes:

    1. SR must be greater than 0 on entry, so that TR [SR - 1] is the last TOS
       register in use.

    2. SM and SR must not be modified within the call to cpu_read_memory.  For
       example, SR++ cannot be passed as a parameter.

    3. The cpu_queue_up routine isn't used, as that routine checks for a stack
       underflow after each word is moved rather than only after the last word.
*/

void cpu_adjust_sr (uint32 target)
{
const uint32 is_imb = sys_configuration & CPU_IS_IMB;   /* non-zero if the CPU is an IMB system */

do {
    cpu_read_memory (stack, SM, &TR [SR]);              /* read the value from memory into a TOS register */

    SM = SM - 1 & R_MASK;                               /* decrement the stack memory register */
    SR = SR + 1;                                        /*   and increment the register-in-use count */
    }
while (SR < target);                                    /* queue up until the requested number of registers are in use */

if (SM < (is_imb ? Q : DB) && NPRV)                     /* if SM is not above Q/DB and the mode is non-privileged */
    MICRO_ABORT (trap_Stack_Underflow);                 /*   then trap with a Stack Underflow */

return;
}


/* Write a stack marker to memory.

   This routine implements the STMK microcode subroutine that writes a four-word
   marker to the stack.  The format of the marker is as follows:

       0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |                       X register value                        |  [Q - 3]  X
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | T | M |          PB-relative return address                   |  [Q - 2]  P + 1 - PB
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |                     Status register value                     |  [Q - 1]  STA
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |                         Delta Q value                         |  [Q - 0]  S - Q
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   After the values are written, the Q register is set to point to the marker.

   This routine is always entered with SR = 0, i.e., with the TOS registers
   flushed to the memory stack.  It does not check for stack overflow.


   Implementation notes:

    1. The PB-relative return address points to the instruction after the point
       of the call.  Conceptually, this is location P + 1, but because the CPU
       uses a two-instruction prefetch, the location is actually P - 1.

    2. The CSMAP register holds the mapping state of the current code segment.
       It is either STMK_M for physical mapping or 0 for logical mapping.
       Therefore, it can be merged into the stack marker directly if the CST
       expansion firmware is installed.
*/

void cpu_mark_stack (void)
{
HP_WORD return_address;

return_address = STMK_RETURN (P - 1 - PB);              /* determine the return address */

if (sys_configuration & FW_CSTX)                        /* if the CST expansion firmware is present */
    return_address |= CSMAP;                            /*   then add the code segment mapping */

SM = SM + 4 & R_MASK;                                   /* adjust the stack pointer */

cpu_write_memory (stack, SM - 3, X);                    /* push the index register */
cpu_write_memory (stack, SM - 2, return_address);       /*   and delta P */
cpu_write_memory (stack, SM - 1, STA);                  /*     and the status register */
cpu_write_memory (stack, SM - 0, SM - Q & LA_MASK);     /*       and delta Q */

Q = SM;                                                 /* set Q to point to the new stack marker */

return;
}


/* Calculate an effective memory address.

   This routine calculates the effective address for a memory reference or
   branch instruction.  On entry, "mode_disp" contains the mode, displacement,
   index, and indirect fields of the instruction, "classification" and "offset"
   point to variables to receive the corresponding values, and "selector" points
   to a variable to receive the byte selection ("upper" or "lower") for byte-
   addressable instructions or is NULL for word-addressable instructions.  On
   exit, "classification" is set to the memory access classification, "offset"
   is set to the address offset within the memory bank implied by the
   classification, and "selector" is set to indicate the byte referenced if the
   pointer is non-NULL.

   The mode and displacement fields of the instruction encode an address
   relative to one of the base registers P, DB, Q, or S, as follows:

       0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |   memory op   | X | I |         mode and displacement         |
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
                             | 0 | 0 |     P+ displacement 0-255     |
                             +---+---+---+---+---+---+---+---+---+---+
                             | 0 | 1 |     P- displacement 0-255     |
                             +---+---+---+---+---+---+---+---+---+---+
                             | 1 | 0 |    DB+ displacement 0-255     |
                             +---+---+---+---+---+---+---+---+---+---+
                             | 1 | 1 | 0 |   Q+ displacement 0-127   |
                             +---+---+---+---+---+---+---+---+---+---+
                             | 1 | 1 | 1 | 0 | Q- displacement 0-63  |
                             +---+---+---+---+---+---+---+---+---+---+
                             | 1 | 1 | 1 | 1 | S- displacement 0-63  |
                             +---+---+---+---+---+---+---+---+---+---+

   The displacement encoded in the instruction is an unsigned value that is
   added to or subtracted from the indicated base register.

   If the X and I fields are both 0, the addressing is direct.  If the X field
   is 1, the addressing is indexed.  If the I field is 1, the addressing is
   indirect.  If both fields are 1, the addressing is indirect indexed, with
   indirection performed before indexing.

   To improve execution speed in hardware, a preadder is implemented that sums
   the offset contained in certain bits of the CIR with the index register (if
   enabled).  The primary use is to save a microinstruction cycle during memory
   reference instructions, which must add a base register, the offset in the
   CIR, and potentially the X register (either directly or shifted left or right
   by one place for LDD/STD or LDB/STB, respectively).  The preadder also serves
   to hold other counts obtained from the CIR, e.g., shift counts, although in
   these cases, the addition function is not used.

   This routine simulates the preadder as part of the effective address
   calculation.  The calculations employed for word addressing are:

     Direct word addressing:
       ea = PBANK.(P + displacement)
       ea = DBANK.(DB + displacement)
       ea = SBANK.(Q,S + displacement)

     Direct indexed word addressing:
       ea = PBANK.(P + displacement + X)
       ea = DBANK.(DB + displacement + X)
       ea = SBANK.(Q,S + displacement + X)

     Indirect word addressing:
       ea = PBANK.(P + displacement + M [PBANK.(P + displacement)])
       ea = DBANK.(DB + M [DBANK.(DB + displacement)])
       ea = DBANK.(DB + M [SBANK.(Q,S + displacement)])

     Indirect indexed word addressing:
       ea = PBANK.(P + displacement + M [PBANK.(P + displacement)] + X)
       ea = DBANK.(DB + M [DBANK.(DB + displacement)] + X)
       ea = DBANK.(DB + M [SBANK.(Q,S + displacement)] + X)

   The indirect cell contains either a self-relative, P-relative address or a
   DB-relative address, even for S or Q-relative modes.  Indirect branches with
   DB/Q/S-relative mode are offsets from PB, not DB.

   The effective address calculations employed for byte addressing are:

     Direct byte addressing:
       ea = DBANK.(DB + displacement).byte [0]
       ea = SBANK.(Q,S + displacement).byte [0]

     Direct indexed byte addressing:
       ea = DBANK.(DB + displacement + X / 2).byte [X & 1]
       ea = SBANK.(Q,S + displacement + X / 2).byte [X & 1]

     Indirect byte addressing:
       ea,I = DBANK.(DB + M [DBANK.(DB + displacement)] / 2).byte [cell & 1]
       ea,I = DBANK.(DB + M [SBANK.(Q,S + displacement)] / 2).byte [cell & 1]

     Indirect indexed byte addressing:
       ea,I = DBANK.(DB + (M [DBANK.(DB + displacement)] + X) / 2).byte [cell + X & 1]
       ea,I = DBANK.(DB + (M [SBANK.(Q,S + displacement)] + X) / 2).byte [cell + X & 1]

   For all modes, the displacement is a word address, whereas the indirect cell
   and index register contain byte offsets.  For direct addressing, the byte
   selected is byte 0.  For all other modes, the byte selected is the byte at
   (offset & 1), where the offset is the index register value, the indirect cell
   value, or the sum of the two.

   Byte offsets into data segments present problems, in that negative offsets
   are permitted (to access the DL-to-DB area), but there are not enough bits to
   represent all locations unambiguously in the potential -32K to +32K word
   offset range.  Therefore, a byte offset with bit 0 = 1 can represent either a
   positive or negative word offset from DB, depending on the interpretation.
   The HP 3000 adopts the convention that if the address resulting from a
   positive-offset interpretation does not fall within the DL-to-S range, then
   32K is added to the address, effectively changing the interpretation from a
   positive to a negative offset.  If this new address does not fall within the
   DL-to-S range, a Bounds Violation trap occurs if the mode is non-privileged.

   The reinterpretation as a negative offset is performed only if the CPU is not
   in split-stack mode (where either DBANK is different from SBANK, or DB does
   not lie between DL and Z), as extra data segments do not permit negative-DB
   addressing.  Reinterpretation is also not used for code segments, as negative
   offsets from PB are not permitted.


   Implementation notes:

    1. On entry, the program counter points to the instruction following the
       next instruction (i.e., the NIR location + 1).  However, P-relative
       offsets are calculated from the current instruction (CIR location).
       Therefore, we decrement P by two before adding the offset.

       In hardware, P-relative addresses obtained from the preadder are offset
       from P + 1, whereas P-relative addresses obtained by summing with
       P-register values obtained directly from the S-Bus are offset from P + 2.
       This is because the P-register increment that occurs as part of a NEXT
       micro-order is coincident with the R-Bus and S-Bus register loads; both
       operations occur when the NXT+1 signal asserts.  Therefore, the microcode
       handling P-relative memory reference address resolution subtracts one to
       get the P value corresponding to the CIR, whereas branches on overflow,
       carry, etc. subtract two.

    2. If the mode is indirect, this routine handles bounds checks and TOS
       register accesses on the initial address.

    3. The System Reference Manual states that byte offsets are interpreted as
       negative if the effective address does not lie between DL and Z.
       However, the Series II microcode actually uses DL and S for the limits.

    4. This routine calculates the effective address for the memory address
       instructions.  These instructions always bounds-check their accesses, and
       data and stack accesses are mapped to the TOS registers if the effective
       address lies between SM and SM + SR within the stack bank.
*/

void cpu_ea (HP_WORD mode_disp, ACCESS_CLASS *classification, HP_WORD *offset, BYTE_SELECTOR *selector)
{
HP_WORD      base, displacement;
ACCESS_CLASS class;

switch ((mode_disp & MODE_MASK) >> MODE_SHIFT) {        /* dispatch on the addressing mode */

    case 000:
    case 001:
    case 002:
    case 003:                                           /* positive P-relative displacement */
        base = P - 2 + (mode_disp & DISPL_255_MASK);    /* add the displacement to the base */
        class = program_checked;                        /*   and classify as a program reference */
        break;

    case 004:
    case 005:
    case 006:
    case 007:                                           /* negative P-relative displacement */
        base = P - 2 - (mode_disp & DISPL_255_MASK);    /* subtract the displacement from the base */
        class = program_checked;                        /*   and classify as a program reference */
        break;

    case 010:
    case 011:
    case 012:
    case 013:                                           /* positive DB-relative displacement */
        base = DB + (mode_disp & DISPL_255_MASK);       /* add the displacement to the base */
        class = data_mapped_checked;                    /*   and classify as a data reference */
        break;

    case 014:
    case 015:                                           /* positive Q-relative displacement */
        base = Q + (mode_disp & DISPL_127_MASK);        /* add the displacement to the base */
        class = stack_checked;                          /*   and classify as a stack reference */
        break;

    case 016:                                           /* negative Q-relative displacement */
        base = Q - (mode_disp & DISPL_63_MASK);         /* subtract the displacement from the base */
        class = stack_checked;                          /*   and classify as a stack reference */
        break;

    case 017:                                           /* negative S-relative displacement */
        base = SM + SR - (mode_disp & DISPL_63_MASK);   /* subtract the displacement from the base */
        class = stack_checked;                          /*   and classify as a stack reference */
        break;
    }                                                   /* all cases are handled */


if (not (mode_disp & I_FLAG_BIT_5))                         /* if the mode is direct */
    displacement = 0;                                       /*   then there's no displacement */

else {                                                      /* otherwise the mode is indirect */
    cpu_read_memory (class, base & LA_MASK, &displacement); /*   so get the displacement value */

    if ((CIR & BR_MASK) == BR_DBQS_I) {                 /* if this a DB/Q/S-relative indirect BR instruction */
        base = PB;                                      /*   then PB is the base for the displacement */
        class = program_checked;                        /* reclassify as a program reference */
        }

    else if (class != program_checked) {                /* otherwise if it is a data or stack reference */
        base = DB;                                      /*   then DB is the base for the displacement */
        class = data_mapped_checked;                    /* reclassify as a data reference */
        }
                                                        /* otherwise, this is a program reference */
    }                                                   /*   which is self-referential */

if ((CIR & LSDX_MASK) == LDD_X                          /* if the mode */
  || (CIR & LSDX_MASK) == STD_X)                        /*   is double-word indexed */
    displacement = displacement + X * 2 & DV_MASK;      /*     then add the doubled index to the displacement */

else if (mode_disp & X_FLAG)                            /* otherwise if the mode is indexed */
    displacement = displacement + X & DV_MASK;          /*   then add the index to the displacement */

if (selector == NULL)                                   /* if a word address is requested */
    base = base + displacement;                         /*   then add in the word displacement */

else if ((mode_disp & (X_FLAG | I_FLAG_BIT_5)) == 0)    /* otherwise if a direct byte address is requested */
    *selector = upper;                                  /*   then it references the upper byte */

else {                                                  /* otherwise an indexed or indirect byte address is requested */
    if (displacement & 1)                               /*   so if the byte displacement is odd */
        *selector = lower;                              /*     then the lower byte was requested */
    else                                                /*   otherwise it is even */
        *selector = upper;                              /*     and the upper byte was requested */

    base = base + (displacement >> 1) & LA_MASK;        /* convert the displacement from byte to word and add */

    if (DBANK == SBANK && DL <= DB && DB <= Z           /* if not in split-stack mode */
     && (base < DL || base > SM + SR))                  /*   and the word address is out of range */
        base = base ^ D16_SIGN;                         /*     then add 32K to swap the offset polarity */
    }

*offset = base & LA_MASK;                               /* set the */
*classification = class;                                /*   return values */

return;
}


/* Convert a data- or program-relative byte address to a word effective address.

   The supplied byte offset from DB or PB is converted to a memory address,
   bounds-checked if requested, and returned.  If the supplied block length is
   not zero, the converted address is assumed to be the starting address of a
   block, and an ending address, based on the block length, is calculated and
   bounds-checked if requested.  If either address lies outside of the segment
   associated with the access class, a Bounds Violation trap occurs, unless a
   privileged data access is requested.

   The "class" parameter indicates whether the offset is from DB or PB and
   must be "data", "data_checked", "program", or "program_checked".  No other
   access classes are supported.  In particular, "data_mapped" and
   "data_mapped_checked" are not allowed, as callers of this routine never map
   accesses to the TOS registers.

   Byte offsets into data segments present problems, in that negative offsets
   are permitted (to access the DL-to-DB area), but there are not enough bits to
   represent all locations unambiguously in the potential -32K to +32K word
   offset range.  Therefore, a byte offset with bit 0 = 1 can represent either a
   positive or negative word offset from DB, depending on the interpretation.
   The HP 3000 adopts the convention that if the address resulting from a
   positive-offset interpretation does not fall within the DL-to-S range, then
   32K is added to the address, effectively changing the interpretation from a
   positive to a negative offset.  If this new address does not fall within the
   DL-to-S range, a Bounds Violation trap occurs if the mode is non-privileged.

   The reinterpretation as a negative offset is performed only if the CPU is not
   in split-stack mode (where either DBANK is different from SBANK, or DB does
   not lie between DL and Z), as extra data segments do not permit negative-DB
   addressing.  Reinterpretation is also not used for code segments, as negative
   offsets from PB are not permitted.


   Implementation notes:

    1. This routine implements the DBBC microcode subroutine.
*/

uint32 cpu_byte_ea (ACCESS_CLASS class, uint32 byte_offset, uint32 block_length)
{
uint32 starting_word, ending_word, increment;

if (block_length & D16_SIGN)                            /* if the block length is negative */
    increment = 0177777;                                /*   then the memory increment is negative also */
else                                                    /* otherwise */
    increment = 1;                                      /*   the increment is positive */

if (class == program || class == program_checked) {     /* if this is a program access */
    starting_word = PB + (byte_offset >> 1) & LA_MASK;  /*   then determine the starting word address */

    if (class == program_checked                        /* if checking is requested */
      && starting_word < PB || starting_word > PL)      /*   and the starting address is out of range */
        MICRO_ABORT (trap_Bounds_Violation);            /*     then trap for a bounds violation */

    if (block_length != 0) {                            /* if a block length was supplied */
        ending_word =                                   /*   then determine the ending address */
           starting_word + ((block_length - increment + (byte_offset & 1)) >> 1) & LA_MASK;

        if (class == program_checked                    /* if checking is requested */
          && ending_word < PB || ending_word > PL)      /*   and the ending address is out of range */
            MICRO_ABORT (trap_Bounds_Violation);        /*     then trap for a bounds violation */
        }
    }

else {                                                  /* otherwise this is a data address */
    starting_word = DB + (byte_offset >> 1) & LA_MASK;  /*   so determine the starting word address */

    if (DBANK == SBANK && DL <= DB && DB <= Z           /* if not in split-stack mode */
      && (starting_word < DL || starting_word > SM)) {  /*   and the word address is out of range */
        starting_word = starting_word ^ D16_SIGN;       /*     then add 32K and try again */

        if (class == data_checked && NPRV                   /* if checking is requested and non-privileged */
          && (starting_word < DL || starting_word > SM))    /*   and still out of range */
            MICRO_ABORT (trap_Bounds_Violation);            /*     then trap for a bounds violation */
        }

    if (block_length != 0) {                            /* if a block length was supplied */
        ending_word =                                   /*   then determine the ending word address */
           starting_word + ((block_length - increment + (byte_offset & 1)) >> 1) & LA_MASK;

        if (class == data_checked && NPRV               /* if checking is requested and non-privileged */
          && (ending_word < DL || ending_word > SM))    /*   and the address is out of range */
            MICRO_ABORT (trap_Bounds_Violation);        /*     then trap for a bounds violation */
        }
    }


return starting_word;                                   /* return the starting word address */
}


/* Set up the entry into an interrupt handler.

   This routine prepares the CPU state to execute an interrupt handling
   procedure.  On entry, "class" is the classification of the current interrupt,
   and "parameter" is the parameter associated with the interrupt.  On exit, the
   stack has been set up correctly, and the PB, P, PL, and status registers have
   been set up for entry into the interrupt procedure.

   Run-mode interrupts are classified as external or internal and ICS or
   non-ICS.  External interrupts are those originating with the device
   controllers, and internal interrupts are conditions detected by the microcode
   (e.g., a bounds violation or arithmetic overflow).  ICS interrupts execute
   their handlers on the system's Interrupt Control Stack.  Non-ICS interrupts
   execute on the user's stack.

   Of the run-mode interrupts, the External, System Parity Error, Address
   Parity Error, Data Parity Error, and Module interrupts execute on the ICS.
   All other interrupts execute on the user's stack.  The routine begins by
   determining whether an ICS or non-ICS interrupt is indicated.  The
   appropriate stack is established, and the stack marker is written to preserve
   the state of the interrupted routine.  The label of the handler procedure is
   obtained, and then the procedure designated by the label is set up.  On
   return, the first instruction of the handler is ready to execute.


   Implementation notes:

    1. This routine implements various execution paths through the microcode
       labeled as INT0 through INT7.

    2. This routine is also called directly by the IXIT instruction executor if
       an external interrupt is pending.  This is handled as an external
       interrupt but is classified differently so that the teardown and rebuild
       of the stack may be avoided to improve performance.

    3. The IOP and IMB machines use different DRT layouts, which must be
       accommodated here.  Specifically, the interrupt handler program label and
       DB value entries are interchanged in the table.
*/

void cpu_setup_irq_handler (IRQ_CLASS class, HP_WORD parameter)
{
HP_WORD label;
PLABEL  plab;

tprintf (cpu_dev, TRACE_INSTR | TRACE_IRQ, BOV_FORMAT "%s interrupt\n",
         PBANK, P - 1 & R_MASK, parameter, interrupt_name [class]);

if (class == irq_External || class == irq_IXIT) {       /* if entry is for an external interrupt */
    if (class == irq_External)                          /*   then if it was detected during normal execution */
        cpu_setup_ics_irq (class, 0);                   /*     then set it up on the ICS */
    else                                                /*   otherwise it was detected during IXIT */
        SM = Q + 2 & R_MASK;                            /*     so the ICS is already set up */

    DBANK = 0;                                          /* all handlers are in bank 0 */
    STA = STATUS_M | STATUS_I;                          /* enter privileged mode with interrupts enabled */

    if (sys_configuration & CPU_IS_IOP) {                   /* if this is an IOP system */
        cpu_read_memory (stack, parameter * 4 + 2, &DB);    /*   then read the DB value */
        cpu_read_memory (stack, parameter * 4 + 1, &label); /*     and the procedure label from the DRT */
        }

    else {                                                  /* otherwise for IMB machines */
        cpu_read_memory (stack, parameter * 4 + 1, &DB);    /*   read the DB value */
        cpu_read_memory (stack, parameter * 4 + 2, &label); /*     and the procedure label from the DRT */
        }
    }

else if (class >= irq_System_Parity                     /* otherwise if entry is for */
  && class <= irq_System_Clock) {                       /*   another ICS interrupt */
    cpu_setup_ics_irq (class, 0);                       /*     then set it up on the ICS */

    label = TO_ISR_LABEL (class);                       /* form the label for the specified classification */

    STA = STATUS_M;                                     /* clear status and enter privileged mode */
    }

else {                                                  /* otherwise entry is for a non-ICS interrupt */
    if (class == irq_Integer_Overflow)                  /* if this is an integer overflow interrupt */
        label = TO_ISR_LABEL (trap_User);               /*   then form the label for a user trap */
    else                                                /* otherwise form the label */
        label = TO_ISR_LABEL (class);                   /*   for the specified classification */

    cpu_flush ();                                       /* flush the TOS registers to memory */
    cpu_mark_stack ();                                  /*   and write a stack marker */

    STA = STATUS_M;                                     /* clear status and enter privileged mode */
    }

SM = SM + 1 & R_MASK;                                   /* increment the stack pointer */
cpu_write_memory (stack, SM, parameter);                /*   and push the parameter on the stack */

X = CIR;                                                /* save the CIR in the index register */

plab.label = label;                                     /* create an external label descriptor */
plab.mode  = External;                                  /*   for the interrupt service routine */

cpu_call_procedure (plab, 0);                           /* set up to call the interrupt handling procedure */

return;
}


/* Set up an interrupt on the Interrupt Control Stack.

   This routine prepares the Interrupt Control Stack (ICS) to support interrupt
   processing.  It is called from the run-time interrupt routine for ICS
   interrupts, the microcode abort routine for ICS traps, and from the DISP and
   PSEB instruction executors before entering the dispatcher.  On entry, "class"
   is the interrupt classification, and, if the class is "irq_Trap", then "trap"
   is the trap classification.  The trap classification is ignored for
   interrupts, including the dispatcher start interrupt.

   Unless entry is for a Cold Load trap, the routine begins by writing a
   six-word stack marker.  This special ICS marker extends the standard marker
   by adding the DBANK and DB values as follows:

       0   1   2   3   4   5   6   7   8   9  10  11  12  13  14  15
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |                       X register value                        |  [Q - 3]  X
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |                  PB-relative return address                   |  [Q - 2]  P + 1 - PB
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |                     Status register value                     |  [Q - 1]  STA
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     | D |                     Delta Q value                         |  [Q - 0]  S - Q
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |                         DB-Bank value                         |  [Q + 1]  DBANK
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
     |                           DB value                            |  [Q + 2]  DB
     +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     D = the dispatcher was interrupted

   After the values are written, the Q register is set to point to the marker.
   The stack bank register is then cleared, as the ICS is always located in
   memory bank 0.

   If the interrupt or trap occurred while executing on the ICS, and the
   dispatcher was running at the time, the "dispatcher running" bit in the CPX1
   register is cleared, and the D-bit is set in the delta-Q value word of the
   stack marker.  This bit will be used during interrupt exit to decide whether
   to restart the dispatcher.

   If the CPU was executing on the user's stack, the "ICS flag" bit in CPX1 is
   set, the Q register is reset to point at the permanent dispatcher stack
   marker established by the operating system, and the Z register is reset to
   the stack limit established by the OS for the ICS; the values are obtained
   from reserved memory locations 5 and 6, respectively.  The ICS DB value is
   read from the ICS global area that precedes the dispatcher stack marker and
   is used to write the stack-DB-relative S value back to the global area.

   Finally, the stack pointer is set to point just above the stack marker.


   Implementation notes:

    1. This routine implements various execution paths through the microcode
       labeled as INT1 through INT5.
*/

void cpu_setup_ics_irq (IRQ_CLASS class, TRAP_CLASS trap)
{
HP_WORD delta_q, stack_db;

if (class != irq_Trap                                   /* if this is not */
  || trap != trap_Cold_Load && trap != trap_Power_On) { /*   a cold load or power on trap entry */
    cpu_flush ();                                       /*     then flush the TOS registers to memory */
    cpu_mark_stack ();                                  /*       and write a four-word stack marker */

    cpu_write_memory (stack, SM + 1 & LA_MASK, DBANK);  /* add DBANK and DB to the stack */
    cpu_write_memory (stack, SM + 2 & LA_MASK, DB);     /*   to form a six-word ICS marker */
    }

SBANK = 0;                                              /* the ICS is always located in bank 0 */

if (CPX1 & cpx1_ICSFLAG) {                              /* if execution is currently on the ICS */
    if (CPX1 & cpx1_DISPFLAG) {                         /*   then if the dispatcher was interrupted */
        CPX1 &= ~cpx1_DISPFLAG;                         /*     then clear the dispatcher flag */

        cpu_read_memory (stack, Q, &delta_q);           /* get the delta Q value from the stack marker */
        cpu_write_memory (stack, Q, delta_q | STMK_D);  /*   and set the dispatcher-interrupted flag */
        }
    }

else {                                                  /* otherwise execution is on the user's stack */
    CPX1 |= cpx1_ICSFLAG;                               /*   so set the ICS flag */

    cpu_read_memory (stack, ICS_Q, &Q);                 /* set Q = QI */
    cpu_read_memory (stack, ICS_Z, &Z);                 /* set Z = ZI */

    cpu_read_memory (stack, Q - 4 & LA_MASK,            /* read the stack DB value */
                     &stack_db);

    cpu_write_memory (stack, Q - 6 & LA_MASK,           /* write the stack-DB-relative S value */
                      SM + 2 - stack_db & DV_MASK);     /*   which is meaningless for a cold load or power on */

    SR = 0;                                             /* invalidate the stack registers for a cold load */

    if (sys_configuration & CPU_IS_IOP)                 /* if this is an IOP system */
        DL = D16_UMAX;                                  /*   then set the data limit to -1 */
    else                                                /* otherwise for IMB systems */
        DL = Q;                                         /*   set the data limit to QI */
    }

SM = Q + 2 & R_MASK;                                    /* set S above the stack marker */

return;
}


/* Set up a code segment.

   This routine is called to set up a code segment in preparation for calling or
   exiting a procedure located in a segment different from the currently
   executing segment.  On entry, "plab" is a program label descriptor that
   indicates the segment number containing the procedure.  On exit, the new
   status register value and the first word of the Code Segment Table entry are
   returned to the variables pointed to by "status" and "entry_0", respectively.

   The Code Segment Table is split into two parts: a base table ("CST"), and an
   extension table ("CSTX").  The base table holds entries for system segments
   and user SL segments.  The extension table holds entries for the code
   segments that make up the set of currently allocated programs.  The first
   step is determining in which table the target CST entry resides.

   The routine begins by isolating the segment number from the supplied label.
   If the CST expansion firmware is installed, then if the physical mapping bit
   is set in the label, then the entry is for a system segment and resides in
   the CST.  If the mapping bit is clear, then SYSGLOB location %223 is read;
   this contains the number of code segments in the currently executing program.
   If the segment number is beyond that count, then the entry is for a user SL
   segment and resides in the CST; otherwise, it's a program segment, and the
   entry resides in the CSTX.

   If the firmware is not installed, then the table to use is determined by the
   segment number, with segments 1-191 residing in the CST, and segments
   193-255 residing in the CSTX.  Segment numbers 0 and 192, corresponding to
   the first entries of the two tables, are reserved and cause a CST Violation
   trap if specified.

   Next, the table base address and then the table header are read.  If the
   header indicates a zero size, then a System Halt 5 occurs.

   If the firmware is installed, then if the physical mapping bit is set in the
   label or the segment number is a program segment, then the segment
   number is used as the table entry index.  Otherwise, the Logical Segment
   Transform Table (LSTT) is consulted.

   The LSTT is divided into three major sections: the header, the segment map,
   and the STT map.  The first two words are the header and contain the number
   of logical segments in the segment map and the total length of the table in
   words.

   The segment map consists of two-word entries, one for each logical program
   segment and user SL segment referenced by the program.  Each entry contains
   the physical segment number, which is an index into the CSTX (for program
   segments) or CST (for user SL segments), and the offset to the STT map for
   the segment.  STT maps are only used for user SL segments; the offset will be
   zero if the STT in the code segment is to be used.

   The STT map consists of one STT for each user SL segment.  The STT contains
   only external labels from the code segment's STT; the number of local labels
   field will always be zero.

   The LSTT bank and offset for the currently executing program are read from
   SYSGLOB locations %221 and %222.  If both are zero, the table is undefined,
   and a CST Violation trap occurs.  Otherwise, the LSTT header is read.  This
   gives the number of segment entries in the table.  If the target segment is
   not contained in the table, then an STT Violation trap occurs.  If the LSTT
   contains the target segment, the corresponding segment map entry is read to
   obtain the entry index.

   If the entry index is zero or greater than the number of entries in the
   CST(X), then a CST Violation trap occurs.  Otherwise, the offset to the
   four-word table entry is determined, and the first, third, and fourth words
   are read.  The "segment-is-referenced" bit in the first word is set and
   written back to the table entry.  The program bank, base, and limit registers
   are set up (the segment length stored in the table is number of quad-words,
   which must be multiplied by four to get the size in words).  The new status
   register value is set up and returned, along with the first word of the CST
   entry.


   Implementation notes:

    1. This routine implements the microcode SSEG subroutine.

    2. Passing D16_UMAX as a parameter to trap_CST_Violation ensures that the
       segment number > 1 check will pass and the trap handler will be invoked.

    3. The Series II microcode sets PBANK and PB unilaterally but sets PL only
       if the code segment is not absent.  An absent segment entry contains the
       disc address in words 3 and 4 instead of the bank address and base
       address, so PBANK and PB will contain invalid values in this case.  It is
       not clear why the microcode avoids setting PL; the microinstruction in
       question also sets Flag 2, so conditioning PL may be just a side effect.
       In any case, we duplicate the firmware behavior here.

    4. This routine is only used locally, but we leave it as a global entry to
       support future firmware extensions that may need to call it.
*/

void cpu_setup_code_segment (PLABEL plab, HP_WORD *status, HP_WORD *entry_0)
{
HP_WORD cst_base, cst_pointer, cst_size, cst_entry, cst_bank;
HP_WORD program_count, segment_number, entry_number;
HP_WORD lstt_bank, lstt_offset, lstt_base, lstt_count;

segment_number = LABEL_SEGN (plab.label);               /* isolate the segment number from the label */

if (sys_configuration & FW_CSTX)                        /* if CST Expansion firmware is installed */
    if (plab.label & LABEL_PHYSICAL)                    /*   then if the segment is physically mapped */
        cst_base = CSTB_POINTER;                        /*     then use the CST as the base */

    else {                                              /* otherwise it's logically mapped */
        cpu_read_memory (absolute, CODE_SEG_COUNT,      /*   so get the number of program code segments */
                         &program_count);

        if (segment_number > program_count)             /* if the segment is not a program segment */
            cst_base = CSTB_POINTER;                    /*   then use the CST as the base */
        else                                            /* otherwise */
            cst_base = CSTX_POINTER;                    /*   use the CST extension as the base */
        }

else if (segment_number < CST_RESERVED) {               /* otherwise if the target segment is in the base table */
    cst_base = CSTB_POINTER;                            /*   then use the CST as the base */
    entry_number = segment_number;                      /*     and set the entry number */
    }

else {                                                  /* otherwise it is in the extension table */
    cst_base = CSTX_POINTER;                            /*   so use the CST extension as the base */
    entry_number = segment_number - CST_RESERVED;       /*     and set the adjusted entry number */
    }

cpu_read_memory (absolute, cst_base, &cst_pointer);     /* read the CST base pointer */
cpu_read_memory (absolute, cst_pointer, &cst_size);     /*   and the table size */

if (cst_size == 0)                                      /* if the CST length is zero */
    MICRO_ABORT (trap_SysHalt_CST_Length);              /*   then a System Halt occurs */

if (sys_configuration & FW_CSTX)                        /* if CST Expansion firmware is installed */
    if (plab.label & LABEL_PHYSICAL                     /*   then if the segment is physically mapped */
      || segment_number <= program_count)               /*   or the segment is a program segment */
        entry_number = segment_number;                  /*     then the segment number is the entry number */

    else {                                                      /* otherwise it's logically mapped */
        cpu_read_memory (absolute, LSTT_BANK, &lstt_bank);      /*   so read the LSTT bank */
        cpu_read_memory (absolute, LSTT_OFFSET, &lstt_offset);  /*     and offset */

        lstt_base = TO_PA (lstt_bank, lstt_offset);     /* get the absolute address of the base of the LSTT */

        if (lstt_base == 0)                             /* if the LSTT is undefined */
            MICRO_ABORTP (trap_CST_Violation, STA);     /*   then trap for a CST violation */

        cpu_read_memory (absolute, lstt_base,           /* read the LSTT header */
                         &lstt_count);                  /*   to get the number of segments in the table */

        if (segment_number > lstt_count)                /* if the segment number exceeds the segment count */
            MICRO_ABORTP (trap_STT_Violation, STA);     /*   then trap for an STT violation */

        cpu_read_memory (absolute,                      /* read the entry number for the current segment */
                         lstt_base + segment_number * 2,
                         &entry_number);
        }

tprintf (cpu_dev, TRACE_OPND, BOV_FORMAT "  %slabel %06o resolves to CST%sentry %04o\n",
         0, cst_base, cst_size,
         (sys_configuration & FW_CSTX ? (plab.label & LABEL_PHYSICAL ? "physical " : "logical ") : ""),
         plab.label,
         (cst_base ? "X " : " "), entry_number);

if (entry_number == 0)                                  /* segment number 0 does not exist */
    MICRO_ABORTP (trap_CST_Violation, D16_UMAX);        /*   so trap for a violation if it is specified */

else if (entry_number > cst_size)                       /* otherwise if the entry is outside of the table */
    MICRO_ABORTP (trap_CST_Violation, entry_number);    /*   then trap for a violation */

cst_entry = cst_pointer + entry_number * 4;             /* get the address of the target CST entry */

cpu_read_memory  (absolute, cst_entry, entry_0);                /* get the first word of the entry */
cpu_write_memory (absolute, cst_entry, *entry_0 | CST_R_BIT);   /*   and set the segment reference bit */

cpu_read_memory (absolute, cst_entry + 2, &cst_bank);   /* read the bank address word */
PBANK = cst_bank & cpu_bank_mask;                       /*   and mask to just the bank number */

cpu_read_memory (absolute, cst_entry + 3, &PB);         /* read the segment's base address */

PL = (*entry_0 & CST_SEGLEN_MASK) * 4 - 1;              /* set PL to the segment length - 1 */

*status = STA & ~STATUS_CS_MASK | segment_number;       /* set the segment number in the new status word */

if (*entry_0 & CST_M_BIT)                               /* if the segment executes in privileged mode */
    *status |= STATUS_M;                                /*   then set up to enter privileged mode */

if (not (*entry_0 & CST_A_BIT))                         /* if the segment is not absent */
    PL = PL + PB;                                       /*   then set the segment limit */

return;
}


/* Set up a data segment.

   This routine is called to set up a data segment for access.  It is called by
   the MDS, MFDS, and MTDS instruction executors to obtain the bank and offset
   of specified segments from the Data Segment Table.  On entry,
   "segment_number" indicates the number of the desired data segment.  On exit,
   the memory bank number and offset of the data segment base are returned to
   the variables pointed to by "bank" and "address", respectively.

   The routine begins by reading the DST pointer.  Segment number 0,
   corresponding to the first entry of the table, is reserved and causes a DST
   Violation trap if specified.

   The DST entry corresponding to the segment number is examined to obtain the
   bank and base address.  If the segment is absent, a Data Segment Absent trap
   is taken.  Otherwise, the bank and address values are returned.


   Implementation notes:

    1. This routine implements the microcode DSEG subroutine.
*/

void cpu_setup_data_segment (HP_WORD segment_number, HP_WORD *bank, HP_WORD *address)
{
HP_WORD dst_pointer, dst_size, dst_entry, entry_0;

cpu_read_memory (absolute, DST_POINTER, &dst_pointer);  /* read the DST base pointer */

if (segment_number == 0)                                /* segment number 0 does not exist */
    MICRO_ABORT (trap_DST_Violation);                   /*   so trap for a violation if it is specified */

cpu_read_memory (absolute, dst_pointer, &dst_size);     /* read the table size */

if (segment_number > dst_size)                          /* if the entry is outside of the table */
    MICRO_ABORT (trap_DST_Violation);                   /*   then trap for a violation */

dst_entry = dst_pointer + segment_number * 4;           /* get the address of the target DST entry */

cpu_read_memory (absolute, dst_entry, &entry_0);                /* get the first word of the entry */
cpu_write_memory (absolute, dst_entry, entry_0 | DST_R_BIT);    /*   and set the segment reference bit */

if (entry_0 & DST_A_BIT)                                /* if the segment is absent */
    MICRO_ABORTP (trap_DS_Absent, segment_number);      /*   then trap for an absentee violation */

cpu_read_memory (absolute, dst_entry + 2, bank);        /* read the segment bank number */
cpu_read_memory (absolute, dst_entry + 3, address);     /*   and base address */

*bank = *bank & cpu_bank_mask;                          /* mask off the reserved bits */

return;
}


/* Read a label from the STT.

   This routine is called to resolve an STT index into a program label.  On
   entry "stt_index" contains the (positive) index into the current code
   segment's STT.  The routine returns a program label descriptor consisting of
   the label word and the label accessing mode (internal or external).

   The routine begins by reading the STT header, located at the program limit
   (PL).  If the supplied index is greater than the number of labels given in
   the header, then an STT Violation occurs.  Otherwise, the label at PL - N is
   read.

   If the CST expansion firmware is not installed, then the external bit is
   checked to determine whether the label is external or local.  If the firmware
   is present, then if the index is within the range of local labels, the mode
   is local; otherwise, it's external.  If it is external, and the segment
   number in the label is zero, then the mapping bit in the label is checked.
   If the mapping is physical, an STT Violation occurs, because segment 0 does
   not exist.

   If the mapping is logical, then the LSTT must be consulted to obtain the
   label containing the actual segment number.  First, the LSTT bank and offset
   for the currently executing program are read from SYSGLOB locations %221 and
   %222.  If both are zero, the table is undefined, and a CST Violation trap
   occurs.  Otherwise, the LSTT header is read.  This gives the number of
   segment entries in the table.  If the target segment is not contained in the
   table, then an STT Violation trap occurs; otherwise, the corresponding
   segment map entry is read to obtain the STT offset.  If that offset is zero,
   then an STT Violation occurs.  Otherwise, the STT entry containing the
   replacement label is read.

   Once the label is obtained, the descriptor containing it and its access mode
   is returned.


   Implementation notes:

    1. See the comments for the "cpu_setup_code_segment" routine for a
       description of the LSTT organization.
*/

PLABEL cpu_read_label (HP_WORD stt_index)
{
PLABEL  plab;
HP_WORD bank, offset, stt_header, lstt_header, lstt_offset;
uint32  lstt_base;

cpu_read_memory (program, PL, &stt_header);             /* read the STT length */

if (stt_index > STT_LABELS (stt_header))                /* if the STT index is not within the STT */
    MICRO_ABORTP (trap_STT_Violation, STA);             /*   then trap for an STT violation */

cpu_read_memory (program_checked,                           /* check the bounds */
                 PL - stt_index & LA_MASK, &plab.label);    /*   when reading the label */

if (sys_configuration & FW_CSTX)
    tprintf (cpu_dev, TRACE_OPND, BOV_FORMAT "  STT %03o of %03o is %s label %06o\n",
             PBANK, PL, stt_header,
             stt_index, STT_LABELS (stt_header),
             (stt_index > STT_LOCALS (stt_header) ? "external" : "local"),
             plab.label);
else
    tprintf (cpu_dev, TRACE_OPND, BOV_FORMAT "  STT %03o of %03o is %s label %06o%s\n",
             PBANK, PL, stt_header,
             stt_index, STT_LABELS (stt_header),
             (plab.label & LABEL_EXTERNAL ? "external" : "local"),
             plab.label,
             (not (plab.label & LABEL_EXTERNAL) && plab.label & LABEL_UNCALLABLE ? " uncallable" : ""));

if (not (sys_configuration & FW_CSTX))                  /* if the CST expansion firmware is absent */
    if (plab.label & LABEL_EXTERNAL)                    /*   then if the external bit is set */
        plab.mode = External;                           /*     then the label is external */
    else                                                /*   otherwise */
        plab.mode = Local;                              /*     it's a local label */

else if (stt_index <= STT_LOCALS (stt_header))          /* otherwise if the index is within the local range */
    plab.mode = Local;                                  /*   then the label is local */

else {                                                  /* otherwise */
    plab.mode = External;                               /*   it's an external label */

    if (LABEL_SEGN (plab.label) == 0)                   /* if the segment number is zero */
        if (plab.label & LABEL_PHYSICAL)                /*   then if the label is physically mapped */
            MICRO_ABORTP (trap_STT_Violation, STA);     /*     then trap for an STT violation */

        else {                                                  /* otherwise the label is logically mapped */
            cpu_read_memory (absolute, LSTT_BANK, &bank);       /*   so read the LSTT bank */
            cpu_read_memory (absolute, LSTT_OFFSET, &offset);   /*     and offset */

            lstt_base = TO_PA (bank, offset);           /* get the absolute address of the base of the LSTT */

            if (lstt_base == 0)                         /* if the LSTT is undefined */
                MICRO_ABORTP (trap_CST_Violation, STA); /*   then trap for a CST violation */

            cpu_read_memory (absolute, lstt_base,       /* read the LSTT header */
                             &lstt_header);             /*   to get the number of segments in the table */

            if (STATUS_CS (STA) > lstt_header)          /* if the current segment exceeds the segment count */
                MICRO_ABORTP (trap_STT_Violation, STA); /*   then trap for an STT violation */

            cpu_read_memory (absolute,                  /* read the LSTT offset for the current segment */
                             lstt_base + STATUS_CS (STA) * 2 + 1,
                             &lstt_offset);

            if (lstt_offset == 0)                       /* if there is no replacement STT */
                MICRO_ABORTP (trap_STT_Violation, STA); /*   then trap for an STT violation */

            cpu_read_memory (absolute,                  /* read the replacement STT label */
                             lstt_base + lstt_offset
                               - (stt_index - STT_LOCALS (stt_header)),
                             &plab.label);

            tprintf (cpu_dev, TRACE_OPND, BOV_FORMAT "  STT %03o resolves to external %s label %06o\n",
                     bank, offset, lstt_offset, stt_index,
                     (plab.label & LABEL_PHYSICAL ? "physical" : "logical"),
                     plab.label);
            }
    }

return plab;                                            /* return the label descriptor */
}


/* Call a procedure.

   This routine sets up the PB, P, PL, and status registers (and code segment
   mapping register, if the CST extension firmware is installed) to enter a
   procedure.  It is called by the PCAL instruction executor, by the interrupt
   and trap routines, and by the XBR, PARC, and ENDP instructions of the
   COBOL-II Instruction Set to set up the target procedure.  On entry, "plab" is
   a program label descriptor containing an external program label indicating
   the segment number and Segment Transfer Table entry number describing the
   procedure, or a local program label indicating the starting address of the
   procedure.  The "offset" parameter contains an offset to be added to the
   starting address obtained from the label.  On exit, the registers are set up
   for execution to resume with the first instruction of the procedure.

   If the label is a local label, the PB-relative address is obtained from the
   label and stored in the P register, and the Next Instruction Register is
   loaded with the first instruction of the procedure.  The Status Register and
   the Code Segment Mapping Register do not change when a local procedure is
   called.

   If the label is external, the new mapping mode is obtained from the label,
   and the code segment referenced by the label is set up.  If the "trace" or
   "absent" bits are set, the stack is marked, and the corresponding trap is
   taken.  Otherwise, the label in the new segment corresponding to the STT
   index number present in the supplied label is read.

   If the target procedure is not in the designated code segment or is
   uncallable from user mode, the appropriate traps are taken.  If the STT entry
   contains a local label, it is used to set up the P register and NIR as above.
   If the supplied label references STT 0, then the offset applies to the start
   of the segment, i.e., PB, rather than to the start of a given procedure.

   When referencing an external procedure, the XBR, PARC, and ENDP instructions
   from the COBOL-II Instruction Set specify STT 0 in the external label and use
   the "offset" parameter to specify the procedure starting address.  These
   instructions cause a Privilege Violation trap if they call into a privileged
   segment from user mode.

   When the CST Expansion firmware is not installed, or when STT > 0 is
   referenced, PCAL instructions check the uncallable bit in the referenced STT
   entry (including the STT header for STT 0) and cause an STT Uncallable trap
   if privilege is violated.  When the firmware is installed, the STT header no
   longer has an uncallable bit, so a user-mode PCAL to STT 0 causes an
   unconditional STT Uncallable trap.


   Implementation notes:

    1. This routine implements the microcode PCL3 and PCL5 subroutines.
*/

void cpu_call_procedure (PLABEL plab, HP_WORD offset)
{
PLABEL  new_plab;
HP_WORD new_status, new_p, new_csmap, cst_entry, stt_index;

if (plab.mode == Local) {                               /* if the label is local */
    new_status = STA;                                   /*   then the status register */
    new_csmap = CSMAP;                                  /*     and code segment mapping remain unchanged */
    }

else {                                                  /* otherwise the label is external */
    new_csmap = TO_STMK_MAP (plab.label);               /*   so get the new mapping */

    cpu_setup_code_segment (plab, &new_status, &cst_entry); /* set up the new code segment */

    stt_index = LABEL_STTN (plab.label);                /* get the STT entry number from the label */

    if (cst_entry & (CST_A_BIT | CST_T_BIT)) {          /* if the code segment is absent or being traced */
        STA = new_status;                               /*   then set the new status before trapping */
        CSMAP = TO_STMK_MAP (plab.label);               /*     and the new mapping mode */

        cpu_mark_stack ();                              /* write a stack marker to memory */

        if (cst_entry & CST_A_BIT)                      /* if the code segment is absent */
            MICRO_ABORTP (trap_CS_Absent, plab.label);  /*   then trap to load it */
        else                                            /* otherwise */
            MICRO_ABORTP (trap_Trace, plab.label);      /*   trap to trace it */
        }

    new_plab = cpu_read_label (stt_index);              /* read the target STT entry */

    if (new_plab.mode == External)                      /* if the procedure is not in the target segment */
        MICRO_ABORTP (trap_STT_Violation, new_status);  /*   then trap for a violation */

    if (NPRV)                                                   /* if running in user mode */
        if ((CIR & CIS_XBR_MASK) == CIS_XBR_GROUP) {            /*   then if the instruction is a CIS external branch */
            if (cst_entry & CST_M_BIT)                          /*     then if the segment is privileged */
                MICRO_ABORT (trap_Privilege_Violation);         /*       then abort with a privilege violation */
            }

        else if (stt_index == 0 && sys_configuration & FW_CSTX  /* otherwise if STT is 0 with CST expansion */
          || new_plab.label & LABEL_UNCALLABLE)                 /*   or the procedure is uncallable */
            MICRO_ABORTP (trap_Uncallable, plab.label);         /*     then trap for an uncallable violation */

    if (stt_index == 0)                                 /* if the STT number is zero in an external label */
        plab.label = 0;                                 /*   then the starting address is PB */
    else                                                /* otherwise */
        plab.label = new_plab.label;                    /*   the PB offset is contained in the new label */

    cpu_base_changed = TRUE;                            /* the program base registers have changed for tracing */
    }

new_p = PB + LABEL_OFFSET (plab.label)                  /* get the procedure starting address */
          + STMK_OFFSET (offset) & LA_MASK;

cpu_read_memory (fetch_checked, new_p, &NIR);           /* check the bounds and get the next instruction */
P = new_p + 1 & R_MASK;                                 /* the bounds are valid, so set the new P value */

STA = new_status;                                       /* set the new status value */
CSMAP = new_csmap;                                      /*   and the new mapping mode */

cpu_base_changed = TRUE;                                /* one or more base registers have changed */

return;
}



/**********************************
 *                                *
 * CPU local SCP support routines *
 *                                *
 **********************************/


/* Service the CPU process and system clocks.

   The process clock is used by the operating system to time per-process CPU
   usage.  It is always enabled and running, although the PCLK register only
   increments if the CPU is not executing on the ICS.

   IMB systems use an internal clock counter instead of a hardware clock for
   MPE's time-of-day clock.  This counter is updated once per millisecond and is
   set to the negative number of ticks desired per interrupt.  when the counter
   overflows, an interrupt is generated if interrupts are enabled and the timer
   is on.  If either condition is not met, the interrupt is deferred, and a
   total count of the "missing" ticks us updated.

   The process clock may be calibrated to wall-clock time or set to real time.
   In hardware, the process clock has a one-millisecond period.  Setting the
   mode to real time schedules clock events based on the number of event ticks
   equivalent to one millisecond.  Because the simulator is an order of
   magnitude faster than the hardware, this short period precludes idling.

   In the calibrated mode, the short period would still preclude idling.
   Therefore, in this mode, the clock is scheduled with a ten-millisecond
   service time, and the PCLK register is incremented by ten for each event
   service.  To present the correct value when PCLK is read, the
   "cpu_update_pclk" routine is called by the RCLK instruction executor to
   increment the count by an amount proportional to the fraction of the service
   interval that has elapsed.  In addition, that routine is called by the CPU
   instruction postlude, so that PCLK will have the correct value if it is
   examined from the SCP command prompt.


   Implementation notes:

    1. If the process clock is calibrated, it is synchronized with the keyboard
       poll timer to permit idling.

    2. The current CPU speed, expressed as a multiple of the original hardware
       speed, is calculated for each service entry.  It may be displayed at the
       SCP prompt with the SHOW CPU SPEED command.  The speed is only
       representative when the process clock is calibrated, and the CPU is not
       executing a PAUS instruction (which suspends the normal fetch/execute
       instruction cycle).

    3. The Series 58 executes instructions approximately 2.83 times faster than
       the Series III, equivalent to an average instruction execution time of
       0.91 microseconds (source: "Relative Performance Matrix for Various
       HP3000 Systems" by AICS Research).

    4. The IMB clock counter register, CR, is typically initialized to -100,
       representing an interrupt interval of 100 milliseconds.  With the PCLK
       interval set to 10 milliseconds, the addition at each event service
       should cause the CR to overflow to zero.  However, when resetting the CR,
       we add any overflow amount back in, so that the next interval is
       shortened by that amount to correct for the overage.
*/

static t_stat cpu_service (UNIT *uptr)
{
static uint32 last_cpu_speed = 0;               /* last CPU execution rate multiple */
const HP_WORD ics_exec = (CPX1 & cpx1_ICSFLAG); /* non-zero if the CPU is executing on the ICS */
HP_WORD       limit, total;

tprintf (cpu_dev, TRACE_PSERV, "Process clock delay %d service entered on the %s\n",
         uptr->wait, (ics_exec ? "ICS" : "user stack"));

if (not ics_exec)                                       /* if the CPU is not executing on the ICS */
    PCLK = PCLK + pclk_increment & R_MASK;              /*   then increment the process clock */

if (sys_configuration & CPU_IS_IMB) {                   /* if this is an IMB system */
    CR = CR + pclk_increment & R_MASK;                  /*   then increment the clock counter register */

    if (CR < D16_SIGN) {                                    /* if the counter overflowed and is now positive */
        cpu_read_memory (absolute, LIMIT_REGISTER, &limit); /*   then get the limit register value */
        CR = CR - limit & R_MASK;                           /*     and reset the counter register */

        SIR |= sir_CPUTIMER;                            /* set the timer service bit */

        if (not (cpu_sir_irq_set & sir_CPUTIMER)) {             /* if the timer interrupt is disabled */
            cpu_read_memory (absolute, TOTAL_REGISTER, &total); /*   then get the count of missed ticks */
            total = total + limit & R_MASK;                     /*     and update it */
            cpu_write_memory (absolute, TOTAL_REGISTER, total); /*       and save it for the next interrupt */
            }
        }
    }

if (uptr->flags & UNIT_CALTIME) {                       /* if the process clock is tracking wall-clock time */
    uptr->wait = hp_sync_poll (Service);                /*   then maintain poll unit synchronization */
    pclk_increment = PCLK_MULTIPLIER;                   /*     and set the increment to the multiplier */
    }

else {                                                  /* otherwise */
    uptr->wait = PCLK_PERIOD;                           /*   set the delay as an event tick count */
    pclk_increment = 1;                                 /*     and set the increment without multiplying */
    }

sim_activate (uptr, uptr->wait);                        /* reschedule the timer */

if (sys_configuration & CPU_IS_5X                               /* if this is a Series 52 or 58 */
  && uptr->flags & UNIT_CALTIME)                                /*   and the process clock is calibrated */
    cpu_speed = uptr->wait / (PCLK_5X_PERIOD * pclk_increment); /*     then calculate the Series 5x CPU speed */
else                                                            /* otherwise */
    cpu_speed = uptr->wait / (PCLK_PERIOD * pclk_increment);    /*   calculate the Series III CPU speed */

if (TRACING (cpu_dev, TRACE_PSERV))
    if (last_cpu_speed != cpu_speed) {
        hp_trace (&cpu_dev, TRACE_PSERV, "Simulation rate %ux\n", cpu_speed);
        last_cpu_speed = cpu_speed;
        }

return SCPE_OK;                                         /* return the success of the service */
}


/* Reset the CPU.

   This routine is called for a RESET, RESET CPU, or BOOT CPU command.  It is the
   simulation equivalent of the CPURESET signal, which is asserted by the front
   panel LOAD switch.  In hardware, this causes a microcode restart in addition
   to clearing certain registers.

   If this is the first call after simulator startup, the initial memory array
   is allocated, the default CPU and memory size configuration is set, and the
   SCP-required program counter pointer is set to point to the REG array element
   corresponding to the P register.

   If this is a power-on reset ("RESET -P"), any LOAD or DUMP request in
   progress is cleared.

   The micromachine is halted, the process clock is scheduled, and several
   registers are cleared.


   Implementation notes:

    1. The CPU unit pointer cannot be a "const" pointer because the parameter to
       "sim_activate" is not declared constant.

    2. Setting the sim_PC value at run time accommodates changes in the register
       order automatically.  A fixed setting runs the risk of it not being
       updated if a change in the register order is made.

    3. The Status and Interrupt Register (SIR) of the IMB machines is not
       cleared in hardware by a reset.  Only the four LSBs are registered, and
       the device used does not have a reset input.
*/

static t_stat cpu_reset (DEVICE *dptr)
{
UNIT *uptr = dptr->units;                       /* a pointer to the CPU process clock unit */

if (sim_PC == NULL) {                                   /* if this is the first call after simulator start */
    if (mem_initialize (PA_MAX)) {                      /*   then if memory allocation succeeds */
        set_model (uptr, CPU_III, NULL, NULL);          /*     then establish the default CPU model */

        for (sim_PC = dptr->registers;                  /* find the P register entry */
             sim_PC->loc != &P && sim_PC->loc != NULL;  /*   in the register array */
             sim_PC++);                                 /*     for the SCP interface */

        if (sim_PC == NULL)                             /* if the P register entry is not present */
            return SCPE_NXREG;                          /*   then there is a serious problem! */
        }

    else                                                /* otherwise memory initialization failed */
        return SCPE_MEM;                                /*   so report the error and abort the simulation */
    }

if (sim_switches & SWMASK ('P'))                        /* if this is a power-on reset */
    CPX2 &= ~(cpx2_LOADSWCH | cpx2_DUMPSWCH);           /*   then clear any cold load or dump request */

cpu_micro_state = Halted;                               /* halt the micromachine */

if (uptr->flags & UNIT_CALTIME) {                       /* if the process clock is tracking wall-clock time */
    uptr->wait = hp_sync_poll (Initial);                /*   then synchronize with the poll timer */
    pclk_increment = PCLK_MULTIPLIER;                   /*     and set the increment to the multiplier */
    }

else {                                                  /* otherwise */
    uptr->wait = PCLK_PERIOD;                           /*   set the delay as an event tick count */
    pclk_increment = 1;                                 /*     and set the increment without multiplying */
    }

sim_activate_abs (uptr, uptr->wait);                    /* schedule the process clock */

PCLK = 0;                                               /* clear the process clock counter */
SP0  = 0;                                               /*   and scratch pad 0 (system halt code) */

CPX1 = 0;                                               /* clear all run-mode signals */
CPX2 &= ~(cpx2_RUN | cpx2_SYSHALT);                     /*   and the run and system halt flip-flops */

SIR &= ~SIR_IRQ_SET;                                    /* clear the run-time interrupts */
cpu_sir_irq_set = SIR_IRQ_SET & ~sir_CPUTIMER;          /*   and disable timer interrupts */

if (sys_configuration & SYS_ARS)                        /* if the PF/ARS switch position is ENBL */
    CPX2 &= ~cpx2_INHPFARS;                             /*   then clear the auto-restart inhibit flag */
else                                                    /* otherwise the position is DSBL */
    CPX2 |= cpx2_INHPFARS;                              /*   so set the auto-restart inhibit flag */

CNTR = SR;                                              /* copy the stack register to the counter */
cpu_flush ();                                           /*   and flush the TOS registers to memory */

hp_reset_poll ();                                       /* restart the keyboard poll */

return SCPE_OK;                                         /* indicate that the reset succeeded */
}


/* Set the CPU simulation stop conditions.

   This validation routine is called to configure the set of CPU stop
   conditions.  The "option" parameter is 0 to clear the stops and 1 to set
   them, and "cptr" points to the first character of the name of the stop to be
   cleared or set.  The unit and description pointers are not used.

   The routine processes commands of the form:

     SET CPU STOP
     SET CPU STOP=<stopname>[;<stopname>...]
     SET CPU NOSTOP
     SET CPU NOSTOP=<stopname>[;<stopname>...]

   The valid <stopname>s are contained in the debug table "cpu_stop".  If names
   are not specified, all stop conditions are enabled or disabled.


   Implementation notes:

    1. The CPU simulator maintains a private and a public set of simulator
       stops.  This routine sets the private set.  The private set is copied to
       the public set as part of the instruction execution prelude, unless the
       "-B" ("bypass") command-line switch is used with the run command.  This
       allows the stops to be bypassed conveniently for the first instruction
       execution only.
*/

static t_stat set_stops (UNIT *uptr, int32 option, char *cptr, void *desc)
{
char   gbuf [CBUFSIZE];
uint32 stop;

if (cptr == NULL) {                                     /* if there are no arguments */
    sim_stops = 0;                                      /*   then clear all of the stop flags */

    if (option == 1)                                            /* if we're setting the stops */
        for (stop = 0; cpu_stop [stop].name != NULL; stop++)    /*   then loop through the flags */
            sim_stops |= cpu_stop [stop].mask;                  /*     and add each one to the set */
    }

else if (*cptr == '\0')                                 /* otherwise if the argument is empty */
    return SCPE_MISVAL;                                 /*   then report the missing value */

else                                                    /* otherwise at least one argument is present */
    while (*cptr) {                                     /* loop through the arguments */
        cptr = get_glyph (cptr, gbuf, ';');             /* get the next argument */

        for (stop = 0; cpu_stop [stop].name != NULL; stop++)    /* loop through the flags */
            if (strcmp (cpu_stop [stop].name, gbuf) == 0) {     /*   and if the argument matches */
                if (option == 1)                                /*     then if it's a STOP argument */
                    sim_stops |= cpu_stop [stop].mask;          /*       then add the stop flag */
                else                                            /*     otherwise it's a NOSTOP argument */
                    sim_stops &= ~cpu_stop [stop].mask;         /*       so remove the flag */

                break;                                  /* this argument has been processed */
                }

        if (cpu_stop [stop].name == NULL)               /* if the argument was not found */
            return SCPE_ARG;                            /*   then report it */
        }

return SCPE_OK;                                         /* the stops were successfully processed */
}


/* Change the instruction execution trace criteria.

   This validation routine is called to configure the criteria that select
   instruction execution tracing.  The routine processes commands of the form:

     SET CPU EXEC=<match>[;<mask>]
     SET CPU NOEXEC

   If the <mask> value is not supplied, a mask of 177777 octal is used.  The
   values are entered in the current CPU data radix, which defaults to octal,
   unless an override switch is present on the command line.

   On entry, the "option" parameter is 0 to clear and 1 to set the criteria, and
   "cptr" points to the first character of the match value to be set.  The unit
   and description pointers are not used.
*/

static t_stat set_exec (UNIT *uptr, int32 option, char *cptr, void *desc)
{
char   gbuf [CBUFSIZE];
uint32 match, mask, radix;
t_stat status;

if (option == 0)                                        /* if this is a NOEXEC request */
    if (cptr == NULL) {                                 /*   then if there are no arguments */
        exec_match = D16_UMAX;                          /*     then set the match and mask values */
        exec_mask  = 0;                                 /*       to prevent matching */
        return SCPE_OK;                                 /*         and return success */
        }

    else                                                /*   otherwise there are extraneous characters */
        return SCPE_2MARG;                              /*     so report that there are too many arguments */

else if (cptr == NULL || *cptr == '\0')                 /* otherwise if the EXEC request supplies no arguments */
    return SCPE_MISVAL;                                 /*   then report a missing value */

else {                                                  /* otherwise at least one argument is present */
    cptr = get_glyph (cptr, gbuf, ';');                 /*   so get the match argument */

    if (sim_switches & SWMASK ('O'))                    /* if an octal override is present */
        radix = 8;                                      /*   then parse the value in base 8 */
    else if (sim_switches & SWMASK ('D'))               /* otherwise if a decimal override is present */
        radix = 10;                                     /*   then parse the value in base 10 */
    else if (sim_switches & SWMASK ('H'))               /* otherwise if a hex override is present */
        radix = 16;                                     /*   then parse the value in base 16 */
    else                                                /* otherwise */
        radix = cpu_dev.dradix;                         /*   use the current CPU data radix */

    match = (uint32) get_uint (gbuf, radix, D16_UMAX, &status); /* parse the match value */

    if (status != SCPE_OK)                              /* if a parsing error occurred */
        return status;                                  /*   then return the error status */

    else if (*cptr == '\0') {                           /* otherwise if no more characters are present */
        exec_match = match;                             /*   then set the match value */
        exec_mask  = D16_MASK;                          /*     and default the mask value */
        return SCPE_OK;                                 /*       and return success */
        }

    else {                                              /* otherwise another argument is present */
        cptr = get_glyph (cptr, gbuf, ';');             /*   so get the mask argument */

        mask = (uint32) get_uint (gbuf, radix, D16_UMAX, &status);  /* parse the mask value */

        if (status != SCPE_OK)                          /* if a parsing error occurred */
            return status;                              /*   then return the error status */

        else if (*cptr == '\0')                         /* if no more characters are present */
            if (mask == 0)                              /*   then if the mask value is zero */
                return SCPE_ARG;                        /*     then the match will never succeed */

            else {                                      /*   otherwise */
                exec_match = match;                     /*     set the match value */
                exec_mask  = mask;                      /*       and the mask value */
                return SCPE_OK;                         /*         and return success */
                }

        else                                            /* otherwise extraneous characters are present */
            return SCPE_2MARG;                          /*   so report that there are too many arguments */
        }
    }
}


/* Set the CPU cold dump configuration jumpers.

   This validation routine is called to configure the set of jumpers on the
   Series II/III system control panel that preset the device number and control
   value for the cold dump process.  The routine processes commands of the form:

     SET CPU DUMPDEV=<devno>
     SET CPU DUMPCTL=<cntlval>

   The device number is a decimal value between 3 and 127, and the control value
   is an octal number between 0 and 377.  Values outside of these ranges are
   rejected.  Parsing and validation are handled automatically by the expanded
   MTAB entries.

   The command is not allowed unless the CPU is a Series II or III.
*/

static t_stat set_dump (UNIT *uptr, int32 value, char *cptr, void *desc)
{
if (sys_configuration & CPU_IS_IMB)                     /* if this is an IMB system */
    return SCPE_NOFNC;                                  /*   then the jumpers do not exist */
else                                                    /* otherwise */
    return SCPE_OK;                                     /*   store the valid parsed value */
}


/* Set a CPU firmware option.

   This validation routine is called to enable a firmware option for the current
   CPU model.  The routine processes commands of the form:

     SET CPU <option>[,<option>...]

   The "option" parameter is set to the option to be added or removed and will
   be one of the system configuration option bits.  The other parameters are not
   used.

   The option must be a valid optional feature for the current CPU model, or the
   command is rejected.  If the option is valid, SCP will add or remove the
   selected option bit when this routine returns.
*/

static t_stat set_firmware (UNIT *uptr, int32 option, char *cptr, void *desc)
{
const uint32 model = CPU_MODEL (sys_configuration);     /* the current CPU model index */

if (cpu_features [model].opt & option)                  /* if the feature is optional */
    return SCPE_OK;                                     /*   then authorize the change */
else                                                    /* otherwise */
    return SCPE_NOFNC;                                  /*   reject the change */
}


/* Set a CPU auto-restart option.

   This validation routine is called to enable or disable the power-fail/
   auto-restart option for the current CPU model.  The routine processes
   commands of the form:

     SET CPU { ARS | NOARS }

   The "option" parameter is set to SYS_ARS to enable the option or to zero to
   disable the option.  The other parameters are not used.

   The auto-restart option must be a valid optional feature for the current CPU
   model, or the command is rejected.

   If the PF/ARS switch, located behind the Series III system control panel, is
   being set to the ENBL (enable) position, the CPU will perform an auto-restart
   when power is restored after a failure.  If the switch is being set to the
   DSBL (disable) position, the CPU will remain halted after power restoration;
   execution may be continued by pressing the RUN button.  The ARS or NOARS
   setting is reflected in "inhibit auto-restart" bit of the CPX2 register.
*/

static t_stat set_ars (UNIT *uptr, int32 option, char *cptr, void *desc)
{
const uint32 model = CPU_MODEL (sys_configuration);     /* the current CPU model index */

if (not (cpu_features [model].opt & SYS_ARS))           /* if the ARS feature is required */
    return SCPE_NOFNC;                                  /*   then reject the change */

else if (option == SYS_ARS)                             /* otherwise if the option is auto-restart */
    CPX2 |= cpx2_INHPFARS;                              /*   then set the auto-restart inhibit flag */

else                                                    /* otherwise */
    CPX2 &= ~cpx2_INHPFARS;                             /*   clear the auto-restart inhibit flag */

return SCPE_OK;                                         /* the change is successful */
}


/* Set a bootstrap thumbwheel switch.

   This validation routine is called to configure one of the three thumbwheel
   switches on the Series 4x/5x system control panel that preset the channel and
   device numbers for the bootstrap process.

   The routine processes commands of the form:

     SET CPU START=<channel>[;<device>]
     SET CPU DUMP=<channel>[;<device>]
     SET CPU LOAD=<channel>[;<device>]

   The channel number is a decimal value between 4 and 11, and the optional
   device number is a decimal number between 0 and 7.  Values outside of these
   ranges are rejected.  If the device number is omitted, device number 0 is
   used.

   On entry, the "thumbwheel" parameter is "Cold_Start" to set the Start
   thumbwheel, "Cold_Dump" to set the Dump thumbwheel, or "Cold_Load" to set the
   Load thumbwheel.  The "cptr" parameter points to the first character of the
   value(s) to be set.  The unit and description pointers are not used.

   The command is not allowed unless the CPU is a Series 4x/5x.
*/

static t_stat set_thumbwheel (UNIT *uptr, int32 thumbwheel, char *cptr, void *desc)
{
char    gbuf [CBUFSIZE];
t_value channel, device;
t_stat  status;

if (sys_configuration & CPU_IS_IOP)                     /* if this is an IOP system */
    status = SCPE_NOFNC;                                /*   then the thumbwheels do not exist */

else if (cptr == NULL || *cptr == '\0')                 /* otherwise if the expected value is missing */
    status = SCPE_MISVAL;                               /*   then report the error */

else {                                                  /* otherwise */
    cptr = get_glyph (cptr, gbuf, ';');                 /*   get the first argument */

    channel = get_uint (gbuf, CHAN_BASE, CHAN_MAX, &status);    /*   parse the supplied value */

    if (status == SCPE_OK)                              /* if it is valid */
        if (channel < 4)                                /*   but out of the proper range */
            status = SCPE_ARG;                          /*     then the argument is invalid */

        else if (*cptr == '\0')                         /* otherwise if the device number is omitted */
            device = 0;                                 /*   then default to device 0 */

        else {                                          /* otherwise */
            cptr = get_glyph (cptr, gbuf, 0);           /*   get the second argument */

            device = get_uint (gbuf, BUS_BASE, BUS_MAX, &status);   /* parse the supplied value */

            if (status == SCPE_OK && *cptr != '\0')     /* if the value is valid but more characters follow */
                status = SCPE_ARG;                      /*   then report a bad argument */
            }

        else                                            /* otherwise */
            status = SCPE_ARG;                          /*   report a bad argument */

    if (status == SCPE_OK)                              /* if the value(s) are good */
        switch ((PANEL_SWITCH) thumbwheel) {            /*   then dispatch the thumbwheel switch */

            case Cold_Start:
                start_device = TO_DEVNO (channel, device);  /* change the Start thumbwheel setting */
                break;

            case Cold_Dump:
                dump_device = TO_DEVNO (channel, device);   /* change the Dump thumbwheel setting */
                break;

            case Cold_Load:
                load_device = TO_DEVNO (channel, device);   /* change the Load thumbwheel setting */
                break;

            default:                                    /* any other operation */
                status = SCPE_IERR;                     /*   should never occur */
            }
    }

return status;                                          /* return the operation status */
}


/* Change the CPU memory size.

   This validation routine is called to configure the CPU memory size.  The
   routine processes commands of the form:

     SET [-F] CPU <memsize>

   If the new memory size is smaller or larger than the supported sizes for the
   CPU model currently selected, the routine returns an error.  If the new size
   is smaller than the previous size, and if the area that would be lost
   contains non-zero data, the user is prompted to confirm that memory should be
   truncated.  If the user denies the request, the change is rejected.
   Otherwise, the new size is set.  The user may omit the confirmation request
   and force truncation by specifying the "-F" switch on the command line.

   On entry, the "new_size" parameter is set to the size desired and will be one
   of the discrete sizes supported by the machine.  If "desc" is not NULL, then
   it points to an integer containing the CPU model bit.  Otherwise, the model
   is derived from the system configuration word.  The "uptr" and "cptr"
   parameters are not used.


   Implementation notes:

    1. The memory access routines return a zero value for locations beyond the
       currently defined memory size.  Therefore, the unused area need not be
       explicitly zeroed.

    2. The CPU unit must be referenced explicitly because the "uptr" parameter
       may be NULL (actually, it will be NULL if the routine is called by the
       expanded MTAB handler).
*/

static t_stat set_size (UNIT *uptr, int32 new_size, char *cptr, void *desc)
{
static char confirm [] = "Really truncate memory [N]?";
uint32 model;

if (desc == NULL)                                       /* if the descriptor is NULL */
    model = CPU_MODEL (sys_configuration);              /*   then get the current CPU model index */
else                                                    /* otherwise */
    model = CPU_MODEL (*(uint32 *) desc);               /*   get the CPU model index to use */

if ((uint32) new_size < cpu_features [model].minmem     /* if the new memory size */
  || (uint32) new_size > cpu_features [model].maxmem)   /*   is not supported on current model */
    return SCPE_NOFNC;                                  /*     then report the error */

if (not (sim_switches & SWMASK ('F'))                   /* if truncation is not explicitly forced */
  && not mem_is_empty (new_size)                        /*   and the truncated part is not empty */
  && get_yn (confirm, FALSE) == FALSE)                  /*     and the user denies confirmation */
    return SCPE_INCOMP;                                 /*       then abort the command */

cpu_unit [0].capac = mem_size = new_size;               /* set the new memory size */

return SCPE_OK;                                         /* confirm that the change is OK */
}


/* Change the CPU model.

   This validation routine is called to configure the CPU model.  The routine
   processes commands of the form:

     SET [-F] CPU <model>

   Setting the model establishes the set of standard hardware features.  It also
   verifies that the current memory size is supported by the new model.  If it
   is not, the size is increased to the minimum or reduced to the maximum
   supported memory configuration.  If the area that would be lost by reduction
   contains non-zero data, the user is prompted to confirm that memory should be
   truncated.  If the user denies the request, the change is rejected.
   Otherwise, the new size is set.  The user may omit the confirmation request
   and force truncation by specifying the "-F" switch on the command line.

   The "new_model" parameter is set to the model desired and will be one of the
   system configuration CPU model bits.  The other parameters are not used.

   This routine is also called once from the CPU reset routine to establish the
   initial CPU model.  The current memory size will be 0 when this call is made.
   In this case, the memory size is set to the maximum available for that model.
*/

static t_stat set_model (UNIT *uptr, int32 new_model, char *cptr, void *desc)
{
const uint32 new_index = CPU_MODEL (new_model); /* the new index into the CPU features table */
uint32       new_memsize, index;
t_stat       status;

if (mem_size == 0                                       /* if this is the initial establishing call */
  || mem_size > cpu_features [new_index].maxmem)        /*   or if the current memory size is unsupported */
    new_memsize = cpu_features [new_index].maxmem;      /*     then set the new size to the maximum supported size */

else if (mem_size < cpu_features [new_index].minmem)    /* otherwise if the current memory size is too small */
    new_memsize = cpu_features [new_index].minmem;      /*   then set the new size to the minimum supported size */

else                                                    /* otherwise the current size is valid for the new model */
    new_memsize = mem_size;                             /*   so leave it unchanged */

status = set_size (uptr, new_memsize, NULL, &new_model);    /* set the new memory size */

if (status == SCPE_OK) {                                    /* if the change succeeded */
    sys_configuration = new_model                           /*   then set the new CPU model bit */
                          | cpu_features [new_index].std;   /*     and the standard features bits */

    hp_configure_devices (sys_configuration, Set_Devices);  /* enable compatible devices and disable others */

    cpu_bank_mask =                                         /* set the new CPU bank register */
      mem_bank_mask = cpu_features [new_index].bank_mask;   /*   and memory bank register masks */

    for (index = 0; cpu_reg [index].loc != NULL; index++)   /* scan the CPU register list */
        if (cpu_reg [index].loc == &PBANK                   /*   and set the maximum allowed value */
          || cpu_reg [index].loc == &DBANK                  /*     for the three bank registers */
          || cpu_reg [index].loc == &SBANK)                 /*       to the highest bank number */
            cpu_reg [index].maxval = mem_bank_mask;         /*         for the new CPU model */
    }

return status;                                          /* return the validation result */
}


/* Show the CPU simulation stop conditions.

   This display routine is called to show the set of CPU stop conditions.  The
   "st" parameter is the open output stream.  The other parameters are not used.

   If at least one stop condition is enabled, the routine searches through the
   stop table for flag bits that are set in the stop set.  For each one it
   finds, the routine prints the corresponding stop name.

   This routine services an extended modifier entry, so it must add the trailing
   newline to the output before returning.
*/

static t_stat show_stops (FILE *st, UNIT *uptr, int32 value, void *desc)
{
uint32 stop;
t_bool need_spacer = FALSE;

if (sim_stops == 0)                                     /* if no simulation stops are set */
    fputs ("Stops disabled", st);                       /*   then report that all are disabled */

else {                                                  /* otherwise at least one stop is valid */
    fputs ("Stop=", st);                                /*   so prepare to report the list of conditions */

    for (stop = 0; cpu_stop [stop].name != NULL; stop++)    /* loop through the set of stops in the table */
        if (cpu_stop [stop].mask & sim_stops) {             /* if the current stop is enabled */
            if (need_spacer)                                /*   then if a spacer is needed */
                fputc (';', st);                            /*     then add it first */

            fputs (cpu_stop [stop].name, st);           /* report the stop name */

            need_spacer = TRUE;                         /* a spacer will be needed next time */
            }
        }

fputc ('\n', st);                                       /* add the trailing newline */

return SCPE_OK;                                         /* report the success of the display */
}


/* Show the instruction execution trace criteria.

   This display routine is called to show the criteria that select instruction
   execution tracing.  The "st" parameter is the open output stream.  The other
   parameters are not used.

   This routine services an extended modifier entry, so it must add the trailing
   newline to the output before returning.
*/

static t_stat show_exec (FILE *st, UNIT *uptr, int32 value, void *desc)
{
uint32 radix;

if (exec_mask == 0)                                     /* if the instruction is entirely masked */
    fputs ("Execution trace disabled\n", st);           /*   then report that matching is disabled */

else {                                                  /* otherwise */
    if (sim_switches & SWMASK ('O'))                    /*   if an octal override is present */
        radix = 8;                                      /*     then print the value in base 8 */
    else if (sim_switches & SWMASK ('D'))               /*   otherwise if a decimal override is present */
        radix = 10;                                     /*     then print the value in base 10 */
    else if (sim_switches & SWMASK ('H'))               /*   otherwise if a hex override is present */
        radix = 16;                                     /*     then print the value in base 16 */
    else                                                /*   otherwise */
        radix = cpu_dev.dradix;                         /*     use the current CPU data radix */

    fputs ("Execution trace match = ", st);                         /* print the label */
    fprint_val (st, exec_match, radix, cpu_dev.dwidth, PV_RZRO);    /*   and the match value */

    fputs (", mask = ", st);                                        /* print a separator */
    fprint_val (st, exec_mask, radix, cpu_dev.dwidth, PV_RZRO);     /*   and the mask value */

    fputc ('\n', st);                                   /* tie off the line */
    }

return SCPE_OK;                                         /* report the success of the display */
}


/* Show the CPU cold dump configuration jumpers.

   This display routine is called to show the device number and control byte
   that are preset on the rear of the Series II/III system control panel for the
   cold dump process.  The "st" parameter is the open output stream.  The other
   parameters are not used.

   The command is not allowed unless the CPU is a Series II or III.


   Implementation notes:

    1. The SHOW command ignores the return status from regular and extended
       display routines, so in order to report the refusal to display for
       non-IMB systems, we must display the error message here.
*/

static t_stat show_dump (FILE *st, UNIT *uptr, int32 value, void *desc)
{
if (sys_configuration & CPU_IS_IMB)                     /* if this is an IMB system */
    fprintf (st, "%s\n", sim_error_text (SCPE_NOFNC));  /*   then the jumpers do not exist */

else
    fprintf (st, "Dump device = %u, dump control = %03o\n",
                  LOWER_BYTE (dump_control), UPPER_BYTE (dump_control));

return SCPE_OK;
}


/* Show the front panel thumbwheel settings.

   This display routine is called to show the current Series 4x/5x system front
   panel settings.  The "st" parameter is the open output stream.  The other
   parameters are not used.

   The command is not allowed unless the CPU is a Series 4x/5x.


   Implementation notes:

    1. The SHOW command ignores the return status from regular and extended
       display routines, so in order to report the refusal to display for
       non-IMB systems, we must display the error message here.
*/

static t_stat show_panel (FILE *st, UNIT *uptr, int32 value, void *desc)
{
if (sys_configuration & CPU_IS_IOP)                     /* if this is an IOP system */
    fprintf (st, "%s\n", sim_error_text (SCPE_NOFNC));  /*   then the switches do not exist */

else {                                                  /* otherwise this is an IMB system */
    update_SIR ();                                      /*   so update the Status and Interrupt Register */

    fprintf (st, "%s, %s, Start CHL/DEV = %u/%u, Dump CHL/DEV = %u/%u, Load CHL/DEV = %u/%u\n",
                 (SIR & sir_CPUOK  ? "Power on" : "Battery"),
                 (SIR & sir_CPURUN ? "Run"      : "Halt"),
                 IO_CHANNEL (start_device), IO_DEVICE (start_device),
                 IO_CHANNEL (dump_device),  IO_DEVICE (dump_device),
                 IO_CHANNEL (load_device),  IO_DEVICE (load_device));
    }

return SCPE_OK;
}


/* Show the current CPU simulation speed.

   This display routine is called to show the current simulation speed.  The
   "st" parameter is the open output stream.  The other parameters are not used.

   The CPU speed, expressed as a multiple of the Series III hardware speed, is
   calculated by the process clock service routine.  It is only representative
   when the process clock is calibrated, and the CPU is not executing a PAUS
   instruction (which suspends the normal fetch/execute instruction cycle).
*/

static t_stat show_speed (FILE *st, UNIT *uptr, int32 value, void *desc)
{
fprintf (st, "Simulation speed = %ux\n", cpu_speed);    /* display the current CPU speed */
return SCPE_OK;                                         /*   and report success */
}



/******************************
 *                            *
 * CPU local utility routines *
 *                            *
 ******************************/


/* Check for device conflicts.

   This routine is called from the instruction execution prelude to verify that
   there are no conflicts in device configuration.  All enabled IOP devices are
   checked for conflicting device number, interrupt priority, or service request
   priority assignments, and all enabled IMB devices are checked for conflicting
   channel number assignments.  The routine returns TRUE to inhibit execution if
   any conflicts exist and FALSE to allow execution to proceed.

   Each channel number, device number, interrupt priority number, and service
   request number must be unique among the enabled devices.  These requirements
   are checked as part of the instruction execution prelude; this allows the
   user to exchange two device numbers (e.g.) simply by setting each device to
   the other's device number.  If conflicts were enforced instead at the time
   the numbers were entered, the first device would have to be set to an unused
   number before the second could be set to the first device's number.

   The routine begins by scanning the device information block (DIB) of each
   enabled device to record the device pointer, channel number, interrupt
   priority number, service request number, and device number in a DIB
   assignments array.

   For each recorded DIB, the assignments are also recorded in bitmaps
   corresponding to each assignment type.  All of the assignments other than the
   device number can be recorded in single-word bitmaps.  The assigned device
   number, which can range from 3-127, is recorded in an eight-word bitmap by
   dividing the value by 32 and using the quotient to determine the bitmap word
   and the remainder to determine the bit within the word.  Conflicts are
   identified by setting the corresponding bit in a conflicts array whenever a
   given value has been assigned to a prior device.  A cumulative conflict
   indicator is also maintained to provide a simple check after all devices are
   processed.

   If the interface defines a run-time initializer routine, it is called.  If it
   returns FALSE, then it detected some problem that prevents execution.

   If there are no conflicts after scanning all devices, the routine returns
   FALSE to permit execution to being.

   If any conflicts exist, the conflict array is scanned to find the bits
   associated with the multiply-assigned values.  For each bit set in the array,
   the conflicting type and numerical value are determined.  Then a scan of the
   DIB array is made to find those devices having the conflicting assignments.
   For each conflicting value, the value and the device names to which the value
   is assigned are reported to the user.

   After all conflicts are reported, the routine returns TRUE to indicate that a
   conflict exists and execution must be inhibited.


   Implementation notes:

    1. When this routine is called, the console and optional log file have
       already been put into "raw" output mode.  Therefore, newlines are not
       translated to the correct line ends on systems that require it.  Before
       reporting a conflict, "sim_ttcmd" is called to restore the console and
       log file translation.  This is OK because a conflict will abort the run
       and return to the command line anyway.

    2. sim_dname is called instead of using dptr->name directly to ensure that
       we pick up any assigned logical device name.

    3. Because IOP and IMB devices are enabled only with their corresponding CPU
       models, conflicts that might exist for disabled devices are not reported.

    4. Three arrays are used to process the device assignments.  The DIB array
       is populated by DIBs belonging to enabled devices.  The assignments array
       contains bitmaps where each bit corresponds to a value assigned to some
       device.  The conflicts array contains bitmaps where each bit corresponds
       to a value that has been assigned to more than one device.

    5. For ease of processing, the bitmaps for all conflict types are kept in a
       single array.  The first three conflict types use single-word bitmaps
       (i.e., the range of values fits in 32 bits), and so the conflict type
       identifier can be used as an index into this three-element array.  The
       fourth type, device numbers, requires an eight-word bitmap to cover the
       range of values.  This is appended to the three-element array, producing
       an eleven-element array, of which elements three through ten form the
       eight-word device number bitmap.  This latter bitmap uses a sub-index
       (rank) that is added to the start of the sub-array to access the proper
       32-bit subset.
*/

static t_bool device_conflicts (void)
{
typedef enum {                                  /* conflict types */
    Channel,                                    /*   (IMB) channel number conflict */
    Interrupt,                                  /*   (IOP) interrupt priority conflict */
    Service,                                    /*   (IOP) service request number conflict */
    Device                                      /*   (IOP) device number conflict */
    } CONFLICT_TYPE;

typedef struct {                                /* table of DIB assignments */
    DEVICE  *dptr;                              /*   a pointer to the device */
    uint32  assignment [4];                     /*   assignments, indexed by CONFLICT_TYPE */
    } DIB_RECORD;

typedef uint32 BITMAP_ARRAY [3 + 8];            /* bitmaps for Channel (1), Interrupt (1), Service (1) and Device (8) */

static const char *conflict_label [] = {        /* conflict names, indexed by CONFLICT_TYPE */
    "Channel number",
    "Interrupt priority",
    "Service request number",
    "Device number"
    };

const size_t  conflict_size = sizeof (BITMAP_ARRAY) / sizeof (uint32);
DEVICE        *dptr;
DIB           *dibptr;
uint32        dib, dev, rank, bit, index, value, clash;
DIB_RECORD    dibs [SIM_DEVICE_SIZE];
BITMAP_ARRAY  assignments, conflicts;
CONFLICT_TYPE conflict;

memset (dibs,        0, sizeof dibs);                   /* clear the DIB value array */
memset (assignments, 0, sizeof assignments);            /*   and the assignments array */
memset (conflicts,   0, sizeof conflicts);              /*     and the conflicts array */

dib   = 0;                                              /* initialize the DIB index */
clash = 0;                                              /*   and clashing assignments bitmap */

for (dev = 0; sim_devices [dev] != NULL; dev++) {       /* loop through all of the devices */
    dptr = sim_devices [dev];                           /*   and get a pointer to the device */

    if (not (dptr->flags & DEV_DIS)                     /* if the device is enabled */
      && dptr->ctxt != NULL) {                          /*   and the DIB is defined */
        dibptr = (DIB *) dptr->ctxt;                    /*     then get a pointer to the device's DIB */

        dibs [dib].dptr               = dptr;                               /* load the device pointer */
        dibs [dib].assignment [Channel]   = dibptr->channel_number;         /*   and IMB channel number */
        dibs [dib].assignment [Interrupt] = dibptr->interrupt_priority;     /*   and interrupt priority number */
        dibs [dib].assignment [Service]   = dibptr->service_request_number; /*   and service request number */
        dibs [dib].assignment [Device]    = dibptr->device_number;          /*   and device number into the entry */

        for (conflict = Channel; conflict <= Device; conflict++) {  /* check for conflicts */
            value = dibs [dib].assignment [conflict];               /* get the assigned value */

            if (value < 256) {                          /* if the value is assigned */
                bit   = BIT (value % 32);               /*   then get the assignment bit */
                rank  = value / 32;                     /*     and element rank for device numbers */
                index = (uint32) conflict + rank;       /*       and index to the target bitmap */

                clash |=                                            /* a clash exists */
                  conflicts [index] |= assignments [index] & bit;   /*   if the channel is already assigned */

                assignments [index] |= bit;             /* add the assigned value bit to the set of assignments */
                }
            }

        dib++;                                          /* increment the DIB array index */

        if (dibptr->io_initializer != NULL                      /* if the device has an initializer */
          && not (sim_switches & SWMASK ('P'))                  /*   and this is not the power-on clear */
          && dibptr->io_initializer (dptr, dibptr) == FALSE)    /*     then call the run-time initializer; if it fails */
            return TRUE;                                        /*       then inhibit execution */
        }
    }

if (clash == 0)                                         /* if no assignment classer are present */
    return FALSE;                                       /*   then there are no device conflicts */

else {                                                  /* otherwise */
    sim_ttcmd ();                                       /*   restore the console and log I/O modes */

    for (index = 0; index < conflict_size; index++)     /* look through the conflict array */
        while (conflicts [index] != 0) {                /*   for the next conflict to report */
            if (index <= (uint32) Device) {             /* if within the initial set of bitmaps */
                conflict = (CONFLICT_TYPE) index;       /*   then set the conflict type */
                rank = 0;                               /*     and there are no bitmap ranks */
                }

            else {                                      /* otherwise */
                conflict = Device;                      /*   this is within the device bitmap array */
                rank = index - (uint32) Device;         /*     and bias the rank to start at 0 */
                }

            value = LSB_NUMBER (conflicts [index]);     /* get the next conflicting assignment (modulo 32) */
            bit = BIT (value);                          /*   and its associated bit */

            value = value + rank * 32;                  /* adjust the assignment to its proper value */

            cprintf ("%s %u conflict (",                /* report the conflicting value */
                     conflict_label [conflict], value);

            for (dib = 0; dibs [dib].dptr != NULL; dib++)           /* locate the conflicting devices */
                if (dibs [dib].assignment [conflict] == value) {    /*   by matching the assignment */
                    if (conflicts [index] & bit)                    /*     and if this is the first device */
                        conflicts [index] ^= bit;                   /*       then clear the assigned bit */
                    else                                            /*     otherwise */
                        cputs (" and ");                            /*       this is yet another device */

                    cputs (sim_dname (dibs [dib].dptr));    /* report the conflicting device name */
                    }

            cputs (")\n");                              /* tie off the line */
            }                                           /*   and continue to look for other conflicts */

    return TRUE;                                        /* the device conflict inhibits execution */
    }
}


/* Test for a pending interrupt.

   This routine is called from the instruction prelude to initialize the
   interrupting device number, and from the main instruction loop and the
   instruction executors for interruptible instructions to check for a pending
   interrupt and, if found, to set up the interrupt handler.  It returns FALSE
   if no interrupt is pending, TRUE with SCPE_OK status if an interrupt is
   detected, or TRUE with an error status if an event service call failed.

   An initialization call is made to pass in the device number of an
   interrupting device that had not been serviced when simulation was stopped.
   If there were no "leftover" interrupts, then the device number is initialized
   to zero.

   For instruction executors, the routine counts an event tick and initializes
   the status return value.  If an interrupt or error condition is detected, the
   P register is backed up so that the interrupted instruction will be rerun
   once the interrupt is serviced, and the routine returns TRUE to indicate that
   the instruction should yield.

   For the main instruction loop, checks are also made for halt-mode interrupts
   and breakpoints.  IOP interrupts are also deferred if the CIR holds a SED 1
   instruction.  None of these tests are required for interruptible
   instructions.


   Implementation notes:

    1. The routine marked "inline" to place copies in the main instruction loop
       and in the "cpu_interrupt_pending" global routine that is called by the
       instruction executors.  With optimization, the tests for the specific
       invocations are removed, as all calls pass constants for the "invocation"
       parameter.  This effectively customizes each placement while maintaining
       only a single source routine for simpler program maintenance.

    2. When called by the instruction executors, the event interval decrement
       that occurs in the main instruction loop after each instruction execution
       is cancelled here if "sim_process_event" returns an error code.  This is
       done so that a STEP command does not decrement "sim_interval" twice.
       Note that skipping the initial decrement here does not help, as it's the
       sim_interval value AFTER the call to "sim_process_event" that must be
       preserved.
*/

static INLINE t_bool interrupt_pending (INVOCATION_STATE invocation, t_stat *status)
{
static uint32 device;                           /* the device number of a device with a pending interrupt */

if (invocation == Executor) {                           /* if this call is from an instruction executor */
    *status = SCPE_OK;                                  /*   then preset good status from the service */
    sim_interval = sim_interval - 1;                    /*     and count the cycle */
    }

else if (invocation == Initialize) {                    /* otherwise if this is an initialization call */
    device = (uint32) *status;                          /*   then save any "leftover" device interrupt */
    return TRUE;                                        /*     and that's all that is needed */
    }

if (sim_interval <= 0) {                                /* if an event timeout has expired */
    *status = sim_process_event ();                     /*   then call the event service */

    if (*status != SCPE_OK) {                           /* if the service failed */
        if (invocation == Executor) {                   /*   and this is a call from an instruction executor */
            P = P - 1 & R_MASK;                         /*     then back up to reenter the instruction */
            sim_interval = sim_interval + 1;            /*       and cancel the instruction loop increment */
            }

        return TRUE;                                    /* report the failure */
        }
    }

if (cpp_request > 0) {                                  /* if a channel program processor request is pending */
    cpp_request--;                                      /*   then decrement the delay */

    if (cpp_request == 0)                               /* if the delay has expired */
        cpp_service (1);                                /*   then service the request */
    }

if (sys_configuration & CPU_IS_IOP) {                   /* if this is an IOP system */
    if (sel_request)                                    /*   then if a selector channel request is pending */
        sel_service (1);                                /*     then service it */

    if (mpx_request_set)                                /* if a multiplexer channel request is pending */
        mpx_service (1);                                /*   then service it */

    if (imb_interrupt_request_set && cpp_irq_enable)    /* if an IMB interrupt is pending */
        imb_interrupt_poll ();                          /*   then poll for an interrupt */

     if (iop_interrupt_request_set                      /* if an IOP interrupt request is pending */
       && STA & STATUS_I                                /*   and interrupts are enabled */
       && (invocation == Executor || CIR != SED_1))     /*     and not deferred by a SED 1 instruction */
        device = iop_interrupt_poll ();                 /*       then poll to acknowledge the request */
    }

else                                                    /* otherwise it is an IMB system */
    if (imb_interrupt_request_set && STA & STATUS_I)    /*   so if an IMB interrupt is pending */
        device = imb_interrupt_poll ();                 /*     then poll to acknowledge the request */

if (invocation == Main_Loop) {                          /* if this call is from the main instruction loop */
    if (cpu_micro_state == Halted                       /*   then if the micromachine is halted */
      || cpu_micro_state == Waiting) {                  /*     or waiting for a cold load/dump to complete */
        if (CPX2 & CPX2_IRQ_SET)                        /*       then if a halt-mode interrupt is pending */
            *status = halt_mode_interrupt (device);     /*         then service it */

        return TRUE;                                    /* skip instruction execution */
        }

    else if (sim_brk_summ                               /* otherwise if a breakpoint exists */
      && sim_brk_test (TO_PA (PBANK, P - 1 & LA_MASK),  /*   at the next location */
                       BP_EXEC)) {                      /*     to be executed */
        *status = STOP_BRKPNT;                          /*       then stop the simulation */
        sim_interval = sim_interval + 1;                /*         and don't count the cycle */

        return TRUE;                                    /* skip instruction execution */
        }
    }

if (CPX1 & CPX1_IRQ_SET || SIR & cpu_sir_irq_set) {     /* if an interrupt is present */
    if (invocation == Executor)                         /*   then if this call is from an instruction executor */
        P = P - 1 & R_MASK;                             /*     then back up to reenter the instruction */

    run_mode_interrupt (device);                        /* set up the interrupt service routine */
    return TRUE;                                        /*   and skip instruction execution */
    }

else                                                    /* otherwise */
    return FALSE;                                       /*   continue with instruction execution */
}


/* Process a run-mode interrupt.

   This routine is called when one or more of the interrupt request bits are set
   in the CPX1 or SIR registers.  The highest-priority request is identified and
   cleared, the interrupt classification and parameter are established, and the
   associated interrupt handler is set up for execution.  On return, the CPU has
   been configured and is ready to execute the first instruction in the
   interrupt procedure.

   On entry, the power state is checked.  If power is failing, interrupt
   recognition is inhibited.  Otherwise, if the simulator was stopped while the
   CPU was executing a PAUS instruction, and a pending interrupt is present,
   then adjust P to return to the instruction following the PAUS.  The
   micromachine is set to the running state in case it had been halted.

   The interrupt source is then classified.  The first ten classes correspond
   one-for-one with the first ten bits of the CPX1 register, so those are
   examined first (except for power-fail, which has lowest priority).  Then the
   IMB and CPU timer interrupts are examined.  Finally, power-fail is examined.
   The highest-priority interrupt has its associated request bit cleared.

   Based on the classification, the parameter and STT index into segment 1 is
   established.  Then entry into the handler is set up.  Upon return, the first
   instruction of the handler is ready to be executed.


   Implementation notes:

    1. In hardware, halting the CPU while a PAUS instruction is executing leaves
       P pointing to the following instruction, which is executed when the RUN
       button is pressed.  In simulation, this action occurs only when execution
       is resumed with the "-B" option to bypass the pause.  Otherwise, resuming
       (or stepping) continues with the PAUS instruction.

       If an interrupt occurs while PAUS is executing, the interrupt handler
       will return to the instruction following the PAUS, as though HALT and RUN
       had been pressed.  This occurs because the P register normally points two
       instructions past the instruction currently executing.  When an interrupt
       occurs, P is decremented to point at the instruction after the current
       instruction, which is the correct point of return after the interrupt
       service routine completes.

       When the simulator is stopped, P is backed up to point at the next
       instruction to execute.  In the case of a PAUS instruction, the "next
       instruction" is the same PAUS instruction.  When simulation resumes, the
       PAUS instruction is fetched into the NIR, and P is incremented.  If no
       interrupt is pending, the main instruction execution loop copies the NIR
       into the CIR, prefetches the instruction following the PAUS into the NIR,
       and increments P again, so that it points two instructions beyond the
       current instruction.  At this point, everything is set up properly as
       before the simulation stop.

       However, if an interrupt is pending when simulation resumes, this routine
       is called before the NIR-to-CIR operation is performed, so P still points
       one instruction beyond the PAUS.  Stacking the usual P - 1 value would
       cause the interrupt handler to return to the PAUS instruction instead of
       to the instruction following, as would have occurred had a simulation
       stop not been involved.

       Therefore, when resuming from a simulator stop, the SS_PAUSE_RESUMED flag
       is set in the "cpu_stop_flags" variable by the "halt_mode_interrupt"
       routine if a PAUS instruction is in the NIR after reloading.  If an
       interrupt is pending, this routine will be entered with the flag set, and
       we then increment P so that it is correctly set to point two instructions
       beyond the PAUS before handling the interrupt.  If an interrupt is not
       pending on resumption, the P adjustment is performed as described above,
       and P will be set properly when the next interrupt occurs.

       A special case occurs when resuming into a PAUS from a simulator stop if
       a higher-priority interrupt occurs immediately after handling a pending
       lower-priority interrupt.  In this case, this routine will be entered a
       second time before the instruction execution loop performs the NIR-to-CIR
       operation.  Although the PAUS is still in the CIR, we must not increment
       P a second time, because the instruction now being interrupted is not the
       PAUS but is the first instruction of the lower-priority interrupt routine
       that never had a chance to execute.  We do this by clearing the
       SS_PAUSE_RESUMED flag after the first increment.

    2. In order to use the HP_MSB_NUMBER, which reduces to a single machine
       instruction on x86 systems, to identify interrupt sources quickly, we
       must ensure that at least one source is present.  Therefore, we guard the
       call by including the unused CPX1 bit (bit 10) to the test.  If no CPX1
       interrupts are present, this bit number will be returned.  This is
       required so that SIR interrupts and the power-fail interrupt are checked
       in the proper priority order before deciding that no run-time interrupts
       are present.

    3. This routine is entered only when one of the interrupt bits in CPX1 or
       SIR is set.  As all cases are covered in the code, it should not be
       possible to discover that no bits are set.  However, should we be entered
       improperly, we simply return.
*/

static void run_mode_interrupt (HP_WORD device_number)
{
HP_WORD   request_set, sir_set, parameter, limit, total;
IRQ_CLASS class;

if (cpu_power_state == Power_Failing)                   /* if power is failing */
    return;                                             /*   then all interrupts are inhibited */

else if (cpu_stop_flags & SS_PAUSE_RESUMED) {           /* otherwise if we are resuming into a PAUS instruction */
    P = P + 1 & R_MASK;                                 /*   then set the return to the instruction following */
    cpu_stop_flags &= ~SS_PAUSE_RESUMED;                /*     and clear the flag in case of back-to-back interrupts */
    }

cpu_micro_state = Running;                              /* the micromachine may be paused but is no longer */

request_set = CPX1 & CPX1_IRQ_SET | cpx1_UNUSED;        /* get the (guarded) set of CPX interrupt sources */
sir_set = SIR & cpu_sir_irq_set;                        /*   and the set of SIR interrupt sources */

class = (IRQ_CLASS) HP_MSB_NUMBER (request_set);        /* determine the interrupt classification */

if (class <= irq_External)                              /* if it's not power failure (or none) */
    CPX1 &= ~HP_BIT (class);                            /*   then clear the associated CPX request bit */

else if (sir_set & sir_IMBIRQ) {                        /* otherwise if it's an IMB interrupt request */
    class = irq_External;                               /*   then classify as an external interrupt */
    SIR &= ~sir_IMBIRQ;                                 /*     and clear the bit */
    }

else if (sir_set & sir_CPUTIMER) {                      /* otherwise if the system clock ticked */
    class = irq_System_Clock;                           /*   then classify it */
    SIR &= ~sir_CPUTIMER;                               /*     and clear the bit */
    }

else if (class == irq_Power_Fail)                       /* otherwise if it's a power-fail interrupt */
    CPX1 &= ~cpx1_PFINTR;                               /*   then clear the bit */

else                                                    /* otherwise something is wrong */
    return;                                             /*   because an interrupt must be present */


switch (class) {                                        /* dispatch the interrupt classification */

    case irq_Integer_Overflow:                          /* integer overflow */
        parameter = 1;                                  /* set the parameter to 1 */
        STA &= ~STATUS_O;                               /*   and clear the overflow flag */
        break;

    case irq_Module:                                    /* module interrupt */
        parameter = UPPER_BYTE (MOD);                   /* set the parameter to the module number */
        MOD = 0;                                        /*   and clear the register to prevent a second interrupt */
        break;

    case irq_External:                                  /* external interrupt */
        parameter = device_number;                      /* set the parameter to the device number */
        break;

    case irq_Power_Fail:                                /* power-fail interrupt */
        parameter = TO_ISR_LABEL (class);               /* the parameter is the procedure label */
        cpu_power_state = Power_Failing;                /*   and system power is now failing */
        break;

    case irq_System_Clock:                              /* system clock */
        cpu_sir_irq_set &= ~sir_CPUTIMER;               /* disable timer interrupts */

        cpu_read_memory (absolute, LIMIT_REGISTER, &limit); /* get the clock limit */
        cpu_read_memory (absolute, TOTAL_REGISTER, &total); /*   and the total count */

        parameter = total + limit & R_MASK;             /* the parameter is the tick count */

        cpu_write_memory (absolute, TOTAL_REGISTER, 0); /* clear the total count */
        break;

    default:                                            /* for all other interrupts */
        parameter = TO_ISR_LABEL (class);               /*   the parameter is the procedure label */
        break;
    }

cpu_setup_irq_handler (class, parameter);               /* set up the entry into the interrupt handler */

return;
}


/* Process a halt-mode interrupt.

   This routine is called when one or more of the interrupt request bits are set
   in the CPX2 register.  These bits represent switch closures on the Series
   II/III CPU front panel or optional maintenance display panel, or on the
   Series 4x/5x system control panel.  The specific RUN, LOAD, or DUMP switch
   closure is identified, and the corresponding microcode routine is executed.
   If multiple bits are set in CPX2, the microcode recognizes them in order from
   MSB to LSB.

   For the IMB machines, the RUN switch operates identically to the IOP
   machines.  The START and LOAD switches both perform a cold load and differ
   only in the value in the SWCH register.  They are implemented by setting the
   LOAD bit in CPX2.

   An IMB cold dump is performed by a standalone dump program that is
   cold-loaded from the specified device, rather than by a microcoded cold dump
   routine as implemented by the IOP machines.  The DUMP switch performs the
   cold load after saving the machine registers in memory.  It is implemented by
   setting the DUMP bit in CPX2.


   If the RUN switch is set, a test is made for a System Halt, which inhibits
   recognition of the switch.  If the System Halt flag is set, the CPU must be
   reset before execution may be resumed.  Otherwise, the NIR is reloaded in
   case P was changed during the simulation stop.  If the R-bit (right stack op
   pending) flag in the status register is set, but the NIR no longer contains a
   stack instruction, R is cleared.  If a stack instruction is present, and the
   R-bit is set, then the CIR is set, and the following instruction is fetched.
   The micromachine state is set to "running", and one event tick is added back
   to the accumulator to ensure that a single step won't complete without
   executing an instruction.


   If the DUMP switch is pressed on an IOP machine, an I/O Reset is performed on
   all devices except the CPU, which is skipped to preserve the register state.
   The dump device number is obtained from the SWCH register and tested to
   ensure that a tape is mounted with a write ring and the unit is online.  Then
   the contents of the CPU registers are written to the reserved memory area
   starting at address 1400 octal in bank 0.  This is followed by two SIO
   programs.  The main program at addresses 1430-1437 writes 4K-word blocks of
   memory to the dump device.  The error recovery program at addresses 1422-1427
   is invoked when a write fails.  It does a Backspace Record followed by a
   Write Gap to skip the bad spot on the tape, and then the write is retried.

   The DUMP switch remains set, so after each pass through the main execution
   loop to run a channel cycle, this routine is reentered.  The "waiting" state
   causes the second part of the process to check for device completion.  When
   it occurs, the expected external interrupt is cleared, and if the dump is
   complete, the DUMP switch is reset, the original SIO pointer in the DRT is
   restored, and the micromachine is halted.  Otherwise, the SIO pointer is
   read.  If it points at the end of the program, then the operation completed
   normally.  In this case, the dump address is advanced, and, if all of memory
   has been dumped, the SIO pointer is reset to the recovery program, and that
   program is changed to finish up with Write File Mark and Rewind/Offline
   commands.  Otherwise, the pointer is reset to the main program in preparation
   for the next 4K write.  The SIO program is then restarted.

   If SIO pointer failed to complete, the pointer is reset to point at the error
   recovery program, and the memory address is unchanged; the same 4K write will
   be performed once the recovery program runs.  If the recovery program fails,
   the dump is terminated at that point with a failure indication.

   During the dump operation, the CIR register is continually updated with the
   current memory bank number.  If the dump runs to completion, CIR will contain
   the number of 64K memory banks installed in the machine.  A value less than
   the installed memory value indicates a dump failure.  Except for the CIR, the
   machine state is restored, so that another dump may be attempted.


   If the DUMP switch is pressed on an IMB machine, the contents of the CPU
   registers are written to the reserved memory area starting at address 1401
   octal in bank 0.  The system interrupt mask word if read from memory location
   7 and written to location 1515 octal.  The original DRT entry for the dump
   device is copied to locations 1516-1521, and then the Channel Program
   Processor cold-load routine is called to load the dump program into memory.

   The DUMP switch remains set, so after each pass through the main execution
   loop to run a channel cycle, this routine is reentered.  The "waiting" state
   causes the second part of the process to check for the expected IMB
   interrupt, indicating the end of the channel program.  When that occurs, the
   interrupt is cleared, and the CPP cold load routine is called to complete or
   continue the cold-load process.  As long as it returns FALSE, another
   intermediate interrupt is expected.  Once cold-load execution is complete,
   the routine returns TRUE, the DUMP switch is reset, the micromachine state is
   set to "running", and the Cold Load trap is taken to complete the process.


   If the LOAD switch is pressed, the cold load process is initiated.  The cold
   load device number is obtained from the lower byte of the SWCH register to
   determine if this is an SIO cold load or a Starfish cold load; the latter is
   indicated by the presence of the device number of the IMBA in the lower byte
   of the register.  If SWCH register bit 8 is clear or this is a Starfish load,
   the process begins by filling memory with HALT 10 instructions.

   For an IOP cold load, the first part of the process clears the TOS and STA
   registers, stores the initial channel program in memory, and executes an SIO
   instruction to start the channel.  Once the device starts, interrupts are
   enabled, and the micromachine state is set to "waiting" in preparation for
   executing the second part of the cold load process once the channel program
   ends.  The routine then exits to begin channel execution.

   The LOAD switch remains set, so after each pass through the main execution
   loop to run a channel cycle, this routine is reentered.  The "waiting" state
   causes the second part of the process to check for device completion.  The
   expected external interrupt is cleared, the LOAD switch is cleared, the
   micromachine state is set to "running", and the Cold Load trap is taken to
   complete the process.


   For an IMB or Starfish cold load, the load device number if obtained from the
   switch register and passed to the Channel Program Processor routine to handle
   the load.  The routine returns TRUE when the cold load is complete and FALSE
   for intermediate returns that wait for a device interrupt.  The latter will
   occur several times as the process brings in successively larger channel
   programs from the load device.

   As with the IOP load, the LOAD switch remains set.  The CPP cold load routine
   is called when a pass through the main execution loop detects the IMB
   interrupt, and as long as it returns FALSE, another intermediate interrupt is
   expected.  Once the cold load process is complete, the routine returns TRUE,
   the LOAD switch is cleared, the micromachine state is set to "running", and
   the Cold Load trap is taken to complete the process.


   Implementation notes:

    1. After processing the RUN switch and returning to the instruction loop, if
       no interrupt is present, and the R-bit in the status register is clear,
       then CIR will be set from NIR, and NIR will be reloaded.  If the R-bit is
       set, then CIR and NIR will not be changed, i.e., the NEXT action will be
       skipped, so we perform it here.  If an interrupt is pending, then the
       interrupt will be processed using the old value of the CIR, and the
       instruction in the NIR will become the first instruction executed after
       the interrupt handler completes.

    2. The SIO cold-load microcode is shared with the cold dump process.  The
       Series II dump process saves memory locations DRT + 0 through DRT + 3 in
       the TOS registers.  The load process uses the same microcode but does not
       perform the memory read, so the TOS registers are loaded with the
       previous contents of the OPND register, which is effectively a random
       value.  In simulation, the TOS registers are cleared.

    3. The SIO cold load and dump microcode waits forever for an interrupt from
       the cold load device.  If it doesn't occur, the microcode hangs until a
       system reset is performed (it tests CPX1 bit 8 and does a JMP *-1 if the
       bit is not set).  The IMB and Starfish cold load and dump microcode waits
       for one second for the read or write to complete.  If a timeout occurs,
       the CPP performs a System Halt.  The simulation follows the microcode
       behaviors.

    4. Series III front panel diagnostics and direct I/O cold loading is not
       implemented.
*/

static const MEMORY_WORD sio_loader [] = {              /* SIO cold load program */
    SIO_SBANK, 0000000,                                 /*   SETBNK 0 */
    SIO_CNTL,  0000000,                                 /*   CONTRL 0,<SWCH-upper> */
    SIO_READ,  0001400,                                 /*   READ   #16,001400 */
    SIO_JUMP,  0001400                                  /*   JUMP   001400 */
    };

static const MEMORY_WORD sio_dumper [] = {              /* SIO cold dump program */
    SIO_CNTL,  MS_CN_BSR,                               /*   CONTRL 0,BSR */
    SIO_CNTL,  MS_CN_GAP,                               /*   CONTRL 0,GAP */
    SIO_JUMP,  001436,                                  /*   JUMP   001436 (ENDINT) */

    SIO_SBANK, 000000,                                  /*   SETBNK 0 */
    SIO_CNTL,  000000,                                  /*   CONTRL 0,<SWCH-upper> */
    SIO_WRITE, 000000,                                  /*   WRITE  #4096,000000 */
    SIO_ENDIN, 000000                                   /*   ENDINT */
    };


static t_stat halt_mode_interrupt (HP_WORD device_number)
{
static HP_WORD cold_device, sio_pointer, status, offset, frame;
static uint32  address, index;
static t_bool  error_recovery;
HP_WORD        pointer, drt, mask;

if (CPX2 & cpx2_RUNSWCH) {                              /* if the RUN switch is pressed */
    if (CPX2 & cpx2_SYSHALT) {                          /*   then if the System Halt flip-flop is set */
        CPX2 &= ~CPX2_IRQ_SET;                          /*     then clear all switches */
        return STOP_SYSHALT;                            /*       as the CPU cannot run until it is reset */
        }

    else                                                /*   otherwise */
        CPX2 = CPX2 & ~cpx2_RUNSWCH | cpx2_RUN;         /*     clear the switch and set the Run flip-flop */

    cpu_read_memory (fetch, P, &NIR);                   /* load the next instruction to execute */
    P = P + 1 & R_MASK;                                 /*   and point to the following instruction */

    if ((NIR & PAUS_MASK) == PAUS)                      /* if resuming into a PAUS instruction */
        cpu_stop_flags |= SS_PAUSE_RESUMED;             /*   then defer any interrupt until after PAUS is executed */

    if ((NIR & SUBOP_MASK) != 0)                        /* if the instruction is not a stack instruction */
        STA &= ~STATUS_R;                               /*   then clear the R-bit in case it had been set */

    else if (STA & STATUS_R) {                          /* otherwise if a right-hand stack op is pending */
        CIR = NIR;                                      /*   then set the current instruction */
        cpu_read_memory (fetch, P, &NIR);               /*     and load the next instruction */
        }

    cpu_micro_state = Running;                          /* start the micromachine */
    sim_interval = sim_interval + 1;                    /*   but don't count this cycle against a STEP count */

    if (cpu_power_state == Power_Returning) {           /* if power is returning after a failure */
        if (CPX2 & cpx2_INHPFARS)                       /*   then if auto-restart is inhibited */
            CPX2 &= ~cpx2_RUN;                          /*     then clear the Run flip-flop */

        MICRO_ABORT (trap_Power_On);                    /* set up the trap to the power-on routine */
        }
    }


else if (CPX2 & cpx2_DUMPSWCH) {                        /* otherwise if the DUMP switch is pressed */
    if (cpu_micro_state != Waiting) {                   /*   then if the dump is not in progress */
        reset_all (IO_RESET);                           /*     then reset all I/O devices */

        cold_device = DEVNO (SWCH);                     /* get the device number from the SWCH register */

        if (sys_configuration & CPU_IS_IMB) {                   /* if this is an IMB system */
            cpu_write_memory (absolute, 01401, cold_device);    /*   then store the channel and device number */

            cpu_write_memory (absolute, 01402, X);      /* store the index register */
            cpu_write_memory (absolute, 01403, DL);     /*   and the data limit */
            cpu_write_memory (absolute, 01404, DBANK);  /*   and the data bank */
            cpu_write_memory (absolute, 01405, DB);     /*   and the data base */

            cpu_write_memory (absolute, 01406, Q);      /* store the frame pointer */
            cpu_write_memory (absolute, 01407, SM);     /*   and the stack pointer */
            cpu_write_memory (absolute, 01410, SBANK);  /*   and the stack bank */
            cpu_write_memory (absolute, 01411, Z);      /*   and the stack limit */

            cpu_write_memory (absolute, 01412, STA);    /* store the status register */
            cpu_write_memory (absolute, 01413, PBANK);  /*   and the program bank */
            cpu_write_memory (absolute, 01414, PB);     /*   and the program base */
            cpu_write_memory (absolute, 01415, P);      /*   and the program counter */
            cpu_write_memory (absolute, 01416, PL);     /*   and the program limit */
            cpu_write_memory (absolute, 01417, CIR);    /*   and the current instruction */

            cpu_write_memory (absolute, 01420, UPPER_WORD (mem_size));  /* store the memory bank count */

            cpu_write_memory (absolute, 01421, SP0);    /* store the system halt number */
            cpu_write_memory (absolute, 01422, SIR);    /*   and the SIR register */

            cpu_read_memory (absolute, INTERRUPT_MASK, &mask);  /* read the system mask value */
            cpu_write_memory (absolute, 01515, mask);           /*   and save it */

            for (index = 0; index < 4; index++) {                               /* save the original */
                cpu_read_memory (absolute, TO_DRT (cold_device) + index, &drt); /*   DRT entry values */
                cpu_write_memory (absolute, 01516 + index, drt);                /*     of the cold dump device */
                }

            cpp_cold_load (cold_device);                /* call the HP-IB cold load routine */

            STA = STATUS_I | STATUS_O;                  /* enable interrupts and set overflow */
            cpu_micro_state = Waiting;                  /*   and set the load-in-progress state */
            }

        else {                                              /* otherwise this is an IOP system */
            status = iop_direct_io (cold_device, ioTIO, 0); /*   so get the device status */

            if ((status & MS_ST_MASK) != MS_ST_READY) { /* if the tape is not ready and unprotected */
                CPX2 &= ~cpx2_DUMPSWCH;                 /*   then clear the dump switch */

                CIR = 0;                                /* clear CIR to indicate a failure */
                return STOP_CDUMP;                      /*   and terminate the dump */
                }

            if (DBANK > DUMP_BANK_MASK                  /* if any of the bank register values */
              || PBANK > DUMP_BANK_MASK                 /*   are larger than the cold dump format permits */
              || SBANK > DUMP_BANK_MASK)                /*     then warn the user of the truncation */
                cprintf ("%s cold dump warning: bank register values truncated\n",
                         sim_name);

            cpu_read_memory (absolute, TO_DRT (cold_device),    /* get the original DRT pointer */
                             &sio_pointer);

            cpu_write_memory (absolute, 01401, sio_pointer);    /* store the original DRT pointer */
            cpu_write_memory (absolute, 01402, SM);             /* store the stack pointer */
            cpu_write_memory (absolute, 01403, 0);              /* store zeros for the scratch pad 1 */
            cpu_write_memory (absolute, 01404, 0);              /*   and scratch pad 2 register values */
            cpu_write_memory (absolute, 01405, DB);             /* store the data base */
            cpu_write_memory (absolute, 01406, (DBANK & DUMP_BANK_MASK) << 12   /* store DBANK in 0:4 */
                                             | (PBANK & DUMP_BANK_MASK) << 8    /*   and PBANK in 4:4 */
                                             | (SBANK & DUMP_BANK_MASK));       /*     and SBANK in 12:4 */
            cpu_write_memory (absolute, 01407, Z);              /* store the stack limit */
            cpu_write_memory (absolute, 01410, DL);             /*   and the data limit */
            cpu_write_memory (absolute, 01411, X);              /*   and the index register */
            cpu_write_memory (absolute, 01412, Q);              /*   and the frame pointer */
            cpu_write_memory (absolute, 01413, CIR);            /*   and the current instruction */
            cpu_write_memory (absolute, 01414, PB);             /*   and the program base */
            cpu_write_memory (absolute, 01415, PL);             /*   and the program limit */
            cpu_write_memory (absolute, 01416, P);              /*   and the program counter */
            cpu_write_memory (absolute, 01417, CPX1);           /* store the CPX1 register */
            cpu_write_memory (absolute, 01420, STA);            /*   and the status register */

            cpu_write_memory (absolute, 01421, TO_WORD (CPX2,           /* store the lower byte of the CPX2 register */
                                               UPPER_WORD (mem_size))); /*  and the memory bank count */

            if (cold_device == IMBA_DEVNO) {                                    /* if the device is the IMBA */
                cpu_write_memory (absolute, 01400, REPLACE_LOWER (SWCH, 1));    /*   then set the machine ID */

                for (index = 0; index < 4; index++) {                               /* save the original */
                    cpu_read_memory (absolute, TO_DRT (cold_device) + index, &drt); /*   DRT entry values */
                    cpu_write_memory (absolute, 01422 + index, drt);                /*     of the cold load device */
                    }
                }

            else                                        /* otherwise it's an SIO dump */
                cpu_write_memory (absolute, 01400, 1);  /*   so set the machine ID to 1 for the Series III */

            mem_copy_loader (01422, sio_dumper, sizeof sio_dumper); /* copy the SIO dump program to memory */

            cpu_write_memory (absolute, 01433, UPPER_BYTE (SWCH));  /* insert the write command into the code */

            address = 0;                                /* clear the address */
            offset = 0;                                 /*   and memory offset counters */

            CIR = 0;                                    /* clear the memory bank counter */

            cpu_write_memory (absolute, TO_DRT (cold_device), 01430);   /* point the DRT at the cold dump program */
            error_recovery = FALSE;

            iop_direct_io (cold_device, ioSIO, 0);      /* start the device */

            if (CPX1 & cpx1_IOTIMER)                    /* if the device did not respond */
                MICRO_ABORT (trap_SysHalt_IO_Timeout);  /*   then a System Halt occurs */

            else {                                      /* otherwise the device has started */
                status = STA;                           /*   so save the original status register value */

                STA = STATUS_I | STATUS_O;              /* enable interrupts and set overflow */
                cpu_micro_state = Waiting;              /*   and set the load-in-progress state */
                }
            }
        }

    else if (CPX1 & cpx1_EXTINTR) {                     /* otherwise if an IOP interrupt is pending */
        CPX1 &= ~cpx1_EXTINTR;                          /*   then clear it */

        iop_direct_io (device_number, ioRIN, 0);        /* reset the device interrupt */

        if (device_number == cold_device)               /* if the expected device interrupted */
            if (address >= mem_size) {                  /*   then if all of memory has been dumped */
                CPX2 &= ~cpx2_DUMPSWCH;                 /*     then reset the DUMP switch */

                STA = status;                           /* restore the original status register value */

                cpu_write_memory (absolute,             /* restore the */
                                  TO_DRT (cold_device), /*   original SIO pointer */
                                  sio_pointer);         /*     to the DRT */

                cpu_micro_state = Halted;               /* clear the dump-in-progress state */
                return STOP_CDUMP;                      /*   and report dump completion */
                }

            else {                                      /* otherwise the dump continues */
                cpu_read_memory (absolute,              /* read the */
                                 TO_DRT (cold_device),  /*   current SIO pointer address */
                                 &pointer);             /*     from the DRT */

                if (pointer == 01440) {                     /* if the SIO program completed normally */
                    cpu_write_memory (absolute,             /*   then reset the pointer */
                                      TO_DRT (cold_device), /*     to the start */
                                      001430);              /*       of the program */

                    if (error_recovery)                 /* if this was a successful error recovery */
                        error_recovery = FALSE;         /*   then clear the flag and keep the current address */

                    else {                                  /* otherwise this was a successful write */
                        address = address + 4096;           /*   so bump the memory address */
                        offset = offset + 4096 & LA_MASK;   /*     and offset to the next 4K block */

                        cpu_write_memory (absolute,         /* store the new write buffer address */
                                          001435, offset);

                        if (offset == 0) {                  /* if the offset wrapped around */
                            CIR = CIR + 1;                  /*   then increment the bank number */
                            cpu_write_memory (absolute,     /*     and store it as the SET BANK target */
                                              001431, CIR);

                            if (address >= mem_size) {              /* if all of memory has been dumped */
                                cpu_write_memory (absolute, 001423, /*   then change the error recovery program */
                                                  MS_CN_WFM);       /*     to write a file mark */
                                cpu_write_memory (absolute, 001425, /*       followed by */
                                                  MS_CN_RST);       /*         a rewind/offline request */

                                cpu_write_memory (absolute,         /* point at the recovery program */
                                                  TO_DRT (cold_device),
                                                  001422);
                                }
                            }
                        }
                    }

                else if (error_recovery) {              /* otherwise if the recover program failed */
                    CPX2 &= ~cpx2_DUMPSWCH;             /*   then reset the DUMP switch */

                    STA = status;                       /* restore the original status register value */

                    cpu_write_memory (absolute,             /* restore the */
                                      TO_DRT (cold_device), /*   original SIO pointer */
                                      sio_pointer);         /*     to the DRT */

                    cpu_micro_state = Halted;           /* clear the dump-in-progress state */
                    return STOP_CDUMP;                  /*   and report dump failure */
                    }

                else {                                      /* otherwise attempt error recovery */
                    cpu_write_memory (absolute,             /*   by setting the SIO pointer */
                                      TO_DRT (cold_device), /*     to the backspace/write gap */
                                      001422);              /*       program */

                    error_recovery = TRUE;              /* indicate that recovery is in progress */
                    }

                iop_direct_io (cold_device, ioSIO, 0);  /* start the device */
                }
        }

    else if (SIR & sir_IMBIRQ) {                        /* otherwise if an IMB interrupt is pending */
        SIR &= ~sir_IMBIRQ;                             /*   then clear it */

        if (cpp_cold_load (cold_device)) {              /* if the cold load routine is finished */
            CPX1 = 0;                                   /*   then clear all run-mode signals */
            CPX2 = CPX2 & cpx2_INHPFARS | cpx2_RUN;     /*     and keep ARS inhibit and set the RUN flip-flop */

            if (sys_configuration & FW_CSTX) {              /* if the CST expansion firmware is present */
                cpu_read_memory (absolute, ICS_Q, &frame);  /*   then set bit 15 */
                cpu_write_memory (absolute, frame - 9, 1);  /*     in Q - 9 to announce it */
                }

            CSMAP = STMK_M;                             /* start off with physical mapping */

            cpu_micro_state = Running;                  /* clear the load-in-progress state */
            MICRO_ABORT (trap_Cold_Load);               /*   and execute the cold load trap handler */
            }
        }
    }                                                   /* otherwise wait for the cold dump device to interrupt */

else if (CPX2 & cpx2_LOADSWCH)                          /* otherwise if the LOAD switch is pressed */
    if (cpu_micro_state != Waiting) {                   /*   then if the load is not in progress */
        reset_all (CPU_IO_RESET);                       /*     then reset the CPU and all I/O devices */

        cold_device = DEVNO (SWCH);                     /* get the device number from the SWCH register */

        if ((SWCH & HP_BIT (8)) == 0                    /* if switch register bit 8 is clear */
          || cold_device == IMBA_DEVNO)                 /*   or the device is the IMB adapter */
            mem_fill (0, HALT_10);                      /*     then fill all of memory with HALT 10 instructions */

        SBANK = 0;                                      /* set the stack bank to bank 0 */

        if (sys_configuration & CPU_IS_IMB                  /* if this is an IMB system */
          || cold_device == IMBA_DEVNO) {                   /*   or the device is the IMB adapter */
            if (cold_device == IMBA_DEVNO)                  /*     then if it is the IMB adapter */
                cold_device = DEVNO (UPPER_BYTE (SWCH));    /*       then get cold load device */

            cpp_cold_load (cold_device);                /* call the HP-IB cold load routine */

            STA = STATUS_I | STATUS_O;                  /* enable interrupts and set overflow */
            cpu_micro_state = Waiting;                  /*   and set the load-in-progress state */
            }

        else if (cold_device < 3) {                     /* otherwise if the device number is between 0 and 2 */
            CPX2 &= ~cpx2_LOADSWCH;                     /*   then reset the LOAD switch */
            return SCPE_INCOMP;                         /*     and execute a front panel diagnostic */
            }

        else if (cold_device > 63) {                    /* otherwise if the device number is > 63 */
            CPX2 &= ~cpx2_LOADSWCH;                     /*   then reset the LOAD switch */
            return SCPE_INCOMP;                         /*     and execute a direct I/O cold load */
            }

        else {                                          /* otherwise the device number is in the channel I/O range */
            RA = 0;                                     /* set the */
            RB = 0;                                     /*   TOS registers */
            RC = 0;                                     /*     to the same */
            RD = 0;                                     /*       (random) value */

            SR = 4;                                     /* mark the TOS registers as valid */
            STA = 0;                                    /*   and clear the status register */

            mem_copy_loader (01430, sio_loader, sizeof sio_loader); /* copy the SIO cold load program */

            cpu_write_memory (absolute, 01433, UPPER_BYTE (SWCH));      /* insert the read command into the code */
            cpu_write_memory (absolute, TO_DRT (cold_device), 01430);   /*   and point the DRT to the program */

            iop_direct_io (cold_device, ioSIO, 0);      /* start the device */

            if (CPX1 & cpx1_IOTIMER)                    /* if the device did not respond */
                MICRO_ABORT (trap_SysHalt_IO_Timeout);  /*   then a System Halt occurs */

            else {                                      /* otherwise the device has started */
                STA = STATUS_I | STATUS_O;              /*   so enable interrupts and set overflow */
                cpu_micro_state = Waiting;              /*     and set the load-in-progress state */
                }
            }
        }

    else if (CPX1 & cpx1_EXTINTR) {                     /* if an IOP interrupt is pending */
        CPX1 &= ~cpx1_EXTINTR;                          /*   then clear it */

        iop_direct_io (device_number, ioRIN, 0);        /* reset the device interrupt */

        if (device_number == cold_device                /* if the expected device interrupted */
          && (sys_configuration & CPU_IS_IOP            /*   and the device is an IOP device */
          && DEVNO (SWCH) != IMBA_DEVNO                 /*   and it is not the Starfish */
          || cpp_cold_load (cold_device))) {            /*   or the IMB cold load routine is finished */
            CPX2 &= ~cpx2_LOADSWCH;                     /*     then reset the LOAD switch */

            cpu_micro_state = Running;                  /* clear the load-in-progress state */
            MICRO_ABORT (trap_Cold_Load);               /*   and execute the cold load trap handler */
            }
        }

    else if (SIR & sir_IMBIRQ) {                        /* otherwise if an IMB interrupt is pending */
        SIR &= ~sir_IMBIRQ;                             /*   then clear it */

        if (device_number == cold_device                /* if the expected device interrupted */
          && cpp_cold_load (cold_device)) {             /*   and the IMB cold load routine is finished */
            CPX1 = 0;                                   /*     then clear all run-mode signals */
            CPX2 = CPX2 & cpx2_INHPFARS | cpx2_RUN;     /*       and keep ARS inhibit and set the RUN flip-flop */

            if (sys_configuration & FW_CSTX) {              /* if the CST expansion firmware is present */
                cpu_read_memory (absolute, ICS_Q, &frame);  /*   then set bit 15 */
                cpu_write_memory (absolute, frame - 9, 1);  /*     in Q - 9 to announce it */
                }

            CSMAP = STMK_M;                             /* start off with physical mapping */

            cpu_micro_state = Running;                  /* clear the load-in-progress state */
            MICRO_ABORT (trap_Cold_Load);               /*   and execute the cold load trap handler */
            }
        }

return SCPE_OK;
}


/* Execute one machine instruction.

   This routine executes the CPU instruction present in the CIR.  The CPU state
   (registers, memory, interrupt status) is modified as necessary, and the
   routine return SCPE_OK if the instruction executed successfully.  Any other
   status indicates that execution should cease, and control should return to
   the simulator console.  For example, a programmed HALT instruction returns
   STOP_HALT status.

   Unimplemented instructions are detected by those decoding branches that
   result from bit patterns not corresponding to legal instructions or
   corresponding to optional instructions not currently enabled.  Normally,
   execution of an unimplemented instruction would result in an Unimplemented
   Instruction trap.  However, for debugging purposes, a simulator stop may be
   requested instead by returning STOP_UNIMPL status if the SS_UNIMPL simulation
   stop flag is set.

   This routine implements the main instruction dispatcher, as well as memory
   address instructions (subopcodes 04-17).  Instructions corresponding to
   subopcodes 00-03 are executed by routines in the base instruction set module.


   Implementation notes:

    1. Each instruction executor begins with a comment listing the instruction
       mnemonic and, following in parentheses, the condition code setting, or
       "none" if the condition code is not altered, and a list of any traps that
       might be generated.

    2. Stack preadjusts are simpler if performed explicitly due to overlapping
       requirements that reduce the number of static preadjusts to 4 of the 16
       entries.

    3. The order of operations for each instruction follows the microcode.  For
       example, the LOAD instruction performs the bounds check on the effective
       address and reads the operand before pushing the stack down and storing
       it in the RA registrer.  Pushing the stack down first and then reading
       the value directly into the RA register would leave the stack in a
       different state if the memory access caused a Bounds Violation trap.

    4. The TBA, MTBA, TBX, and MTBX instructions take longer to execute than the
       nominal 2.5 microseconds assumed for the average Series III instruction
       execution time.  Consequently, the 7905 disc diagnostic fails Step 66
       (the retry counter test) if the DS device is set for REALTIME operation.
       The diagnostic uses the MTBA P+0 instruction in a timing loop, which
       expires before the disc operations complete.

       A workaround for this diagnostic is to decrement sim_interval twice for
       these instructions.  However, doing so causes the 7970 tape diagnostic to
       fail Step 532 (the timing error test) for the opposite reason: a wait
       loop of MTBA P+0 instructions causes the tape data transfer service event
       time to count down twice as fast, while the multiplexer channel data
       transfer polls occur at the usual one per instruction.  This could be
       remedied by having the channel polls execute twice as many I/O cycles for
       these instructions, although the general solution would be to recast
       sim_intervals as microseconds and to decrement sim_interval by differing
       amounts appropriate for each instruction.

    5. When the CST expansion firmware is installed, the behavior of the PCAL 0
       instruction is changed to assume that an external label is on the TOS.
       This is because the new-format label no longer carries the local/external
       indicator.  However, when the PCAL 0 is immediately preceded by an LRA
       instruction, the assumption is reversed, i.e., a local label is on the
       TOS.  This is required to support the BASIC compiler.

       To aid the PCAL instruction executor, we record the LRA presence in the
       Scratch Pad 0 (SP0) register if a PCAL 0 instruction is in the NIR.  If
       the next instruction is not PCAL 0, a NOP is recorded in SP0.  The value
       is checked in the PCAL executor to determine the proper label
       interpretation.
*/

static t_stat machine_instruction (void)
{
HP_WORD       displacement, opcode, offset, operand, operand_1, operand_2, result;
int32         control, limit;
ACCESS_CLASS  class;
BYTE_SELECTOR selector;
t_bool        branch;
t_stat        status = SCPE_OK;

switch (SUBOP (CIR)) {                                  /* dispatch on bits 0-3 of the instruction */

    case 000:                                           /* stack operations */
        status = cpu_stack_op ();                       /* set the status from the instruction executor */
        break;


    case 001:                                           /* shift, branch, and bit test operations */
        status = cpu_shift_branch_bit_op ();            /* set the status from the instruction executor */
        break;


    case 002:                                           /* move, special, firmware, immediate, field, and register operations */
        status = cpu_move_spec_fw_imm_field_reg_op ();  /* set the status from the instruction executor */
        break;


    case 003:                                           /* I/O, control, program, immediate, and memory operations */
        status = cpu_io_cntl_prog_imm_mem_op ();        /* set the status from the instruction executor */
        break;


    case 004:                                           /* LOAD (CCA; STOV, BNDV) */
        cpu_ea (CIR, &class, &offset, NULL);            /* get the effective address */
        cpu_read_memory (class, offset, &operand);      /*   and read the operand */

        cpu_push ();                                    /* push the operand */
        RA = operand;                                   /*   onto the stack */

        SET_CCA (RA, 0);                                /* set the condition code */
        break;


    case 005:                                           /* TBA, MTBA, TBX, MTBX, and STOR */
        if (CIR & M_FLAG) {                             /* STOR (none; STUN, BNDV) */
            cpu_ea (CIR, &class, &offset, NULL);        /* get the effective address */

            PREADJUST_SR (1);                           /* ensure that at least one TOS register is loaded */

            cpu_write_memory (class, offset, RA);       /* write the TOS to memory */
            cpu_pop ();                                 /*   and pop the stack */
            }

        else {                                          /* TBA, MTBA, TBX, or MTBX */
            opcode = CIR & TBR_MASK;                    /* get the test and branch operation */

            if (opcode == TBA || opcode == MTBA) {      /* TBA or MTBA (none; STUN, STOV, BNDV) */
                PREADJUST_SR (3);                       /* ensure that at least three TOS registers are loaded */

                while (SR > 3)                          /* if more than three TOS register are loaded */
                    cpu_queue_down ();                  /*   queue them down until exactly three are left */

                offset = DB + RC & LA_MASK;             /* get the address of the control value */

                if (DL <= offset && offset <= SM || PRIV)       /* if the address is within the segment */
                    cpu_read_memory (data, offset, &operand);   /*   then read the value */

                else                                            /* otherwise */
                    MICRO_ABORT (trap_Bounds_Violation);        /*   trap with a bounds violation if not privileged */

                if (opcode == MTBA) {                           /* if the instruction is MTBA */
                    operand = operand + RB & DV_MASK;           /*   then add the step size */
                    cpu_write_memory (data, offset, operand);   /*     to the control variable */
                    }

                control = SEXT16 (operand);             /* sign-extend the control value */
                }

            else {                                      /* TBX or MTBX (none; STUN, BNDV) */
                PREADJUST_SR (2);                       /* ensure that at least two TOS registers are loaded */

                if (opcode == MTBX)                     /* if the instruction is MTBX */
                    X = X + RB & R_MASK;                /*   then add the step size to the control variable */

                control = SEXT16 (X);                   /* sign-extend the control value */
                }

            limit = SEXT16 (RA);                        /* sign-extend the limit value */

            if (RB & D16_SIGN)                          /* if the step size is negative */
                branch = control >= limit;              /*   then branch if the value is not below the limit */
            else                                        /* otherwise */
                branch = control <= limit;              /*   branch if the value is not above the limit */

            if (branch) {                                       /* if the test succeeded */
                displacement = CIR & DISPL_255_MASK;            /*   then get the branch displacement */

                if (CIR & DISPL_255_SIGN)                       /* if the displacement is negative */
                    offset = P - 2 - displacement & LA_MASK;    /*   then subtract the displacement from the base */
                else                                            /* otherwise */
                    offset = P - 2 + displacement & LA_MASK;    /*   add the displacement to the base */

                if (cpu_stop_flags & SS_LOOP                    /* if the infinite loop stop is active */
                  && displacement == 0                          /*   and the target is the current instruction */
                  && (opcode == TBA || opcode == TBX))          /*     and the instruction must be checked */
                    status = STOP_INFLOOP;                      /*       then stop the simulator */
                else                                            /* otherwise */
                    status = SCPE_OK;                           /*   continue */

                cpu_read_memory (fetch_checked, offset, &NIR);  /* load the next instruction register */
                P = offset + 1 & R_MASK;                        /*   and increment the program counter */
                }

            else {                                      /* otherwise the test failed */
                cpu_pop ();                             /*   so pop the limit */
                cpu_pop ();                             /*     and the step size from the stack */

                if (opcode == TBA || opcode == MTBA)    /* if the instruction is TBA or MTBA */
                    cpu_pop ();                         /*   then pop the variable address too */
                }                                       /*     and continue execution at P + 1 */
            }
        break;


    case 006:                                           /* CMPM (CCC; STUN) */
        cpu_ea (CIR, &class, &offset, NULL);            /* get the effective address */
        cpu_read_memory (class, offset, &operand);      /*   and read the operand */

        PREADJUST_SR (1);                               /* ensure that at least one TOS register is loaded */

        SET_CCC (RA, 0, operand, 0);                    /* set the condition code from the TOS value */
        cpu_pop ();                                     /*   and then pop the value from the stack */
        break;


    case 007:                                           /* ADDM (CCA, C, O; STUN, BNDV, ARITH) */
        cpu_ea (CIR, &class, &offset, NULL);            /* get the effective address */
        cpu_read_memory (class, offset, &operand);      /*   and read the operand */

        PREADJUST_SR (1);                               /* ensure that at least one TOS register is loaded */

        RA = cpu_add_16 (RA, operand);                  /* add the operands */
        SET_CCA (RA, 0);                                /*   and set the condition code */
        break;


    case 010:                                           /* SUBM (CCA, C, O; STUN, BNDV, ARITH) */
        cpu_ea (CIR, &class, &offset, NULL);            /* get the effective address */
        cpu_read_memory (class, offset, &operand);      /*   and read the operand */

        PREADJUST_SR (1);                               /* ensure that at least one TOS register is loaded */

        RA = cpu_sub_16 (RA, operand);                  /* subtract the operands */
        SET_CCA (RA, 0);                                /*   and set the condition code */
        break;


    case 011:                                           /* MPYM (CCA, O; STUN, STOV, BNDV, ARITH) */
        cpu_ea (CIR, &class, &offset, NULL);            /* get the effective address */
        cpu_read_memory (class, offset, &operand);      /*   and read the operand */

        PREADJUST_SR (1);                               /* ensure that at least one TOS register is loaded */

        RA = cpu_mpy_16 (RA, operand);                  /* multiply the operands */
        SET_CCA (RA, 0);                                /*   and set the condition code */
        break;


    case 012:                                           /* INCM, DECM */
        cpu_ea (CIR | M_FLAG, &class, &offset, NULL);   /* get the effective address (forced to data-relative) */
        cpu_read_memory (class, offset, &operand);      /*   and read the operand */

        if (CIR & M_FLAG)                               /* DECM (CCA, C, O; BNDV, ARITH) */
            result = cpu_sub_16 (operand, 1);           /* decrement the operand and set C and O as necessary */
        else                                            /* INCM (CCA, C, O; BNDV, ARITH) */
            result = cpu_add_16 (operand, 1);           /* increment the operand and set C and O as necessary */

        cpu_write_memory (class, offset, result);       /* write the operand back to memory */
        SET_CCA (result, 0);                            /*   and set the condition code */
        break;


    case 013:                                           /* LDX (CCA; BNDV) */
        cpu_ea (CIR, &class, &offset, NULL);            /* get the effective address */
        cpu_read_memory (class, offset, &X);            /*   and read the operand into the X register */

        SET_CCA (X, 0);                                 /* set the condition code */
        break;


    case 014:                                           /* BR (none; BNDV), BCC (none; BNDV) */
        if ((CIR & BR_MASK) != BCC) {                   /* if the instruction is BR */
            cpu_ea (CIR, &class, &offset, NULL);        /*   then get the effective address of the branch */

            if (cpu_stop_flags & SS_LOOP                /* if the infinite loop stop is active */
              && offset == (P - 2 & LA_MASK))           /*   and the target is the current instruction */
                status = STOP_INFLOOP;                  /*     then stop the simulator */
            else                                        /* otherwise */
                status = SCPE_OK;                       /*   continue */

            cpu_read_memory (fetch_checked, offset, &NIR);  /* load the next instruction register */
            P = offset + 1 & R_MASK;                        /*   and increment the program counter */
            }

        else if (TO_CCF (STA) & CIR << BCC_CCF_SHIFT)   /* otherwise if the BCC test succeeds */
            status = cpu_branch_short (TRUE);           /*   then branch to the target address */
        break;                                          /* otherwise continue execution at P + 1 */


    case 015:                                           /* LDD (CCA; STOV, BNDV), LDB (CCB; STOV, BNDV) */
        if (CIR & M_FLAG) {                             /* if the instruction is LDD */
            cpu_ea (CIR, &class, &offset, NULL);        /*   then get the effective address of the double-word */

            cpu_read_memory (class, offset, &operand_1);                /* read the MSW */
            cpu_read_memory (class, offset + 1 & LA_MASK, &operand_2);  /*   and the LSW of the operand */

            cpu_push ();                                /* push the MSW  */
            cpu_push ();                                /*   and the LSW  */
            RB = operand_1;                             /*     of the operand */
            RA = operand_2;                             /*       onto the stack */

            SET_CCA (RB, RA);                           /* set the condition code */
            }

        else {                                                  /* otherwise the instruction is LDB */
            cpu_ea (CIR | M_FLAG, &class,  &offset, &selector); /*   so get the effective word address of the byte */
            cpu_read_memory (class, offset, &operand);          /*     and read the operand */

            cpu_push ();                                /* push the stack down */

            if (selector == upper)                      /* if the upper byte is selected */
                RA = UPPER_BYTE (operand);              /*   then store it in the TOS */
            else                                        /* otherwise */
                RA = LOWER_BYTE (operand);              /*   store the lower byte in the TOS */

            SET_CCB (RA);                               /* set the condition code */
            }
        break;


    case 016:                                           /* STD (none; STUN, BNDV), STB (none; STUN, BNDV) */
        if (CIR & M_FLAG) {                             /* if the instruction is STD */
            cpu_ea (CIR, &class, &offset, NULL);        /*   then get the effective address of the double-word */

            PREADJUST_SR (2);                           /* ensure that at least two TOS registers are loaded */

            cpu_write_memory (class, offset + 1 & LA_MASK, RA); /* write the LSW first to follow the microcode */
            cpu_write_memory (class, offset, RB);               /*   and then write the MSW */

            cpu_pop ();                                 /* pop the TOS */
            cpu_pop ();                                 /*   and the NOS */
            }

        else {                                                  /* otherwise the instruction is STB */
            cpu_ea (CIR | M_FLAG, &class, &offset, &selector);  /*   so get the effective word address of the byte */

            PREADJUST_SR (1);                           /* ensure that at least one TOS register is loaded */

            cpu_read_memory (class, offset, &operand);  /* read the word containing the target byte */

            if (selector == upper)                      /* if the upper byte is targeted */
                operand = REPLACE_UPPER (operand, RA);  /*   then store the TOS into it */
            else                                        /* otherwise */
                operand = REPLACE_LOWER (operand, RA);  /*   store the TOS into the lower byte */

            cpu_write_memory (class, offset, operand);  /* write the word back */
            cpu_pop ();                                 /*   and pop the TOS */
            }
        break;


    case 017:                                           /* LRA (none; STOV, BNDV) */
        cpu_ea (CIR, &class, &offset, NULL);            /* get the effective address */

        if (class == program_checked)                   /* if this a program reference */
            offset = offset - PB;                       /*   then subtract PB to get the relative address */
        else                                            /* otherwise it is a data or stack reference */
            offset = offset - DB;                       /*   so subtract DB to get the address */

        cpu_push ();                                    /* push the relative address */
        RA = offset & R_MASK;                           /*   onto the stack */

        if (NIR == PCAL_0                               /* if the next instruction is PCAL 0 */
          && sys_configuration & FW_CSTX)               /*   and the CST expansion firmware is present */
            SP0 = LRA;                                  /*     then record our instruction */
        else                                            /* otherwise */
            SP0 = NOP;                                  /*   record a dummy instruction */
        break;

    }                                                   /* all cases are handled */

if (status == STOP_UNIMPL                               /* if the instruction is unimplemented */
  && not (cpu_stop_flags & SS_UNIMPL))                  /*   and the unimplemented instruction stop is inactive */
    MICRO_ABORT (trap_Unimplemented);                   /*     then trap to handle it */

return status;
}


/* Intermodule Bus Adapter IMB interface.

   The IMB systems connect their main memory modules to the IMB through a memory
   controller.  In simulation, instruction executors make memory read and write
   calls directly to the memory simulator to save the overhead of performing IMB
   bus cycles.  However, the DMA sequencer on the GIC does request IMB memory
   cycles because the GIC diagnostic tests for parity error and illegal address
   handling.  As there is no memory device simulation, and consequently no DIB
   in which to specify the IMB interface routine, we handle IMB memory cycles
   here.

   This routine also implements (well, fakes) the Memory Command and Status
   (MCS) command.  The MCS operation involves two extra registers on the MCL:
   the Alternate Check register (ACR) and the Alternate Status Register (ASR).
   These registers are loaded on the execution of each MCS operation.

   The MCS operation behaves as a read operation, in that the CPU establishes a
   value on the IMB address bus and receives data on the data bus.  The address,
   however, is decoded into fields, as follows:

     ADRE 8   7 | 6   5   4 | 3   2   1
        +---+---+---+---+---+---+---+---+
        | - | - | - |    ELA Address    |
        +---+---+---+---+---+---+---+---+

     ADR 0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
       +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
       |          ELA Address          | - | - | Alternate Status Reg  |
       +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
       |    Alternate Check Reg    | - | - | - | - | - | - | - | - | - |
       +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   The ELA Address and the Alternate Check Register contents are loaded
   simultaneously from the same address bus value.  The content of the ELA
   specified by the address is read out and returned in Status Register bit 8.

   The Alternate Status Register bits are defined as follows:

     10 = Disable PER
     11 = Disable all error logging
     12 = Disable read error correction
     13 = Use the ACR on subsequent writes
     14 = Write bit 15 into the specified ELA Address
     15 = Data bit to be written to the ELA

   Disabling PER logs single-bit errors but does not log double-bit errors; the
   latter are reported in bit 9 of the status word.

   Disabling all error logging reports single-bit errors in bit 10 and
   double-bit errors in bit 9 of the status word.

   The Alternate Check Register value is substituted for the normal ECC
   generation if ASR bit 13 is set.

   Setting both bits 10 and 12 causes subsequent writes to initialize both upper
   and lower memory values to the value supplied by the data bus and to generate
   correct ECC bits; error logging is inhibited, i.e., no read precedes the
   write.  If bit 13 is also set, then the ACR is used instead of the generated
   ECC check bits.

   Writing into the ELA also returns the previous bit value in bit 8 of the
   status word.

   The MCS operation returns the value of the Status Register on the data bus.
   The register assignments are as follows:

        0 | 1   2   3 | 4   5   6 | 7   8   9 |10  11  12 |13  14  15
      +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
      |      ECC Check Bits       | R | E | U | C | 0   1   0   0 | N |
      +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+

   Where:

     R = RAM type (0/1 = 16K/64K)
     E = Data bit from the specified address in the ELA
     U = Uncorrectable error occurred since last MCS
     C = Correctable error occurred since last MCS
     N = Controller number (0/1 = first/second controller)

   The controller number is 0 for the first (16K RAM SMAs) or only (64K RAM
   SMAs) controller and 1 for the second 16K RAM SMA controller.


   Implementation notes:

    1. We cannot rely on mem_read and mem_write returning FALSE if the address
       is out of range because they also set the Illegal Address interrupt flag
       when the caller is the CPU device.  So we must test for out of range
       conditions here.

    2. We fudge the GIC parity error test by detecting the MCS cycle that is
       used to disable and enable ECC on the memory controller, and then
       asserting the PER signal for any accesses made with ECC off.
       Technically, this is not correct, as only those words changed after ECC
       id disabled will indicate errors.  But our sleight-of-hand is good enough
       to satisfy the diagnostic at little cost.
*/

static IMB_OUTBOUND cpu_imb_interface (DIB *dibptr, IMB_INBOUND inbound)
{
static t_bool ecc_is_off   = FALSE;             /* TRUE if error correction is disabled */
const IMB_ADDRESS asr_mask = HP_BITS (10, 15);  /* mask to isolate the Alternate Status Register */
IMB_OUTBOUND      outbound = { 0, ADN | DDN | PRO };

switch (inbound.opcode) {                               /* dispatch the IMB opcode */

    case IMB_Read:                                      /* Memory Read */
        if (inbound.address >= mem_size)                /* if the address is out of range */
            outbound.signals &= ~DDN;                   /*   then deny the Data Done signal */

        else {                                          /* otherwise */
            mem_read (&cpu_dev, dma,                    /*   issue a DMA read cycle for the GIC */
                      inbound.address, &outbound.data);

            if (ecc_is_off)                             /* if ECC has been turned off */
                outbound.signals |= PER;                /*   then assert the parity error signal */
            }
        break;


    case IMB_Write:                                     /* Memory Write */
        if (inbound.address >= mem_size)                /* if the address is out of range */
            outbound.signals &= ~DDN;                   /*   then deny the Data Done signal */

        else                                            /* otherwise */
            mem_write (&cpu_dev, dma,                   /*   issue a DMA write cycle */
                       inbound.address, inbound.data);
        break;


    case IMB_MCS:                                       /* Memory Command and Status */
        if ((inbound.address & asr_mask) == 054)        /* if the command is "substitute ACR value for ECC generation" */
            ecc_is_off = TRUE;                          /*   then turn error correction off */

        else if ((inbound.address & asr_mask) == 050)   /* otherwise if the command disables the substitute ACR */
            ecc_is_off = FALSE;                         /*   then turn error correction on */

        outbound.data = 000410;                         /* return "64K RAM" status */
        break;


    case IMB_IO_Write:                                  /* not used by this interface */
    case IMB_IO_Read:                                   /* not used by this interface */
    case IMB_RW_Ones:                                   /* not used by this interface */
        break;
    }

return outbound;                                        /* return the outbound signals and data */
}


/* Update the Status and Interrupt Register.

   The SIR is implemented only on IMB systems.  Because many bits duplicate
   those in the Series II/III CPX1 and CPX2 registers, we elect to work off of
   those registers and use the SIR only for the unique bits (IMB IRQ and CSRQ,
   and the system clock tick).  The other bits are set by calling this routine
   when the user might examine the register.
*/

static void update_SIR (void)
{
SIR &= sir_CPUTIMER | sir_IMBIRQ | sir_IMBCSRQ;         /* preserve just the clock, IRQ and CSRQ bits */

if (cpu_power_state == Power_On)                        /* if system power is on */
    SIR |= sir_CPUOK;                                   /*   then set the "released from reset" bit */

else if (cpu_power_state == Power_Failing)              /* otherwise if power is failing */
    SIR |= sir_PFW;                                     /*   then set the "power fail warning" bit */

if (CPX2 & cpx2_RUN)                                    /* if the CPU is running */
    SIR |= sir_CPURUN;                                  /*   then set the run bit */
else                                                    /* otherwise */
    SIR |= sir_RUNL;                                    /*   set the active-low run indicator for the CMP */

if ((STA & STATUS_OVTRAP) == STATUS_OVTRAP)             /* if an integer overflow trap is active */
    SIR |= sir_INTOVF;                                  /*   then set the trap bit */

if (not (CPX1 & cpx1_DISPFLAG))                         /* if not running in the dispatcher */
    SIR |= sir_DISPL;                                   /*   then set the active-low dispatcher bit */

if (not (CPX1 & cpx1_ICSFLAG))                          /* if not running on the ICS */
    SIR |= sir_ICSL;                                    /*   then set the active-low ICS bit */

if (DBANK == SBANK)                                     /* if not running in split-stack mode */
    SIR |= sir_SPLITSTKL;                               /*   then set the active-low split-stack bit */

return;
}


/* Trace the register set.

   This routine outputs the current values of the machine registers to the
   debug log.  The first output line consists of the SBANK, SM, SR, TOS, index,
   and status register values.  If the "cpu_base_changed" flag is set, a second
   line consisting of the DBANK register, code segment number, and code, data,
   and stack base and limit registers is also output.  If the CST expansion
   firmware is installed, the mapping state of the current code segment is
   appended to the second line.

   The routine must not be called unless the debug log is defined, and should
   not be called unless the register trace flag is set.
*/

static void trace_registers (void)
{
static const char *const stack_formats [] = {           /* stack register display formats, indexed by SR */
    BOV_FORMAT "  ",                                    /*   SR = 0 format */
    BOV_FORMAT "  A %06o, ",                            /*   SR = 1 format */
    BOV_FORMAT "  A %06o, B %06o, ",                    /*   SR = 2 format */
    BOV_FORMAT "  A %06o, B %06o, C %06o, ",            /*   SR = 3 format */
    BOV_FORMAT "  A %06o, B %06o, C %06o, D %06o, "     /*   SR = 4 format */
    };

hp_trace (&cpu_dev, TRACE_REG, stack_formats [SR],      /* output the active TOS registers */
          SBANK, SM, SR, RA, RB, RC, RD);

fprintf (sim_deb, "X %06o, %s\n", X, fmt_status (STA)); /* output the index and status registers */

if (cpu_base_changed) {                                 /* if the base registers have been altered */
    hp_trace (&cpu_dev, TRACE_REG,                      /*   then output the base register values */
              BOV_FORMAT "  PB %06o, PL %06o, DL %06o, DB %06o, Q %06o, Z %06o%s\n",
              DBANK, 0, STATUS_CS (STA), PB, PL, DL, DB, Q, Z,
              (sys_configuration & FW_CSTX ? (CSMAP & STMK_M ? " P" : " L") : ""));

    cpu_base_changed = FALSE;                           /* clear the base registers changed flag */
    }

return;
}
